/*
 * Copyright (c) 2001, Silver Egg Technology Co., Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * $Id: Marker.java,v 1.1 2001/04/18 17:39:48 djvanmaren Exp $
 */

package net.sourceforge.e4graph;

/**
 * The <code>Marker</code> class provides the abstraction of a persistent named 
 * marker. A marker is associated with a storage: the <code>Marker</code> class 
 * provides a method to retrieve an instance of <code>Storage</code> that 
 * represents the associated storage. A marker marks a specific node, its 
 * <i>marked node</i>. Each marker is identified by a name that is required to 
 * be unique within the containing storage; each marker is also identified by a 
 * unique integer identifier.
 * <p>
 * More than one marker may mark a specific node. A node is <i>reachable</i> if 
 * it is marked by one or more markers, or if it is reachable from a marked node, 
 * recursively. The <code>Marker</code> class provides methods to retrieve the 
 * marked node and to reuse this marker to mark another node.
 *
 * @author <a href="mailto:ryan@silveregg.co.jp">Ryan Shaw</a>
 */
public interface Marker {

    /**
     * Returns the name with which this marker was created.
     *
     * @return the name of this marker
     * @exception E4GraphException if an error occurs
     */
    String getName() throws E4GraphException;

    /**
     * Returns the storage associated with this marker.
     *
     * @return the storage associated with this marker
     * @exception E4GraphException if an error occurs
     */
    Storage getStorage() throws E4GraphException;

    /**
     * Returns the node marked by this marker.
     *
     * @return the node marked by this marker
     * @exception E4GraphException if an error occurs
     */
    Node getMarkedNode() throws E4GraphException;

    /**
     * Sets the node marked by this marker to <tt>n</tt>. The node previously 
     * marked by this marker may now be unreachable. If so, it is closed and
     * deleted.
     *
     * @param n the new node to be marked by this marker
     * @exception E4GraphException if an error occurs
     */
    void setMarkedNode(Node n) throws E4GraphException;

    /**
     * Returns the integer that uniquely identifies this marker within its 
     * associated storage.
     *
     * @return this marker's unique ID
     * @exception E4GraphException if an error occurs
     */
    int getUniqueID() throws E4GraphException;

    /**
     * Deletes the underlying storage occupied by this marker in its 
     * associated storage and recursively deletes nodes and vertices that 
     * become unreachable as a result of the deletion. Any instances of 
     * this marker or nodes and vertices that became unreachable are closed
     * and deleted.
     *
     * @exception E4GraphException if an error occurs
     */
    void delete() throws E4GraphException;

    /**
     * Tests to see if a marker is closed.
     *
     * @return <tt>true</tt> if the marker is closed; <tt>false</tt> if 
     *         it's still open
     */
    boolean isClosed();
    
    /**
     * Releases a marker's resources and closes it immediately.
     */
    void close();
}
