/*
 * Copyright (c) 2001, Silver Egg Technology Co., Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * $Id: Node.java,v 1.1 2001/04/18 17:39:48 djvanmaren Exp $
 */

package net.sourceforge.e4graph;

import java.util.List;
import java.util.ListIterator;

/**
 * The <code>Node</code> class provides the abstraction of a persistent node
 * in a graph. A node may be marked by one or more markers, and may be the 
 * value of one or more vertices in other nodes. The <code>Node</code> class 
 * provides methods to retrieve any markers that mark this node, and to retrieve 
 * any nodes that contain vertices with this node as their value. Nodes that 
 * contain a vertex with this node as its value are <i>parents</i> of this node.
 * <p>
 * A node is said to be <i>unreachable</i> if no markers mark it and the node is 
 * not the value of any vertex. Unreachable nodes (and all their recursively
 * unreachable vertices) are automatically closed and deleted.
 * <p>
 * A node can contain any number of vertices; the node is said to be the 
 * <i>originating</i> node for these vertices. Vertices are named and indexed in 
 * order of occurrence, and each vertex has a type and value. More than one 
 * vertex with the same name can occur in a given node, and vertices are 
 * identified either by index or by name and occurrence index. Indices are 
 * zero-based, that is, the first vertex in a node has index 0.
 * <p>
 * Each node is identified by an integer that is unique within the storage 
 * containing this node.
 *
 * @author <a href="mailto:ryan@silveregg.co.jp">Ryan Shaw</a>
 */
public interface Node {

    /**
     * Returns the number of vertices in this node.
     *
     * @return the number of vertices in this node
     */
    int getVertexCount();

    /**
     * Appends a new vertex with the specified name and value to the end of this 
     * node's list of vertices.
     *
     * @param name the name of the new vertex
     * @param value the value of the new vertex
     * @return the newly added vertex
     * @exception E4GraphException if the node is closed
     */
    Vertex addVertex(String name, Object value) throws E4GraphException;
    
    /**
     * Inserts a new vertex with the specified name and value into this node's
     * list of vertices at the specified position. Shifts the vertex currently 
     * at that position (if any) and any subsequent vertices to the right (adds 
     * one to their indices).
     *
     * @param index index at which the new vertex is to be inserted
     * @param name the name of the new vertex
     * @param value the value of the new vertex
     * @return the newly added vertex
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    Vertex addVertex(int index, String name, Object value) 
        throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns <tt>true</tt> if there is at least one vertex in this node named 
     * <tt>name</tt>.
     *
     * @param name the name of the vertex to look for
     * @return <tt>true</tt> if there is at least one vertex in this node named 
     *         <tt>name</tt>; <tt>false</tt> otherwise
     * @exception E4GraphException if the node is closed
     */
    boolean containsVertex(String name) throws E4GraphException;

    /**
     * Returns <tt>true</tt> if there are at least <tt>n</tt> vertices in this 
     * node named <tt>name</tt>.
     *
     * @param name the name of the vertices to look for
     * @param n the minimum number of vertices required
     * @return <tt>true</tt> if there are at least <tt>n</tt> vertices in this 
     *         node named <tt>name</tt>; <tt>false</tt> otherwise
     * @exception E4GraphException if the node is closed
     */
    boolean containsVertices(String name, int n) throws E4GraphException;
    
    /**
     * Returns the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the vertex at the specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    Vertex getVertex(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the first vertex named <tt>name</tt>, or <tt>null</tt> if no
     * such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the first vertex named <tt>name</tt>, or <tt>null</tt> if no
     *         such vertex exists
     * @exception E4GraphException if the node is closed
     */
    Vertex getVertex(String name) throws E4GraphException;
    
    /**
     * Returns the <tt>nth</tt> vertex named <tt>name</tt>, or <tt>null</tt> if 
     * no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the <tt>nth</tt> vertex named <tt>name</tt>, or <tt>null</tt> 
     *         if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    Vertex getVertex(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the value of the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the value of the vertex at the specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    Object getVertexValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the value of the first vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the value of the first vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    Object getVertexValue(String name) throws E4GraphException;
    
    /**
     * Returns the value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    Object getVertexValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the node value of the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the node value of the vertex at the specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    Node getVertexNodeValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the node value of the first vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the node value of the first vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    Node getVertexNodeValue(String name) throws E4GraphException;
    
    /**
     * Returns the node value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the node value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    Node getVertexNodeValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the string value of the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the string value of the vertex at the specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    String getVertexStringValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the string value of the first vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the string value of the first vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    String getVertexStringValue(String name) throws E4GraphException;
    
    /**
     * Returns the string value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the string value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    String getVertexStringValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the byte array value of the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the byte array value of the vertex at the specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    byte[] getVertexBytesValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the byte array value of the first vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the byte array value of the first vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    byte[] getVertexBytesValue(String name) throws E4GraphException;
    
    /**
     * Returns the byte array value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the byte array value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    byte[] getVertexBytesValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the integer value of the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the integer value of the vertex at the specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    int getVertexIntValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the integer value of the first vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the integer value of the first vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    int getVertexIntValue(String name) throws E4GraphException;
    
    /**
     * Returns the integer value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the integer value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    int getVertexIntValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the long integer value of the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the long integer value of the vertex at the specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    long getVertexLongValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the long integer value of the first vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the long integer value of the first vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    long getVertexLongValue(String name) throws E4GraphException;
    
    /**
     * Returns the long integer value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the long integer value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     *         or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    long getVertexLongValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the single-precision floating-point value of the vertex at the 
     * specified index.
     *
     * @param index the index of the desired vertex
     * @return the single-precision floating-point value of the vertex at the 
     *         specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    float getVertexFloatValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the single-precision floating-point value of the first vertex 
     * named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the single-precision floating-point value of the first vertex 
     *         named <tt>name</tt>, or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    float getVertexFloatValue(String name) throws E4GraphException;
    
    /**
     * Returns the single-precision floating-point value of the <tt>nth</tt> 
     * vertex named <tt>name</tt>, or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the single-precision floating-point value of the <tt>nth</tt> 
     *         vertex named <tt>name</tt>, or <tt>null</tt> if no such vertex 
               exists
     * @exception E4GraphException if the node is closed
     */
    float getVertexFloatValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the double-precision floating-point value of the vertex at the 
     * specified index.
     *
     * @param index the index of the desired vertex
     * @return the double-precision floating-point value of the vertex at the 
     *         specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    double getVertexDoubleValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the double-precision floating-point value of the first vertex 
     * named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the double-precision floating-point value of the first vertex 
     *         named <tt>name</tt>, or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    double getVertexDoubleValue(String name) throws E4GraphException;
    
    /**
     * Returns the double-precision floating-point value of the <tt>nth</tt> 
     * vertex named <tt>name</tt>, or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the double-precision floating-point value of the <tt>nth</tt> 
     *         vertex named <tt>name</tt>, or <tt>null</tt> if no such vertex 
               exists
     * @exception E4GraphException if the node is closed
     */
    double getVertexDoubleValue(String name, int nth) throws E4GraphException;
    
    /**
     * Returns the boolean value of the vertex at the specified index.
     *
     * @param index the index of the desired vertex
     * @return the boolean value of the vertex at the 
     *         specified index
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if the node is closed
     */
    boolean getVertexBooleanValue(int index) throws IndexOutOfBoundsException, E4GraphException;

    /**
     * Returns the boolean value of the first vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @return the boolean value of the first vertex 
     *         named <tt>name</tt>, or <tt>null</tt> if no such vertex exists
     * @exception E4GraphException if the node is closed
     */
    boolean getVertexBooleanValue(String name) throws E4GraphException;
    
    /**
     * Returns the boolean value of the <tt>nth</tt> vertex named <tt>name</tt>, 
     * or <tt>null</tt> if no such vertex exists.
     *
     * @param name the name of the desired vertex
     * @param nth the occurrence index of the desired vertex
     * @return the boolean value of the <tt>nth</tt> 
     *         vertex named <tt>name</tt>, or <tt>null</tt> if no such vertex 
               exists
     * @exception E4GraphException if the node is closed
     */
    boolean getVertexBooleanValue(String name, int nth) throws E4GraphException;
    
    /**
     * Creates a new node, and appends a new vertex with the specified name 
     * which points to the new node to this node's list of vertices.
     *
     * @param name the name of the new vertex
     * @return the newly added vertex
     * @exception E4GraphException if this node is closed
     */
    Vertex addNode(String name) throws E4GraphException;
    
    /**
     * Creates a new node, and inserts a new vertex with the specified name 
     * which points to the new node into this node's list of vertices at the 
     * specified position. Shifts the vertex currently at that position (if 
     * any) and any subsequent vertices to the right (adds one to their indices).
     *
     * @param index index at which the new vertex is to be inserted
     * @param name the name of the new vertex
     * @return the newly added vertex
     * @exception IndexOutOfBoundsException if the index is out of range 
     *            (<tt>index < 0 || index > getVertexCount()</tt>)
     * @exception E4GraphException if this node is closed
     */
    Vertex addNode(int index, String name) throws E4GraphException;

    /**
     * Returns an iterator over the list of vertices in this node, in index
     * order.
     *
     * @return an iterator over the list of vertices in this node
     * @exception E4GraphException if the node is closed
     */
    ListIterator vertices() throws E4GraphException;

    /**
     * Returns an iterator over the list of vertices in this node accepted
     * by the specified filter, in index order.
     *
     * @param filter a <code>VertexFilter</code> value
     * @return a iterator over a filtered list of vertices in this node
     * @exception E4GraphException if the node is closed
     */
    ListIterator vertices(VertexFilter filter) throws E4GraphException;

    /**
     * Returns an iterator over the list of vertices in this node, in index 
     * order, starting with the specified vertex.
     *
     * @param start the first vertex in the list
     * @return an iterator over the list of vertices in this node starting with the 
     *         specified vertex
     * @exception E4GraphException if the node or starting vertex is closed
     */
    ListIterator vertices(Vertex start) throws E4GraphException;
    
    /**
     * Returns an iterator over the list of vertices in this node accepted 
     * by the specified filter, in index order, starting with the specified vertex.
     *
     * @param start the first vertex in the list
     * @param filter the filter to use
     * @return an iterator over a filtered list of vertices in this node 
     *         starting with the specified vertex
     * @exception E4GraphException if the node or starting vertex is closed
     */
    ListIterator vertices(Vertex start, VertexFilter filter) throws E4GraphException;
    
    /**
     * Returns a list of this node's parent nodes.
     *
     * @return a list of this node's parent nodes
     * @exception E4GraphException if the node is closed
     */
    List getParents() throws E4GraphException;

    /**
     * Returns <tt>true</tt> if this node has no parents.
     *
     * @return <tt>true</tt> if this node has no parents, <tt>false</tt> otherwise
     */
    boolean isRoot();

    /**
     * Returns the number of vertices that contain this node as their value.
     *
     * @return the number of vertices that contain this node as their value
     */
    int getOccurrenceCount();
    
    /**
     * Returns the number of vertices within <tt>parent</tt> that contain this 
     * node as their value.
     *
     * @param parent a parent node
     * @return the number of vertices within <tt>parent</tt> that contain this 
     *         node as their value
     */
    int getOccurrenceCount(Node parent);

    /**
     * Returns a list of the markers marking this node.
     *
     * @return a list of the markers marking this node
     * @exception E4GraphException if the node is closed
     */
    List getMarkers() throws E4GraphException;

    /**
     * Adds a new marker named <tt>name</tt> to the list of markers marking this
     * node.
     *
     * @param name the name of the new marker
     * @exception E4GraphException if the node is closed
     */
    void mark(String name) throws E4GraphException;
    
    /**
     * Adds the marker <tt>m</tt> to the list of markers marking this node.
     *
     * @param m a marker to be added to the list of markers marking this node
     * @exception E4GraphException if the node or specified marker is closed
     */
    void mark(Marker m) throws E4GraphException;
    
    /**
     * Removes the marker <tt>m</tt> from the list of markers marking this node.
     *
     * @param m a marker to be removed from the list of markers marking this node
     * @exception E4GraphException if the node or specified marker is closed
     */
    void unmark(Marker m) throws E4GraphException;

    /**
     * Removes the marker at <tt>index</tt> from the list of markers marking this 
     * node.
     *
     * @param index the index of the marker to be removed
     * @exception E4GraphException if the node is closed
     * @exception IndexOutOfBoundsException if the index is out of range 
     */
    void unmark(int index) throws E4GraphException, IndexOutOfBoundsException;

    /**
     * Returns <tt>true</tt> if there is at least one marker that is marking 
     * this node.
     *
     * @return <tt>true</tt> if there is at least one marker that is marking 
     *         this node, <tt>false</tt> otherwise
     */
    boolean isMarked();
    
    /**
     * Returns the storage containing this node.
     *
     * @return the storage containing this node
     * @exception E4GraphException if the node is closed
     */
    Storage getStorage() throws E4GraphException;

    /**
     * Returns an integer identifier for this node guaranteed to be unique 
     * within the context of the storage containing this node.
     *
     * @return an integer which uniquely identifies this node
     * @exception E4GraphException if the node is closed
     */
    int getUniqueID() throws E4GraphException;

    /**
     * Deletes this node from its persistent storage.
     * 
     * @exception E4GraphException if an error occurs
     */
    boolean delete() throws E4GraphException;

    /**
     * Tests to see if a node is closed.
     *
     * @return <tt>true</tt> if the node is closed; <tt>false</tt> if 
     *         it's still open
     */
    boolean isClosed();

    /**
     * Releases a node's resources and closes it immediately.
     */
    void close();
    
}
