/*
 * Copyright (c) 2001, Silver Egg Technology Co., Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * $Id: Storage.java,v 1.1 2001/04/18 17:39:48 djvanmaren Exp $
 */

package net.sourceforge.e4graph;

import java.util.Iterator;

/**
 * The <code>Storage</code>class provides persistent storage of markers 
 * and elements reachable from those markers. A user application may have 
 * any number of storages open at a given time, limited only by the available 
 * machine resources.
 *
 * @author <a href="mailto:ryan@silveregg.co.jp">Ryan Shaw</a>
 */
public interface Storage {

    int MARKER_SPACE = 0;
    int NODE_SPACE   = 1;
    int VERTEX_SPACE = 2;
    int NAME_SPACE   = 3;
    int STRING_SPACE = 4;
    int INT_SPACE    = 5;
    int FLOAT_SPACE  = 6;
    int BINARY_SPACE = 7;

    int SPACE_USED  = 0;
    int SPACE_AVAIL = 1;
    int SPACE_FREED = 2;
    int SPACE_ALLOC = 3;

    /**
     * Sets this storage's auto-commit mode. If a storage is in auto-commit 
     * mode, changes are automatically committed when the storage is closed.
     *
     * @param autoCommit <tt>true</tt> enables auto-commit; <tt>false</tt> 
     *                   disables auto-commit.
     */
    void setAutoCommit(boolean autoCommit);

    /**
     * Gets the current auto-commit state.
     *
     * @return the current state of auto-commit mode
     */
    boolean getAutoCommit();

    /**
     * Commits any changes that have been made to the storage immediately.
     *
     * @exception E4GraphException if an error occurs
     */
    void commit() throws E4GraphException;
    
    /**
     * Deletes the underlying storage. No events are fired because of the
     * deletion of storage elements.
     *
     * @exception E4GraphException if an error occurs
     */
    void delete() throws E4GraphException;

    /**
     * Tests to see if a storage is closed.
     *
     * @return <tt>true</tt> if the storage is closed; <tt>false</tt> if 
     *         it's still open
     */
    boolean isClosed();

    /**
     * Releases a storage's resources and closes it immediately, committing 
     * if necessary. When a storage is closed, and markers, nodes or vertices
     * referring to elements within that storage become invalid.
     *
     * @exception E4GraphException if an error occurs
     */
    void close() throws E4GraphException;
    
    /**
     * Copies the contents of the receiver to <code>otherStorage</code>. The 
     * previous contents of otherStorage are deleted, and no events are fired 
     * because of the deletion. If <code>forceCommit</code> is <tt>true</tt>, 
     * <code>otherStorage</code> will be be committed after the copy is done.
     * <p>
     * After this operation, the contents of this storage and 
     * <code>otherStorage</code> are identical. Changes made to one storage 
     * after the copy are not reflected in the other storage.
     * <p>
     * Listeners registered with <code>otherStorage</code> stay in effect and 
     * may be called when events fire after the copy, if changes are made to 
     * <code>otherStorage</code>.
     *
     * @param otherStorage the storage into which the contents of this storage 
     *                     will be copied
     * @param forceCommit whether or not to commit the other storage after the 
     *                    copy
     * @exception E4GraphException if an error occurs
     */
    void copyInto(Storage otherStorage, boolean forceCommit) throws E4GraphException;
    
    /**
     * Returns <tt>true</tt>true if there are uncommitted modifications to this 
     * storage; otherwise returns <tt>false</tt>.
     *
     * @return <tt>true</tt> if there are uncommitted modifications, 
     *         <tt>false</tt> otherwise
     */
    boolean isDirty();

    /**
     * Marks the storage as dirty, i.e. it has been modified and not yet 
     * committed.
     */
    void markDirty();

    /**
     * Returns the number of markers existing within this storage.
     *
     * @return number of markers
     */
    int getMarkerCount();

    /**
     * Returns <tt>true</tt> if a marker with the specified name exists in this 
     * storage. Returns <tt>false</tt> if no marker by that name exists.
     *
     * @param name name of the marker to look for
     * @return <tt>true</tt> if a marker by that name exists, <tt>false</tt> 
     *         otherwise
     */
    boolean containsMarker(String name);

    /**
     * Returns the marker with the specified name if it exists, otherwise returns
     * <tt>null</tt>.
     *
     * @param name name of the marker to return
     * @return the named marker or <tt>null</tt> if no marker by that name 
     *         exists
     * @exception E4GraphException if an error occurs
     */
    Marker getMarker(String name) throws E4GraphException;
    
    /**
     * Returns the marker with the specified unique ID if it exists, otherwise
     * returns <tt>null</tt>.
     *
     * @param id a valid unique ID
     * @return the marker identified by <tt>id</tt> or <tt>null</tt> if no
     *         such marker exists
     * @exception E4GraphException if an error occurs
     */
    Marker getMarker(long id) throws E4GraphException;

    /**
     * Returns the node marked by the marker with the specified name. If no
     * marker by that name exists, a new marker with the specified name will
     * be created, and a new node marked by that marker will be created
     * and returned.
     *
     * @param name name of the marker to return or create
     * @return the node marked by the marker with the specified name
     * @exception E4GraphException if an error occurs
     */
    Node getMarkedNode(String name) throws E4GraphException;

    /**
     * Returns the node with the specified unique ID if it exists, otherwise
     * returns <tt>null</tt>.
     *
     * @param id a valid unique ID
     * @return the node identified by <tt>id</tt> or <tt>null</tt> if no
     *         such node exists
     * @exception E4GraphException if an error occurs
     */
    Node getNode(long id) throws E4GraphException;

    /**
     * Returns the vertex with the specified unique ID if it exists, otherwise
     * returns <tt>null</tt>.
     *
     * @param id a valid unique ID
     * @return the vertex identified by <tt>id</tt> or <tt>null</tt> if no
     *         such vertex exists
     * @exception E4GraphException if an error occurs
     */
    Vertex getVertex(long id) throws E4GraphException;

    /**
     * Returns an iterator over all the markers in this storage. The order in
     * which markers are visited is implementation specific.
     *
     * @return an iterator over the markers in this storage
     */
    Iterator markers();
    
    /**
     * Returns an iterator over all the nodes in this storage. The order in
     * which nodes are visited is implementation specific.
     *
     * @return an iterator over the nodes in this storage
     */
    Iterator nodes();
    
    /**
     * Returns an iterator over all the vertices in this storage. The order in
     * which vertices are visited is implementation specific.
     *
     * @return an iterator over the vertices in this storage
     */
    Iterator vertices();
    
    /**
     * Returns an iterator over vertices in this storage accepted by the 
     * specified filter. The order in which vertices are visited is 
     * implementation specific.
     *
     * @param filter visit only vertices accepted by this filter
     * @return an iterator over filtered vertices in this storage
     */
    Iterator vertices(VertexFilter filter);
    
    /**
     * Returns the name of the storage, or <tt>null</tt> if the storage is 
     * closed.
     *
     * @return the name of the storage, or <tt>null</tt> if the storage is 
     *         closed
     */
    String getName();

    /**
     * Returns the name of the driver used to open this storage, or <tt>null</tt> 
     * if the storage is closed. 
     *
     * @return the driver name, or <tt>null</tt> if the storage is closed 
     */
    String getDriver();

    /**
     * Returns <tt>true</tt> if the storage is empty, that is, it contains no 
     * markers or other elements. Otherwise it returns <tt>false</tt>.
     *
     * @return <tt>true</tt> if the storage is empty, <tt>false</tt> otherwise
     */
    boolean isEmpty();
    
    /**
     * Removes all markers and elements reachable from those markers from this 
     * storage.
     *
     * @exception E4GraphException if an error occurs
     */
    void removeAll() throws E4GraphException;

    /**
     * Retrieves statistics about the use of various allocation spaces within 
     * a storage.
     *
     * @param space the allocation space we are interested in
     * @param stat the specific kind of statistic desired
     * @return the desired statistic
     * @exception E4GraphException if an error occurs
     */
    int getStatistic(int space, int stat) throws E4GraphException;

    /**
     * Adds the specified marker addition listener to receive marker addition 
     * events from this storage. Marker addition events occur when markers are 
     * added to this storage. If <tt>l</tt> is <tt>null</tt>, no exception is 
     * thrown and no action is performed.
     *
     * @param l the marker addition listener
     *
     * @see MarkerAdditionListener
     * @see #removeMarkerAdditionListener
     */
    void addMarkerAdditionListener(MarkerAdditionListener l);

    /**
     * Removes the specified marker addition listener so that it no longer 
     * receives marker addition events from this storage. Marker addition events 
     * occur when markers are added to this storage. If <tt>l</tt> is 
     * <tt>null/tt>, no exception is thrown and no action is performed.
     *
     * @param l the marker addition listener
     *
     * @see MarkerAdditionListener
     * @see #addMarkerAdditionListener
     */
    void removeMarkerAdditionListener(MarkerAdditionListener l);

    /**
     * Adds the specified marker deletion listener to receive marker deletion 
     * events from this storage. Marker deletion events occur when markers are 
     * added to this storage. If <tt>l</tt> is <tt>null/tt>, no exception is 
     * thrown and no action is performed.
     *
     * @param l the marker deletion listener
     *
     * @see MarkerDeletionListener
     * @see #removeMarkerDeletionListener
     */
    void addMarkerDeletionListener(MarkerDeletionListener l);

    /**
     * Removes the specified marker deletion listener so that it no longer 
     * receives marker deletion events from this storage. Marker deletion events 
     * occur when markers are added to this storage. If <tt>l</tt> is 
     * <tt>null/tt>, no exception is thrown and no action is performed.
     *
     * @param l the marker deletion listener
     *
     * @see MarkerDeletionListener
     * @see #addMarkerDeletionListener
     */
    void removeMarkerDeletionListener(MarkerDeletionListener l);

    /**
     * Adds the specified node addition listener to receive node addition 
     * events from this storage. Node addition events occur when nodes are 
     * added to this storage. If <tt>l</tt> is <tt>null/tt>, no exception is 
     * thrown and no action is performed.
     *
     * @param l the node addition listener
     *
     * @see NodeAdditionListener
     * @see #removeNodeAdditionListener
     */
    void addNodeAdditionListener(NodeAdditionListener l);

    /**
     * Removes the specified node addition listener so that it no longer 
     * receives node addition events from this storage. Node addition events 
     * occur when nodes are added to this storage. If <tt>l</tt> is 
     * <tt>null/tt>, no exception is thrown and no action is performed.
     *
     * @param l the node addition listener
     *
     * @see NodeAdditionListener
     * @see #addNodeAdditionListener
     */
    void removeNodeAdditionListener(NodeAdditionListener l);

    /**
     * Adds the specified node deletion listener to receive node deletion 
     * events from this storage. Node deletion events occur when nodes are 
     * added to this storage. If <tt>l</tt> is <tt>null/tt>, no exception is 
     * thrown and no action is performed.
     *
     * @param l the node deletion listener
     *
     * @see NodeDeletionListener
     * @see #removeNodeDeletionListener
     */
    void addNodeDeletionListener(NodeDeletionListener l);

    /**
     * Removes the specified node deletion listener so that it no longer 
     * receives node deletion events from this storage. Node deletion events 
     * occur when nodes are added to this storage. If <tt>l</tt> is 
     * <tt>null/tt>, no exception is thrown and no action is performed.
     *
     * @param l the node deletion listener
     *
     * @see NodeDeletionListener
     * @see #addNodeDeletionListener
     */
    void removeNodeDeletionListener(NodeDeletionListener l);

    /**
     * Adds the specified vertex addition listener to receive vertex addition 
     * events from this storage. Vertex addition events occur when vertices are 
     * added to this storage. If <tt>l</tt> is <tt>null/tt>, no exception is 
     * thrown and no action is performed.
     *
     * @param l the vertex addition listener
     *
     * @see VertexAdditionListener
     * @see #removeVertexAdditionListener
     */
    void addVertexAdditionListener(VertexAdditionListener l);

    /**
     * Removes the specified vertex addition listener so that it no longer 
     * receives vertex addition events from this storage. Vertex addition events 
     * occur when vertices are added to this storage. If <tt>l</tt> is 
     * <tt>null/tt>, no exception is thrown and no action is performed.
     *
     * @param l the vertex addition listener
     *
     * @see VertexAdditionListener
     * @see #addVertexAdditionListener
     */
    void removeVertexAdditionListener(VertexAdditionListener l);

    /**
     * Adds the specified vertex modification listener to receive vertex modification 
     * events from this storage. Vertex modification events occur when vertices in 
     * this storage are changed. If <tt>l</tt> is <tt>null/tt>, no exception is 
     * thrown and no action is performed.
     *
     * @param l the vertex modification listener
     *
     * @see VertexModificationListener
     * @see #removeVertexModificationListener
     */
    void addVertexModificationListener(VertexModificationListener l);

    /**
     * Removes the specified vertex modification listener so that it no longer 
     * receives vertex modification events from this storage. Vertex modification events 
     * occur when vertices in this storage are changed. If <tt>l</tt> is 
     * <tt>null/tt>, no exception is thrown and no action is performed.
     *
     * @param l the vertex modification listener
     *
     * @see VertexModificationListener
     * @see #addVertexModificationListener
     */
    void removeVertexModificationListener(VertexModificationListener l);

    /**
     * Adds the specified vertex deletion listener to receive vertex deletion 
     * events from this storage. Vertex deletion events occur when vertices are 
     * added to this storage. If <tt>l</tt> is <tt>null/tt>, no exception is 
     * thrown and no action is performed.
     *
     * @param l the vertex deletion listener
     *
     * @see VertexDeletionListener
     * @see #removeVertexDeletionListener
     */
    void addVertexDeletionListener(VertexDeletionListener l);
    
    /**
     * Removes the specified vertex deletion listener so that it no longer 
     * receives vertex deletion events from this storage. Vertex deletion events 
     * occur when vertices are added to this storage. If <tt>l</tt> is 
     * <tt>null/tt>, no exception is thrown and no action is performed.
     *
     * @param l the vertex deletion listener
     *
     * @see VertexDeletionListener
     * @see #addVertexDeletionListener
     */
    void removeVertexDeletionListener(VertexDeletionListener l);
    
    
}// Storage
