/*
 * Copyright (c) 2001, Silver Egg Technology Co., Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * $Id: Vertex.java,v 1.1 2001/04/18 17:39:48 djvanmaren Exp $
 */

package net.sourceforge.e4graph;

/**
 * The <tt>Vertex</tt> class provides the abstraction of a vertex originating 
 * from a node. A vertex is associated with a node, its <i>originating node</i>.
 * Vertices are identified by name and instance count or by rank within the 
 * originating node. Each vertex stores an instance of class <tt>Object</tt>.
 *
 * @author <a href="mailto:ryan@silveregg.co.jp">Ryan Shaw</a>
 */
public interface Vertex {
    
    /**
     * Returns the value of this vertex.
     *
     * @return the value of this vertex
     * @exception E4GraphException if the vertex is closed
     */
    Object get() throws E4GraphException;

    /**
     * Sets the value of this vertex.
     *
     * @param value the new value of this vertex
     * @exception E4GraphException if the vertex is closed
     */
    void set(Object value) throws E4GraphException;

    /**
     * Returns the index of this vertex within its containing node.
     *
     * @return the index of this vertex within its containing node
     * @exception E4GraphException if the vertex is closed
     */
    int getIndex() throws E4GraphException;
    
    /**
     * Returns the name of this vertex, or <tt>null</tt> if the vertex is closed.
     *
     * @return the name of this vertex, or <tt>null</tt> if the vertex is closed
     */
    String getName();
    
    /**
     * Sets the name of this vertex.
     *
     * @param name the new name of the vertex
     * @exception E4GraphException if the vertex is closed
     */
    void setName(String name) throws E4GraphException;

    /**
     * Returns an integer which uniquely identifies this vertex within its 
     * containing storage.
     *
     * @return an integer which uniquely identifies this vertex
     */
    int getUniqueID();

    /**
     * Returns the storage containing this vertex.
     *
     * @return the storage containing this vertex
     * @exception E4GraphException if the vertex is closed
     */
    Storage getStorage() throws E4GraphException;

    /**
     * Returns the node containing this vertex.
     *
     * @return the node containing this vertex
     * @exception E4GraphException if the vertex is closed
     */
    Node getNode() throws E4GraphException;
    
    /**
     * Deletes this vertex from its persistent storage.
     *
     * @exception E4GraphException if an error occurs
     */
    void delete() throws E4GraphException;
    
    /**
     * Tests to see if a vertex is closed.
     *
     * @return <tt>true</tt> if the vertex is closed; <tt>false</tt> if 
     *         it's still open
     */
    boolean isClosed();

    /**
     * Releases a vertex's resources and closes it immediately.
     */
    void close();
    
}
