/*
 * i4marker.cpp --
 *
 *	This file provides the implementation of the e4_MarkerImpl class 
 *	defined	in e4graphimpl.h. See that file for comments on the
 *	semantics of the e4_MarkerImpl class.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graphimpl.h"

/*
 * Destructor:
 */

e4_MarkerImpl::~e4_MarkerImpl()
{
    if (storage != NULL) {
        storage->ForgetMarker(markerID);
    }
}

/*
 * Constructor:
 */

e4_MarkerImpl::e4_MarkerImpl(e4_StorageImpl *s, int ti)
    : e4_RefCounter(),
      storage(s),
      markerID(ti)
{
}

/*
 * This method is called when the reference count goes to (or below) zero.
 */

void
e4_MarkerImpl::NotReferenced()
{
    delete this;
}

/*
 * Operations to retrieve the various values associated with a marker:
 */

const char *
e4_MarkerImpl::GetName() const 
{
    if (storage == NULL) {
	return NULL;
    }
    return storage->GetMarkerName(markerID);
}

e4_StorageImpl *
e4_MarkerImpl::GetStorage() const 
{
    return storage;
}

e4_NodeImpl *
e4_MarkerImpl::GetMarkedNode() const 
{
    if (storage == NULL) {
	return NULL;
    }
    return storage->GetMarkedNode(markerID);
}

/*
 * Delete this marker.
 */

bool
e4_MarkerImpl::Delete()
{
    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    return storage->DeleteMarkerByID(markerID);
}

/*
 * Operation to return an ID which uniquely identifies this e4_MarkerImpl
 * in its containing e4_StorageImpl.
 */

int
e4_MarkerImpl::GetUniqueID() const
{
    if (storage == NULL) {
	return E4_MARKERNOTFOUND;
    }
    return markerID;
}

/*
 * Is this marker valid?
 */

bool
e4_MarkerImpl::IsValid() const
{
    if (storage == NULL) {
	return false;
    }
    return storage->IsValid();
}

/*
 * Assign the storage field:
 */

void
e4_MarkerImpl::SetStorage(e4_StorageImpl *newstorage)
{
    storage = newstorage;
}

/*
 * Return the marker ID for use by the containing storage implementation.
 */

int
e4_MarkerImpl::GetMarkerID() const
{
    return markerID;
}

/*
 * Get/Set the user data associated with this marker.
 */

bool
e4_MarkerImpl::GetUserData(int &userData) const
{
    if (storage == NULL) {
	return false;
    }
    return storage->GetMarkerUserData(markerID, userData);
}

bool
e4_MarkerImpl::SetUserData(int userData) const
{
    if (storage == NULL) {
	return false;
    }
    return storage->SetMarkerUserData(markerID, userData);
}

