/*
 * i4node.cpp --
 *
 *	This file provides the implementation of the e4_NodeImpl class defined
 *	in e4graphimpl.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graphimpl.h"

/*
 * This hash table is for hashing names of vertices to GUUIDs:
 */

static e4_HashTable *globalNameHash = NULL;
static int globalNameKey = 0;

/*
 * This constructor is for use by e4_Marker only:
 */

e4_NodeImpl::e4_NodeImpl(e4_StorageImpl *s, int ni)
    : e4_RefCounter(),
      storage(s),
      nodeID(ni)
{}

/*
 * The destructor:
 */

e4_NodeImpl::~e4_NodeImpl()
{
    if (storage != NULL) {
        storage->ForgetNode(nodeID);
    }
}

/*
 * This method is called when the reference count goes to (or below) zero.
 */

void
e4_NodeImpl::NotReferenced()
{
    delete this;
}

/*
 * Set an existing vertex or add a vertex with a node value.
 */
        
e4_NodeImpl *
e4_NodeImpl::SetNthNode(const char *nm, int nth)
{
    int i, rank, nameID, hisNodeID;

    if (storage == NULL) {
        return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return NULL;
    }
    hisNodeID = storage->ReserveNodeID();
    storage->AddParent(hisNodeID, nodeID);
    storage->SetVertexByIndexToNode(i, hisNodeID);
    if (storage->HasCallbacks(E4_CBADDNODE)) {
	storage->CallbackEvent(E4_CBADDNODE, 
			       storage->FindOrCreateNode(hisNodeID));
    }
    return storage->FindOrCreateNode(hisNodeID);
}

e4_NodeImpl *
e4_NodeImpl::SetNodeByRank(int rank)
{
    int i, hisNodeID;
    
    if (storage == NULL) {
        return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return NULL;
    }
    hisNodeID = storage->ReserveNodeID();
    storage->AddParent(hisNodeID, nodeID);
    storage->SetVertexByIndexToNode(i, hisNodeID);
    if (storage->HasCallbacks(E4_CBADDNODE)) {
	storage->CallbackEvent(E4_CBADDNODE,
			       storage->FindOrCreateNode(hisNodeID));
    }
    return storage->FindOrCreateNode(hisNodeID);
}

e4_NodeImpl *
e4_NodeImpl::AddNode(const char *nm, e4_InsertOrder io, int &rank)
{
    int i, hisNodeID;

    if (storage == NULL) {
	return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return NULL;
    }
    hisNodeID = storage->ReserveNodeID();
    storage->AddParent(hisNodeID, nodeID);
    storage->SetVertex(i, storage->InternName(nm), E4_VTNODE, hisNodeID);
    if (storage->HasCallbacks(E4_CBADDNODE)) {
	storage->CallbackEvent(E4_CBADDNODE,
			       storage->FindOrCreateNode(hisNodeID));
    }
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }
    return storage->FindOrCreateNode(hisNodeID);
}

bool
e4_NodeImpl::SetNthVertex(const char *nm, int nth, int value)
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, value)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetNthVertex(const char *nm, int nth, double value)
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, value)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetNthVertex(const char *nm, int nth, const char *value)
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, value)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetNthVertex(const char *nm, int nth, 
			  const void *bytes, int nbytes)
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, bytes, nbytes)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetNthVertexToNode(const char *nm, int nth, int childID)
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    storage->AddParent(childID, nodeID);
    if (!storage->SetVertexByIndexToNode(i, childID)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetVertexByRank(int rank, int value)
{
    int i;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, value)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetVertexByRank(int rank, double value)
{
    int i;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, value)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetVertexByRank(int rank, const char *value)
{
    int i;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, value)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetVertexByRank(int rank, const void *bytes, int nbytes)
{
    int i;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    if (!storage->SetVertexByIndex(i, bytes, nbytes)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::SetVertexByRankToNode(int rank, int childID)
{
    int i;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    storage->AddParent(childID, nodeID);
    if (!storage->SetVertexByIndexToNode(i, childID)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

/*
 * Add a vertex to this node at a given position.
 */

bool
e4_NodeImpl::AddVertex(const char *nm, e4_InsertOrder io, 
		       int &rank, int value)
{
    int i, j;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return false;
    }
    j = storage->AddInt(value);
    storage->SetVertex(i, storage->InternName(nm), E4_VTINT, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }
    return true;
}

bool
e4_NodeImpl::AddVertex(const char *nm, e4_InsertOrder io, 
		       int &rank, double value)
{
    int i, j;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return false;
    }
    j = storage->AddDouble(value);
    storage->SetVertex(i, storage->InternName(nm), E4_VTDOUBLE, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return true;
}

bool
e4_NodeImpl::AddVertex(const char *nm, e4_InsertOrder io, int &rank,
		       const char *value)
{
    int i, j;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return false;
    }
    j = storage->AddString(value);
    storage->SetVertex(i, storage->InternName(nm), E4_VTSTRING, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return true;
}

bool
e4_NodeImpl::AddVertex(const char *nm, e4_InsertOrder io, int &rank,
		       const void *bytes, int nbytes)
{
    int i, j;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return false;
    }
    j = storage->AddBinary(bytes, nbytes);
    storage->SetVertex(i, storage->InternName(nm), E4_VTBINARY, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return true;
}

bool
e4_NodeImpl::AddVertexWithNode(const char *nm, e4_InsertOrder io, int &rank,
			       int childID)
{
    int i;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return false;
    }
    storage->AddParent(childID, nodeID);
    storage->SetVertex(i, storage->InternName(nm), E4_VTNODE, childID);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return true;
}

/*
 * Add a vertex and return an e4_VertexImpl for it in one operation.
 */

e4_VertexImpl *
e4_NodeImpl::AddVertexRef(const char *nm, e4_InsertOrder io, 
			  int &rank, int value)
{
    int i, j;

    if (storage == NULL) {
	return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return NULL;
    }
    j = storage->AddInt(value);
    storage->SetVertex(i, storage->InternName(nm), E4_VTINT, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return storage->GetVertex(i);
}

e4_VertexImpl *
e4_NodeImpl::AddVertexRef(const char *nm, e4_InsertOrder io,
			  int &rank, double value)
{
    int i, j;

    if (storage == NULL) {
	return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return NULL;
    }
    j = storage->AddDouble(value);
    storage->SetVertex(i, storage->InternName(nm), E4_VTDOUBLE, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return storage->GetVertex(i);
}

e4_VertexImpl *
e4_NodeImpl::AddVertexRef(const char *nm, e4_InsertOrder io, int &rank, 
			  const char *value)
{
    int i, j;

    if (storage == NULL) {
	return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return NULL;
    }
    j = storage->AddString(value);
    storage->SetVertex(i, storage->InternName(nm), E4_VTSTRING, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return storage->GetVertex(i);
}

e4_VertexImpl *
e4_NodeImpl::AddVertexRef(const char *nm, e4_InsertOrder io, int &rank, 
			  const void *bytes, int nbytes)
{
    int i, j;

    if (storage == NULL) {
	return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return NULL;
    }
    j = storage->AddBinary(bytes, nbytes);
    storage->SetVertex(i, storage->InternName(nm), E4_VTBINARY, j);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return storage->GetVertex(i);
}

e4_VertexImpl *
e4_NodeImpl::AddVertexRefWithNode(const char *nm, e4_InsertOrder io, int &rank,
				  int childID)
{
    int i;

    if (storage == NULL) {
	return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return NULL;
    }
    storage->AddParent(childID, nodeID);
    storage->SetVertex(i, storage->InternName(nm), E4_VTNODE, childID);
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    return storage->GetVertex(i);
}

e4_VertexImpl *
e4_NodeImpl::AddNodeRef(const char *nm, e4_InsertOrder io, int &rank,
			e4_NodeImpl *&node)
{
    int i, hisNodeID;

    if (storage == NULL) {
	return NULL;
    }
    if (storage->IsLocked()) {
	return NULL;
    }
    i = storage->AddVertex(nodeID, io, rank);
    if (i == E4_VERTEXNOTCREATED) {
	return NULL;
    }
    hisNodeID = storage->ReserveNodeID();
    storage->AddParent(hisNodeID, nodeID);
    storage->SetVertex(i, storage->InternName(nm), E4_VTNODE, hisNodeID);
    if (storage->HasCallbacks(E4_CBADDNODE)) {
	storage->CallbackEvent(E4_CBADDNODE,
			       storage->FindOrCreateNode(hisNodeID));
    }
    if (storage->HasCallbacks(E4_CBADDVERTEX)) {
	storage->CallbackEvent(E4_CBADDVERTEX, storage->GetVertex(i));
    }

    node = storage->FindOrCreateNode(hisNodeID);

    return storage->GetVertex(i);
}

/*
 * Get a value from a vertex:
 */

bool
e4_NodeImpl::GetNthVertex(const char *nm, int nth, e4_ValueImpl *&v) const
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, v) == false)) {
	return false;
    }
    return true;
}

bool
e4_NodeImpl::GetNthVertex(const char *nm, int nth, e4_NodeImpl *&n) const
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, n) == false)) {
	return false;
    }
    return true;
}

bool
e4_NodeImpl::GetNthVertex(const char *nm, int nth, int &value) const
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, value) == false)) {
        return false;
    }
    return true;
}

bool
e4_NodeImpl::GetNthVertex(const char *nm, int nth, double &value) const
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, value) == false)) {
	return false;
    }
    return true;
}

bool
e4_NodeImpl::GetNthVertex(const char *nm, int nth, const char *&value) const
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, value) == false)) {
	return false;
    }
    return true;
}

bool
e4_NodeImpl::GetNthVertex(const char *nm, int nth,
			 const void *&bytes, int &nbytes)
  const
{
    int i, rank, nameID;

    if (storage == NULL) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if ((i == E4_VERTEXNOTFOUND) || 
	(storage->GetVertexByIndex(i, bytes, nbytes) == false)) {
	return false;
    }
    return true;
}

bool
e4_NodeImpl::GetVertexByRank(int rank, e4_ValueImpl *&value) const
{
    int i;

    if (storage == NULL) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
	(storage->GetVertexByIndex(i, value) == false)) {
        return false;
    }
    return true;
}

bool
e4_NodeImpl::GetVertexByRank(int rank, e4_NodeImpl *&value) const
{
    int i;

    if (storage == NULL) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, value) == false)) {
        return false;
    }
    return true;
}

bool
e4_NodeImpl::GetVertexByRank(int rank, int &value) const
{
    int i;

    if (storage == NULL) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, value) == false)) {
	return false;
   }
    return true;
}

bool
e4_NodeImpl::GetVertexByRank(int rank, double &value) const
{
    int i;

    if (storage == NULL) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, value) == false)) {
	return false;
    }
    return true;
}

bool
e4_NodeImpl::GetVertexByRank(int rank, const char *&value) const
{
    int i;

    if (storage == NULL) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
        (storage->GetVertexByIndex(i, value) == false)) {
	return false;
    }
    return true;
}

bool
e4_NodeImpl::GetVertexByRank(int rank, const void *&bytes, int &nbytes) const
{
    int i;

    if (storage == NULL) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if ((i == E4_VERTEXNOTFOUND) ||
	(storage->GetVertexByIndex(i, bytes, nbytes) == false)) {
	return false;
    }
    return true;
}

e4_VertexType
e4_NodeImpl::VertexType(const char *nm, int nth) const
{
    int i, rank, nameID;

    if (storage == NULL) {
        return E4_VTUNKNOWN;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return E4_VTUNKNOWN;
    }
    return storage->VertexTypeFromVertexID(i);
}

e4_VertexType
e4_NodeImpl::VertexTypeByRank(int rank) const
{
    int i;

    if (storage == NULL) {
        return E4_VTUNKNOWN;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return E4_VTUNKNOWN;
    }
    return storage->VertexTypeFromVertexID(i);
}

const char *
e4_NodeImpl::VertexName(int rank) const
{
    int i;

    if (storage == NULL) {
	return NULL;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return NULL;
    }
    return storage->VertexNameFromVertexID(i);
}

/*
 * Rename a vertex identified by rank.
 */

bool
e4_NodeImpl::RenameVertex(int rank, const char *newname)
{
    int i;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }

    if (!storage->RenameVertexByVertexID(i, newname)) {
	return false;
    }
    if (storage->HasCallbacks(E4_CBMODVERTEX)) {
	storage->CallbackEvent(E4_CBMODVERTEX, storage->GetVertex(i));
    }
    return true;
}

/*
 * Compute the rank of the nth vertex with a given name.
 */

int
e4_NodeImpl::VertexRank(const char *nm, int nth) const
{
    int i, rank, nameID;

    if (storage == NULL) {
	return E4_VERTEXNOTFOUND;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
        return E4_VERTEXNOTFOUND;
    }
    return rank;
}

/*
 * Get the name of the vertex containing this node in its parent node.
 */

const char *
e4_NodeImpl::GetNameInParent(int nth) const
{
    int parentID;
    int vertexID;

    if (storage == NULL) {
        return NULL;
    }
    parentID = storage->GetParentNodeID(nodeID, nth);
    if (parentID == E4_NODENOTFOUND) {
        return NULL;
    }
    vertexID = storage->GetVertexIDInParent(parentID, nodeID);
    if (vertexID == E4_VERTEXNOTFOUND) {
        return NULL;
    }
    return storage->VertexNameFromVertexID(vertexID);
}

/*
 * Get the rank of the vertex containing this node in its parent node.
 */

int
e4_NodeImpl::GetRankInParent(int nth) const
{
    int parentID, rank;
        
    if (storage == NULL) {
        return E4_NODENOTFOUND;
    }
    parentID = storage->GetParentNodeID(nodeID, nth);
    if (parentID == E4_NODENOTFOUND) {
        return E4_ROOTINDEX;
    }
    rank = storage->GetRankOfChildNode(parentID, nodeID);
    if (rank == E4_VERTEXNOTFOUND) {
	return E4_ROOTINDEX;
    }
    return rank;
}

/*
 * Delete the nth vertex with the given name in this node.
 */

bool
e4_NodeImpl::DeleteVertex(const char *nm, int nth)
{
    int nameID, rank, i;

    if (storage == NULL) {
        return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    return storage->DeleteVertexByID(i);
}

/*
 * Delete the vertex identified by the given rank.
 */

bool
e4_NodeImpl::DeleteVertexByRank(int rank)
{
    int i;

    if (storage == NULL) {
        return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return false;
    }
    return storage->DeleteVertexByID(i);
}

/*
 * Delete the first vertex whose value is the given child node.
 */

bool
e4_NodeImpl::DeleteFirstVertexWithNode(e4_NodeImpl *childImpl)
{
    
    if ((storage == NULL) || (childImpl == NULL) ||
	(childImpl->storage != storage)) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    return storage->DeleteFirstVertexWithNode(nodeID, childImpl->nodeID);
}

/*
 * Get an e4_VertexImpl for a vertex specified by rank.
 */

e4_VertexImpl *
e4_NodeImpl::GetVertexRefByRank(int rank)
{
    int i;

    if (storage == NULL) {
        return NULL;
    }
    i = storage->VertexIDFromRank(nodeID, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return NULL;
    }
    return storage->GetVertex(i);
}

/*
 * Get an e4_VertexImpl from a vertex specified by name and nth.
 */

e4_VertexImpl *
e4_NodeImpl::GetVertexRef(const char *nm, int nth)
{
    int i, rank, nameID;

    if (storage == NULL) {
        return NULL;
    }
    nameID = storage->InternName(nm);
    i = storage->VertexIDFromNthVertex(nodeID, nameID, nth, rank);
    if (i == E4_VERTEXNOTFOUND) {
	return NULL;
    }
    return storage->GetVertex(i);
}

/*
 * Move a vertex identified by a vertex ID into the rank given in this
 * node.
 */

bool
e4_NodeImpl::MoveVertex(int vertexID, e4_InsertOrder order, int rank)
{
    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }

    /*
     * Finally move the vertex.
     */

    switch (order) {
    case E4_IOFIRST:
	return storage->MoveVertex(vertexID, nodeID, 1);
    case E4_IOLAST:
	return storage->MoveVertex(vertexID, nodeID, -1);
    case E4_IOAT:
    case E4_IOBEFORE:
        return storage->MoveVertex(vertexID, nodeID, rank);
    case E4_IOAFTER:
	return storage->MoveVertex(vertexID, nodeID, rank + 1);
    default:
	return false;
    }
}

/*
 * How many vertices are currently in this node?
 */

int
e4_NodeImpl::VertexCount() const
{
    if (storage == NULL) {
	return 0;
    }
    return storage->VertexCountFromNodeID(nodeID);
}

/*
 * Set a vertex to a new node:
 */

e4_NodeImpl *
e4_NodeImpl::SetNode(const char *nm)
{
    return SetNthNode(nm, 1);
}

/*
 * Does a vertex with the given name exist?
 */

bool
e4_NodeImpl::Exists(const char *nm, int nth) const
{
    return (VertexRank(nm, nth) != E4_VERTEXNOTFOUND) ? true : false;
}

/*
 * Get the nth parent node:
 */

e4_NodeImpl *
e4_NodeImpl::GetParent(int nth) const
{
    if (storage == NULL) {
	return NULL;
    }
    return storage->GetParentNode(nodeID, nth);
}

/*
 * How many parents does this node have?
 */

int
e4_NodeImpl::ParentCount() const
{
    if (storage == NULL) {
	return E4_NODENOTFOUND;
    }
    return storage->ParentCount(nodeID);
}

/*
 * In how many vertices is this node the value?
 */

int
e4_NodeImpl::OccurrenceCount() const
{
    if (storage == NULL) {
	return E4_NODENOTFOUND;
    }
    return storage->OccurrenceCount(nodeID);
}

/*
 * How many vertices in this parent have this node as the value?
 */

int
e4_NodeImpl::OccurrenceCount(int parentID) const
{
    if (storage == NULL) {
	return E4_NODENOTFOUND;
    }
    return storage->OccurrenceCount(nodeID, parentID);
}

/*
 * What is the rank of the given parent in the list of parents
 * of this node?
 */

int
e4_NodeImpl::ParentRank(int parentID) const
{
    if (storage == NULL) {
	return E4_NODENOTFOUND;
    }
    return storage->ParentRank(nodeID, parentID);
}

/*
 * Retrieve the storage containing this node:
 */

e4_StorageImpl *
e4_NodeImpl::GetStorage() const
{
    return storage;
}

/*
 * Is this node a root? It is if it has no parents.
 */

bool
e4_NodeImpl::IsRoot() const
{
    if (storage == NULL) {
	return false;
    }
    return (storage->GetParentNodeID(nodeID, 1) == E4_NODENOTFOUND)
      ? true
      : false;
}

/*
 * Retrieve an ID which uniquely identifies this node within its
 * containing storage.
 */

int
e4_NodeImpl::GetUniqueID() const
{
    if (storage == NULL) {
	return E4_NODENOTFOUND;
    }
    return nodeID;
}

/*
 * Delete this node from its containing storage, including recursively
 * accessible nodes and vertices that become inaccessible and any
 * markers that mark this node.
 */

bool
e4_NodeImpl::Delete()
{
    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    return storage->DeleteNodeByID(nodeID);
}

/*
 * Is this node valid?
 */

bool
e4_NodeImpl::IsValid() const
{
    if (storage == NULL) {
	return false;
    }
    return storage->IsValid();
}

/*
 * Is this node marked by any markers?
 */

bool
e4_NodeImpl::IsMarked() const
{
    if (storage == NULL) {
	return false;
    }
    return storage->IsNodeMarked(nodeID);
}

/*
 * How many markers mark this node?
 */

int
e4_NodeImpl::MarkerCount() const
{
    if (storage == NULL) {
	return 0;
    }
    return storage->MarkerCount(nodeID);
}

/*
 * Retrieve the nth marker marking this node.
 */

e4_MarkerImpl *
e4_NodeImpl::GetNthMarker(int nth) const
{
    if (storage == NULL) {
	return NULL;
    }
    return storage->GetNthMarker(nth, nodeID);
}

/*
 * Delete a marker of this node given its marker ID.
 */

bool
e4_NodeImpl::DeleteMarkerID(int markerID) const
{
    e4_MarkerImpl *mip;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }

    /*
     * Get the marker impl if it is a valid marker.
     */

    mip = storage->GetMarker(markerID);
    if (mip == NULL) {
	return false;
    }

    /*
     * Only delete the marker if it is in fact marking this node.
     */

    if (mip->GetMarkedNode() != this) {
	return false;
    }
    return storage->DeleteMarkerByID(markerID);
}

/*
 * Delete the nth marker marking this node.
 */

bool
e4_NodeImpl::DeleteNthMarker(int nth) const
{
    e4_MarkerImpl *mip;

    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }

    /*
     * Get the marker impl if it is valid.
     */

    mip = GetNthMarker(nth);
    if (mip == NULL) {
	return false;
    }

    /*
     * Only delete the marker if in fact it is marking this node.
     */

    if ((const e4_NodeImpl *) mip->GetMarkedNode() != this) {
	return false;
    }
    return storage->DeleteMarkerByID(mip->GetUniqueID());
}

/*
 * Make the marker identified by the given marker ID now mark
 * this node.
 */

bool
e4_NodeImpl::AddMarker(int markerID) const
{
    if (storage == NULL) {
	return false;
    }
    if (storage->IsLocked()) {
	return false;
    }
    return storage->AddNodeMarker(markerID, nodeID);
}

/*
 * Retrieve the ID of the first and last vertex in this node.
 */

int
e4_NodeImpl::GetFirstVertexID() const
{
    if (storage == NULL) {
	return E4_VERTEXNOTFOUND;
    }
    return storage->GetFirstVertexID(nodeID);
}

int
e4_NodeImpl::GetLastVertexID() const
{
    if (storage == NULL) {
	return E4_VERTEXNOTFOUND;
    }
    return storage->GetLastVertexID(nodeID);
}

/*
 * Retrieve the node ID for use by the containing e4_StorageImpl.
 */

int
e4_NodeImpl::GetNodeID() const
{
    return nodeID;
}

/*
 * Set the storage field of this node.
 */

void
e4_NodeImpl::SetStorage(e4_StorageImpl *newstorage)
{
    storage = newstorage;
}

/*
 * Get/Set the user data associated with this node.
 */

bool
e4_NodeImpl::GetUserData(int &userData) const
{
    if (storage == NULL) {
	return false;
    }
    return storage->GetNodeUserData(nodeID, userData);
}

bool
e4_NodeImpl::SetUserData(int userData) const
{
    if (storage == NULL) {
	return false;
    }
    return storage->SetNodeUserData(nodeID, userData);
}
