/*
 * i4storage.cpp --
 *
 *	This file contains the implementation of the e4_StorageImpl class
 *	defined in e4graph.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graphimpl.h"

static bool initialized = false;
static e4_HashTable *activeStorages = NULL;

/*
 * This static function is the only public way to get an e4_StorageImpl.
 */

e4_StorageImpl *
e4_StorageImpl::GetStorage(const char *name, const char *drivername)
{
    e4_HashEntry *hPtr;
    e4_StorageImpl *s;
    e4_StorageConstructor scp;
    int isnew;

    if (!initialized) {
	initialized = true;
	e4_InitializeStorageRegistry();
	activeStorages = e4_NewHashTable(E4_STRING_KEYS);
    }
    hPtr = E4_CREATEHASHENTRY(activeStorages, name, &isnew);
    if (!isnew) {
	return (e4_StorageImpl *) E4_GETHASHVALUE(hPtr);
    }
    scp = e4_GetStorageConstructor(drivername);
    if (scp == NULL) {
	s = NULL;
    } else {
	s = (*scp)(name);
    }
    if (s == NULL) {
	e4_DeleteHashEntry(hPtr);
    } else {
	E4_SETHASHVALUE(hPtr, s);
    }
    return s;
}

/*
 * This constructor opens a storage in write mode.
 */

e4_StorageImpl::e4_StorageImpl(const char *fname, const char *dnm) 
    : e4_RefCounter()
{
    drivername = new char[strlen(dnm) + 1];
    strcpy(drivername, dnm);

    name = new char[strlen(fname) + 1];
    strcpy(name, fname);

    commit = true;
    dirty = false;
    locked = false;

    InitializeNameHash();
    InitializeNodeHash();
    InitializeVertexHash();
    InitializeMarkerHash();
    InitializeCallbackHash();
}

/*
 * The destructor is virtual so it is called before destructors of derived
 * classes to clean up the state belonging to the e4_StorageImpl base class.
 */

e4_StorageImpl::~e4_StorageImpl()
{
    if (name != NULL) {
	/*
	 * We are supposed to delete the storage occupied by the name
	 * but because of bugs in some compilers we cannot delete it.
	 * This is because attempting to delete it with e.g. VC6 causes
	 * an assert at this point. We therefore just leak the string.
	 */
    }
}
    
/*
 * Helper functions to initialize and manage the caches.
 */

void
e4_StorageImpl::InitializeNameHash()
{
    nameHash = e4_NewHashTable(E4_STRING_KEYS);
}

void
e4_StorageImpl::InitializeNodeHash()
{
    activeNodes = e4_NewHashTable(E4_ONE_WORD_KEYS);
}

void
e4_StorageImpl::InitializeMarkerHash()
{
    activeMarkers = e4_NewHashTable(E4_STRING_KEYS);
}

void
e4_StorageImpl::InitializeVertexHash()
{
    activeVertices = e4_NewHashTable(E4_ONE_WORD_KEYS);
}

void
e4_StorageImpl::InitializeCallbackHash()
{
    callbacks = e4_NewHashTable(3);
    callbacksPresent = 0;
}

int
e4_StorageImpl::GetOrCreateMarkerID(const char *nm, bool &created)
{
    int markerID;
    int nameID;

    if ((nm == NULL) || (*nm == '\0')) {
	return E4_MARKERNOTFOUND;
    }
    nameID = InternName(nm);
    markerID = GetMarkerIDFromNameID(nameID);
    if (markerID == E4_MARKERNOTFOUND) {
	if (locked) {
	    return E4_MARKERNOTFOUND;
	}
	markerID = CreateMarkerIDFromNameID(nameID);
	created = true;
    } else {
	created = false;
    }
    return markerID;
}

/*
 * Retrieve an instance of e4_VertexImpl for a given vertex.
 */

e4_VertexImpl *
e4_StorageImpl::GetVertex(int i) const
{
    e4_VertexImpl *f;
    e4_HashEntry *hPtr;
    int isNew;

    if (!IsLegalVertexID(i)) {
	return NULL;
    }
    hPtr = E4_CREATEHASHENTRY(activeVertices, (const char *) i, &isNew);
    if (isNew) {
        f = new e4_VertexImpl((e4_StorageImpl *) this, i);
        E4_SETHASHVALUE(hPtr, f);
    } else {
        f = (e4_VertexImpl *) E4_GETHASHVALUE(hPtr);
    }
    return f;
}

/*
 * Create a marker from a storage.
 */

e4_MarkerImpl *
e4_StorageImpl::GetMarker(const char *markerName) const
{
    int markerID;
    e4_HashEntry *hPtr;
    e4_MarkerImpl *t;
    bool created;
    e4_StorageImpl *ncthis = (e4_StorageImpl *) this;

    /*
     * First check if the requested marker is in-memory; if so, return it.
     */

    hPtr = E4_FINDHASHENTRY(activeMarkers, markerName);
    if (hPtr != NULL) {
        return (e4_MarkerImpl *) E4_GETHASHVALUE(hPtr);
    }

    /*
     * The requested marker is not in-memory yet, so load it, store it in
     * the activeMarkers table and return a pointer.
     */

    markerID =
      ((e4_StorageImpl *) this)->GetOrCreateMarkerID(markerName, created);
    if (markerID == E4_MARKERNOTFOUND) {
	return NULL;
    }

    /*
     * A callback event occurs if the marker was created in this operation.
     */

    if (created) {
	if (ncthis->HasCallbacks(E4_CBADDMARKER)) {
	    ncthis->CallbackEvent(E4_CBADDMARKER, GetMarker(markerID));
	}
	if (ncthis->HasCallbacks(E4_CBADDNODE)) {
	    ncthis->CallbackEvent(E4_CBADDNODE, GetMarkedNode(markerID));
	}      
    }

    /*
     * Create the marker.
     */

    t = new e4_MarkerImpl((e4_StorageImpl *) this, markerID);
    ((e4_StorageImpl *) this)->StoreMarker(markerName, t);

    return t;
}

/*
 * Get an instance of e4_MarkerImpl for a marker given a marker ID.
 */

e4_MarkerImpl *
e4_StorageImpl::GetMarker(int markerID) const
{
    e4_HashEntry *hPtr;
    const char *nm;
    e4_MarkerImpl *t;

    if (!IsLegalMarkerID(markerID)) {
        return NULL;
    }

    /*
     * If the marker is already in-memory, return a pointer to it from the
     * activeMarkers hash table.
     */

    nm = ((e4_StorageImpl *) this)->GetMarkerName(markerID);
    hPtr = E4_FINDHASHENTRY(activeMarkers, nm);
    if (hPtr != NULL) {
	return (e4_MarkerImpl *) E4_GETHASHVALUE(hPtr);
    }

    /*
     * The marker is not in-memory, so load it, store it in the activeMarkers
     * hash table and return a pointer to it.
     */

    t = new e4_MarkerImpl((e4_StorageImpl *) this, markerID);
    ((e4_StorageImpl *) this)->StoreMarker(nm, t);
    return t;
}

void
e4_StorageImpl::ForgetMarker(int markerID)
{
    e4_HashEntry *hPtr;
    const char *nm;
    e4_MarkerImpl *t;

    if (!IsLegalMarkerID(markerID)) {
	return;
    }
    nm = GetMarkerName(markerID);
    hPtr = E4_FINDHASHENTRY(activeMarkers, nm);
    if (hPtr != NULL) {
        t = (e4_MarkerImpl *) E4_GETHASHVALUE(hPtr);
        t->SetStorage(NULL);
	e4_DeleteHashEntry(hPtr);
    }
}

void
e4_StorageImpl::StoreMarker(const char *nm, e4_MarkerImpl *marker) 
{
    e4_HashEntry *hPtr;
    int isNew;

    hPtr = E4_CREATEHASHENTRY(activeMarkers, nm, &isNew);
    if (!isNew) {
        fprintf(stderr, "WHOA: Active marker \"%s\" stored more than once!\n",
                nm);
    }
    E4_SETHASHVALUE(hPtr, marker);
}

/*
 * Store a node in the activeNodes hash table.
 */

void
e4_StorageImpl::StoreNode(int index, e4_NodeImpl *n) const
{
    e4_HashEntry *hPtr;
    int isnew;

    hPtr = E4_CREATEHASHENTRY(activeNodes, (const char *) index, &isnew);
    E4_SETHASHVALUE(hPtr, n);
}

/*
 * Remove a node from the activeNodes hash table. This method should only
 * be called from the e4_NodeImpl destructor, to avoid infinite recursion.
 */

void
e4_StorageImpl::ForgetNode(int index)
{
    e4_HashEntry *hPtr;
    e4_NodeImpl *n;

    hPtr = E4_FINDHASHENTRY(activeNodes, (const char *) index);
    if (hPtr != NULL) {
        n = (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
        n->SetStorage(NULL);
        e4_DeleteHashEntry(hPtr);
    }
}

/*
 * Find a node in the activeNodes hash table given its index, if it is
 * in-memory. If not, check whether the index is legal and create it
 * if it is.
 */

e4_NodeImpl *
e4_StorageImpl::FindNode(int index) const
{
    if (IsLegalNodeID(index)) {
	return FindOrCreateNode(index);
    }
    return NULL;
}

/*
 * Find a node in the activeNodes hash table given its index, if it is
 * in-memory. If not, create it. This method does not check whether
 * the index is legal.
 */

e4_NodeImpl *
e4_StorageImpl::FindOrCreateNode(int index) const
{
    e4_HashEntry *hPtr;
    e4_NodeImpl *n;

    hPtr = E4_FINDHASHENTRY(activeNodes, (const char *) index);
    if (hPtr != NULL) {
	return (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
    }
    n = new e4_NodeImpl((e4_StorageImpl *) this, index);
    StoreNode(index, n);
    
    return n;
}

/*
 * Find a node in the activeNodes hash table given its index, if it is
 * in-memory. If not, create it and store it in the activeNodes table.
 * Also return an indication whether the node was newly created.
 */

e4_NodeImpl *
e4_StorageImpl::FindOrCreateNode(int index, bool *isNew) const
{
    e4_HashEntry *hPtr;
    e4_NodeImpl *n;

    hPtr = E4_FINDHASHENTRY(activeNodes, (const char *) index);
    if (hPtr != NULL) {
        *isNew = false;
	return (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
    }
    *isNew = true;
    n = new e4_NodeImpl((e4_StorageImpl *) this, index);
    StoreNode(index, n);
    
    return n;
}

/*
 * Store a vertex in the activeVertices hash table.
 */

void
e4_StorageImpl::StoreVertex(int vertexID, e4_VertexImpl *f) const
{
    e4_HashEntry *hPtr;
    int isNew;

    hPtr = E4_CREATEHASHENTRY(activeVertices, (const char *) vertexID, &isNew);
    if (!isNew) {

        /*
         * Should abort here.
         */
    }
    E4_SETHASHVALUE(hPtr, f);
}

/*
 * Remove an e4_VertexImpl from the activeVertices hash table.
 */

void
e4_StorageImpl::ForgetVertex(int vertexID)
{
    e4_HashEntry *hPtr;
    e4_VertexImpl *f;

    hPtr = E4_FINDHASHENTRY(activeVertices, (const char *) vertexID);
    if (hPtr != NULL) {
	f = (e4_VertexImpl *) E4_GETHASHVALUE(hPtr);
	f->SetStorage(NULL);
        e4_DeleteHashEntry(hPtr);
    }
}

/*
 * Add a name to the name hash table.
 */

void
e4_StorageImpl::AddNameToNameHash(const char *nm, int idx) const
{
    e4_HashEntry *hPtr;
    int isnew;

    hPtr = E4_CREATEHASHENTRY(nameHash, nm, &isnew);
    E4_SETHASHVALUE(hPtr, idx);
}

/*
 * Return a name given its nameID.
 */

const char *
e4_StorageImpl::NameFromNameID(int nameID) const
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;

    /*
     * Search through the hash table until we find the hash entry with
     * the nameID as its value. If found, return the key which is the
     * name that hashes to this nameID.
     */

    for (hPtr = e4_FirstHashEntry(nameHash, &search);
	 hPtr != NULL;
	 hPtr = e4_NextHashEntry(&search)) {
	if ((int) E4_GETHASHVALUE(hPtr) == nameID) {
	    return (const char *) E4_GETHASHKEY(nameHash, hPtr);
	}
    }

    /*
     * Not found, return NULL.
     */

    return NULL;
}

/*
 * Add a name to the persistent storage and also intern it in the
 * hash table.
 */

int
e4_StorageImpl::InternName(const char *nm) const
{
    e4_HashEntry *hPtr;
    int isnew, idx;
    
    /*
     * See if we already have this name in the hash table; if so, just
     * return the hash value.
     */

    hPtr = E4_FINDHASHENTRY(nameHash, nm);
    if (hPtr != NULL) {
	return (int) E4_GETHASHVALUE(hPtr);
    }

    /*
     * No, we need to add the name to hash table.
     */

    idx = ((e4_StorageImpl *) this)->AddName(nm);
    hPtr = E4_CREATEHASHENTRY(nameHash, nm, &isnew);
    E4_SETHASHVALUE(hPtr, idx);

    return idx;
}

/*
 * This method is called to clean up all references to this e4_StorageImpl
 * from other instances.
 */

void
e4_StorageImpl::CleanUp()
{
    /*
     * If commit is needed and requested, do it now before the state
     * is destroyed.
     */
    
    if (AutoCommit() && IsDirty()) {
	Commit();
    }

    HashCleanup();
}

/*
 * Helper method to clean up the hash tables for various exported entities.
 */

void
e4_StorageImpl::HashCleanup()
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;
    e4_MarkerImpl *marker;
    e4_NodeImpl *node;
    e4_VertexImpl *vertex;

    /*
     * Clean up my own state, consisting of the hash tables for
     * caching in-core instances of markers, nodes and vertices.
     */
    
    for (hPtr = e4_FirstHashEntry(activeMarkers, &search);
         hPtr != NULL;
         hPtr = e4_FirstHashEntry(activeMarkers, &search)) {
        marker = (e4_MarkerImpl *) E4_GETHASHVALUE(hPtr);
        marker->SetStorage(NULL);
        e4_DeleteHashEntry(hPtr);
    }
    for (hPtr = e4_FirstHashEntry(activeNodes, &search);
         hPtr != NULL;
         hPtr = e4_FirstHashEntry(activeNodes, &search)) {
        node = (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
        node->SetStorage(NULL);
        e4_DeleteHashEntry(hPtr);
    }
    for (hPtr = e4_FirstHashEntry(activeVertices, &search);
         hPtr != NULL;
         hPtr = e4_FirstHashEntry(activeVertices, &search)) {
        vertex = (e4_VertexImpl *) E4_GETHASHVALUE(hPtr);
        vertex->SetStorage(NULL);
        e4_DeleteHashEntry(hPtr);
    }
    for (hPtr = e4_FirstHashEntry(nameHash, &search);
         hPtr != NULL;
         hPtr = e4_FirstHashEntry(nameHash, &search)) {
        e4_DeleteHashEntry(hPtr);
    }
}

/*
 * This method is called when the reference count goes to (or below) zero.
 */

void
e4_StorageImpl::NotReferenced()
{
    e4_HashEntry *hPtr;

    CleanUp();

    /*
     * Delete the hash tables associated with this instance so that
     * we do not leak memory.
     */

    e4_DeleteHashTable(activeMarkers);
    e4_DeleteHashTable(activeNodes);
    e4_DeleteHashTable(activeVertices);
    e4_DeleteHashTable(nameHash);
    e4_DeleteHashTable(callbacks);

    /*
     * Remove this e4_StorageImpl from the global cache of 
     * e4_StorageImpl instances.
     */
    
    hPtr = E4_FINDHASHENTRY(activeStorages, name);
    if (hPtr == NULL) {
	return;
    }
    e4_DeleteHashEntry(hPtr);

    if (name != NULL) {
	delete[] name;
    }
    if (drivername != NULL) {
	delete[] drivername;
    }

    /*
     * Call the destructor which is most likely going to be overridden by
     * the derived class.
     */
    
    delete this;
}

/*
 * Retrieve the first e4_StorageImpl.
 */

e4_StorageImpl *
e4_StorageImpl::GetFirstStorageImpl()
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;

    if (activeStorages == NULL) {
	return NULL;
    }
    hPtr = e4_FirstHashEntry(activeStorages, &search);
    if (hPtr == NULL) {
	return NULL;
    }
    return (e4_StorageImpl *) E4_GETHASHVALUE(hPtr);
}

/*
 * Retrieve the next e4_StorageImpl after a given one.
 */

e4_StorageImpl *
e4_StorageImpl::GetNextStorageImpl(e4_StorageImpl *cur)
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;

    if (cur == NULL) {
	return NULL;
    }
    if (activeStorages == NULL) {
	return NULL;
    }
    for (hPtr = e4_FirstHashEntry(activeStorages, &search);
	 hPtr != NULL;
	 hPtr = e4_NextHashEntry(&search)) {
	if (cur == (e4_StorageImpl *) E4_GETHASHVALUE(hPtr)) {
	    hPtr = e4_NextHashEntry(&search);
	    if (hPtr == NULL) {
		return NULL;
	    }
	    return (e4_StorageImpl *) E4_GETHASHVALUE(hPtr);
	}
    }

    /*
     * Current one not found, return NULL.
     */

    return NULL;
}

/*
 * Add a callback.
 */

void
e4_StorageImpl::AddCallback(e4_CallbackKind kind, void *fn, void *clientData)
{
    int key[3];
    e4_HashEntry *ePtr;
    int isnew;

    key[0] = (int) kind;
    key[1] = (int) fn;
    key[2] = (int) clientData;

    ePtr = E4_CREATEHASHENTRY(callbacks, (char *) key, &isnew);
    E4_SETHASHVALUE(ePtr, (void *) NULL);

    callbacksPresent |= (1 << (int) kind);
}

/*
 * Remove a callback.
 */

void
e4_StorageImpl::DelCallback(e4_CallbackKind kind, void *fn, void *clientData)
{
    int key[3], *keyPtr;
    e4_HashEntry *ePtr;
    e4_HashSearch search;

    key[0] = (int) kind;
    key[1] = (int) fn;
    key[2] = (int) clientData;

    ePtr = E4_FINDHASHENTRY(callbacks, (char *) key);
    if (ePtr != NULL) {
	e4_DeleteHashEntry(ePtr);
    }

    /*
     * Now see if there are any other callbacks of this kind left. If not,
     * then remove the notation that this kind of callback is present for
     * this storage.
     */

    for (ePtr = e4_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = e4_NextHashEntry(&search)) {
	keyPtr = (int *) E4_GETHASHKEY(callbacks, ePtr);
	if (keyPtr[0] == (int) kind) {
	    return;
	}
    }

    /*
     * Did not find any, so remove the notation.
     */

    callbacksPresent &=(~(1 << (int) kind));
}

/*
 * Cause callbacks to be called for a specified kind of event on the
 * given reference.
 */

void
e4_StorageImpl::CallbackEvent(e4_CallbackKind kind, e4_RefCounter *ref)
{
    e4_MarkerAddCallbackFunction macbfn;
    e4_MarkerDelCallbackFunction mdcbfn;
    e4_NodeAddCallbackFunction nacbfn;
    e4_NodeDelCallbackFunction ndcbfn;
    e4_VertexAddCallbackFunction vacbfn;
    e4_VertexDelCallbackFunction vdcbfn;
    e4_VertexModCallbackFunction vmcbfn;
    e4_HashEntry *ePtr;
    e4_HashSearch search;
    int *key;
    bool oldlocking;

    /*
     * Lock the storage so no changes are allowed.
     */

    oldlocking = SetLocking(true);

    switch (kind) {
    case E4_CBADDMARKER:
	{
	    e4_Marker m1((e4_MarkerImpl *) ref);

	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBADDMARKER) {
		    macbfn = (e4_MarkerAddCallbackFunction) key[1];
		    (macbfn)((void *) key[2], m1);
		}
	    }
	}
	break;
    case E4_CBDELMARKER:
	{
	    e4_Marker m2((e4_MarkerImpl *) ref);

	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBDELMARKER) {
		    mdcbfn = (e4_MarkerDelCallbackFunction) key[1];
		    (mdcbfn)((void *) key[2], m2);
		}
	    }
	}
	break;
    case E4_CBADDNODE:
	{
	    e4_Node n1((e4_NodeImpl *) ref);

	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBADDNODE) {
		    nacbfn = (e4_NodeAddCallbackFunction) key[1];
		    (nacbfn)((void *) key[2], n1);
		}
	    }
	}
	break;
    case E4_CBDELNODE:
	{
	    e4_Node n2((e4_NodeImpl *) ref);

	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBDELNODE) {
		    ndcbfn = (e4_NodeDelCallbackFunction) key[1];
		    (ndcbfn)((void *) key[2], n2);
		}
	    }
	}
	break;
    case E4_CBADDVERTEX:
	{
	    e4_Vertex v1((e4_VertexImpl *) ref);

	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBADDVERTEX) {
		    vacbfn = (e4_VertexAddCallbackFunction) key[1];
		    (vacbfn)((void *) key[2], v1);
		}
	    }
	}
	break;
    case E4_CBDELVERTEX:
	{
	    e4_Vertex v2((e4_VertexImpl *) ref);

	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBDELVERTEX) {
		    vdcbfn = (e4_VertexDelCallbackFunction) key[1];
		    (vdcbfn)((void *) key[2], v2);
		}
	    }
	}
	break;
    case E4_CBMODVERTEX:
        {
	    e4_Vertex v2((e4_VertexImpl *) ref);

	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBMODVERTEX) {
		    vmcbfn = (e4_VertexModCallbackFunction) key[1];
		    (vmcbfn)((void *) key[2], v2);
		}
	    }
	}
	break;
    }

    (void) SetLocking(oldlocking);
}

/*
 * Make this storage empty.
 */

bool
e4_StorageImpl::MakeEmpty()
{
    e4_VertexImpl *vip;
    e4_NodeImpl *nip;
    e4_MarkerImpl *mip;
    int id;

    /*
     * If the storage is locked, do nothing.
     */

    if (locked) {
	return false;
    }

    /*
     * Cause callback events to happen for the deletion of all entities:
     */

    if (HasCallbacks(E4_CBDELMARKER)) {
	for (mip = FindNextMarker(E4_MARKERNOTFOUND);
	     mip != NULL;
	     mip = FindNextMarker(id)) {
	    id = mip->GetUniqueID();
	    CallbackEvent(E4_CBDELMARKER, mip);
	}
    }
    if (HasCallbacks(E4_CBDELNODE)) {
	for (nip = FindNextNode(E4_NODENOTFOUND);
	     nip != NULL;
	     nip = FindNextNode(id)) {
	    id = nip->GetUniqueID();
	    CallbackEvent(E4_CBDELNODE, nip);
	}
    }
    if (HasCallbacks(E4_CBDELVERTEX)) {
	for (vip = FindNextVertex(E4_VERTEXNOTFOUND, E4_VMSTORAGE, E4_VFNONE,
				  E4_VERTEXNOTFOUND, E4_NODENOTFOUND,
				  E4_VTUNKNOWN);
	     vip != NULL;
	     vip = FindNextVertex(id, E4_VMSTORAGE, E4_VFNONE,
				  E4_VERTEXNOTFOUND, E4_NODENOTFOUND,
				  E4_VTUNKNOWN)) {
	    id = vip->GetUniqueID();
	    CallbackEvent(E4_CBDELVERTEX, vip);
	}
    }

    /*
     * Use storage driver specific method to empty the storage:
     */

    return RemoveEverything();
}

/*
 * Event firing mechanism for deletion callbacks:
 */

void
e4_StorageImpl::FireDeletionEvents()
{
    int id;

    for (id = NextDeletedMarker(E4_NEXTNONE); 
	 id != E4_NEXTNONE; 
	 id = NextDeletedMarker(id)) {
	CallbackEvent(E4_CBDELMARKER, GetMarker(id));
    }
    for (id = NextDeletedNode(E4_NEXTNONE);
	 id != E4_NEXTNONE;
	 id = NextDeletedNode(id)) {
	CallbackEvent(E4_CBDELNODE, FindOrCreateNode(id));
    }
    for (id = NextDeletedVertex(E4_NEXTNONE);
	 id != E4_NEXTNONE;
	 id = NextDeletedVertex(id)) {
	CallbackEvent(E4_CBDELVERTEX, GetVertex(id));
    }
}

/*
 * Lock or unlock the storage.
 */

bool
e4_StorageImpl::SetLocking(bool lockvalue) 
{
    bool oldvalue = locked;

    locked = lockvalue;
    return oldvalue;
}

