/*
 * i4vertex.cpp --
 *
 *	This file provides the implementation of the e4_VertexImpl 
 *	class defined in e4graphimpl.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graphimpl.h"

/*
 * Constructor:
 */

e4_VertexImpl::e4_VertexImpl(e4_StorageImpl *st, int i)
  : e4_RefCounter(),
    vertexID(i),
    s(st)
{
}

/*
 * Destructor:
 */

e4_VertexImpl::~e4_VertexImpl()
{
    if (s != NULL) {
	s->ForgetVertex(vertexID);
    }
}

/*
 * This method is called when the reference count goes to (or below) zero.
 */

void
e4_VertexImpl::NotReferenced()
{
    delete this;
}

/*
 * Set the value of this vertex to a new node.
 */

e4_NodeImpl *
e4_VertexImpl::SetNode() const
{
    int nodeID;
    e4_NodeImpl *nip;

    if (s == NULL) {
	return NULL;
    }
    if (s->IsLocked()) {
	return NULL;
    }
    nodeID = s->ReserveNodeID();
    nip = s->FindOrCreateNode(nodeID);
    s->AddParent(nodeID, s->NodeIDFromVertexID(vertexID));
    s->SetVertex(vertexID, s->NameIDFromVertexID(vertexID), E4_VTNODE, nodeID);

    if (s->HasCallbacks(E4_CBADDNODE)) {
	s->CallbackEvent(E4_CBADDNODE, s->FindOrCreateNode(nodeID));
    }
    if (s->HasCallbacks(E4_CBMODVERTEX)) {
	s->CallbackEvent(E4_CBMODVERTEX, (e4_RefCounter *) this);
    }

    return nip;
}

/*
 * Delete this vertex.
 */

bool
e4_VertexImpl::Delete() 
{
    if (s == NULL) {
	return false;
    }
    if (s->IsLocked()) {
	return false;
    }
    return s->DeleteVertexByID(vertexID);
}

/*
 * Set a vertex to contain an existing node.
 */

bool
e4_VertexImpl::SetToNode(int childID) 
{
    int nodeID;
    e4_NodeImpl *nip;

    if (s == NULL) {
	return false;
    }
    if (s->IsLocked()) {
	return false;
    }

    /*
     * Only assign if the node is not already the value of this
     * vertex!!!
     */

    if ((!Get(nip)) || (nip->GetUniqueID() != childID)) {
	nodeID = s->NodeIDFromVertexID(vertexID);
	if (nodeID == E4_NODENOTFOUND) {
	    return false;
	}
	s->AddParent(childID, nodeID);
	if (!s->SetVertexByIndexToNode(vertexID, childID)) {
	    return false;
	}
    }

    /*
     * Fire the vertex modify event (even in the case where the
     * node is already the value of this vertex).
     */

    if (s->HasCallbacks(E4_CBMODVERTEX)) {
	s->CallbackEvent(E4_CBMODVERTEX, this);
    }
    return true;
}

/*
 * Get/Set user data associated with this vertex.
 */

bool
e4_VertexImpl::GetUserData(int &userData) const
{
    if (s == NULL) {
	return false;
    }
    return s->GetVertexUserData(vertexID, userData);
}

bool
e4_VertexImpl::SetUserData(int userData) const
{
    if (s == NULL) {
	return false;
    }
    return s->SetVertexUserData(vertexID, userData);
}
