/*
 * e4marker.cpp --
 *
 *	Implementation of the e4_Marker class defined in e4graph.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graph.h"
#include "e4graphimpl.h"

/*
 * This instance of e4_Marker can be used (by assigning it to fields or
 * variables of type e4_Marker) to remove remaining references to instances
 * that you want to discard. Especially useful for removing remaining
 * references in malloc'ed memory.
 */

e4_Marker const invalidMarker;

/*
 * Default constructor:
 */

e4_Marker::e4_Marker() : e4_RefCount() {}

/*
 * Constructor which assigns a value to the the implementation pointer.
 */

e4_Marker::e4_Marker(e4_MarkerImpl *ip) : e4_RefCount(ip) {}

/*
 * Copying constructor:
 */

e4_Marker::e4_Marker(const e4_Marker &referrer) : e4_RefCount(referrer) {}

/*
 * Assignment operator:
 */

e4_Marker & e4_Marker::operator=(const e4_Marker &referrer)
{
    return (e4_Marker &) e4_RefCount::operator=(referrer);
}

/*
 * Get the name of this marker.
 *
 * NOTE: The caller must copy the returned value if she wants to hold on to
 * it beyond the next call to any e4Graph API.
 */

const char *
e4_Marker::GetName() const
{
    if (impl == NULL) {
	return NULL;
    }
    return ((e4_MarkerImpl *) impl)->GetName();
}

/*
 * Get the e4_Storage for this marker.
 */

bool
e4_Marker::GetStorage(e4_Storage &s) const
{
    e4_StorageImpl *ssip;

    if (impl == NULL) {
	return false;
    }
    ssip = ((e4_MarkerImpl *) impl)->GetStorage();
    if (ssip == NULL) {
	return false;
    }

    e4_Storage ss(ssip);

    s = ss;

    return true;
}

/*
 * Get the root node for this marker.
 */

bool
e4_Marker::GetMarkedNode(e4_Node &n) const
{
    e4_NodeImpl *nnip;

    if (impl == NULL) {
	return false;
    }
    nnip = ((e4_MarkerImpl *) impl)->GetMarkedNode();
    if (nnip == NULL) {
	return false;
    }

    e4_Node nn(nnip);

    n = nn;

    return true;
}

/*
 * Set the marked node of this marker to the supplied node.
 */

bool
e4_Marker::SetMarkedNode(e4_Node n) const
{
    return n.MarkWith(*this);
}

/*
 * Get a unique identifier (within this storage) for this marker.
 */

bool
e4_Marker::GetUniqueID(e4_MarkerUniqueID &m) const
{
    int id;
    e4_StorageImpl *sp;

    if (impl == NULL) {
	return false;
    }
    id = ((e4_MarkerImpl *) impl)->GetUniqueID();
    if (id == E4_MARKERNOTFOUND) {
	return false;
    }
    sp = ((e4_MarkerImpl *) impl)->GetStorage();
    if (sp == NULL) {
	return false;
    }

    e4_MarkerUniqueID mm(id, sp->HashCode());

    m = mm;

    return true;
}

/*
 * Get the unique ID and return it unwrapped.
 */

int
e4_Marker::GetRawUniqueID() const
{
    if (impl == NULL) {
	return E4_MARKERNOTFOUND;
    }
    return ((e4_MarkerImpl *) impl)->GetUniqueID();
}

/*
 * Delete the underlying storage occuppied by this marker.
 */

bool
e4_Marker::Delete() const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_MarkerImpl *) impl)->Delete();
}

/*
 * Return the kind identifier for this type of instance.
 */

e4_RefKind
e4_Marker::Kind() const
{
    return E4_RKMARKER;
}

/*
 * Get/Set user data associated with this marker.
 */

bool
e4_Marker::GetUserData(int &userData) const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_MarkerImpl *) impl)->GetUserData(userData);
}

bool
e4_Marker::SetUserData(int userData) const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_MarkerImpl *) impl)->SetUserData(userData);
}
