/*
 * e4mvisitor.cpp --
 *
 *	Implementation of the e4_MarkerVisitor class defined in e4graph.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graph.h"
#include "e4graphimpl.h"

/*
 * Default constructor:
 */

e4_MarkerVisitor::e4_MarkerVisitor()
    : done(true)
{
}

/*
 * Copying constructor:
 */

e4_MarkerVisitor::e4_MarkerVisitor(const e4_MarkerVisitor &referrer)
    : done(referrer.done),
      s(referrer.s),
      t(referrer.t)
{
}

/*
 * Assignment operator:
 */

e4_MarkerVisitor &
e4_MarkerVisitor::operator=(const e4_MarkerVisitor &referrer)
{
    done = referrer.done;
    t = referrer.t;
    s = referrer.s;

    return * this;
}

/*
 * Comparison operators:
 */

bool
e4_MarkerVisitor::operator==(const e4_MarkerVisitor &compared) const
{
    if ((done == compared.done) &&
	(t == compared.t) &&
	(s == compared.s)) {
	return true;
    }
    return false;
}

bool
e4_MarkerVisitor::operator!=(const e4_MarkerVisitor &compared) const
{
    if ((done == compared.done) &&
	(t == compared.t) &&
	(s == compared.s)) {
	return false;
    }
    return true;
}

/*
 * Constructors that initialize the visitor from a storage,
 * marker, node or vertex.
 */

e4_MarkerVisitor::e4_MarkerVisitor(const e4_Storage &ss)
{
    SetStorage(ss);
}

e4_MarkerVisitor::e4_MarkerVisitor(const e4_Marker &tt)
{
    SetMarker(tt);
}

e4_MarkerVisitor::e4_MarkerVisitor(const e4_Node &nn)
{
    SetNode(nn);
}

e4_MarkerVisitor::e4_MarkerVisitor(const e4_Vertex &ff)
{
    SetVertex(ff);
}

/*
 * Destructor:
 */

e4_MarkerVisitor::~e4_MarkerVisitor()
{
}

/*
 * Reset the visitor from a storage, marker, node or vertex.
 */

bool
e4_MarkerVisitor::SetStorage(const e4_Storage &ss)
{
    if (!ss.IsValid()) {
	return false;
    }

    s = ss;
    done = (!s.FindNextMarker(E4_MARKERNOTFOUND, t));
    return true;
}

bool
e4_MarkerVisitor::SetMarker(const e4_Marker &tt)
{
    if (!tt.IsValid()) {
	return false;
    }
    done = (!tt.GetStorage(s));
    if (done) {
	t = invalidMarker;
    } else {
	done = (!s.FindNextMarker(E4_MARKERNOTFOUND, t));
    }
    return true;
}

bool
e4_MarkerVisitor::SetNode(const e4_Node &nn)
{
    if (!nn.IsValid()) {
	return false;
    }
    done = (!nn.GetStorage(s));
    if (done) {
	t = invalidMarker;
    } else {
	done = (!s.FindNextMarker(E4_MARKERNOTFOUND, t));
    }
    return true;
}

bool
e4_MarkerVisitor::SetVertex(const e4_Vertex &ff)
{
    if (!ff.IsValid()) {
	return false;
    }
    done = (!ff.GetStorage(s));
    if (done) {
	t = invalidMarker;
    } else {
	done = (!s.FindNextMarker(E4_MARKERNOTFOUND, t));
    }
    return true;
}

/*
 * Get the current marker.
 */

bool
e4_MarkerVisitor::CurrentMarker(e4_Marker &tt)
{
    if (done) {
	return false;
    }
    tt = t;
    return true;
}

/*
 * Get the current marker and advance to the next one.
 */

bool
e4_MarkerVisitor::CurrentMarkerAndAdvance(e4_Marker &tt)
{
    e4_Marker dm;

    if (!CurrentMarker(tt)) {
	return false;
    }
    (void) NextMarker(dm);

    return true;
}

/*
 * Get the next marker.
 */

bool
e4_MarkerVisitor::NextMarker(e4_Marker &tt)
{
    if (done) {
	return false;
    }
    if (!s.FindNextMarker(t.GetRawUniqueID(), t)) {
	done = true;
	return false;
    }
    tt = t;
    return true;
}

/*
 * Is this instance done?
 */

bool
e4_MarkerVisitor::IsDone()
{
    if (done) {
	return true;
    }
    if (!s.IsValid() || !t.IsValid()) {
	done = true;
    }
    return done;
}

/*
 * Is this instance valid?
 */

bool
e4_MarkerVisitor::IsValid()
{
    if (!s.IsValid()) {
	done = true;
	return false;
    }
    return true;
}
