/*
 * e4storage.cpp --
 *
 *	Implementation of the e4_Storage class defined in e4graph.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graph.h"
#include "e4graphimpl.h"

/*
 * This instance of e4_Storage can be used (by assigning it to fields or
 * variables of type e4_Storage) to remove remaining references to instances
 * that you want to discard. Especially useful for removing remaining
 * references in malloc'ed memory.
 */

e4_Storage const invalidStorage;

/*
 * Default constructor:
 */

e4_Storage::e4_Storage() : e4_RefCount() {}

/*
 * Constructor which assigns a value to the the implementation pointer.
 */

e4_Storage::e4_Storage(e4_StorageImpl *ip) : e4_RefCount(ip) {}

/*
 * Copying constructor:
 */

e4_Storage::e4_Storage(const e4_Storage &referrer) : e4_RefCount(referrer) {}

/*
 * Assignment operator:
 */

e4_Storage & e4_Storage::operator=(const e4_Storage &referrer)
{
    return (e4_Storage &) e4_RefCount::operator=(referrer);
}

/*
 * Constructor that chooses a concrete representation for the underlying
 * implementation.
 */

e4_Storage::e4_Storage(const char *nm, const char *drivername)
    : e4_RefCount()
{
    impl = e4_StorageImpl::GetStorage(nm, drivername);
    if (impl != NULL) {
	impl->IncrRefCount();
    }
}

/*
 * Turn auto-commit on or off.
 */

void
e4_Storage::AutoCommit(bool on) const
{
    if (impl != NULL) {
	((e4_StorageImpl *) impl)->AutoCommit(on);
    }
}

/*
 * Is auto-commit turned on?
 */

bool
e4_Storage::AutoCommit() const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->AutoCommit();
}

/*
 * Commit the storage now.
 */

bool
e4_Storage::Commit() const
{
    if (impl != NULL) {
	return ((e4_StorageImpl *) impl)->Commit();
    }
    return false;
}

/*
 * Copy this storage to another storage.
 */

bool
e4_Storage::CopyTo(e4_Storage otherStorage, bool forceCommit) const
{
    if (!IsValid() || !otherStorage.IsValid()) {
	return false;
    }
    if (!((e4_StorageImpl *) impl)->
		CopyTo((e4_StorageImpl *) otherStorage.impl)) {
	return false;
    }
    if (forceCommit) {
	return otherStorage.Commit();
    }
    return true;
}

/*
 * Destroy this storage. If auto-commit is set, first commit.
 */

bool
e4_Storage::Delete()
{
    if (impl != NULL) {
	if (((e4_StorageImpl *) impl)->AutoCommit()) {
	    ((e4_StorageImpl *) impl)->Commit();
	}
	((e4_StorageImpl *) impl)->Destroy();
	impl->DecrRefCount();
    }
    impl = NULL;

    return true;
}
    
/*
 * How many markers are there in this storage?
 */

int
e4_Storage::MarkerCount() const
{
    if (impl == NULL) {
	return 0;
    }
    return ((e4_StorageImpl *) impl)->MarkerCount();
}

/*
 * Get a marker out of this storage.
 */

bool
e4_Storage::GetMarker(const char *nm, e4_Marker &t) const
{
    e4_MarkerImpl *ttip;

    if (impl == NULL) {
	return false;
    }
    ttip = ((e4_StorageImpl *) impl)->GetMarker(nm);
    if (ttip == NULL) {
	return false;
    }

    e4_Marker tt(ttip);

    t = tt;

    return true;
}

/*
 * Obtain the node marked by the named marker. Create the marker if it
 * does not exist.
 */

bool
e4_Storage::GetMarkedNode(const char *nm, e4_Node &n) const
{
    e4_Marker m;

    if (!GetMarker(nm, m)) {
	return false;
    }
    if (!m.GetMarkedNode(n)) {
	return false;
    }
    return true;
}

/*
 * Does a marker with the give name exist in this storage?
 */

bool
e4_Storage::MarkerExists(const char *nm)
{
   int nameID;
   int markerID;

   if (impl == NULL) {
	return false;
   }
   nameID = ((e4_StorageImpl *) impl)->InternName(nm);
   markerID = ((e4_StorageImpl *) impl)->GetMarkerIDFromNameID(nameID);
   if (markerID == E4_MARKERNOTFOUND) {
	return false;
   }
   return true;
}

/*
 * Given a unique ID, obtain the referenced marker.
 */

bool
e4_Storage::GetMarkerFromID(e4_MarkerUniqueID mid, e4_Marker &m) const
{
    e4_MarkerImpl *mmip;

    if (impl == NULL) {
	return false;
    }
    mmip = ((e4_StorageImpl *) impl)->GetMarker(mid.GetUniqueID());
    if (mmip == NULL) {
	return false;
    }
    e4_Marker mm(mmip);

    m = mm;

    return true;
}

/*
 * Given a unique ID, obtain the referenced node.
 */

bool
e4_Storage::GetNodeFromID(e4_NodeUniqueID nid, e4_Node &n) const
{
    e4_NodeImpl *nnip;

    if (impl == NULL) {
	return false;
    }
    nnip = ((e4_StorageImpl *) impl)->FindNode(nid.GetUniqueID());
    if (nnip == NULL) {
	return false;
    }
    e4_Node nn(nnip);

    n = nn;

    return true;
}

/*
 * Given a unique ID, obtain the vertex referenced by the ID.
 */

bool
e4_Storage::GetVertexFromID(e4_VertexUniqueID vid, e4_Vertex &v) const
{
    e4_VertexImpl *vvip;

    if (impl == NULL) {
	return false;
    }
    vvip = ((e4_StorageImpl *) impl)->GetVertex(vid.GetUniqueID());
    if (vvip == NULL) {
	return false;
    }

    e4_Vertex vv(vvip);

    v = vv;

    return true;
}

/*
 * Given a unique ID for a vertex, get the next vertex after it in the same
 * node.
 */

bool
e4_Storage::FindNextVertex(int vertexID, e4_VisitMethod vm, int vf,
			   int nameID, int nodeID,
			   e4_VertexType typeID, e4_Vertex &f) const
{
    e4_VertexImpl *fi;

    if (impl == NULL) {
	return false;
    }
    fi = ((e4_StorageImpl *) impl)->FindNextVertex(vertexID, vm, vf,
						   nameID, nodeID, typeID);
    if (fi == NULL) {
	return false;
    }
    e4_Vertex ff(fi);

    f = ff;

    return true;
}

/*
 * Given a unique ID for a marker, find the next one after it in
 * the storage.
 */

bool
e4_Storage::FindNextMarker(int markerID, e4_Marker &t) const
{
    e4_MarkerImpl *ti;

    if (impl == NULL) {
	return false;
    }
    ti = ((e4_StorageImpl *) impl)->FindNextMarker(markerID);
    if (ti == NULL) {
	return false;
    }

    e4_Marker tt(ti);

    t = tt;

    return true;
}

/*
 * Given a unique ID for a node, find the next one after it in
 * the storage.
 */

bool
e4_Storage::FindNextNode(int nodeID, e4_Node &n) const
{
    e4_NodeImpl *ni;

    if (impl == NULL) {
	return false;
    }
    ni = ((e4_StorageImpl *) impl)->FindNextNode(nodeID);
    if (ni == NULL) {
	return false;
    }

    e4_Node nn(ni);

    n = nn;

    return true;
}

/*
 * Is this storage dirty?
 */

bool
e4_Storage::IsDirty() const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->IsDirty();
}

void
e4_Storage::MarkDirty() const
{
    if (impl != NULL) {
	((e4_StorageImpl *) impl)->MarkDirty();
    }
}

/*
 * Is this storage locked?
 */

bool
e4_Storage::IsLocked() const
{
    if (impl == NULL) {
	return true;
    }
    return ((e4_StorageImpl *) impl)->IsLocked();
}

/*
 * Get the name of this storage.
 */

const char *
e4_Storage::GetName() const
{
    if (impl == NULL) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->GetName();
}

/*
 * Get the ID of the storage driver used by this storage.
 */

const char *
e4_Storage::GetDriver() const
{
    if (impl == NULL) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->GetDriver();
}

/*
 * Get a value for a statistic being collected for this storage.
 */

bool
e4_Storage::GetStatistic(e4_Space sp, e4_SpaceStat st, int &v) const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->GetStatistic(sp, st, v);
}

/*
 * Intern a name.
 */

int
e4_Storage::InternName(const char *nm) const
{
    if (impl == NULL) {
	return E4_VERTEXNOTFOUND;
    }
    return ((e4_StorageImpl *) impl)->InternName(nm);
}

/*
 * Get a marker given its marker ID.
 */

e4_MarkerImpl *
e4_Storage::GetMarker(int markerID) const
{
    if (!IsValid()) {
      return NULL;
    }
    return ((e4_StorageImpl *) impl)->GetMarker(markerID);
}

/*
 * Get a node given its node ID:
 */

e4_NodeImpl *
e4_Storage::GetNode(int nodeID) const
{
    if (!IsValid()) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->FindOrCreateNode(nodeID);
}

/*
 * Get the name of a vertex given the ID of the name.
 */

const char *
e4_Storage::GetName(int nameID) const
{
    if (!IsValid()) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->NameFromNameID(nameID);
}

/*
 * Return the storage implementation for this storage:
 */

e4_StorageImpl *
e4_Storage::GetStorageImpl() const
{
    if (!IsValid()) {
	return NULL;
    }
    return (e4_StorageImpl *) impl;
}

/*
 * Make this storage empty.
 */

bool
e4_Storage::MakeEmpty() const
{
    if (!IsValid()) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->MakeEmpty();
}

/*
 * Is this storage empty?
 */

bool
e4_Storage::IsEmpty() const
{
    if (!IsValid()) {
	return true;
    }
    if (MarkerCount() == 0) {
	return true;
    }
    return false;
}

/*
 * Callback facility: Declare or remove callbacks for specific kinds of
 * events.
 */

bool
e4_Storage::DeclareMarkerAddCallback(e4_MarkerAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBADDMARKER, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteMarkerAddCallback(e4_MarkerAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBADDMARKER, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareNodeAddCallback(e4_NodeAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBADDNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteNodeAddCallback(e4_NodeAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBADDNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareVertexAddCallback(e4_VertexAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBADDVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteVertexAddCallback(e4_VertexAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBADDVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareMarkerDelCallback(e4_MarkerDelCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBDELMARKER, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteMarkerDelCallback(e4_MarkerDelCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBDELMARKER, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareNodeDelCallback(e4_NodeDelCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBDELNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteNodeDelCallback(e4_NodeDelCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBDELNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareVertexDelCallback(e4_VertexDelCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBDELVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteVertexDelCallback(e4_VertexDelCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBDELVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareVertexModCallback(e4_VertexModCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBMODVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteVertexModCallback(e4_VertexModCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBMODVERTEX, (void *) fn, cd);
    return true;
}

/*
 * Return the kind identifier for this instance.
 */

e4_RefKind
e4_Storage::Kind() const
{
    return E4_RKSTORAGE;
}
