/*
 * mkstorage.cpp --
 *
 *	This file contains the implementation of the 
 *	e4_MetakitStorageImpl class.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Include the e4Graph types used directly here.
 */

#include "e4graphimpl.h"

/*
 * Include the type information specific to this storage driver.
 */

#include "mkstorage.h"

/*
 * These constants define bits for recording state information about
 * an entity (vertex, node or marker) in a storage.
 */

#define E4_INUSE			(1<<0)
#define E4_REACHABLE			(1<<1)
#define E4_DELETED			(1<<2)

/*
 * Properties for indexing entities in a storage (common across 1.0
 * and 1.1):
 */

static c4_IntProp pUserData("userData");
static c4_IntProp pNameID("nameID");
static c4_IntProp pNext("next");
static c4_IntProp pPrev("prev");

static c4_ViewProp pMarkers("e4GraphMarkers");
  static c4_IntProp pRoot("root");

static c4_ViewProp pNodes("e4GraphNodes");
  static c4_IntProp pFirstVertex("firstVertex");
  static c4_IntProp pLastVertex("lastVertex");
  static c4_IntProp pParentID("parentID");
  static c4_IntProp pVertexCount("vertexCount");
  static c4_IntProp pRefCount("refCount");
  static c4_IntProp pNodeMarkers("nodeMarkers");

static c4_ViewProp pVertices("e4GraphVertices");
  static c4_IntProp pNodeID("nodeID");
  static c4_IntProp pVertexType("vertexType");
  static c4_IntProp pRowID("row");

static c4_ViewProp pInts("e4GraphInts");
  static c4_IntProp pIntVal("i");

static c4_ViewProp pDoubles("e4GraphDoubles");
  static c4_DoubleProp pDoubleVal("d");

static c4_ViewProp pStrings("e4GraphStrings");
  static c4_StringProp pStringVal("s");

static c4_ViewProp pNames("e4GraphNames");
  static c4_StringProp pNameVal("n");

static c4_ViewProp pBinary("e4Graphbinary");
  static c4_BytesProp pBinaryVal("b");

static c4_ViewProp pUnused("e4GraphUnused");
  static c4_IntProp pFirst("first");

static c4_ViewProp pParents("e4GraphParents");
  static c4_IntProp pCount("count");

/*
 * Properties present only in 1.1:
 */

static c4_IntProp pFlags("flags");

/*
 * Properties present only in 1.0:
 */

static c4_IntProp pUsed("used");

/*
 * This static method is used to create a new
 * instance of e4_MetakitStorageImpl.
 */

e4_StorageImpl *
e4_MetakitStorageImpl::GetStorage(const char *fname)
{
    e4_StorageImpl *s = new e4_MetakitStorageImpl(fname);

    return s;
}

/*
 * This constructor opens a storage in write mode.
 */

e4_MetakitStorageImpl::e4_MetakitStorageImpl(const char *fname) 
    : e4_StorageImpl(fname, E4_METAKIT)
{
    storage = new c4_Storage(fname, true);
    (void) Initialize();
}

/*
 * Helper function for initializing a storage.
 */

bool
e4_MetakitStorageImpl::Initialize()
{
    int i, j, cnt;

    unused = storage->GetAs(E4_GRAPHUNUSED1_1);
    if (unused.GetSize() == 0) {
        unused.SetSize(E4_GRAPHUNUSEDSIZE);
	pFirst(unused[E4_GRAPHSTORAGEMAJORVER]) = (int) E4_MKSTORAGE_MAJORVER;
	pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = (int) E4_MKSTORAGE_MINORVER;
        for (i = E4_GRAPHFIRSTUNUSEDMARKER; i < E4_GRAPHUNUSEDSIZE; i++) {
            pFirst(unused[i]) = (int) E4_NEXTNONE;
        }

	/*
	 * Need better source of randomness than the address of the storage!
	 */

	pFirst(unused[E4_GRAPHHASHCODE]) = (int) this;

	/*
	 * The storage is initially clean -- no leaked data.
	 */

	pFirst(unused[E4_GRAPHUNREACHABLENODES]) = 0;
    }

    /*
     * Ensure the data stored in the storage just opened is in a
     * format compatible with the current version of the driver.
     */

    if ((int) pFirst(unused[E4_GRAPHSTORAGEMAJORVER]) != 
	E4_MKSTORAGE_MAJORVER) {
	fprintf(stderr, 
		"**Warning**: version mismatch: storage version %d.%d, ",
		(int) pFirst(unused[E4_GRAPHSTORAGEMAJORVER]),
		(int) pFirst(unused[E4_GRAPHSTORAGEMINORVER]));
	fprintf(stderr, 
		"expected version: %d.X. ABORTING\n", 
		E4_MKSTORAGE_MAJORVER);
	return false;
    }

    /*
     * Check if the minor version is older than what we expect. If it is
     * then fix the storage transparently to be of the format expected by
     * the newer minor version, if possible.
     */

    switch ((int) pFirst(unused[E4_GRAPHSTORAGEMINORVER])) {
    case 0:
	/*
	 * Attempt to convert format 1.0 to format 1.1:
	 */

	UpdateFormat1_0to1_1();

	/*
	 * Fall through:
	 */

    case 1:
	/*
	 * Get the storage structure as expected by the current minor
	 * version.
	 */

	markers = storage->GetAs(E4_GRAPHMARKERS1_1);
	nodes = storage->GetAs(E4_GRAPHNODES1_1);
	vertices = storage->GetAs(E4_GRAPHVERTICES1_1);
	doubles = storage->GetAs(E4_GRAPHDOUBLES1_1);
	strings = storage->GetAs(E4_GRAPHSTRINGS1_1);
	binary = storage->GetAs(E4_GRAPHBINARY1_1);
	names = storage->GetAs(E4_GRAPHNAMES1_1);
	parents = storage->GetAs(E4_GRAPHPARENTS1_1);
	break;

    default:
	fprintf(stderr,
		"Expected version %d.%d or earlier, got %d.%d. ABORTING\n",
		E4_MKSTORAGE_MAJORVER, E4_MKSTORAGE_MINORVER,
		E4_MKSTORAGE_MAJORVER, 
		(int) pFirst(unused[E4_GRAPHSTORAGEMINORVER]));
	return false;
    }

    /*
     * Hash all the names.
     */

    PopulateNameHash();

    for (cnt = 0, i = 0; i < nodes.GetSize(); i++) {
      if (((int) pFlags(nodes[i]) & E4_INUSE) == E4_INUSE) {
	cnt++;
      }
    }

    /*
     * Initialize the array of statistics.
     */

    for (i = 0; i < (int) E4_SPLAST; i++) {
	for (j = 0; j < (int) E4_SSLAST; j++) {
	    statistics[i][j] = 0;
	}
    }
    statistics[(int) E4_SPMARKER][(int) E4_SSAVAIL] = markers.GetSize();
    statistics[(int) E4_SPNODE][(int) E4_SSAVAIL] = nodes.GetSize();
    statistics[(int) E4_SPVERTEX][(int) E4_SSAVAIL] = vertices.GetSize();
    statistics[(int) E4_SPNAME][(int) E4_SSAVAIL] = names.GetSize();
    statistics[(int) E4_SPSTRING][(int) E4_SSAVAIL] = strings.GetSize();
    statistics[(int) E4_SPDOUBLE][(int) E4_SSAVAIL] = doubles.GetSize();
    statistics[(int) E4_SPBINARY][(int) E4_SSAVAIL] = binary.GetSize();

    statistics[(int) E4_SPMARKER][(int) E4_SSUSED] = MarkerCount();
    for (i = 0, cnt = 0; i < nodes.GetSize(); i++) {
      if (((int) pFlags(nodes[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPNODE][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < vertices.GetSize(); i++) {
	if (((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPVERTEX][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < names.GetSize(); i++) {
	if (((int) pFlags(names[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPNAME][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < strings.GetSize(); i++) {
	if (((int) pFlags(strings[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPSTRING][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < doubles.GetSize(); i++) {
	if (((int) pFlags(doubles[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < binary.GetSize(); i++) {
	if (((int) pFlags(binary[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPBINARY][(int) E4_SSUSED] = cnt;

    /*
     * Initialize the deletion mechanism to NULL.
     */

    idStack1 = NULL;
    idStack2 = NULL;

    return true;
}

/*
 * Helper method to update the format from 1.0 to 1.1:
 */

void
e4_MetakitStorageImpl::UpdateFormat1_0to1_1()
{
    int i, l;

    markers = storage->GetAs(E4_GRAPHMARKERS1_0);
    nodes = storage->GetAs(E4_GRAPHNODES1_0);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_0);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_0);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_0);
    binary = storage->GetAs(E4_GRAPHBINARY1_0);
    names = storage->GetAs(E4_GRAPHNAMES1_0);
    parents = storage->GetAs(E4_GRAPHPARENTS1_0);

    /*
     * Add the new property "flags":
     */

    markers.AddProperty(pFlags);
    nodes.AddProperty(pFlags);
    vertices.AddProperty(pFlags);
    doubles.AddProperty(pFlags);
    strings.AddProperty(pFlags);
    binary.AddProperty(pFlags);
    names.AddProperty(pFlags);
    parents.AddProperty(pFlags);

    /*
     * For each entry, check if its "used" propert was set -- if so, set
     * the "flags" property to E4_INUSE.
     */

    for (i = 0, l = markers.GetSize(); i < l; i++) {
	if ((int) pUsed(markers[i]) == 1) {
	    pFlags(markers[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	if ((int) pUsed(nodes[i]) == 1) {
	    pFlags(nodes[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = vertices.GetSize(); i < l; i++) {
	if ((int) pUsed(vertices[i]) == 1) {
	    pFlags(vertices[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = doubles.GetSize(); i < l; i++) {
	if ((int) pUsed(doubles[i]) == 1) {
	    pFlags(doubles[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = strings.GetSize(); i < l; i++) {
	if ((int) pUsed(strings[i]) == 1) {
	    pFlags(strings[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = binary.GetSize(); i < l; i++) {
	if ((int) pUsed(binary[i]) == 1) {
	    pFlags(binary[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = names.GetSize(); i < l; i++) {
	if ((int) pUsed(names[i]) == 1) {
	    pFlags(names[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = parents.GetSize(); i < l; i++) {
	if ((int) pUsed(parents[i]) == 1) {
	    pFlags(parents[i]) = (int) E4_INUSE;
	}
    }

    /*
     * Now discard the old "used" property by getting the views with
     * the new format:
     */

    markers = storage->GetAs(E4_GRAPHMARKERS1_1);
    nodes = storage->GetAs(E4_GRAPHNODES1_1);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_1);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_1);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_1);
    binary = storage->GetAs(E4_GRAPHBINARY1_1);
    names = storage->GetAs(E4_GRAPHNAMES1_1);
    parents = storage->GetAs(E4_GRAPHNAMES1_1);

    pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = E4_MKSTORAGE_MINORVER;

    storage->Commit();
}

/*
 * The destructor:
 */

e4_MetakitStorageImpl::~e4_MetakitStorageImpl()
{
    if (storage != NULL) {
        delete storage;
    }

    /*
     * Clean up the deletion hash tables if they're present.
     */

    if (idStack1 != NULL) {
	delete idStack1;
    }
    if (idStack2 != NULL) {
	delete idStack2;
    }
}
        
/*
 * Commit changes to this storage.
 */

bool
e4_MetakitStorageImpl::Commit()
{
    if (!IsDirty()) {
	return true;
    }
    storage->Commit();
    MarkClean();

    return true;
}

/*
 * Compute how many markers there are in this storage.
 */

int
e4_MetakitStorageImpl::MarkerCount() const
{
    int i, lim, cnt;

    for (i = 0, lim = markers.GetSize(), cnt = 0;
         i < lim;
         i++) {
        if (((int) pFlags(markers[i]) & E4_INUSE) == E4_INUSE) {
            cnt++;
        }
    }
    return cnt;
}

/*
 * Helper functions to initialize and manage the free lists.
 */

void
e4_MetakitStorageImpl::PopulateNameHash()
{
    int i, lim;

    for (i = 0, lim = names.GetSize(); i < lim; i++) {
	if (((int) pFlags(names[i]) & E4_INUSE) == E4_INUSE) {
            AddNameToNameHash(pNameVal(names[i]), i);
	}
    }
}

void
e4_MetakitStorageImpl::MakeParentSpace()
{
    int next = parents.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    parents.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) = next;
    for (; next < newsize; next++) {
	pNext(parents[next]) = next + 1;
	pFlags(parents[next]) = 0;
    }
    pNext(parents[newsize - 1]) = E4_NEXTNONE;
    pFlags(parents[newsize - 1]) = 0;
}

void
e4_MetakitStorageImpl::MakeMarkerSpace()
{
    int next = markers.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    markers.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDMARKER]) = next;
    for (; next < newsize; next++) {
        pNext(markers[next]) = next + 1;
	pFlags(markers[next]) = 0;
    }
    pNext(markers[newsize - 1]) = E4_NEXTNONE;
    pFlags(markers[newsize - 1]) = 0;

    statistics[(int) E4_SPMARKER][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeNodeSpace()
{
    int next = nodes.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    nodes.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) = next;
    for (; next < newsize; next++) {
        pNext(nodes[next]) = next + 1;
	pFlags(nodes[next]) = 0;
    }
    pNext(nodes[newsize - 1]) = E4_NEXTNONE;
    pFlags(nodes[newsize - 1]) = 0;

    statistics[(int) E4_SPNODE][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeVertexSpace()
{
    int next = vertices.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    vertices.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) = next;
    for (; next < newsize; next++) {
        pNext(vertices[next]) = next + 1;
	pFlags(vertices[next]) = 0;
    }
    pNext(vertices[newsize - 1]) = E4_NEXTNONE;
    pFlags(vertices[newsize - 1]) = 0;

    statistics[(int) E4_SPVERTEX][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeDoubleSpace()
{
    int next = doubles.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    doubles.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) = next;
    for (; next < newsize; next++) {
        pNext(doubles[next]) = next + 1;
	pFlags(doubles[next]) = 0;
    }
    pNext(doubles[newsize - 1]) = E4_NEXTNONE;
    pFlags(doubles[newsize - 1]) = 0;

    statistics[(int) E4_SPDOUBLE][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeStringSpace()
{
    int next = strings.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    strings.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) = next;
    for (; next < newsize; next++) {
        pNext(strings[next]) = next + 1;
	pFlags(strings[next]) = 0;
    }
    pNext(strings[newsize - 1]) = E4_NEXTNONE;
    pFlags(strings[newsize - 1]) = 0;

    statistics[(int) E4_SPSTRING][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeBinarySpace()
{
    int next = binary.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    binary.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) = next;
    for (; next < newsize; next++) {
        pNext(binary[next]) = next + 1;
	pFlags(binary[next]) = 0;
    }
    pNext(binary[newsize - 1]) = E4_NEXTNONE;
    pFlags(binary[newsize - 1]) = 0;

    statistics[(int) E4_SPBINARY][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeNameSpace()
{
    int next = names.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    names.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) = next;
    for (; next < newsize; next++) {
	pNext(names[next]) = next + 1;
	pFlags(names[next]) = 0;
    }
    pNext(names[newsize - 1]) = E4_NEXTNONE;
    pFlags(names[newsize - 1]) = 0;

    statistics[(int) E4_SPNAME][(int) E4_SSAVAIL] = newsize;
}

int
e4_MetakitStorageImpl::CreateMarkerIDFromNameID(int nameID)
{
    int idx;
    int nodeID;
    
    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDMARKER]) == E4_NEXTNONE) {
        MakeMarkerSpace();
    }

    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDMARKER]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDMARKER]) = (int) pNext(markers[idx]);

    /*
     * Update the number of used markers.
     */

    statistics[(int) E4_SPMARKER][(int) E4_SSUSED]++;
    statistics[(int) E4_SPMARKER][(int) E4_SSALLOC]++;

    /*
     * Initialize the marker.
     */

    nodeID = ReserveNodeID();

    pFlags(markers[idx]) = E4_INUSE;
    pUserData(markers[idx]) = 0;
    pNameID(markers[idx]) = nameID;

    /*
     * We set the root here to E4_NEXTNONE. The root will be
     * set by AddNodeMarker below.
     */

    pRoot(markers[idx]) = E4_NEXTNONE;
    pNext(markers[idx]) = E4_NEXTNONE;

    /*
     * Also add the new marker to the list of markers for the
     * new node.
     */

    if (!AddNodeMarker(idx, nodeID)) {

	ForgetNode(nodeID);
	AddToUnusedNodesList(nodeID);

	ForgetMarker(idx);
	AddToUnusedMarkersList(idx);

	return E4_MARKERNOTFOUND;
    }

    MarkDirty();

    return idx;
}

int
e4_MetakitStorageImpl::GetMarkerIDFromNameID(int nameID)
{
    int i, lim;
    
    for (i = 0, lim = markers.GetSize(); i < lim; i++) {
        if ((((int) pFlags(markers[i]) & E4_INUSE) == E4_INUSE) &&
	    (pNameID(markers[i]) == nameID)) {
            return i;
        }
    }
    return E4_MARKERNOTFOUND;
}

/*
 * Compute the rank of a child node in the vertex list of a node.
 */

int
e4_MetakitStorageImpl::GetRankOfChildNode(int parentID, int childID) const
{
    int f, rank;
    
    if ((parentID < 0) || (parentID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[parentID]) & E4_INUSE) == 0) ||
        (childID < 0) || (childID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[childID]) & E4_INUSE) == 0)) {
        return E4_VERTEXNOTFOUND;
    }
    for (rank = 1, f = (int) pFirstVertex(nodes[parentID]);
         f != E4_NEXTNONE;
         f = (int) pNext(vertices[f]), rank++) {
        if (((e4_VertexType) (int) pVertexType(vertices[f]) == E4_VTNODE) &&
            ((int) pRowID(vertices[f]) == childID)) {
            return rank;
        }
    }
    return E4_VERTEXNOTFOUND;
}

/*
 * Helper function to set a vertex value:
 */

bool
e4_MetakitStorageImpl::SetVertex(int index, int nameID, 
				 int vertexType, int itemID)
{
    c4_RowRef r = vertices[index];

    pVertexType(r) = vertexType;
    pNameID(r) = nameID;
    pRowID(r) = itemID;

    /*
     * Do not change the "next" and "used" property.
     */

    return true;
}

/*
 * Insert a new vertex in the node identified by nodeID at the position
 * indicated by "io" and "rank". The ID of the new vertex is returned.
 */

int
e4_MetakitStorageImpl::AddVertex(int nodeID, e4_InsertOrder io, int &rank)
{
    int i, j, k;

    /*
     * First of all grab a free row in the "vertices" view.
     */

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) == E4_NEXTNONE) {
	MakeVertexSpace();
    }
    i = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) = (int) pNext(vertices[i]);

    /*
     * Update the count of used vertices.
     */

    statistics[(int) E4_SPVERTEX][(int) E4_SSUSED]++;
    statistics[(int) E4_SPVERTEX][(int) E4_SSALLOC]++;

    /*
     * Initialize the values in the newly allocated vertex.
     */

    c4_RowRef r = vertices[i];
    pNameID(r) = -1;
    pNodeID(r) = nodeID;
    pRowID(r) = -1;
    pFlags(r) = E4_INUSE;
    pUserData(r) = 0;

    /*
     * Now use the row we grabbed as the vertex indicated by the "io" and
     * "rank" arguments.
     */

    switch (io) {
    case E4_IOFIRST:
	pNext(vertices[i]) = (int) pFirstVertex(nodes[nodeID]);
	pPrev(vertices[i]) = E4_NEXTNONE;
	if ((int) pFirstVertex(nodes[nodeID]) != E4_NEXTNONE) {
	    pPrev(vertices[pFirstVertex(nodes[nodeID])]) = i;
	}
        pFirstVertex(nodes[nodeID]) = i;
        if ((int) pLastVertex(nodes[nodeID]) == E4_NEXTNONE) {
            pLastVertex(nodes[nodeID]) = i;
	}
	rank = 1;
	break;

    case E4_IOLAST:
	pNext(vertices[i]) = E4_NEXTNONE;
        if ((int) pFirstVertex(nodes[nodeID]) == E4_NEXTNONE) {
            pFirstVertex(nodes[nodeID]) = i;
        }
        if ((int) pLastVertex(nodes[nodeID]) != E4_NEXTNONE) {
	    pPrev(vertices[i]) = (int) pLastVertex(nodes[nodeID]);
            pNext(vertices[pLastVertex(nodes[nodeID])]) = i;
	} else {
	    pPrev(vertices[i]) = E4_NEXTNONE;
	}
        pLastVertex(nodes[nodeID]) = i;
        rank = (int) pVertexCount(nodes[nodeID]) + 1;
	break;

    case E4_IOAT:
    case E4_IOBEFORE:
	if (rank == 1) {
	    pNext(vertices[i]) = (int) pFirstVertex(nodes[nodeID]);
	    pPrev(vertices[i]) = E4_NEXTNONE;
            if ((int) pLastVertex(nodes[nodeID]) == E4_NEXTNONE) {
                pLastVertex(nodes[nodeID]) = i;
	    }
	    if ((int) pFirstVertex(nodes[nodeID]) != E4_NEXTNONE) {
		pPrev(vertices[pFirstVertex(nodes[nodeID])]) = i;
	    }
	    pFirstVertex(nodes[nodeID]) = i;
	    rank = 1;
	} else {
	    j = VertexIDFromRank(nodeID, rank - 1);
	    if (j == E4_VERTEXNOTFOUND) {
                goto error;
            }
            k = (int) pNext(vertices[j]);
	    pNext(vertices[j]) = i;
	    pNext(vertices[i]) = k;
	    pPrev(vertices[k]) = i;
	    pPrev(vertices[i]) = j;
            if ((int) pLastVertex(nodes[nodeID]) == j) {
                pLastVertex(nodes[nodeID]) = i;
            }
	    rank = i;
	}
	break;

    case E4_IOAFTER:
        j = VertexIDFromRank(nodeID, rank);
	if (j == E4_VERTEXNOTFOUND) {
	    goto error;
	}
	pNext(vertices[i]) = (int) pNext(vertices[j]);
	pPrev(vertices[i]) = j;
	if ((int) pNext(vertices[i]) != E4_NEXTNONE) {
	    pPrev(vertices[pNext(vertices[i])]) = i;
	}
	pNext(vertices[j]) = i;
        if ((int) pLastVertex(nodes[nodeID]) == j) {
            pLastVertex(nodes[nodeID]) = i;
        }
	rank = j + 1;
	break;
    }

    /*
     * Update the vertex count for this node.
     */

    pVertexCount(nodes[nodeID]) = (int) pVertexCount(nodes[nodeID]) + 1;

    return i;

error:
    /*
     * Some kind of error occurred, after we allocated the new row.
     * Free up the allocated row so that it can be used in a future
     * operation.
     */

    AddToUnusedVerticesList(i);

    return E4_VERTEXNOTCREATED;
}

/*
 * Free a row occupied by a 64-bit floating point value.
 */

bool
e4_MetakitStorageImpl::FreeDouble(int doubleID)
{
    if ((doubleID < 0) || 
	(doubleID >= doubles.GetSize()) ||
        (((int) pFlags(doubles[doubleID]) & E4_INUSE) == 0)) {
        return false;
    }

    AddToUnusedDoublesList(doubleID);
    MarkDirty();

    return true;
}

/*
 * Free a row occupied by a string value. Release the storage occupied by
 * the actual string value also.
 */

bool
e4_MetakitStorageImpl::FreeString(int stringID)
{
    if ((stringID < 0) || 
	(stringID >= strings.GetSize()) ||
        (((int) pFlags(strings[stringID]) & E4_INUSE) == 0)) {
        return false;
    }

    AddToUnusedStringsList(stringID);
    pStringVal(strings[stringID]) = "";
    MarkDirty();

    return true;
}

/*
 * Free a row occupied by a binary value. Release the storage occupied by
 * the actual value also.
 */

bool
e4_MetakitStorageImpl::FreeBinary(int binaryID)
{
    if ((binaryID < 0) || 
	(binaryID >= binary.GetSize()) ||
        (((int) pFlags(binary[binaryID]) & E4_INUSE) == 0)) {
        return false;
    }

    AddToUnusedBinaryList(binaryID);

    c4_Bytes bytes(NULL, 0);

    pBinaryVal(binary[binaryID]) = bytes;
    MarkDirty();

    return true;
}

/*
 * Return a node from a designated row.
 */

bool
e4_MetakitStorageImpl::GetNode(int index, e4_NodeImpl *&value) const
{
    if ((index < 0) ||
	(index >= nodes.GetSize()) ||
        (((int) pFlags(nodes[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = FindOrCreateNode(index);
    return true;
}

/*
 * Specialized internal version to return a node from a designated row.
 */

bool
e4_MetakitStorageImpl::GetNode(int index, e4_NodeImpl *&value, 
			       bool *isNew) const
{
    if ((index < 0) ||
	(index >= nodes.GetSize()) ||
        (((int) pFlags(nodes[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = FindOrCreateNode(index, isNew);
    return true;
}

/*
 * Return an integer value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetInt(int index, int &value) const 
{
    /*
     * Integers are represented directly in the vertex, so the index
     * *is* the vertex value.
     */

    value = index;

    return true;
}

/*
 * Return a 64-bit floating point value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetDouble(int index, double &value) const
{
    if ((index < 0) ||
	(index >= doubles.GetSize()) ||
        (((int) pFlags(doubles[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = (double) pDoubleVal(doubles[index]);

    return true;
}

/*
 * Return a string value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetString(int index, char *&value) const
{
    if ((index < 0) ||
	(index >= strings.GetSize()) ||
        (((int) pFlags(strings[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = (char *) (const char *) pStringVal(strings[index]);

    return true;
}

/*
 * Return a binary value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetBinary(int index, void *&value,
				 int &nbytes) const
{
    if ((index < 0) ||
	(index >= binary.GetSize()) ||
        (((int) pFlags(binary[index]) & E4_INUSE) == 0)) {
        return false;
    }

    c4_Bytes bytes = pBinaryVal(binary[index]);

    value = (void *) (const void *) bytes.Contents();
    nbytes = bytes.Size();

    return true;
}

/*
 * Retrieve a vertex name by the name ID.
 */

bool
e4_MetakitStorageImpl::GetName(int index, const char *&nm) const
{
    if ((index < 0) ||
	(index >= names.GetSize()) ||
	(((int) pFlags(names[index]) & E4_INUSE) == 0)) {
	return false;
    }

    nm = pNameVal(names[index]);
    return true;
}

/*
 * Add a row for a new node.
 */

int
e4_MetakitStorageImpl::ReserveNodeID()
{
    int idx;

    if (pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) == E4_NEXTNONE) {
        MakeNodeSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) = (int) pNext(nodes[idx]);

    statistics[(int) E4_SPNODE][(int) E4_SSUSED]++;
    statistics[(int) E4_SPNODE][(int) E4_SSALLOC]++;

    c4_RowRef r = nodes[idx];
    pFlags(r) = E4_INUSE;
    pUserData(r) = 0;
    pNext(r) = E4_NEXTNONE;
    pVertexCount(r) = 0;
    pFirstVertex(r) = E4_NEXTNONE;
    pLastVertex(r) = E4_NEXTNONE;
    pParentID(r) = E4_NEXTNONE;
    pRefCount(r) = 0;
    pNodeMarkers(r) = E4_NEXTNONE;

    MarkDirty();
    return idx;
}

/*
 * Add an integer to the storage -- nothing to do because the
 * integer is represented directly in the vertex.
 */

int
e4_MetakitStorageImpl::AddInt(int value)
{
    MarkDirty();
    return value;
}

/*
 * Add a row for a new 64-bit floating point value.
 */

int
e4_MetakitStorageImpl::AddDouble(double value)
{
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) == E4_NEXTNONE) {
        MakeDoubleSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) = (int) pNext(doubles[idx]);

    statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED]++;
    statistics[(int) E4_SPDOUBLE][(int) E4_SSALLOC]++;

    pFlags(doubles[idx]) = E4_INUSE;
    pDoubleVal(doubles[idx]) = value;
    MarkDirty();
    return idx;
}

/*
 * Add a row for a new string value.
 */

int
e4_MetakitStorageImpl::AddString(const char *value)
{
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) == E4_NEXTNONE) {
        MakeStringSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) = (int) pNext(strings[idx]);

    statistics[(int) E4_SPSTRING][(int) E4_SSUSED]++;
    statistics[(int) E4_SPSTRING][(int) E4_SSALLOC]++;

    pFlags(strings[idx]) = E4_INUSE;
    pStringVal(strings[idx]) = value;
    MarkDirty();
    return idx;
}

/*
 * Add a row for a new binary value.
 */

int
e4_MetakitStorageImpl::AddBinary(const void *value, int nbytes)
{
    c4_Bytes bytes(value, nbytes);
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) == E4_NEXTNONE) {
        MakeBinarySpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) = (int) pNext(binary[idx]);

    statistics[(int) E4_SPBINARY][(int) E4_SSUSED]++;
    statistics[(int) E4_SPBINARY][(int) E4_SSALLOC]++;

    pFlags(binary[idx]) = E4_INUSE;
    pBinaryVal(binary[idx]) = bytes;
    MarkDirty();
    return idx;
}

/*
 * Add a row for a new vertex, node or marker name.
 */

int
e4_MetakitStorageImpl::AddName(const char *nm)
{
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) == E4_NEXTNONE) {
	MakeNameSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) = (int) pNext(names[idx]);

    statistics[(int) E4_SPNAME][(int) E4_SSUSED]++;
    statistics[(int) E4_SPNAME][(int) E4_SSALLOC]++;

    pFlags(names[idx]) = E4_INUSE;
    pNameVal(names[idx]) = nm;
    MarkDirty();
    return idx;
}

/*
 * Is this instance valid?
 */

bool
e4_MetakitStorageImpl::IsValid() const
{
    return (storage == NULL) ? false : true;
}

/*
 * Convert between a marker name and a marker ID.
 */

int
e4_MetakitStorageImpl::GetMarkerID(const char *nm)
{
    int nameID = InternName(nm);
    int i, lim;

    for (i = 0, lim = markers.GetSize(); i < lim; i++) {
        if ((int) pNameID(markers[i]) == nameID) {
            return i;
        }
    }
    return E4_MARKERNOTFOUND;
}

const char *
e4_MetakitStorageImpl::GetMarkerName(int markerID)
{
    if ((markerID < 0) ||
	(markerID >= markers.GetSize()) ||
        (((int) pFlags(markers[markerID]) & E4_INUSE) == 0)) {
        return NULL;
    }
    return pNameVal(names[pNameID(markers[markerID])]);
}

int
e4_MetakitStorageImpl::VertexIDFromNthVertex(int nodeID, int nameID,
					     int nth, int &rank) const
{
    int i, r, cnt;

    /*
     * Its not cached, so we must search for it the slow way. Note
     * that it's also not known whether there is an nth vertex with
     * name nameID in this node.
     */

    for (r = 1, i = (int) pFirstVertex(nodes[nodeID]), cnt = 0; 
	 i != E4_NEXTNONE; 
	 i = (int) pNext(vertices[i]), r++) {
        if ((int) pNameID(vertices[i]) == nameID) {
            cnt++;
            if (cnt == nth) {
                break;
            }
        }
    }
    if ((cnt != nth) || (i == E4_NEXTNONE) || 
	((int) pNameID(vertices[i]) != nameID)) {
        return E4_VERTEXNOTFOUND;
    }
    rank = r;
    return i;
}

int
e4_MetakitStorageImpl::VertexIDFromRank(int nodeID, int rank) const
{
    int i, r;

    /*
     * Quick check to see if the node has at least that many
     * vertices. If not, we do not need to search.
     */

    if ((int) pVertexCount(nodes[nodeID]) < rank) {
	return E4_VERTEXNOTFOUND;
    }

    /*
     * The data we were looking for is not cached, so we
     * must search for it the slow way. But we know at this
     * point that the search will be successful.
     */

    for (i = (int) pFirstVertex(nodes[nodeID]), r = 1; 
	 r < rank;
         i = (int) pNext(vertices[i]), r++) {
    }
    return i;
}

int
e4_MetakitStorageImpl::RankFromVertexID(int nodeID, int index) const
{
    int i, r;

    for (i = (int) pFirstVertex(nodes[nodeID]), r = 1;
         (i != E4_NEXTNONE) && (i != index);
         i = (int) pNext(vertices[i]), r++) {
    }
    if (i != index) {
        r = E4_VERTEXNOTFOUND;
    }
    return r;
}

int
e4_MetakitStorageImpl::GetVertexIDInParent(int parentID, int childID) const 
{
    int i;

    if ((parentID < 0) ||
	(parentID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[parentID]) & E4_INUSE) == 0) ||
        (childID < 0) ||
	(childID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[childID]) & E4_INUSE) == 0)) {
        return E4_VERTEXNOTFOUND;
    }
    for (i = (int) pFirstVertex(nodes[parentID]); 
	 i != E4_NEXTNONE;
	 i = (int) pNext(vertices[i])) {
        if (((e4_VertexType) (int) pVertexType(vertices[i]) == E4_VTNODE) &&
            ((int) pRowID(vertices[i]) == childID)) {
            return i;
        }
    }
    return E4_VERTEXNOTFOUND;
}

/*
 * Free the storage occupied by an old value of a vertex:
 */

void
e4_MetakitStorageImpl::FreeVertexValue(int vertexID)
{
    switch (pVertexType(vertices[vertexID])) {
    case E4_VTNODE:
	DetachParent((int) pRowID(vertices[vertexID]),
		     (int) pNodeID(vertices[vertexID]));
	break;
    case E4_VTINT:

        /*
	 * Nothing to do because the value is represented directly
	 * in the vertex.
	 */

        break;
    case E4_VTDOUBLE:
        FreeDouble((int) pRowID(vertices[vertexID]));
        break;
    case E4_VTSTRING:
        FreeString((int) pRowID(vertices[vertexID]));
        break;
    case E4_VTBINARY:
        FreeBinary((int) pRowID(vertices[vertexID]));
        break;
    default:
        /*
         * Should abort here.
         */

        break;
    }
}

/*
 * Get the value of a vertex given its index.
 */

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, e4_ValueImpl *&v) const
{
    e4_ValueImpl *vp;
    e4_NodeImpl *n;
    double d;
    char *s;
    void *bytes;
    int nbytes;

    vp = new e4_ValueImpl;
    vp->vertexType = (e4_VertexType) (int)  pVertexType(vertices[i]);
    switch(vp->vertexType) {
    case E4_VTUNKNOWN:
	return false;
    case E4_VTNODE:
        if (GetNode((int) pRowID(vertices[i]), n) != true) {
	    return false;
	}
	vp->u.n = n;
	break;
    case E4_VTINT:

        /*
	 * Integers are represented directly in a vertex.
	 */

	vp->u.i = (int) pRowID(vertices[i]);
	break;
    case E4_VTDOUBLE:
        if (GetDouble((int) pRowID(vertices[i]), d) != true) {
	    return false;
	}
	vp->u.d = d;
	break;
    case E4_VTSTRING:
	if (GetString((int) pRowID(vertices[i]), s) != true) {
	    return false;
	}
	vp->u.s = s;
	break;
    case E4_VTBINARY:
	if (GetBinary((int) pRowID(vertices[i]), bytes, nbytes) != true) {
	    return false;
	}
	vp->u.b.bytes = bytes;
	vp->u.b.nbytes = nbytes;
	break;
    default:
	return false;
    }

    v = vp;
    return true;
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, e4_NodeImpl *&n) const
{
    if ((int) pVertexType(vertices[i]) != E4_VTNODE) {
	return false;
    }
    return GetNode((int) pRowID(vertices[i]), n);
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, int &value) const
{
    if ((int) pVertexType(vertices[i]) != E4_VTINT) {
	return false;
    }

    /*
     * Integers are represented directly in the vertex.
     */

    value = (int) pRowID(vertices[i]);
    return true;
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, double &value) const
{
    if ((int) pVertexType(vertices[i]) != E4_VTDOUBLE) {
	return false;
    }
    return GetDouble((int) pRowID(vertices[i]), value);
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, const char *&value) const
{
    char *lvalue;

    if ((int) pVertexType(vertices[i]) != E4_VTSTRING) {
	return false;
    }
    if (GetString((int) pRowID(vertices[i]), lvalue)) {
	value = lvalue;
	return true;
    }
    return false;
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, const void *&bytes,
                                   int &nbytes) const
{
    void *lbytes;

    if ((int) pVertexType(vertices[i]) != E4_VTBINARY) {
	return false;
    }
    if (GetBinary((int) pRowID(vertices[i]), lbytes, nbytes)) {
	bytes = lbytes;
        return true;
    }
    return false;
}

/*
 * Set a vertex whose index is given:
 */

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, int value)
{
    if (pVertexType(vertices[i]) != E4_VTINT) {
        FreeVertexValue(i);
    }
    SetVertex(i, pNameID(vertices[i]), E4_VTINT, value);
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, double value)
{
    int j;

    if ((int) pVertexType(vertices[i]) != E4_VTDOUBLE) {
	FreeVertexValue(i);
	j = AddDouble(value);
	SetVertex(i, (int) pNameID(vertices[i]), E4_VTDOUBLE, j);
    } else {
	pDoubleVal(doubles[pRowID(vertices[i])]) = value;
    }
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, const char *value)
{
    int j;

    if ((int) pVertexType(vertices[i]) != E4_VTSTRING) {
	FreeVertexValue(i);
	j = AddString(value);
	SetVertex(i, (int) pNameID(vertices[i]), E4_VTSTRING, j);
    } else {

        /*
	 * NOTE: This assumes that string values are not
	 * shared.
	 */

	pStringVal(strings[pRowID(vertices[i])]) = value;
    }
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, const void *bs, int nbs)
{
    int j;

    if ((int) pVertexType(vertices[i]) != E4_VTBINARY) {
	FreeVertexValue(i);
	j = AddBinary(bs, nbs);
	SetVertex(i, pNameID(vertices[i]), E4_VTBINARY, j);
    } else {
        c4_Bytes bytes(bs, nbs);
	pBinaryVal(binary[pRowID(vertices[i])]) = bytes;
    }
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndexToNode(int i, int childID)
{
    FreeVertexValue(i);
    SetVertex(i, pNameID(vertices[i]), E4_VTNODE, childID);
    MarkDirty();
    return true;
}

/*
 * Get and set the user data for different entities:
 */

bool
e4_MetakitStorageImpl::SetMarkerUserData(int markerID, int userData) const
{
    if ((markerID < 0) ||
	(markerID >= markers.GetSize()) ||
	(((int) pFlags(markers[markerID]) & E4_INUSE) == 0)) {
	return false;
    }

    pUserData(markers[markerID]) = userData;

    return true;
}

bool
e4_MetakitStorageImpl::SetNodeUserData(int nodeID, int userData) const
{
    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }

    pUserData(nodes[nodeID]) = userData;

    return true;
}

bool
e4_MetakitStorageImpl::SetVertexUserData(int vertexID, int userData) const
{
    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return false;
    }

    pUserData(vertices[vertexID]) = userData;

    return true;
}

bool
e4_MetakitStorageImpl::GetMarkerUserData(int markerID, int &userData) const
{
    if ((markerID < 0) || 
	(markerID >= markers.GetSize()) ||
	(((int) pFlags(markers[markerID]) & E4_INUSE) == 0)) {
	return false;
    }

    userData = (int) pUserData(markers[markerID]);

    return true;
}

bool
e4_MetakitStorageImpl::GetNodeUserData(int nodeID, int &userData) const
{
    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }

    userData = (int) pUserData(nodes[nodeID]);

    return true;
}

bool
e4_MetakitStorageImpl::GetVertexUserData(int vertexID, int &userData) const
{
    if ((vertexID < 0) || 
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return false;
    }

    userData = (int) pUserData(vertices[vertexID]);

    return true;
}

/*
 * Given a node ID, retrieve the parent node or its ID.
 */

e4_NodeImpl *
e4_MetakitStorageImpl::GetParentNode(int childID, int nth) const 
{
    e4_NodeImpl *n;
    int parentID;

    if ((childID < 0) ||
	(childID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[childID]) & E4_INUSE) == 0)) {
	return NULL;
    }
    parentID = GetParentNodeID(childID, nth);
    if (parentID == E4_NODENOTFOUND) {
	return NULL;
    }
    if (GetNode(parentID, n) == false) {
	return NULL;
    }
    return n;
}

int 
e4_MetakitStorageImpl::GetParentNodeID(int childID, int nth) const 
{
    int i, pid;

    /*
     * If this node has no parents then return an indication
     * that we could not find an ID for its parent.
     */

    if ((int) pParentID(nodes[childID]) == E4_NEXTNONE) {
	return E4_NODENOTFOUND;
    }

    /*
     * If nth is less than one, we want the last parent, whatever its rank.
     */

    if (nth < 1) {
	for (pid = (int) pParentID(nodes[childID]);
	     (int) pNext(parents[pid]) != E4_NEXTNONE;
	     pid = (int) pNext(parents[pid]));
	return (int) pNodeID(parents[pid]);
    }

    /*
     * Get the nth parent in the list of parents.
     */

    for (pid = (int) pParentID(nodes[childID]), i = 1; 
	 (i < nth) && ((int) pNext(parents[pid]) != E4_NEXTNONE);
	 i++, pid = (int) pNext(parents[pid]));

    /*
     * If this node does not have an nth parent, return failure.
     */

    if (i < nth) {
	return E4_NODENOTFOUND;
    }

    /*
     * We found the nth parent.
     */

    return (int) pNodeID(parents[pid]);
}

/*
 * Get the vertex after one identified by vertexID.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::NextVertexAfter(int num, int vertexID) const
{
    int nvID;
    int i;

    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return NULL;
    }
    if (num < 1) {
	return NULL;
    }
    for (nvID = vertexID, i = 0; i < num; i++) {
	nvID = (int) pNext(vertices[nvID]);
	if ((nvID == E4_NEXTNONE) ||
	    (((int) pFlags(vertices[nvID]) & E4_INUSE) == 0)) {
	    return NULL;
	}
    }
    return GetVertex(nvID);
}

/*
 * Get the vertex before one identified by vertexID.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::PrevVertexBefore(int num, int vertexID) const
{
    int nvID;
    int i;

    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return NULL;
    }
    if (num < 1) {
	return NULL;
    }
    for (nvID = vertexID, i = 0; i < num; i++) {
	nvID = (int) pPrev(vertices[nvID]);
	if ((nvID == E4_NEXTNONE) || 
	    (((int) pFlags(vertices[nvID]) & E4_INUSE) == 0)) {
	    return NULL;
	}
    }
    return GetVertex(nvID);
}

/*
 * Given a marker ID, retrieve the node it marks.
 */

e4_NodeImpl *
e4_MetakitStorageImpl::GetMarkedNode(int markerID) const 
{
    e4_NodeImpl *v;

    if (GetNode(pRoot(markers[markerID]), v) == false) {
        return NULL;
    }
    return v;
}

/*
 * How many markers mark a node identified by nodeID?
 */

int
e4_MetakitStorageImpl::MarkerCount(int nodeID) const
{
    int count, m;

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return 0;
    }
    for (count = 0, m = (int) pNodeMarkers(nodes[nodeID]);
	 m != E4_NEXTNONE;
	 count++, m = (int) pNext(markers[m]));
    return count;
}

/*
 * Is the node identified by nodeID marked?
 */

bool
e4_MetakitStorageImpl::IsNodeMarked(int nodeID) const
{
    /*
     * Sanity checks:
     */

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }

    if (pNodeMarkers(nodes[nodeID]) != E4_NEXTNONE) {
	return true;
    }
    return false;
}

/*
 * Is the node identified by index valid?
 */

bool
e4_MetakitStorageImpl::IsLegalNodeID(int index) const
{
    if ((index < 0) ||
	(index >= nodes.GetSize()) ||
	(((int) pFlags(nodes[index]) & E4_INUSE) == 0)) {
	return false;
    }
    return true;
}

/*
 * Is the node identified by nodeID reachable from any parent?
 */

bool
e4_MetakitStorageImpl::IsReachableNode(int nodeID) const
{
    /*
     * Sanity checks:
     */

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }

    /*
     * A node is reachable if it is marked by any markers, or if it
     * has parents (because these parents must themselves be reachable
     * from some marker).
     */

    if (((int) pParentID(nodes[nodeID]) != E4_NEXTNONE) ||
	((int) pNodeMarkers(nodes[nodeID]) != E4_NEXTNONE)) {
	return true;
    }
    return false;
}

/*
 * Retrieve the nth marker of a node identified by nodeID.
 */

e4_MarkerImpl *
e4_MetakitStorageImpl::GetNthMarker(int nth, int nodeID) const
{
    int m, hold, i;

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return NULL;
    }

    /*
     * If "nth" is less than one, we want the last marker, whatever its rank.
     */

    if (nth <= 0) {
	for (m = (int) pNodeMarkers(nodes[nodeID]), hold = E4_NEXTNONE;
	     m != E4_NEXTNONE;
	     m = (int) pNext(markers[m])) {
	    hold = m;
	}
	if (hold == E4_NEXTNONE) {
	    return NULL;
	}
	return GetMarker(m);
    }

    /*
     * We want the "nth" marker, where "nth" is one or larger.
     */

    for (i = 1, m = (int) pNodeMarkers(nodes[nodeID]);
	 (i < nth) && (m != E4_NEXTNONE);
	 i++, m = (int) pNext(markers[m]));
    if ((i < nth) || (m == E4_NEXTNONE)) {
	return NULL;
    }
    return GetMarker(m);
}

/*
 * Retrieve the ID of the first and last vertex in a node.
 */

int
e4_MetakitStorageImpl::GetFirstVertexID(int nodeID) const 
{
    return (int) pFirstVertex(nodes[nodeID]);
}

int 
e4_MetakitStorageImpl::GetLastVertexID(int nodeID) const 
{
    return (int) pLastVertex(nodes[nodeID]);
}

/*
 * Retrieve an e4_VertexImpl instance for the vertex following the one
 * identified by the given ID in its node.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertex(int vertexID, e4_VisitMethod vm,
				      int vf, int nameID, int nodeID,
				      e4_VertexType typeID) const
{
    switch (vm) {
    case E4_VMSTORAGE:
        return FindNextVertexStorage(vertexID, vf, nameID, typeID);
    case E4_VMNODE:
        return FindNextVertexNode(vertexID, vf, nameID, typeID, nodeID);
    case E4_VMNODERANDOM:
        return FindNextVertexNodeRandom(vertexID, vf, nameID, typeID, nodeID);
    case E4_VMUNKNOWN:
    case E4_VMLASTMETHOD:
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices of a storage according to some given
 * constraints.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorage(int vertexID, int vf, int nameID,
					     e4_VertexType typeID) const
{
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    switch (vf) {
    case E4_VFNONE:
        return FindNextVertexStorageNone(vertexID);
    case E4_VFNAME:
        return FindNextVertexStorageName(vertexID, nameID);
    case E4_VFTYPE:
        return FindNextVertexStorageType(vertexID, typeID);
    case (E4_VFNAME | E4_VFTYPE):
        return FindNextVertexStorageBoth(vertexID, nameID, typeID);
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices of a storage.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageNone(int vertexID) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if (((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices of a storage with a given name.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageName(int vertexID, 
						 int nameID) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
	    ((int) pNameID(vertices[i]) == nameID)) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices of a storage with a given type.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageType(int vertexID,
						 e4_VertexType vt) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
            return GetVertex(i);
        }
    }

    return NULL;
}

/*
 * Helper method to visit all vertices of a storage with a given name and type.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageBoth(int vertexID,
						 int nameID,
						 e4_VertexType vt) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNameID(vertices[i]) == nameID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNode(int vertexID, int vf, int nameID,
					  e4_VertexType vt, int nodeID) const
{
    switch (vf) {
    case E4_VFNONE:
        return FindNextVertexNodeNone(vertexID, nodeID);
    case E4_VFNAME:
        return FindNextVertexNodeName(vertexID, nameID, nodeID);
    case E4_VFTYPE:
        return FindNextVertexNodeType(vertexID, vt, nodeID);
    case (E4_VFNAME | E4_VFTYPE):
        return FindNextVertexNodeBoth(vertexID, nameID, vt, nodeID);
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices in a node without constraints.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeNone(int vertexID,
                                             int nodeID) const
{
    vertexID = FindNextVertexIndexInNode(vertexID, nodeID);
    if (vertexID == E4_NEXTNONE) {
        return NULL;
    }
    return GetVertex(vertexID);
}

/*
 * Helper method to visit all vertices in a node with a given name.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeName(int vertexID, int nameID,
					      int nodeID) const
{
    int i;
    
    for (i = FindNextVertexIndexInNode(vertexID, nodeID);
         i != E4_NEXTNONE;
         i = (int) pNext(vertices[i])) {
        if ((int) pNameID(vertices[i]) == nameID) {
	    return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node with a given type.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeType(int vertexID, e4_VertexType vt,
					      int nodeID) const
{
    int i;

    for (i = FindNextVertexIndexInNode(vertexID, nodeID);
         i != E4_NEXTNONE;
         i = (int) pNext(vertices[i])) {
        if ((e4_VertexType) (int) pVertexType(vertices[i]) == vt) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node with a given type and name.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeBoth(int vertexID, int nameID,
					      e4_VertexType vt, 
					      int nodeID) const
{
    int i;
    
    for (i = FindNextVertexIndexInNode(vertexID, nodeID);
         i != E4_NEXTNONE;
         i = (int) pNext(vertices[i])) {
        if (((int) pNameID(vertices[i]) == nameID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeRandom(int vertexID, int vf,
						int nameID, e4_VertexType vt,
						int nodeID) const
{
    switch (vf) {
    case E4_VFNONE:
        return FindNextVertexNRNone(vertexID, nodeID);
    case E4_VFNAME:
        return FindNextVertexNRName(vertexID, nameID, nodeID);
    case E4_VFTYPE:
        return FindNextVertexNRType(vertexID, vt, nodeID);
    case (E4_VFNAME | E4_VFTYPE):
        return FindNextVertexNRBoth(vertexID, nameID, vt, nodeID);
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices in a node in some random order,
 * with no constraints.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRNone(int vertexID, int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
	    ((int) pNodeID(vertices[i]) == nodeID)) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices with a given name in a supplied node
 * in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRName(int vertexID, int nameID, 
					    int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNodeID(vertices[i]) == nodeID) &&
            ((int) pNameID(vertices[i]) == nameID)) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices with a given type in a supplied node
 * in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRType(int vertexID, e4_VertexType vt,
					    int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNodeID(vertices[i]) == nodeID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices with a given name and type in a supplied
 * node in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRBoth(int vertexID, int nameID, 
					    e4_VertexType vt, int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
	if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNodeID(vertices[i]) == nodeID) &&
            ((int) pNameID(vertices[i]) == nameID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
            return GetVertex(i);
        }
    }
    return NULL;
}
    
/*
 * Helper method to find the vertexID with the next rank after a given
 * vertexID in a node identified by a supplied nodeID.
 */

int
e4_MetakitStorageImpl::FindNextVertexIndexInNode(int vertexID, 
						 int nodeID) const
{
    if (vertexID == E4_VERTEXNOTFOUND) {
        if (nodeID == E4_NODENOTFOUND) {
            return E4_NEXTNONE;
        }
        return pFirstVertex(nodes[nodeID]);
    }
    if (((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0) {
        return E4_NEXTNONE;
    }
    if ((int) pNodeID(vertices[vertexID]) != nodeID) {
	return E4_NEXTNONE;
    }
    return (int) pNext(vertices[vertexID]);
}

/*
 * Given a marker ID, retrieve the instance of the next marker after
 * the given one in this storage.
 */

e4_MarkerImpl *
e4_MetakitStorageImpl::FindNextMarker(int markerID) const
{
    int i, l;

    if (markerID == E4_MARKERNOTFOUND) {
	markerID = -1;
    }
    for (l = markers.GetSize(), i = markerID + 1;
	 i < l;
	 i++) {
	if (((int) pFlags(markers[i]) & E4_INUSE) == E4_INUSE) {
	    return GetMarker(i);
	}
    }
    return NULL;
}

/*
 * Given a node ID, retrieve the instance of the next node after
 * the given one in this storage.
 */

e4_NodeImpl *
e4_MetakitStorageImpl::FindNextNode(int nodeID) const
{
    int i, l;
    e4_NodeImpl *n;

    if (nodeID == E4_NODENOTFOUND) {
	nodeID = -1;
    }
    for (l = nodes.GetSize(), i = nodeID + 1;
	 i < l;
	 i++) {
	if (((int) pFlags(nodes[i]) & E4_INUSE) == E4_INUSE) {
	    (void) GetNode(i, n);
	    return n;
	}
    }
    return NULL;
}

/*
 * Given a node ID, retrieve the number of vertices in that node.
 */

int 
e4_MetakitStorageImpl::VertexCountFromNodeID(int nodeID) const
{
    return (int) pVertexCount(nodes[nodeID]);
}

/*
 * Given a vertex ID retrieve the node containing that vertex.
 */

e4_NodeImpl *
e4_MetakitStorageImpl::ContainingNodeFromVertexID(int vertexID) const 
{
    e4_NodeImpl *n;
    int i;

    i = (int) pNodeID(vertices[vertexID]);
    if (GetNode(pNodeID(vertices[vertexID]), n) == false) {
	return NULL;
    }
    return n;
}

/*
 * Check whether a marker ID is legal.
 */

bool
e4_MetakitStorageImpl::IsLegalMarkerID(int index) const
{
    return ((index < 0) || (index >= markers.GetSize()) ||
            (((int) pFlags(markers[index]) & E4_INUSE) == 0))
      ? false
      : true;
}

/*
 * Check whether a vertex ID is legal.
 */

bool
e4_MetakitStorageImpl::IsLegalVertexID(int index) const
{
    return ((index < 0) || (index >= vertices.GetSize()) ||
	    (((int) pFlags(vertices[index]) & E4_INUSE) == 0))
      ? false
      : true;
}

/*
 * Retrieve the vertex type, the ID of the containing node or the name
 * of the vertex, given a vertex ID.
 */

e4_VertexType
e4_MetakitStorageImpl::VertexTypeFromVertexID(int index) const 
{
    return (e4_VertexType) (int) pVertexType(vertices[index]);
}

int 
e4_MetakitStorageImpl::NodeIDFromVertexID(int index) const 
{
    return (int) pNodeID(vertices[index]);
}

const char *
e4_MetakitStorageImpl::VertexNameFromVertexID(int index) const
{
    const char *nm;

    if (GetName(pNameID(vertices[index]), nm) != true) {
	return NULL;
    }
    return nm;
}

/*
 * Find the name ID of a vertex with a given vertex ID.
 */

int
e4_MetakitStorageImpl::NameIDFromVertexID(int vertexID) const 
{
    return (int) pNameID(vertices[vertexID]);
}

/*
 * Rename a vertex given a vertex ID.
 */

bool
e4_MetakitStorageImpl::RenameVertexByVertexID(int vertexID, 
					      const char *newname)
{
    pNameID(vertices[vertexID]) = InternName(newname);
    return true;
}

/*
 * Add the marker identified by "markerID" to the list of markers that
 * mark the node identified by "nodeID".
 */

bool
e4_MetakitStorageImpl::AddNodeMarker(int markerID, int nodeID)
{
    int m;

    /*
     * The marker is already allocated in the "e4GraphMarkers" view.
     * We do not need to allocate more space.
     */

    /*
     * If the marker is currently marking a node, ensure that that
     * node is unmarked.
     */

    if (!DetachMarker(markerID)) {
	return false;
    }

    /*
     * Add the marker as the last one in the list of markers that mark
     * this node. First find the last marker in the list.
     */

    for (m = (int) pNodeMarkers(nodes[nodeID]);
	 m != E4_NEXTNONE;
	 m = (int) pNext(markers[m])) {
	if ((int) pNext(markers[m]) == E4_NEXTNONE) {
	    break;
	}
    }

    /*
     * If there are no markers in the list of markers for this node,
     * simply add this marker as the first. Otherwise add the marker
     * as the last element in the list.
     */

    if (m == E4_NEXTNONE) {
	pNodeMarkers(nodes[nodeID]) = markerID;
    } else {
	pNext(markers[m]) = markerID;
    }
    pRoot(markers[markerID]) = nodeID;
    pNext(markers[markerID]) = E4_NEXTNONE;

    return true;
}

/*
 * Detach a marker from the node it was marking. If this is the last
 * marker marking the node and the node has no parents, then we free
 * the node recursively.
 */

bool
e4_MetakitStorageImpl::DetachMarker(int markerID)
{
    int oldNodeID = (int) pRoot(markers[markerID]);
    int m, hold;

    /*
     * If the marker is not marking any node, just return.
     */

    if (oldNodeID == E4_NEXTNONE) {
	return true;
    }

    /*
     * Sanity checks:
     */

    if ((oldNodeID < 0) ||
	(oldNodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[oldNodeID]) & E4_INUSE) == 0) ||
	(markerID < 0) ||
	(markerID >= markers.GetSize()) ||
	(((int) pFlags(markers[markerID]) & E4_INUSE) == 0)) {
	return false;
    }

    /*
     * Splice this marker out of the list of markers for this node.
     */

    for (m = (int) pNodeMarkers(nodes[oldNodeID]), hold = E4_NEXTNONE;
	 m != E4_NEXTNONE;
	 m = (int) pNext(markers[m])) {
	if (m == markerID) {
	    if (hold == E4_NEXTNONE) {
		pNodeMarkers(nodes[oldNodeID]) = (int) pNext(markers[m]);
	    } else {
		pNext(markers[hold]) = (int) pNext(markers[m]);
	    }

	    /*
	     * The marker has to stay valid so that it can be recycled
	     * or reassigned. Therefore only unset the "root" and "next"
	     * fields.
	     */

	    pRoot(markers[markerID]) = E4_NEXTNONE;
	    pNext(markers[markerID]) = E4_NEXTNONE;
	    
	    return true;
	}

	hold = m;
    }

    /*
     * Did not find the marker in the list.
     */

    return false;
}

/*
 * Detach the nth marker from the list of markers for the node identified
 * by nodeID.
 */

int
e4_MetakitStorageImpl::DetachNthMarker(int nth, int nodeID)
{
    int i, m, hold;

    if ((nodeID < 0) || 
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return E4_MARKERNOTFOUND;
    }

    for (i = 1, m = (int) pNodeMarkers(nodes[nodeID]), 
	   hold = E4_MARKERNOTFOUND;
	 (i < nth) && (m != E4_NEXTNONE);
	 m = (int) pNext(markers[m]), i++) {
	hold = m;
    }

    /*
     * If there are not at least "nth" markers marking this node,
     * fail.
     */

    if ((i < nth) || (m == E4_NEXTNONE)) {
	return E4_MARKERNOTFOUND;
    }

    /*
     * OK, we found the marker. Now splice it out of the list of
     * markers for this node.
     */

    if (hold == E4_MARKERNOTFOUND) {
	pNodeMarkers(nodes[nodeID]) = (int) pNext(markers[m]);
    } else {
	pNext(markers[hold]) = (int) pNext(markers[m]);
    }

    /*
     * The marker has to stay valid so that it can be recycled
     * or reassigned. Therefore only unset the "root" and "next"
     * fields.
     */

    pRoot(markers[m]) = E4_NEXTNONE;
    pNext(markers[m]) = E4_NEXTNONE;

    return m;
}

/*
 * Add the node identified by "parentID" to the list of nodes that
 * contain the node identified by "childID".
 *
 * NOTE: A parent node can contain a child node multiple times, hence we
 * maintain a reference count to efficienty represent this multiple
 * containment.
 */

void
e4_MetakitStorageImpl::AddParent(int childID, int parentID) 
{
    int i, back, pid;

    /*
     * Increment the reference count for the child node because
     * it gets a new parent.
     */

    pRefCount(nodes[childID]) = (int) pRefCount(nodes[childID]) + 1;

    /*
     * See if the parentID is already present in the list
     * of parents for the node denoted by childID. In that case
     * increment the number of times the node is a child of this
     * parent, and return.
     */

    for (i = (int) pParentID(nodes[childID]);
	 i != E4_NEXTNONE;
	 i = (int) pNext(parents[i])) {
	if ((int) pNodeID(parents[i]) == parentID) {
	    pCount(parents[i]) = (int) pCount(parents[i]) + 1;
	    MarkDirty();
	    return;
	}
    }

    /*
     * Did not find the "parentID" in the current list, so
     * add a new row in the "parents" view for this node.
     */

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) == E4_NEXTNONE) {
	MakeParentSpace();
    }

    i = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) = (int) pNext(parents[i]);

    /*
     * Must add the new parent as the last one in the list
     * of parents for the child node.
     */

    pFlags(parents[i]) = E4_INUSE;
    pNodeID(parents[i]) = parentID;
    pCount(parents[i]) = 1;
    pNext(parents[i]) = E4_NEXTNONE;

    /*
     * After this loop, back will contain the last parent in
     * the chain of parents for the child node.
     */

    for (back = E4_NEXTNONE, pid = (int) pParentID(nodes[childID]);
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid])) {
	back = pid;
    }

    /*
     * If back == E4_NEXTNONE it means that the child node does
     * not yet have any parents, hence make this entry the first
     * one. Otherwise, modify the chain pointer of back to point
     * to the new entry.
     */

    if (back == E4_NEXTNONE) {
	pParentID(nodes[childID]) = i;
    } else {
	pNext(parents[back]) = i;
    }

    MarkDirty();
}

/*
 * Remove a parent from the list of parents of a child, and if requested,
 * decrement the reference count for the child.
 */

bool
e4_MetakitStorageImpl::RemoveParent(int childID, int parentID, bool decref)
{
    int cur, prev;

    /*
     * If the parent ID is present in the list of parents of the
     * child, retrieve the parent record. Otherwise, bail out.
     */

    for (cur = (int) pParentID(nodes[childID]), prev = E4_NEXTNONE;
	 cur != E4_NEXTNONE;
	 cur = (int) pNext(parents[cur])) {
	if ((int) pNodeID(parents[cur]) == parentID) {
	    break;
	}
	prev = cur;
    }
    if (cur == E4_NEXTNONE) {
	return false;
    }

    /*
     * We are making changes to the storage, so mark it dirty.
     */

    MarkDirty();

    /*
     * If requested, decrement the reference count for the child node
     * because it loses a parent.
     */

    if (decref) {
	pRefCount(nodes[childID]) = (int) pRefCount(nodes[childID]) - 1;
    }

    /*
     * If the number of times the child node appears in the parent
     * node is greater than one, decrement the count, and remove the
     * first vertex in the parent that has the child as its value.
     */

    if ((int) pCount(parents[cur]) > 1) {
	pCount(parents[cur]) = (int) pCount(parents[cur]) - 1;
	return true;
    }

    /*
     * No, this parent had only one reference to this child, so we must
     * splice out the record from the list of parents. If this is the
     * first record, then update the pParentID vertex in the child
     * node to the next record, otherwise splice it out of the list. In
     * either case, return the record to the freelist of parent records.
     */

    if (prev == E4_NEXTNONE) {
	pParentID(nodes[childID]) = (int) pNext(parents[cur]);
    } else {
	pNext(parents[prev]) = (int) pNext(parents[cur]);
    }
    AddToUnusedParentList(cur);

    return true;
}

/*
 * Remove a parent from the list of parents for a node.
 *
 * NOTE: A parent node can contain a child node multiple times, hence we
 * maintain a reference count to efficiently represent this multiple
 * containment.
 */

bool
e4_MetakitStorageImpl::RemoveParent(int childID, int parentID)
{
    return RemoveParent(childID, parentID, true);
}

/*
 * Detach a parentID from the list of parent IDs for a given
 * node.
 */

bool
e4_MetakitStorageImpl::DetachParent(int childID, int parentID)
{
    /*
     * Sanity checks:
     */

    if ((childID < 0) ||
	(childID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[childID]) & E4_INUSE) == 0) ||
	(parentID < 0) ||
	(parentID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[parentID]) & E4_INUSE) == 0)) {
	return false;
    }

    /*
     * Remove this parent from the child's list of parents.
     */

    return RemoveParent(childID, parentID, true);
}

/*
 * Add items to each specific free list.
 */

void 
e4_MetakitStorageImpl::AddToUnusedMarkersList(int i) 
{
    pNext(markers[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDMARKER]);
    pFlags(markers[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDMARKER]) = i;

    statistics[(int) E4_SPMARKER][(int) E4_SSUSED]--;
    statistics[(int) E4_SPMARKER][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedNodesList(int i) 
{
    pNext(nodes[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]);
    pFlags(nodes[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) = i;

    statistics[(int) E4_SPNODE][(int) E4_SSUSED]--;
    statistics[(int) E4_SPNODE][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedVerticesList(int i) 
{
    pNext(vertices[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]);
    pFlags(vertices[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) = i;

    statistics[(int) E4_SPVERTEX][(int) E4_SSUSED]--;
    statistics[(int) E4_SPVERTEX][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedIntsList(int i)
{
    /*
     * Nothing to do here.
     */
}

void 
e4_MetakitStorageImpl::AddToUnusedDoublesList(int i) 
{
    pNext(doubles[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]);
    pFlags(doubles[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) = i;

    statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED]--;
    statistics[(int) E4_SPDOUBLE][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedStringsList(int i) 
{
    pNext(strings[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]);
    pFlags(strings[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) = i;

    statistics[(int) E4_SPSTRING][(int) E4_SSUSED]--;
    statistics[(int) E4_SPSTRING][(int) E4_SSFREED]++;
}

void 
e4_MetakitStorageImpl::AddToUnusedNamesList(int i) 
{
    pNext(names[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]);
    pFlags(names[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) = i;

    statistics[(int) E4_SPNAME][(int) E4_SSUSED]--;
    statistics[(int) E4_SPNAME][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedBinaryList(int i)
{
    pNext(binary[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]);
    pFlags(binary[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) = i;

    statistics[(int) E4_SPBINARY][(int) E4_SSUSED]--;
    statistics[(int) E4_SPBINARY][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedParentList(int i)
{
    pNext(parents[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]);
    pFlags(parents[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) = i;
}

/*
 * Move the vertex indentified by vertexID into the node identified by
 * nodeID at the given rank.
 */

bool
e4_MetakitStorageImpl::MoveVertex(int vertexID, int nodeID, int rank)
{
    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0) || 
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0) ||
	(rank == 0)) {
	return false;
    }
    if (rank > 1) {
	/*
	 * If we are trying to insert beyond the current end of the
	 * node, prevent it.
	 */

	if ((int) pVertexCount(nodes[nodeID]) < rank - 1) {
	    return false;
	}

	/*
	 * If the requested rank is exactly the new last vertex in
	 * the node, convert it to the "insert as last" case.
	 */

	if ((int) pVertexCount(nodes[nodeID]) == rank - 1) {
	    rank = -1;
	}
    }

    /*
     * Move the vertex.
     */

    SpliceIn(vertexID, SpliceOut(vertexID), nodeID, rank);

    return true;
}

/*
 * Helper functions for moving vertices around.
 */

int
e4_MetakitStorageImpl::SpliceOut(int vertexID)
{
    int nodeID = (int) pNodeID(vertices[vertexID]);
    int nextID, prevID;

    /*
     * Splice the vertex out of its containing node.
     */

    nextID = (int) pNext(vertices[vertexID]);
    prevID = (int) pPrev(vertices[vertexID]);
    if ((int) pFirstVertex(nodes[nodeID]) == vertexID) {
	pFirstVertex(nodes[nodeID]) = nextID;
    }
    if ((int) pLastVertex(nodes[nodeID]) == vertexID) {
	pLastVertex(nodes[nodeID]) = prevID;
    }
    if (nextID != E4_NEXTNONE) {
	pPrev(vertices[nextID]) = prevID;
    }
    if (prevID != E4_NEXTNONE) {
	pNext(vertices[prevID]) = nextID;
    }

    /*
     * Decrement the vertex count for the node out of which we spliced
     * the vertex.
     */

    pVertexCount(nodes[nodeID]) = (int) pVertexCount(nodes[nodeID]) - 1;

    /*
     * Finally return the nodeID for the parent from which we just
     * spliced this vertex out.
     */

    return nodeID;
}

void
e4_MetakitStorageImpl::SpliceIn(int vertexID, int oldParent,
				int newParent, int rank)
{
    int cnt, childID, nextID, prevID;

    switch (rank) {
    case 1:
	/*
	 * Insert the new vertex as the first vertex in this node.
	 */

	pNext(vertices[vertexID]) = nextID = 
	  (int) pFirstVertex(nodes[newParent]);
	if (nextID != E4_NEXTNONE) {
	    pPrev(vertices[nextID]) = vertexID;
	}
	pPrev(vertices[vertexID]) = E4_NEXTNONE;
	if ((int) pLastVertex(nodes[newParent]) == E4_NEXTNONE) {
	    pLastVertex(nodes[newParent]) = vertexID;
	}
	pFirstVertex(nodes[newParent]) = vertexID;
	break;

    case -1: 
	/* 
	 * Insert the new vertex as the last vertex in this node.
	 */

	pNext(vertices[vertexID]) = E4_NEXTNONE;
	pPrev(vertices[vertexID]) = prevID = 
	  (int) pLastVertex(nodes[newParent]);
	if (prevID == E4_NEXTNONE) {
	    pFirstVertex(nodes[newParent]) = vertexID;
	} else {
	    pNext(vertices[prevID]) = vertexID;
	}
	pLastVertex(nodes[newParent]) = vertexID;
	break;

    default:
	/*
	 * Insert the new vertex at the specified rank position.
	 */

	for (prevID = (int) pFirstVertex(nodes[newParent]), cnt = 1;
	     cnt < rank - 1;
	     prevID = (int) pNext(vertices[prevID]), cnt++) {
	}
	pNext(vertices[vertexID]) = nextID = (int) pNext(vertices[prevID]);
	pPrev(vertices[vertexID]) = prevID;
	if (nextID != E4_NEXTNONE) {
	    pPrev(vertices[nextID]) = vertexID;
	}
	pNext(vertices[prevID]) = vertexID;
	break;
    }

    /*
     * Special handling if the vertex contains a node: in that case
     * we must set its parent to the node into which the vertex was
     * spliced. Order is important here: we want to add the new parent
     * before removing the old parent so that the parent count for
     * the child node never reaches zero.
     */

    if ((oldParent != newParent) &&
	((e4_VertexType) (int) pVertexType(vertices[vertexID]) == E4_VTNODE)) {
	childID = (int) pRowID(vertices[vertexID]);
	AddParent(childID, newParent);
	(void) RemoveParent(childID, oldParent, true);
    }

    /*
     * Update the newParent for the moved vertex.
     */

    pNodeID(vertices[vertexID]) = newParent;

    /*
     * Finally increment the vertex count in this node to account for
     * the newly added vertex.
     */

    pVertexCount(nodes[newParent]) = (int) pVertexCount(nodes[newParent]) + 1;
}

/*
 * Copy all e4Graph data from this storage to another storage.
 */

bool
e4_MetakitStorageImpl::CopyTo(e4_StorageImpl *osp)
{
    e4_MetakitStorageImpl *omsp;
    int i, l;

    /*
     * Check the other storage is using the Metakit storage driver. If
     * not, bail out.
     */

    if (strcmp(GetDriver(), E4_METAKIT) != 0) {
	return false;
    }
    omsp = (e4_MetakitStorageImpl *) osp;

    /*
     * Empty out the cache and data of the other storage.
     */

    omsp->CleanUp();
    if (!omsp->MakeEmpty()) {
	return false;
    }

    /*
     * Copy all tables from this storage to the other storage.
     */

    omsp->markers.SetSize(markers.GetSize());
    omsp->nodes.SetSize(nodes.GetSize());
    omsp->vertices.SetSize(vertices.GetSize());
    omsp->doubles.SetSize(doubles.GetSize());
    omsp->strings.SetSize(strings.GetSize());
    omsp->binary.SetSize(binary.GetSize());
    omsp->names.SetSize(names.GetSize());
    omsp->unused.SetSize(unused.GetSize());
    omsp->parents.SetSize(parents.GetSize());

    for (i = 0, l = markers.GetSize(); i < l; i++) {
	omsp->markers[i] = markers[i];
    }
    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	omsp->nodes[i] = nodes[i];
    }
    for (i = 0, l = vertices.GetSize(); i < l; i++) {
	omsp->vertices[i] = vertices[i];
    }
    for (i = 0, l = doubles.GetSize(); i < l; i++) {
	omsp->doubles[i] = doubles[i];
    }
    for (i = 0, l = strings.GetSize(); i < l; i++) {
	omsp->strings[i] = strings[i];
    }
    for (i = 0, l = binary.GetSize(); i < l; i++) {
	omsp->binary[i] = binary[i];
    }
    for (i = 0, l = names.GetSize(); i < l; i++) {
	omsp->names[i] = names[i];
    }
    for (i = 0, l = unused.GetSize(); i < l; i++) {
	omsp->unused[i] = unused[i];
    }
    for (i = 0, l = parents.GetSize(); i < l; i++) {
	omsp->parents[i] = parents[i];
    }

    /*
     * Finally initialize the new data in the other storage.
     */

    omsp->Initialize();

    return true;
}

/*
 * Destroy the physical storage and make this instance invalid.
 */

void
e4_MetakitStorageImpl::Destroy()
{
    /*
     * First of all, clean up the state associated with this storage.
     */

    CleanUp();

    /*
     * Destroying a Metakit storage is a two step process. First
     * we have to discard the last reference to the storage, to
     * give Metakit a chance to commit the storage. Then, we unlink
     * the file containing the storage so that it will no longer
     * be present on the disk. The order of these operations is
     * important: if you reverse the order the effect of unlinking
     * the file will be undone when Metakit commits the file and
     * makes it appear again!
     */

    if (storage != NULL) {
	delete storage;
	storage = NULL;
    }
    if (name != NULL) {
	unlink(name);
    }
}

/*
 * How many parents does the node with the given ID have?
 */

int
e4_MetakitStorageImpl::ParentCount(int nodeID) const
{
    int count, pid;

    if ((nodeID < 0) || (nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }

    /*
     * Count the number of different nodes that are parents of
     * the node with the given node ID.
     */

    for (count = 0, pid = (int) pParentID(nodes[nodeID]);
	 pid != E4_NEXTNONE;
	 count++, pid = (int) pNext(parents[pid]));

    return count;
}

/*
 * How many vertices contain the node with the given nodeID as their value?
 */

int
e4_MetakitStorageImpl::OccurrenceCount(int nodeID) const
{
    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }
    return (int) pRefCount(nodes[nodeID]);
}

/*
 * How many vertices in the node identified by the given parent ID contain
 * as their value the node denoted by the child ID?
 */

int
e4_MetakitStorageImpl::OccurrenceCount(int nodeID, int parentID) const
{
    int pid;

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0) ||
	(parentID < 0) || 
	(parentID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[parentID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }

    /*
     * Search for the parent record for the child node that contains
     * the parent node information:
     */

    for (pid = (int) pParentID(nodes[nodeID]);
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid])) {
	if ((int) pNodeID(parents[pid]) == parentID) {
	    return (int) pCount(parents[pid]);
	}
    }

    return E4_NODENOTFOUND;
}

/*
 * What is the rank of the parent node with the given parent ID for the
 * child node with the given child ID?
 */

int
e4_MetakitStorageImpl::ParentRank(int nodeID, int parentID) const
{
    int pid, rank;

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0) ||
	(parentID < 0) || 
	(parentID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[parentID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }

    /*
     * Search for the parent record for the child node that contains
     * the parent node information:
     */

    for (pid = (int) pParentID(nodes[nodeID]), rank = 1;
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid]), rank++) {
	if ((int) pNodeID(parents[pid]) == parentID) {
	    return rank;
	}
    }

    return E4_NODENOTFOUND;
}

/*
 * Return (initialize on demand) the hash code for this storage.
 */

int
e4_MetakitStorageImpl::HashCode() const
{
    return (int) pFirst(unused[E4_GRAPHHASHCODE]);
}

/*
 * Make this storage empty.
 */

bool
e4_MetakitStorageImpl::RemoveEverything()
{
    int i;

    if (storage != NULL) {
	markers.SetSize(0);
	nodes.SetSize(0);
	vertices.SetSize(0);
	doubles.SetSize(0);
	strings.SetSize(0);
	binary.SetSize(0);
	names.SetSize(0);
	parents.SetSize(0);

	pFirst(unused[E4_GRAPHSTORAGEMAJORVER]) = E4_MKSTORAGE_MAJORVER;
	pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = E4_MKSTORAGE_MINORVER;

	/*
	 * Do not reset the vertex that determines whether we allow
	 * directed graphs, so that this setting persists across
	 * emptying the storage.
	 */

	for (i = E4_GRAPHFIRSTUNUSEDMARKER; i < E4_GRAPHUNUSEDSIZE; i++) {
	    pFirst(unused[i]) = E4_NEXTNONE;
	}

	MarkDirty();
	HashCleanup();

	statistics[(int) E4_SPMARKER][(int) E4_SSFREED] =
	  statistics[(int) E4_SPMARKER][(int) E4_SSUSED];
	statistics[(int) E4_SPMARKER][(int) E4_SSUSED] = 0;
	statistics[(int) E4_SPMARKER][(int) E4_SSAVAIL] = 0;

	statistics[(int) E4_SPNODE][(int) E4_SSFREED] =
	  statistics[(int) E4_SPNODE][(int) E4_SSUSED];
	statistics[(int) E4_SPNODE][(int) E4_SSUSED] = 0;
	statistics[(int) E4_SPNODE][(int) E4_SSAVAIL] = 0;

	statistics[(int) E4_SPVERTEX][(int) E4_SSFREED] =
	  statistics[(int) E4_SPVERTEX][(int) E4_SSUSED];
	statistics[(int) E4_SPVERTEX][(int) E4_SSUSED] = 0;
	statistics[(int) E4_SPVERTEX][(int) E4_SSAVAIL] = 0;

	statistics[(int) E4_SPNAME][(int) E4_SSFREED] =
	  statistics[(int) E4_SPNAME][(int) E4_SSUSED];
	statistics[(int) E4_SPNAME][(int) E4_SSUSED] = 0;
	statistics[(int) E4_SPNAME][(int) E4_SSAVAIL] = 0;

	statistics[(int) E4_SPSTRING][(int) E4_SSFREED] =
	  statistics[(int) E4_SPSTRING][(int) E4_SSUSED];
	statistics[(int) E4_SPSTRING][(int) E4_SSUSED] = 0;
	statistics[(int) E4_SPSTRING][(int) E4_SSAVAIL] = 0;

	statistics[(int) E4_SPDOUBLE][(int) E4_SSFREED] =
	  statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED];
	statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED] = 0;
	statistics[(int) E4_SPDOUBLE][(int) E4_SSAVAIL] = 0;

	statistics[(int) E4_SPBINARY][(int) E4_SSFREED] =
	  statistics[(int) E4_SPBINARY][(int) E4_SSUSED];
	statistics[(int) E4_SPBINARY][(int) E4_SSUSED] = 0;
	statistics[(int) E4_SPBINARY][(int) E4_SSAVAIL] = 0;
    }
    return true;
}

/*
 * Get a value for a statistic being collected for this storage.
 */

bool
e4_MetakitStorageImpl::GetStatistic(e4_Space sp, e4_SpaceStat st, int &v)
{
    /*
     * Select the statistic to retrieve.
     */

    if (((int) sp < 0) || ((int) sp >= (int) E4_SPLAST) ||
	((int) st < 0) || ((int) st >= (int) E4_SSLAST)) {
	return false;
    }
    v = statistics[(int) sp][(int) st];
    return true;
}

/*
 * Compute the number of slots available in a specific space in the storage.
 */

void
e4_MetakitStorageImpl::UpdateAvailable(e4_Space sp)
{
    /* TODO: Implement! */
}

/*
 * Compute the number of slots used in a specific space in the storage.
 */

void
e4_MetakitStorageImpl::UpdateUsed(e4_Space sp)
{
    /* TODO: Implement! */
}

/*
 * Deletion marking can start from one of three starting points:
 *
 * - A marker is being deleted.
 * - A node is being deleted.
 * - A vertex is being deleted.
 *
 * The following three methods implement the entry points from the
 * generic e4Graph level into the deletion mechanism.
 */

bool
e4_MetakitStorageImpl::DeleteMarkerByID(int markerID)
{
    if (storage == NULL) {
	return false;
    }
    PrepareDeletionStacks();
    SeedDeletionsFromMarkerID(markerID);
    SpanReachableNodes();
    MarkDeletedVertices();
    FireDeletionEvents();
    DeleteAllMarkedEntities();
    return true;
}

bool
e4_MetakitStorageImpl::DeleteNodeByID(int nodeID)
{
    if (storage == NULL) {
	return false;
    }
    PrepareDeletionStacks();
    SeedDeletionsFromNodeID(nodeID);
    SpanReachableNodes();
    MarkDeletedVertices();
    FireDeletionEvents();
    DeleteAllMarkedEntities();
    return true;
}

bool
e4_MetakitStorageImpl::DeleteVertexByID(int vertexID)
{
    if (storage == NULL) {
	return false;
    }
    PrepareDeletionStacks();
    SeedDeletionsFromVertexID(vertexID);
    SpanReachableNodes();
    MarkDeletedVertices();
    FireDeletionEvents();
    DeleteAllMarkedEntities();
    return true;
}

/*
 * Delete the first vertex in the parent node whose value is the child node.
 */

bool
e4_MetakitStorageImpl::DeleteFirstVertexWithNode(int parentID, int childID)
{
    int i;

    if (storage == NULL) {
	return false;
    }

    /*
     * Find the vertex.
     */

    for (i = (int) pFirstVertex(nodes[parentID]);
	 i != E4_NEXTNONE;
	 i = (int) pNext(vertices[i])) {
	if (((e4_VertexType) (int) pVertexType(vertices[i]) == E4_VTNODE) &&
	    ((int) pRowID(vertices[i]) == childID)) {
	    break;
	}
    }

    /*
     * If we didn't find the vertex, bail out.
     */

    if (i == E4_NEXTNONE) {
	return false;
    }

    /*
     * Found it, so delete this vertex.
     */

    return DeleteVertexByID(i);
}

/*
 * Add this marker to the set of markers to be deleted and
 * add its node to the set of potentially unreachable nodes.
 */

void
e4_MetakitStorageImpl::SeedDeletionsFromMarkerID(int markerID)
{
    int mstate;

    /*
     * Record the deletion of the marker.
     */

    mstate = (int) pFlags(markers[markerID]);
    mstate |= E4_DELETED;
    pFlags(markers[markerID]) = mstate;
}

/*
 * Add this node to the list of nodes to delete, all its markers to the
 * list of markers to delete, and all its vertices to the list of vertices
 * to delete. Remove this node from all its parents.
 *
 * If any of the vertices has an unmarked node as its value, the child
 * node is added to the set of potentially unreachable nodes, for further
 * examination.
 */

void
e4_MetakitStorageImpl::SeedDeletionsFromNodeID(int nodeID)
{
    int markerID, vertexID, pid, parentID, mstate, nstate, vstate;

    /*
     * Record the deletion of this node.
     */

    nstate = (int) pFlags(nodes[nodeID]);
    nstate |= E4_DELETED;
    pFlags(nodes[nodeID]) = nstate;

    /*
     * Delete all the markers marking this node.
     */

    for (markerID = (int) pNodeMarkers(nodes[nodeID]);
	 markerID != E4_NEXTNONE;
	 markerID = (int) pNext(markers[markerID])) {

	/*
	 * Record its deletion.
	 */

	mstate = (int) pFlags(markers[markerID]);
	mstate |= E4_DELETED;
	pFlags(markers[markerID]) = mstate;
    }

    /*
     * Now remove this node from all its parents.
     */

    for (pid = (int) pParentID(nodes[nodeID]);
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid])) {
	parentID = (int) pNodeID(parents[pid]);
	for (vertexID = (int) pFirstVertex(nodes[parentID]);
	     vertexID != E4_NEXTNONE;
	     vertexID = (int) pNext(vertices[vertexID])) {
	    if (((e4_VertexType) (int) pVertexType(vertices[vertexID]) == 
		 E4_VTNODE) &&
		((int) pRowID(vertices[vertexID]) == nodeID)) {

		/*
		 * Mark this vertex as deleted.
		 */

		vstate = (int) pFlags(vertices[vertexID]);
		vstate |= E4_DELETED;
		pFlags(vertices[vertexID]) = vstate;
	    }
	}
    }
}

/*
 * Add this vertex to the set of vertices to be deleted. If the vertex value
 * is a node, add it to the set of potentially unreachable vertices to be
 * checked further.
 */

void
e4_MetakitStorageImpl::SeedDeletionsFromVertexID(int vertexID)
{
    int nodeID, state;

    /*
     * Record the deletion of this vertex.
     */

    state = (int) pFlags(vertices[vertexID]);
    state |= E4_DELETED;
    pFlags(vertices[vertexID]) = state;

    /*
     * If the vertex value is a node, classify the node.
     */

    if ((e4_VertexType) (int) pVertexType(vertices[vertexID]) == E4_VTNODE) {
	nodeID = (int) pRowID(vertices[vertexID]);

	/*
	 * If the node has no legal markers and no legal parents,
	 * it is unreachable. Therefore set the flag indicating that
	 * unreachable nodes may be present (in this case we know that
	 * there definitely is at least one unreachable nodes, this one).
	 */

	if (!NodeHasLegalMarkers(nodeID) &&
	    !ChildIsReachableFromParents(nodeID)) {
	    state = (int) pFlags(nodes[nodeID]);
	    state |= E4_DELETED;
	    pFlags(nodes[nodeID]) = state;
	}
    }
}

/*
 * Helper method that determines if a node is reachable by looking for
 * legal markers on this node.
 */

bool
e4_MetakitStorageImpl::NodeHasLegalMarkers(int nodeID)
{
    int markerID, state;

    for (markerID = (int) pNodeMarkers(nodes[nodeID]);
	 markerID != E4_NEXTNONE;
	 markerID = (int) pNext(markers[markerID])) {
	state = (int) pFlags(markers[markerID]);

	/*
	 * Found a marker that's not deleted, so it's a legal
	 * marker.
	 */

	if ((state & E4_DELETED) == 0) {
	    return true;
	}
    }

    /*
     * No legal marker found.
     */

    return false;
}

/*
 * Method to determine if a child node is reachable from any of its
 * parents.
 */

bool
e4_MetakitStorageImpl::ChildIsReachableFromParents(int childID)
{
    int parentID, pid;

    for (pid = (int) pParentID(nodes[childID]);
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid])) {
	parentID = (int) pNodeID(parents[pid]);
	if (ChildIsReachableFromParent(childID, parentID)) {
	    return true;
	}
    }
    return false;
}

/*
 * Helper method to determine whether a child node is reachable from
 * a parent node.
 */

bool
e4_MetakitStorageImpl::ChildIsReachableFromParent(int childID, int parentID)
{
    int vertexID;

    for (vertexID = (int) pFirstVertex(nodes[parentID]);
	 vertexID != E4_NEXTNONE;
	 vertexID = (int) pNext(vertices[vertexID])) {
      
	/*
	 * Do not look at deleted vertices.
	 */

	if (((int) pFlags(vertices[vertexID]) & E4_DELETED) == E4_DELETED) {
	    continue;
	}

	/*
	 * If this is a node-valued vertex and its value is childID,
	 * we found that childID is reachable from parentID.
	 */

	if (((e4_VertexType) (int) pVertexType(vertices[vertexID]) == 
	     E4_VTNODE) &&
	    ((int) pRowID(vertices[vertexID]) == childID)) {
	    return true;
	}
    }

    return false;
}

/*
 * Delete all marked entities.
 */

void
e4_MetakitStorageImpl::DeleteAllMarkedEntities()
{
    /*
     * Order is important: first delete all vertices and markers, only then
     * all marked nodes, because vertex and marker deletion needs access to
     * the nodes.
     */

    DeleteMarkedVertices();
    DeleteMarkedMarkers();
    DeleteMarkedNodes();
}

/*
 * Delete all markers that have been marked as deleted.
 */

void
e4_MetakitStorageImpl::DeleteMarkedMarkers()
{
    int l, m, markerID, nodeID, hold;

    for (markerID = 0, l = markers.GetSize(); markerID < l; markerID++) {
	if (((int) pFlags(markers[markerID]) & E4_DELETED) == 0) {
	    continue;
	}

	/*
	 * Found a marker being deleted.
	 *
	 * If the node marked by this marker is also being deleted, do not
	 * bother splicing the marker out of its marker list.
	 */

	nodeID = (int) pRoot(markers[markerID]);
	if (((int) pFlags(nodes[nodeID]) & (E4_REACHABLE | E4_INUSE)) ==
	    (E4_REACHABLE | E4_INUSE)) {

	    /*
	     * No, we must splice this marker out of the list of markers for
	     * the node.
	     */

	    for (m = (int) pNodeMarkers(nodes[nodeID]), hold = E4_NEXTNONE;
		 m != E4_NEXTNONE;
		 m = (int) pNext(markers[m])) {

		/*
		 * If we found it, splice it out.
		 */

		if (m == markerID) {
		    if (hold == E4_NEXTNONE) {
			pNodeMarkers(nodes[nodeID]) = 
			  (int) pNext(markers[markerID]);
		    } else{
			pNext(markers[hold]) = (int) pNext(markers[markerID]);
		    }

		    break;
		}
		hold = m;
	    }
	}

	/*
	 * Recycle the row occuppied by this marker.
	 */

	ForgetMarker(markerID);
	AddToUnusedMarkersList(markerID);
    }
}

/*
 * Delete all nodes that have been marked as deleted.
 */

void
e4_MetakitStorageImpl::DeleteMarkedNodes()
{
    int c, l, nodeID, parentID, nextParentID, state;

    for (c = 0, nodeID = 0, l = nodes.GetSize(); nodeID < l; nodeID++) {
	state = (int) pFlags(nodes[nodeID]);

	/*
	 * Check if the node is in use now.
	 */

	if ((state & E4_INUSE) == 0) {
	    continue;
	}

	/*
	 * Clear reachable state if set.
	 */

	if ((state & E4_REACHABLE) == E4_REACHABLE) {
	    state &= (~(E4_REACHABLE));
	    pFlags(nodes[nodeID]) = state;
	    continue;
	}

	/*
	 * Found an in-use node being deleted.
	 *
	 * Recycle the parent records of this node.
	 */

	for (parentID = (int) pParentID(nodes[nodeID]);
	     parentID != E4_NEXTNONE;
	     parentID = nextParentID) {
	    nextParentID = (int) pNext(parents[parentID]);
	    AddToUnusedParentList(parentID);
	}

	/*
	 * Recycle the row occuppied by this node.
	 */

	ForgetNode(nodeID);
	AddToUnusedNodesList(nodeID);

	c++;
    }
}

/*
 * Delete all vertices that have been marked as deleted.
 */

void
e4_MetakitStorageImpl::DeleteMarkedVertices()
{
    int c, l, vertexID;

    for (c = 0, vertexID = 0, l = vertices.GetSize();
	 vertexID < l;
	 vertexID++) {

	if (((int) pFlags(vertices[vertexID]) & (E4_DELETED | E4_INUSE)) !=
	    (E4_DELETED | E4_INUSE)) {
	    continue;
	}

	/*
	 * Found a vertex being deleted.
	 */

	c++;

	switch ((e4_VertexType) (int) pVertexType(vertices[vertexID])) {
	case E4_VTINT:
	    /*
	     * Nothing to do -- value is inline.
	     */

	    break;

	case E4_VTNODE:
	    /*
	     * Remove the parent node from the list of parents for
	     * the child node.
	     */

	    RemoveParent((int) pRowID(vertices[vertexID]),
			 (int) pNodeID(vertices[vertexID]),
			 true);
	    break;

	case E4_VTDOUBLE:
	    /*
	     * Recycle the row containing the 64-bit floating point value.
	     */

	    (void) FreeDouble((int) pRowID(vertices[vertexID]));
	    break;

	case E4_VTSTRING:
	    /*
	     * Recycle the row containing the string value.
	     */

	    (void) FreeString((int) pRowID(vertices[vertexID]));
	    break;

	case E4_VTBINARY:
	    /*
	     * Recycle the row containing the binary value.
	     */

	    (void) FreeBinary((int) pRowID(vertices[vertexID]));
	    break;
	}

	/*
	 * If the node containing this vertex is marked as deleted, don't
	 * bother splicing the vertex out of its vertices list. Otherwise
	 * splice this vertex out of the list of vertices in its node.
	 */

	if (((int) pFlags(nodes[(int) pNodeID(vertices[vertexID])]) &
	     E4_DELETED) ==
	    0) {

	    (void) SpliceOut(vertexID);
	}

	/*
	 * Recycle the row occuppied by this vertex.
	 */

	ForgetVertex(vertexID);
	AddToUnusedVerticesList(vertexID);
    }
}

/*
 * Helper method to ensure that deletion stacks have been allocated.
 */

void
e4_MetakitStorageImpl::PrepareDeletionStacks()
{
    if (idStack1 == NULL) {
	idStack1 = new e4_IntStack();
    }
    if (idStack2 == NULL) {
	idStack2 = new e4_IntStack();
    }
}
    
/*
 * Span all reachable nodes. At the end of this procedure all nodes in
 * reachableNodeIDs have had their child nodes added to the set.
 */

void
e4_MetakitStorageImpl::SpanReachableNodes()
{
    int i, l, nodeID, childID, vertexID, nstate, cstate, mstate, vstate;
    bool changed;
    e4_IntStack *tmpstk;

    /*
     * First stage: seed reachable nodes from the valid markers in
     * this storage.
     */

    for (i = 0, l = markers.GetSize();
	 i < l;
	 i++) {
	mstate = (int) pFlags(markers[i]);
	if ((mstate & E4_INUSE) == E4_INUSE) {

	    /*
	     * Found an in-use marker. See if it is has been deleted.
	     */

	    if ((mstate & E4_DELETED) == 0) {

		/*
		 * This marker is legal, so use its marked node as
		 * a reachable root, if this node is not being
		 * deleted.
		 */

		nodeID = (int) pRoot(markers[i]);
		nstate = (int) pFlags(nodes[nodeID]);
		if ((nstate & E4_DELETED) == 0) {
		    idStack1->Push(nodeID);
		}
	    }
	}
    }

    /*
     * Second stage: iterate until changed == false. If a node is found
     * that has not had its child nodes added to the set of reachable nodes
     * then change is set to true and the child nodes are added.
     */

    do {
	/*
	 * No change so far.
	 */

	changed = false;

	/*
	 * Iterate over all newly discovered reachable nodes.
	 */

	while (idStack1->Pop(&nodeID)) {

	    /*
	     * Check if this node is not yet in the reachableNodeIDs table.
	     */

	    nstate = (int) pFlags(nodes[nodeID]);
	    if ((nstate & E4_REACHABLE) == 0) {

		/*
		 * Found a node whose children have not yet been
		 * added to the spanning set. Add it to the set of
		 * reachable nodes and add its children to the set
		 * of nodes to be examined.
		 */

		changed = true;

		nstate |= E4_REACHABLE;
		pFlags(nodes[nodeID]) = nstate;

		/*
		 * Add the child nodes to the table of nodes to examine
		 * in the next iteration.
		 */

		for (vertexID = (int) pFirstVertex(nodes[nodeID]);
		     vertexID != E4_NEXTNONE;
		     vertexID = (int) pNext(vertices[vertexID])) {

		    /*
		     * If this vertex is deleted, don't check its value.
		     */

		    vstate = (int) pFlags(vertices[vertexID]);
		    if ((vstate & E4_DELETED) == E4_DELETED) {
			continue;
		    }

		    /*
		     * See if the value is a node.
		     */

		    if ((e4_VertexType) (int) pVertexType(vertices[vertexID])
			== E4_VTNODE) {

			/*
			 * Found a node. Check if it's being deleted.
			 * If not, add it to the set of nodes to be
			 * explored further.
			 */

			childID = (int) pRowID(vertices[vertexID]);
			cstate = (int) pFlags(nodes[childID]);
			if ((cstate & E4_DELETED) == 0) {
			    idStack2->Push(childID);
			}
		    }
		}
	    }
	}

	/*
	 * Switch stacks. At this point idStack2 contains the newly
	 * reachable nodes whose children may not yet have been
	 * explored, and idStack1 is empty.
	 */

	tmpstk = idStack1;
	idStack1 = idStack2;
	idStack2 = tmpstk;

    } while (changed);
}

/*
 * This procedure collects all vertices that are in unreachable nodes.
 */

void
e4_MetakitStorageImpl::MarkDeletedVertices()
{
    int l, nodeID, vertexID, nstate, vstate;

    for (nodeID = 0, l = nodes.GetSize(); nodeID < l; nodeID++) {

	/*
	 * See if this node is being deleted.
	 */

	nstate = (int) pFlags(nodes[nodeID]);
	if ((nstate & E4_INUSE) == 0) {
	    continue;
	}

	if ((nstate & E4_REACHABLE) == E4_REACHABLE) {
	    continue;
	}

	/*
	 * Found a node being deleted. Mark all its vertices as deleted.
	 */
	
	for (vertexID = (int) pFirstVertex(nodes[nodeID]);
	     vertexID != E4_NEXTNONE;
	     vertexID = (int) pNext(vertices[vertexID])) {
	    vstate = (int) pFlags(vertices[vertexID]);
	    vstate |= E4_DELETED;
	    pFlags(vertices[vertexID]) = vstate;
	}
    }
}

/*
 * Find the ID of the next deleted marker after the given one.
 */

int
e4_MetakitStorageImpl::NextDeletedMarker(int id)
{
    int l, state;

    if (id == E4_NEXTNONE) {
	id = 0;
    } else {
	id++;
    }

    for (l = markers.GetSize(); id < l; id++) {
	state = (int) pFlags(markers[id]);
	if ((state & E4_DELETED) == E4_DELETED) {
	    return id;
	}
    }

    return E4_NEXTNONE;
}

/*
 * Find the ID of the next deleted node after the given one.
 */

int
e4_MetakitStorageImpl::NextDeletedNode(int id)
{
    int l, state;

    if (id == E4_NEXTNONE) {
	id = 0;
    } else {
	id++;
    }

    for (l = nodes.GetSize(); id < l; id++) {
	state = (int) pFlags(nodes[id]);
	if ((state & E4_INUSE) == 0) {
	    continue;
	}
	if ((state & E4_REACHABLE) == E4_REACHABLE) {
	    continue;
	}

	return id;
    }

    return E4_NEXTNONE;
}

/*
 * Find the ID of the next deleted vertex after the given one.
 */

int
e4_MetakitStorageImpl::NextDeletedVertex(int id)
{
    int l, state;

    if (id == E4_NEXTNONE) {
	id = 0;
    } else {
	id++;
    }

    for (l = vertices.GetSize(); id < l; id++) {
	state = (int) pFlags(vertices[id]);
	if ((state & (E4_INUSE|E4_DELETED)) == (E4_INUSE|E4_DELETED)) {
	    return id;
	}
    }

    return E4_NEXTNONE;
}
