/*
 * t4command.cpp --
 *
 *	This file contains the implementation of procedures for
 *	creation of Tcl commands that invoke operations on T4Graph
 *	objects.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * Pattern to create a new namespace for a storage:
 */

static char *nspcPattern = (char *) "namespace eval ::tgraph::%s {}";
static char *mPattern = (char *) "namespace eval ::tgraph::%s::markers {}";
static char *nPattern = (char *) "namespace eval ::tgraph::%s::nodes {}";
static char *vPattern = (char *) "namespace eval ::tgraph::%s::vertices {}";

/*
 * Static functions declared in this file:
 */

static void StorageCmdDeleteProc(ClientData cd);
static void MarkerCmdDeleteProc(ClientData cd);
static void NodeCmdDeleteProc(ClientData cd);
static void VertexCmdDeleteProc(ClientData cd);

/*
 * MakeStorageCommand --
 *
 *	Given a T4Storage object, make a Tcl command that will
 *	invoke operations on that object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May create a new Tcl command. May create a new namespace for
 *	stored procedure commands.
 */

int
MakeStorageCommand(Tcl_Interp *interp, T4Storage *s)
{
    char *nspc;
    char *sn = s->GetName();
    Tcl_Command res;

    nspc = Tcl_Alloc(strlen(sn) + strlen(vPattern) + 10);
    sprintf(nspc, nspcPattern, sn);    
    if (Tcl_Eval(interp, nspc) == TCL_ERROR) {
	Tcl_Free(nspc);
	return TCL_ERROR;
    }
    sprintf(nspc, mPattern, sn);
    if (Tcl_Eval(interp, nspc) == TCL_ERROR) {
	Tcl_Free(nspc);
	return TCL_ERROR;
    }
    sprintf(nspc, nPattern, sn);
    if (Tcl_Eval(interp, nspc) == TCL_ERROR) {
	Tcl_Free(nspc);
	return TCL_ERROR;
    }
    sprintf(nspc, vPattern, sn);
    if (Tcl_Eval(interp, nspc) == TCL_ERROR) {
	Tcl_Free(nspc);
	return TCL_ERROR;
    }
    Tcl_Free(nspc);
    res = Tcl_CreateObjCommand(interp, sn, StorageCmdProc, s,
			       StorageCmdDeleteProc);
    if (res != NULL) {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), sn, -1);
	return TCL_OK;
    }
    return TCL_ERROR;
}

/*
 * MakeMarkerCommand --
 *
 *	Given a T4Marker object ID, make a Tcl command that will
 *	invoke operations on that object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May create a new Tcl command.
 */

char *
MakeMarkerCommand(Tcl_Interp *interp, T4Marker *mp, e4_MarkerUniqueID muid)
{
    int id = muid.GetUniqueID();
    int len;
    char *cn;
    char *sn = mp->GetStorage()->GetName();

    len = strlen("::tgraph::") + strlen(sn) + strlen("::markers::marker") +
	  14 + 1;
    cn = new char[len];
    sprintf(cn, "::tgraph::%s::markers::marker%d", sn, id);
    (void) Tcl_CreateObjCommand(interp, cn, MarkerCmdProc, mp,
				MarkerCmdDeleteProc);
    return cn;
}

/*
 * MakeNodeCommand --
 *
 *	Given a T4Node object ID, make a Tcl command that will
 *	invoke operations on that object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May create a new Tcl command.
 */

char *
MakeNodeCommand(Tcl_Interp *interp, T4Node *np, e4_NodeUniqueID nuid)
{
    int id = nuid.GetUniqueID();
    int len;
    char *cn;
    char *sn = np->GetStorage()->GetName();

    len = strlen("::tgraph::") + strlen(sn) + strlen("::nodes::node") +
	  14 + 1;
    cn = new char[len];
    sprintf(cn, "::tgraph::%s::nodes::node%d", sn, id);
    (void) Tcl_CreateObjCommand(interp, cn, NodeCmdProc, np,
				NodeCmdDeleteProc);
    return cn;
}

/*
 * MakeVertexCommand --
 *
 *	Given a T4Vertex object ID, make a Tcl command that will
 *	invoke operations on that object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May create a new Tcl command.
 */

char *
MakeVertexCommand(Tcl_Interp *interp, T4Vertex *vp, e4_VertexUniqueID vuid)
{
    int id = vuid.GetUniqueID();
    int len;
    char *cn;
    char *sn = vp->GetStorage()->GetName();

    len = strlen("::tgraph::") + strlen(sn) + strlen("::vertices::vertex") +
	  14 + 1;
    cn = new char[len];
    sprintf(cn, "::tgraph::%s::vertices::vertex%d", sn, id);
    (void) Tcl_CreateObjCommand(interp, cn, VertexCmdProc, vp,
				VertexCmdDeleteProc);
    return cn;
}

/*
 * StorageCmdDeleteProc --
 *
 *	This procedure is called when the Tcl command for a storage
 *	is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 */

static void
StorageCmdDeleteProc(ClientData cd)
{
    /* Nothing to do. */
}

/*
 * MarkerCmdDeleteProc --
 *
 *	This procedure is called when the Tcl command for a marker is
 *	deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Removes the cached T4Marker object from the cache in its enclosing
 *	T4Storage object.
 */

static void
MarkerCmdDeleteProc(ClientData cd)
{
    T4Marker *mp = (T4Marker *) cd;
    T4Storage *s = mp->GetStorage();
    e4_Marker m;
    e4_MarkerUniqueID muid;

    mp->ExternalizeMarker(m);
    (void) m.GetUniqueID(muid);
    s->RemoveMarker(muid);
}

/*
 * NodeCmdDeleteProc --
 *
 *	This procedure is called when the Tcl command for a node is
 *	deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Removes the cached T4Node object from the cache in its enclosing
 *	T4Storage object.
 */

static void
NodeCmdDeleteProc(ClientData cd)
{
    T4Node *np = (T4Node *) cd;
    T4Storage *s = np->GetStorage();
    e4_Node n;
    e4_NodeUniqueID nuid;

    np->ExternalizeNode(n);
    (void) n.GetUniqueID(nuid);
    s->RemoveNode(nuid);
}

/*
 * VertexCmdDeleteProc --
 *
 *	This procedure is called when the Tcl command for a vertex
 *	is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Removes the cached T4Vertex object from the cahce in its enclosing
 *	T4Storage object.
 */

static void
VertexCmdDeleteProc(ClientData cd)
{
    T4Vertex *vp = (T4Vertex *) cd;
    T4Storage *s = vp->GetStorage();
    e4_Vertex v;
    e4_VertexUniqueID vuid;

    vp->ExternalizeVertex(v);
    (void) v.GetUniqueID(vuid);
    s->RemoveVertex(vuid);
}

/*
 * DeleteStorageCommand --
 *
 *	Given a T4Storage object that is about to be deleted, delete the
 *	Tcl command created with MakeStorageCommand to operate on it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May delete a Tcl command. May delete the namespace for stored
 *	procedure commands associated with this storage.
 */

void
DeleteStorageCommand(Tcl_Interp *interp, T4Storage *s)
{
    char *sn = s->GetName();
    char *dscmd;
    int len;

    len = strlen(sn) + strlen("tgraph::") + strlen("namespace delete ") + 1;
    dscmd = (char *) Tcl_Alloc(len);

    sprintf(dscmd, "namespace delete tgraph::%s", sn);
    (void) Tcl_Eval(interp, dscmd);

    Tcl_Free(dscmd);

    Tcl_DeleteCommand(interp, sn);
    Tcl_ResetResult(interp);
}

/*
 * DeleteMarkerCommand --
 *
 *	Given a T4Marker object that is about to be deleted, delete the
 *	Tcl command created with MakeMarkerCommand to operate on it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May delete a Tcl command.
 */

void
DeleteMarkerCommand(Tcl_Interp *interp, T4Marker *t)
{
    char *tn = t->GetName();

    Tcl_DeleteCommand(interp, tn);
    Tcl_ResetResult(interp);
}

/*
 * DeleteNodeCommand --
 *
 *	Given a T4Node object that is about to be deleted, delete the
 *	Tcl command created with MakeNodeCommand to operate on it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May delete a Tcl command.
 */

void
DeleteNodeCommand(Tcl_Interp *interp, T4Node *n)
{
    char *nn = n->GetName();

    Tcl_DeleteCommand(interp, nn);
    Tcl_ResetResult(interp);
}

/*
 * DeleteVertexCommand --
 *
 *	Given a T4Vertex object that is about to be deleted, delete the
 *	Tcl command created with MakeVertexCommand to operate on it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May delete a Tcl command.
 */

void
DeleteVertexCommand(Tcl_Interp *interp, T4Vertex *f)
{
    char *fn = f->GetName();

    Tcl_DeleteCommand(interp, fn);
    /*    Tcl_ResetResult(interp); */
}
