/*
 * t4init.cpp --
 *
 *	Implementation of the T4Graph initialization procedures.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * Initialization script for the T4Graph package:
 */

static char *initScript = (char *) "namespace eval ::tgraph {}";

/*
 * Tgraph_RealInit --
 *
 *	This procedure initializes the T4Graph package in an interpreter.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Creates a "t4graph" namespace and several commands within the
 *	namespace in the given interpreter.
 */

extern "C" int
Tgraph_RealInit(Tcl_Interp *interp)
{
    Tcl_HashTable *tblPtr;

    if (Tcl_Eval(interp, initScript) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * Check if there is already a registry of open storages. This can
     * happen if first a storage is transferred into this interpreter and
     * later the TGraph package is loaded into this interpreter.
     */

    if (Tcl_GetAssocData(interp, T4_ASSOCKEY, NULL) == NULL) {

	/*
	 * Create the hash table that contains the known storages in this
	 * interpreter.
	 */

	tblPtr = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(tblPtr, TCL_STRING_KEYS);

	/*
	 * Associate this table with the interpreter, so that the open
	 * command can return the same object each time if multiple open
	 * requests are made on the same storage.
	 */

	Tcl_SetAssocData(interp, T4_ASSOCKEY, OpenStorageInterpDelProc, 
			 tblPtr);
    }

    /*
     * If the interpreter is not safe, create the "open" command:
     */

    if (!Tcl_IsSafe(interp)) {
	(void) Tcl_CreateObjCommand(interp, (char *) "::tgraph::open",
				    OpenStorageProc, NULL,
				    (Tcl_CmdDeleteProc *) NULL);
    }

    /*
     * Create a command to determine whether a storage by a given
     * name is open in this interpreter.
     */

    (void) Tcl_CreateObjCommand(interp, (char *) "::tgraph::isopen",
				IsOpenStorageProc, (ClientData) NULL,
				(Tcl_CmdDeleteProc *) NULL);

    /*
     * Create a command to transfer a storage to a sub-interpreter.
     */

    (void) Tcl_CreateObjCommand(interp, (char *) "::tgraph::transfer",
				TransferStorageProc, (ClientData) NULL,
				(Tcl_CmdDeleteProc *) NULL);

    /*
     * All went well, so declare that we provide the T4Graph package.
     */

    return Tcl_PkgProvide(interp, (char *) "TGraph", (char *) "1.0.0");
}

/*
 * OpenStorageInterpDelProc --
 *
 *	This procedure is called when an interpreter is deleted, to clean
 *	up the state associated with the TGraph package. It closes all open
 *	storages and recycles all objects.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May release memory. May close some storages.
 */

void
OpenStorageInterpDelProc(ClientData cd, Tcl_Interp *interp)
{
    Tcl_HashTable *tblPtr = (Tcl_HashTable *) cd;

    Tcl_DeleteHashTable(tblPtr);
}

