/*
 * t4marker.cpp --
 *
 *	This file implements the command procedure invoked when the
 *	Tcl procedure "t4graph::marker" is called.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * The various sub-commands for "t4graph::marker":
 */

static char *subCommands[] = {
    (char *) "kind", 
    (char *) "name",
    (char *) "storage",
    (char *) "node", 
    (char *) "setnode",
    (char *) "delete",
    (char *) "dispose",
    (char *) "isvalid",
    (char *) "id",
    (char *) "userdata",
    (char *) NULL
};
typedef enum TSubCommands {
     TKind = 0, 
     TName,
     TStorage,
     TNode,
     TSetNode,
     TDelete,
     TDispose,
     TIsValid,
     TId,
     TUserData
} TSubCommands;

/*
 * MarkerCmdProc --
 *
 *	This procedure is invoked when the "t4graph::marker" command
 *	is called.
 *
 * Results:
 *	A standard Tcl result. The interpreter result may contain various
 *	results depending on the sub-command selected.
 *
 * Side effects:
 *	Whatever the selected sub-command does.
 */

int
MarkerCmdProc(ClientData cd, Tcl_Interp *interp, int objc,
	      Tcl_Obj *CONST objv[])
{
    T4Marker *t;
    TSubCommands index;

    /*
     * There must be at least two arguments.
     */

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$marker cmd ?arg ...?");
        return TCL_ERROR;
    }

    /*
     * The client data is the T4Marker for this instance.
     */

    t = (T4Marker *) cd;

    /*
     * Figure out what operation was requested. No default is provided,
     * so the caller must select an operation.
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], subCommands, (char *) "cmd",
			    0, (int *) &index) != TCL_OK) {
	return TCL_ERROR;
    }

    objc -= 2;
    objv += 2;

    /*
     * Invoke the requested operation:
     */

    switch (index) {
    case TKind:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "marker", -1);
	return TCL_OK;
    case TName:
	return t->Name(interp, objc, objv);
    case TStorage:
	Tcl_SetStringObj(Tcl_GetObjResult(interp),
			 t->GetStorage()->GetName(), -1);
	return TCL_OK;
    case TNode:
	return t->MarkedNode(interp, objc, objv);
    case TSetNode:
	return t->SetMarkedNode(interp, objc, objv);
    case TDelete:
	return t->Delete(interp, objc, objv);
    case TDispose:
	return t->Dispose(interp, objc, objv);
    case TIsValid:
	return t->IsValid(interp, objc, objv);
    case TId:
	return t->Id(interp, objc, objv);
    case TUserData:
	return t->UserData(interp, objc, objv);
    default:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "MarkerProc: unreachable code!",
			       (char *) NULL);
	return TCL_ERROR;
    }
}
