/*
 * t4markerrep.cpp --
 *
 *	This file contains the implementation of the class T4Marker
 *	which is defined in .../include/t4graphrep.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * Constructor:
 */

T4Marker::T4Marker(e4_Marker tt, char *name, T4Storage *ss)

{
    SetName(name);
    t = tt;
    s = ss;
}

/*
 * Destructor:
 */

T4Marker::~T4Marker()
{
    t = invalidMarker;
    SetName(NULL);
}

/*
 * ExternalizeMarker --
 *
 *	This procedure gives the caller access to the underlying
 *	e4_Marker object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The caller can now manipulate the underlying e4_Marker
 *	object directly.
 */

void
T4Marker::ExternalizeMarker(e4_Marker &tt)
{
   tt = t;
}

/*
 ****************************************************************************
 *
 * The following methods implement Tcl sub-commands on a T4Marker
 * object.
 *
 ****************************************************************************
 */

/*
 * Name --
 *
 *	Sets the interpreter result to a string, the name of this
 *	marker.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name of this marker.
 *
 * Side effects:
 *	None.
 */

int
T4Marker::Name(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$marker name");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!t.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to the name of this marker.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) t.GetName(), -1);

    return TCL_OK;
}

/*
 * MarkedNode --
 *
 *	Set the interpreter result to the name assigned to the node
 *	marked by this marker.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name assigned to the node marked by
 *	this marker.
 *
 * Side effects:
 *	May export a T4Node object to Tcl.
 */

int
T4Marker::MarkedNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node n;
    T4Node *np;
    e4_NodeUniqueID nuid;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$marker node");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!t.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to obtain the marked node from this marker.
     */

    if ((!t.GetMarkedNode(n)) || (!n.IsValid())) {
	Tcl_AppendResult(interp, 
			 "could not obtain root of marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * See if we have exported this node to Tcl before. If not, create
     * a new T4Node and make it ready to be exported.
     */

    (void) n.GetUniqueID(nuid);
    np = s->GetNodeById(nuid);
    if (np == NULL) {
	np = new T4Node(n, NULL, s);
	np->SetName(MakeNodeCommand(interp, np, nuid));
	s->StoreNode(np, nuid);
    }
    Tcl_SetStringObj(Tcl_GetObjResult(interp), np->GetName(), -1);
    return TCL_OK;
}

/*
 * SetMarkedNode --
 *
 *	Set the interpreter result to the name assigned to the node
 *	given as its argument. Sets the marker to mark that node in
 *	the persistent storage.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name assigned to the given node.
 *
 * Side effects:
 *	The node marked by this marker in the persistent storage may be
 *	modified.
 */

int
T4Marker::SetMarkedNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node n;
    T4Node *np;
    char *nn;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$marker setnode node");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!t.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the node named by the supplied argument.
     */

    nn = Tcl_GetString(objv[0]);
    np = s->GetNodeByName(nn);
    if (np == NULL) {
	Tcl_AppendResult(interp, "invalid node ", nn, NULL);
	return TCL_ERROR;
    }
    np->ExternalizeNode(n);
    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "invalid node ", nn, NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to set this marker to the obtained node.
     */

    if (!t.SetMarkedNode(n)) {
	Tcl_AppendResult(interp, "could not set marked node of ",
			 GetName(), NULL);
	return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 * Delete --
 *
 *	Deletes the e4_Marker object associated with this T4Graph 
 *	marker object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May delete a marker in the persistent storage. If successful, T4Graph
 *	objects that are reachable from this marker and which have been
 *	exported to Tcl become invalid.
 */

int
T4Marker::Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node n;
    e4_NodeUniqueID nuid;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$marker delete");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!t.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to delete the marker.
     */

    if (!t.Delete()) {
	Tcl_AppendResult(interp, "could not delete marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Do not leave any remnants in the interpreter result!
     */

    Tcl_ResetResult(interp);

    return TCL_OK;
}

/*
 * Dispose --
 *
 *	Disposes of the Tcl command for this marker.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Disposes of the Tcl command for this marker. After this call,
 *	the Tcl command for this marker can no longer be used by Tcl
 *	programs. The marker still exists in its persistent storage.
 */

int
T4Marker::Dispose(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$marker dispose");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!t.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Delete the Tcl command that refers to this marker and ask the storage
     * to forget this instance.
     */

    DeleteMarkerCommand(interp, this);

    /*
     * Dispose of the tgraph instance.
     */

    delete this;

    /*
     * Do not leave any remnants in the interpreter result!
     */

    Tcl_ResetResult(interp);

    return TCL_OK;
}

/*
 * IsValid --
 *
 *	Returns a boolean result indicating whether the T4Graph marker
 *	is valid.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a boolean value indicating whether this T4Graph
 *	marker object is valid.
 *
 * Side effects:
 *	None.
 */

int
T4Marker::IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$marker isvalid");
	return TCL_ERROR;
    }

    if (t.IsValid()) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    }

    return TCL_OK;
}

/*
 * Id --
 *
 *	Returns an identifier that is unique among all markers in the
 *	storage containing this marker.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result contains the integer ID.
 *
 * Side effects:
 *	None.
 */

int
T4Marker::Id(Tcl_Interp *interp, int objc, Tcl_Obj *CONST obvj[])
{
    e4_MarkerUniqueID muid;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$marker id");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!t.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", GetName(), NULL);
	return TCL_ERROR;
    }
    
    (void) t.GetUniqueID(muid);
    Tcl_SetIntObj(Tcl_GetObjResult(interp), muid.GetUniqueID());

    return TCL_OK;
}

/*
 * UserData --
 *
 *	Sets or returns the user data associated with this marker.
 *
 * Results:
 *	A standard Tcl result. When retrieving the user data, upon
 *	success, the interpreter result contains an integer representing
 *	the stored user data.
 *
 * Side effects:
 *	May modify the user data associated with this marker in its
 *	storage.
 */

int
T4Marker::UserData(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int i;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$marker userdata ?newvalue?");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!t.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * If one argument is given, it must be an integer (the new value).
     * Attempt to extract the new value.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &i) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (!t.SetUserData(i)) {
	    Tcl_AppendResult(interp, "cannot set user data for marker ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	return TCL_OK;
    }

    /*
     * No arguments given, so we want to retrieve the value.
     */

    if (!t.GetUserData(i)) {
	Tcl_AppendResult(interp, "cannot retrieve user data for marker ",
			 GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to the retrieved value.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), i);

    return TCL_OK;
}
