/*
 * t4node.cpp --
 *
 *	This file implements the command procedure invoked when the
 *	Tcl procedure "t4graph::node" is called.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * The various sub-commands for "t4graph::node":
 */

static char *subCommands[] = {
    (char *) "kind",
    (char *) "vertexcount",
    (char *) "set",
    (char *) "add",
    (char *) "get",
    (char *) "setnode",
    (char *) "addnode",
    (char *) "getvertex",
    (char *) "movevertex",
    (char *) "deletevertex", 
    (char *) "vertextype",
    (char *) "vertexrank",
    (char *) "vertexname",
    (char *) "renamevertex", 
    (char *) "exists",
    (char *) "parent",
    (char *) "parentcount",
    (char *) "occurrencecount",
    (char *) "parentrank",
    (char *) "isroot",
    (char *) "rankinparent",
    (char *) "nameinparent",
    (char *) "addmarker",
    (char *) "markwith",
    (char *) "unmark",
    (char *) "markercount",
    (char *) "ismarked",
    (char *) "marker",
    (char *) "storage",
    (char *) "delete",
    (char *) "dispose",
    (char *) "isvalid",
    (char *) "method",
    (char *) "call",
    (char *) "foreach",
    (char *) "id",
    (char *) "userdata",
    (char *) "deletefirstvertexwithnode",
    (char *) NULL
};
typedef enum NSubCommands {
    NKind = 0, 
    NVertexCount,
    NSet,
    NAdd,
    NGet,
    NSetNode,
    NAddNode,
    NGetVertex,
    NMoveVertex,
    NDeleteVertex,
    NVertexType,
    NVertexRank, 
    NVertexName, 
    NRenameVertex,
    NExists, 
    NParent,
    NParentCount,
    NOccurrenceCount,
    NParentRank,
    NIsRoot,
    NRankInParent,
    NNameInParent,
    NAddMarker,
    NMarkWith,
    NUnmark,
    NMarkerCount,
    NIsMarked,
    NMarker,
    NStorage,
    NDelete,
    NDispose,
    NIsValid,
    NMethod,
    NCall,
    NForeach,
    NId,
    NUserData,
    NDeleteVWN
} NSubCommands;

/*
 * NodeCmdProc --
 *
 *	This procedure is invoked when the "tgraph::node" command is
 *	called.
 *
 * Results:
 *	A standard Tcl result. The interpreter result may contain various
 *	results depending on the sub-command selected.
 *
 * Side effects:
 *	Whatever the selected sub-command does.
 */

int
NodeCmdProc(ClientData cd, Tcl_Interp *interp, int objc,
	    Tcl_Obj *CONST objv[])
{
    T4Node *n;
    NSubCommands index;

    /*
     * There must be at least two arguments.
     */

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node cmd ?arg ...?");
	return TCL_ERROR;
    }

    /*
     * The client data is the T4Node for this instance.
     */

    n = (T4Node *) cd;

    /*
     * Figure out what operation was requested.
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], subCommands, (char *) "cmd",
			    0, (int *) &index) != TCL_OK) {
	return TCL_ERROR;
    }

    objc -= 2;
    objv += 2;

    /*
     * Invoke the requested operation:
     */

    switch (index) {
    case NKind:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "node", -1);
	return TCL_OK;
    case NVertexCount:
	return n->VertexCount(interp, objc, objv);
    case NSet:
	return n->Set(interp, objc, objv);
    case NAdd:
	return n->Add(interp, objc, objv);
    case NGet:
	return n->Get(interp, objc, objv);
    case NSetNode:
	return n->SetNode(interp, objc, objv);
    case NAddNode:
	return n->AddNode(interp, objc, objv);
    case NGetVertex:
	return n->GetVertex(interp, objc, objv);
    case NMoveVertex:
	return n->MoveVertex(interp, objc, objv);
    case NDeleteVertex:
	return n->DeleteVertex(interp, objc, objv);
    case NVertexType:
	return n->VertexType(interp, objc, objv);
    case NVertexRank:
	return n->VertexRank(interp, objc, objv);
    case NVertexName:
	return n->VertexName(interp, objc, objv);
    case NRenameVertex:
	return n->RenameVertex(interp, objc, objv);
    case NExists:
	return n->Exists(interp, objc, objv);
    case NParent:
	return n->Parent(interp, objc, objv);
    case NParentCount:
	return n->ParentCount(interp, objc, objv);
    case NOccurrenceCount:
	return n->OccurrenceCount(interp, objc, objv);
    case NParentRank:
	return n->ParentRank(interp, objc, objv);
    case NIsRoot:
	return n->IsRoot(interp, objc, objv);
    case NRankInParent:
	return n->RankInParent(interp, objc, objv);
    case NNameInParent:
	return n->NameInParent(interp, objc, objv);
    case NAddMarker:
	return n->AddMarker(interp, objc, objv);
    case NMarkWith:
	return n->MarkWith(interp, objc, objv);
    case NUnmark:
	return n->Unmark(interp, objc, objv);
    case NMarkerCount:
	return n->MarkerCount(interp, objc, objv);
    case NIsMarked:
	return n->IsMarked(interp, objc, objv);
    case NMarker:
	return n->Marker(interp, objc, objv);
    case NStorage:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), 
			 n->GetStorage()->GetName(), -1);
	return TCL_OK;
    case NDelete:
	return n->Delete(interp, objc, objv);
    case NDispose:
	return n->Dispose(interp, objc, objv);
    case NIsValid:
	return n->IsValid(interp, objc, objv);
    case NMethod:
	return n->Method(interp, objc, objv);
    case NCall:
	return n->Call(interp, objc, objv);
    case NForeach:
	return n->Foreach(interp, objc, objv);
    case NId:
	return n->Id(interp, objc, objv);
    case NUserData:
	return n->UserData(interp, objc, objv);
    case NDeleteVWN:
	return n->DeleteFirstVertexWithNode(interp, objc, objv);
    default:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "NodeProc: unreachable code!",
			       (char *) NULL);
	return TCL_ERROR;
    }
}
