/*
 * t4noderep.cpp --
 *
 *	This file contains the implementation of the class T4Node
 *	which is defined in .../include/t4graphrep.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * Constructor:
 */

T4Node::T4Node(e4_Node nn, char *name, T4Storage *ss)
{
    SetName(name);
    n = nn;
    s = ss;
    vertexcache = NULL;
}

/*
 * Destructor:
 */

T4Node::~T4Node()
{
    n = invalidNode;
    SetName(NULL);
    if (vertexcache != NULL) {
	Tcl_DeleteHashTable(vertexcache);
	Tcl_Free((char *) vertexcache);
    }
}

/*
 * ExternalizeNode --
 *
 *	This procedure gives the caller access to the underlying
 *	e4_Node object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The caller can now manipulate the underlying e4_Node
 *	object directly.
 */

void
T4Node::ExternalizeNode(e4_Node &nn)
{
    nn = n;
}

/*
 * MaybeCreateVertex --
 *
 *	This procedure creates a new vertex if the given name is
 *	a simple one, and sets it to an integer value of zero.
 *
 * Results:
 *	The new T4Vertex object.
 *
 * Side effects:
 *	May modify the underlying storage, may create a new vertex
 *	and may export a new vertex to Tcl.
 */

T4Vertex *
T4Node::MaybeCreateVertex(Tcl_Interp *interp, char *vn)
{
    T4VertexNameKinds vnk;
    T4Vertex *vp;
    int index;
    int rank = 0;
    char *fnp;

    /*
     * Parse the vertex name.
     */

    if (ParseVertexName(interp, vn, &fnp, &index, &vnk) == TCL_ERROR) {
	return NULL;
    }

    /*
     * If the name is not simple, then set an error into the interpreter
     * and return NULL.
     */

    if ((vnk != T4VNK_INDEX) || (index != 1)) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "can not set vertex ", vn, " in node ",
			 GetName(), NULL);
	return NULL;
    }

    /*
     * Try to add the new vertex with an integer value of zero.
     */

    if (!n.AddVertex(fnp, E4_IOLAST, rank, (int) 0)) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "can not create vertex ", vn, " in node ",
			 GetName(), NULL);
	return NULL;
    }

    /*
     * No longer need "fnp", so we can reset the interpreter result here.
     */

    Tcl_ResetResult(interp);

    /*
     * Now obtain the vertex and export it to Tcl.
     */

    vp = ObtainVertex(interp, vn);
    if (vp == NULL) {
	return NULL;
    }
    Tcl_ResetResult(interp);

    return vp;
}

/*
 * ObtainVertex --
 *
 *	Retrieve a T4Vertex given a vertex name.
 *
 * Results:
 *	A pointer to a T4Vertex or NULL if the vertex can not be
 *	obtained. In that case the interpreter result will contain
 *	an error message.
 *
 * Side effects:
 *	A new object command might be created for the vertex because
 *	it is now exported to Tcl.
 */

T4Vertex *
T4Node::ObtainVertex(Tcl_Interp *interp, char *vn)
{
    char *fn;
    int i;
    T4VertexNameKinds vnk;
    e4_Vertex f;
    e4_VertexUniqueID vuid;
    T4Vertex *fp;

    /*
     * Parse the name of the vertex to rename.
     */

    if (ParseVertexName(interp, vn, &fn, &i, &vnk) == TCL_ERROR) {
	return NULL;
    }

    if (vnk == T4VNK_INDEX) {
	if ((!n.GetVertexRef(fn, i, f)) || (!f.IsValid())) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "can not retrieve vertex ",
			     vn, " in node ", GetName(), NULL);
	    return NULL;
	}
    } else {
	if ((!n.GetVertexRefByRank(i, f)) || (!f.IsValid())) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "can not retrieve vertex of rank ",
			     vn, " in node ", GetName(), NULL);
	    return NULL;
	}
    }

    /*
     * No longer need "fnp", so we can reset the interpreter result here.
     */

    Tcl_ResetResult(interp);

    /*
     * See if this vertex has already been exported to Tcl.
     */

    (void) f.GetUniqueID(vuid);
    fp = s->GetVertexById(vuid);

    /*
     * If this vertex has not yet been exported to Tcl, wrap it in
     * a new T4Vertex object and export it to Tcl.
     */

    if (fp == NULL) {
	fp = new T4Vertex(f, NULL, s);
	fp->SetName(MakeVertexCommand(interp, fp, vuid));
	s->StoreVertex(fp, vuid);
    }

    /*
     * Finally return the obtained vertex.
     */

    return fp;
}

/*
 ****************************************************************************
 *
 * The following methods implement Tcl sub-commands on a T4Node
 * object.
 *
 ****************************************************************************
 */

/*
 * VertexCount --
 *
 *	Returns the number of vertices in this node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains an integer representing the number of vertices in
 *	this node.
 *
 * Side effects:
 *	None.
 */

int
T4Node::VertexCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node vertexcount");
	return TCL_ERROR;
    }

    /*
     * Check that the node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Return the number of vertices as an integer:
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), n.VertexCount());

    return TCL_OK;
}

/*
 * Set --
 *
 *	Sets the value of an existing vertex in a node to a
 *	new value.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May modify an existing vertex in the persistent storage.
 */

int
T4Node::Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;
    bool created = false;

    /*
     * Expecting two or three arguments.
     */

    if ((objc != 2) && (objc != 3)) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
			 (char *) "$node set vertexspec value ?astype?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {
	    vp = MaybeCreateVertex(interp, vn);
	    created = true;
	}
	if (vp == NULL) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXUSED);

    /*
     * Now call the Set method of the obtained vertex.
     */

    if (vp->Set(interp, objc-1, objv+1) == TCL_ERROR) {

        /*
	 * If the vertex was created for this method, delete it,
	 * to ensure there are no side effects.
	 */

	if (created) {
	    (void) vp->Delete(interp, 0, NULL);
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "can not set vertex ",
			     vn, " in node ", GetName(), NULL);
	}

	return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 * The various (optional) type selector values for the T4Vertex::Set
 * operation.
 */

static char *typeSelectors[] = {
    (char *) "-int",
    (char *) "-double",
    (char *) "-string",
    (char *) "-node",
    (char *) "-binary",
    NULL
};
typedef enum TSelector {
    TSInteger = 0,
    TSDouble,
    TSString,
    TSNode,
    TSBinary
} TSelector;    

/*
 * Add --
 *
 *	Adds a new vertex with a given value to this node.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May add a new vertex to this node in the persistent storage.
 */

int
T4Node::Add(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    TSelector ts;
    Tcl_Obj *value;
    Tcl_Obj *astype;
    char *fn;
    e4_InsertOrder io;
    int rank = 0;
    bool rankgiven = false, done = false;
    int i, len;
    double d;
    char *ss;
    e4_Node nn;
    T4Node *np;

    /*
     * Expecting three, four or five arguments.
     */

    if ((objc != 3) && (objc != 4) && (objc != 5)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node add vn io ?rank? val ?astype?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Get the vertex name:
     */

    fn = Tcl_GetString(objv[0]);

    /*
     * Get the insertOrder.
     */

    if (ParseInsertOrder(interp, objv[1], &io) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * If the insert order is not E4_IOLAST, then we need to clear
     * the vertex cache because things might change incompatibly due
     * to adding a new vertex somewhere in the middle of this node.
     */

    if (io != E4_IOLAST) {
	ClearVertexCache();
    }

    /*
     * If the insert order is E4_IOLAST or E4_IOFIRST, then the next
     * argument is the value. Otherwise it is a rank.
     */

    if ((io != E4_IOLAST) && (io != E4_IOFIRST)) {
	if (Tcl_GetIntFromObj(interp, objv[2], &rank) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	rankgiven = true;
    }

    /*
     * If a rank was given, the value is in objv[3], otherwise it is
     * in objv[2]. If present, the argument after that is a type selector
     * for the type of value to install.
     */

    if (rankgiven) {
	value = objv[3];
	astype = NULL;
	if (objc == 5) {
	    astype = objv[4];
	}
    } else {
	value = objv[2];
	astype = NULL;
	if (objc == 4) {
	    astype = objv[3];
	}
    }

    /*
     * Attempt to use the type selector if given.
     */

    if (astype != NULL) {
	if (Tcl_GetIndexFromObj(interp, astype, typeSelectors,
				(char *) "typeselector", 0, (int *) &ts)
	    != TCL_OK) {
	    return TCL_ERROR;
	}
	switch (ts) {
	case TSInteger:
	    if (Tcl_GetIntFromObj(interp, value, &i) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	    if (!n.AddVertex(fn, io, rank, i)) {
		Tcl_AppendResult(interp, "can not add vertex ", fn,
				 " to node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	    break;
	case TSDouble:
	    if (Tcl_GetDoubleFromObj(interp, value, &d) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	    if (!n.AddVertex(fn, io, rank, d)) {
		Tcl_AppendResult(interp, "can not add vertex ", fn,
				 " to node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	    break;
	case TSString:
	    ss = Tcl_GetString(value);
	    if (!n.AddVertex(fn, io, rank, ss)) {
		Tcl_AppendResult(interp, "can not add vertex ", fn,
				 " to node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	    break;
	case TSNode:
	    ss = Tcl_GetString(value);
	    np = s->GetNodeByName(ss);
	    if (np == NULL) {
		Tcl_AppendResult(interp, "could not retrieve node named ", ss,
				 NULL);
		return TCL_ERROR;
	    }
	    np->ExternalizeNode(nn);
	    if (!nn.IsValid()) {
		Tcl_AppendResult(interp, "node name ", ss, 
				 " is invalid", NULL);
		return TCL_ERROR;
	    }
	    if (!n.AddVertex(fn, io, rank, nn)) {
		Tcl_AppendResult(interp, "can not add vertex ", fn,
				 " to node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	    break;
	case TSBinary:
	    ss = Tcl_GetStringFromObj(value, &len);
	    if (!n.AddVertex(fn, io, rank, (void *) ss, len)) {
		Tcl_AppendResult(interp, "can not add vertex ", fn,
				 " to node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	    break;
	}
    } else {

	/*
	 * Extract the value as a string first.
	 */

	ss = Tcl_GetString(value);

	/*
	 * First of all see if the new value contains a '.'; if it does then
	 * attempt to convert it to a double.
	 */

	if (Tcl_StringMatch(ss, "?*.*")) {
	    if (Tcl_GetDoubleFromObj(interp, value, &d) == TCL_OK) {
		if (!n.AddVertex(fn, io, rank, d)) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "cannot add vertex ", fn,
				     " to node ", GetName(), NULL);
		    return TCL_ERROR;
		}
	    }
	    done = true;
	}

	/*
	 * Attempt to get an integer value out of the argument. If that
	 * works, set the vertex value to that integer.
	 */

	Tcl_ResetResult(interp);
	if ((!done) && (Tcl_GetIntFromObj(interp, value, &i) == TCL_OK)) {
	    if (!n.AddVertex(fn, io, rank, i)) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "cannot add vertex ", fn,
				 " to node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	    done = true;
	}

	/*
	 * Attempt to convert the string into the name of a node.
	 */

	if (!done) {
	    Tcl_ResetResult(interp);
	    np = s->GetNodeByName(ss);
	    if (np != NULL) {
		np->ExternalizeNode(nn);
		if ((!nn.IsValid()) || (!n.AddVertex(fn, io, rank, nn))) {
		    Tcl_AppendResult(interp, "cannot add vertex ", fn,
				     " to node ", GetName(), NULL);
		    return TCL_ERROR;
		}
		done = true;
	    }
	}

	/*
	 * Last resort - set the value as a string.
	 */

	if (!done) {
	    if (!n.AddVertex(fn, io, rank, ss)) {
		Tcl_AppendResult(interp, "cannot add vertex ", fn,
				 " to node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	}
    }

    return TCL_OK;
}

/*
 * Get --
 *
 *	Retrieve the value of an existing vertex.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains an object representing the value of the requested
 *	vertex.
 *
 * Side effects:
 *	None.
 */

int
T4Node::Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;
    
    /*
     * Expecting one argument.
     */

    if ((objc != 1) && (objc != 2) && (objc != 3)) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
			 (char *) "$node get vertexspec ?createval? ?astype?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertex we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_FindHashEntry(vertexcache, vn);
    if (hPtr == NULL) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {

	    /*
	     * If an optional create value was given, the vertex is
	     * created in this node as the last vertex and initialized
	     * to that value.
	     */

	    if (objc > 1) {
		vp = MaybeCreateVertex(interp, vn);
		if (vp == NULL) {
		    return TCL_ERROR;
		}

		/*
		 * Set the vertex to the supplied value.
		 */

		if (vp->Set(interp, objc-1, objv+1) == TCL_ERROR) {
		    (void) vp->Delete(interp, 0, NULL);
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "can not find vertex ", vn,
				     " in node ", GetName(), NULL);

		    return TCL_ERROR;
		}
	    } else {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "can not find vertex ", vn,
				 " in node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	}
	vp->SetFlag(T4_VERTEXUSED);
	Tcl_ResetResult(interp);

	/*
	 * Create the hash entry for this vertex, and set its value.
	 */

	hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXUSED);

    /*
     * Now call the Get method of the obtained vertex.
     */

    return vp->Get(interp, 0, NULL);
}

/*
 * SetNode --
 *
 *	Sets the value of an existing vertex to a new node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the name by which the new node is known in Tcl.
 *
 * Side effects:
 *	May create a new node in the persistent storage and may modify
 *	the value of a vertex in the persistent storage.
 */

int
T4Node::SetNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node setnode vertexspec");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);

	/*
	 * If the vertex could not be found, attempt to create it.
	 */

	if (vp == NULL) {
	    vp = MaybeCreateVertex(interp, vn);
	}
	if (vp == NULL) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXUSED);

    /*
     * Now call the SetNode method of the obtained vertex.
     */

    return vp->SetNode(interp, objc-1, objv+1);
}

/*
 * AddNode --
 *
 *	Adds a new vertex with a node as its value, to this node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the name by which the new node is known in Tcl.
 *
 * Side effects:
 *	May add a new vertex and a new node in the persistent storage.
 */

int
T4Node::AddNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *fn;
    int rank = 0;
    e4_InsertOrder io;
    e4_Node nn;
    e4_Vertex vv;
    e4_NodeUniqueID nuid;
    T4Node *np;
    char buf[32];

    /*
     * Expecting two or three arguments.
     */

    if ((objc != 2) && (objc != 3)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node addnode name io ?rank?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Retrieve the vertex name to add:
     */

    fn = Tcl_GetString(objv[0]);

    /*
     * Get the insertOrder.
     */

    if (ParseInsertOrder(interp, objv[1], &io) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * If the insert order is not E4_IOLAST, then we need to clear
     * the vertex cache because things might change incompatibly due
     * to adding a new vertex somewhere in the middle of this node.
     */

    if (io != E4_IOLAST) {
	ClearVertexCache();
    }

    /*
     * If there are three arguments, the third is the requested rank:
     */

    if ((objc == 3) &&
	(Tcl_GetIntFromObj(interp, objv[2], &rank) == TCL_ERROR)) {
	return TCL_ERROR;
    }

    /*
     * Attempt to add the vertex and obtain the new node.
     */

    if ((!n.AddNodeRef(fn, io, rank, nn, vv)) || (!nn.IsValid()) ||
	(!vv.IsValid())) {
	sprintf(buf, "%d", rank);
	Tcl_AppendResult(interp, "can not add vertex ", fn,
			 " at ", Tcl_GetString(objv[1]),
			 ", offset ", buf, ", to node ",
			 GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Force the mod-vertex callback.
     */

    vv.Set(nn);

    /*
     * Create the new T4Node to wrap around the new e4_Node object and
     * export the T4Node object to Tcl.
     */

    (void) nn.GetUniqueID(nuid);
    np = new T4Node(nn, NULL, s);
    np->SetName(MakeNodeCommand(interp, np, nuid));
    s->StoreNode(np, nuid);
    Tcl_SetStringObj(Tcl_GetObjResult(interp), np->GetName(), -1);
    return TCL_OK;
}

/*
 * GetVertex --
 *
 *	Export a T4Vertex object for the selected vertex to Tcl.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the name by which the vertex is known in Tcl.
 *
 * Side effects:
 *	None.
 */

int
T4Node::GetVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * One argument expected.
     */

    if ((objc != 1) && (objc != 2)) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
		 (char *) "$node getvertex vertexspec ?createval? ?astype?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_FindHashEntry(vertexcache, vn);
    if (hPtr == NULL) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {

	    /*
	     * If a create value was given, create the vertex as the last
	     * one in this node and set it to the supplied value. Then,
	     * return the name of the newly created vertex.
	     */

	    if (objc > 1) {
		vp = MaybeCreateVertex(interp, vn);
		if (vp == NULL) {
		    return TCL_ERROR;
		}
		if (vp->Set(interp, objc-1, objv+1) == TCL_ERROR) {
		    (void) vp->Delete(interp, 0, NULL);
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "can not find vertex ", vn,
				     " in node ", GetName(), NULL);

		    return TCL_ERROR;
		}
		Tcl_SetStringObj(Tcl_GetObjResult(interp), vp->GetName(), -1);
	    } else {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "can not find vertex ", vn,
				 " in node ", GetName(), NULL);
		return TCL_ERROR;
	    }
	}
	Tcl_ResetResult(interp);

	/*
	 * Create the hash entry for this vertex, and set its value.
	 */

	hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXEXPORTED | T4_VERTEXUSED);

    /*
     * Now return the name of the vertex as the Tcl result.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), vp->GetName(), -1);
    return TCL_OK;
}

/*
 * MoveVertex --
 *
 *	Move a vertex identified by a vertex name to another location.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May move a vertex to another location in the persistent storage.
 */

int
T4Node::MoveVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *fn;
    T4Vertex *fp;
    e4_Vertex f;
    e4_InsertOrder io;
    int offset = 0;
    char buf[32];

    /*
     * Expecting two or three arguments.
     */

    if ((objc != 3) && (objc != 2)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node movevertex vn io ?offset?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Locate the T4Vertex object for the vertex to move.
     */

    fn = Tcl_GetString(objv[0]);
    fp = s->GetVertexByName(fn);
    if (fp == NULL) {
	Tcl_AppendResult(interp, "can not find vertex ", fn, NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the enclosed e4_Vertex object.
     */

    fp->ExternalizeVertex(f);
    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "invalid vertex ", fn, NULL);
	return TCL_ERROR;
    }

    /*
     * Clear the vertex cache of the node containing the vertex being
     * moved. We must do this before the move because the vertex may
     * move to a different node (this one).
     */

    fp->NodeClearVertexCache();

    /*
     * Parse the insertOrder.
     */

    if (ParseInsertOrder(interp, objv[1], &io) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * If there are three arguments, the third is a requested offset.
     */

    if ((objc == 3) &&
	(Tcl_GetIntFromObj(interp, objv[2], &offset) == TCL_ERROR)) {
	return TCL_ERROR;
    }

    /*
     * Attempt to move the vertex.
     */

    if (!n.MoveVertex(f, io, offset)) {
	sprintf(buf, "%d", offset);
	Tcl_AppendResult(interp, "can not move vertex ", fn, " to ",
			 Tcl_GetString(objv[1]),
			 " ", buf, " in node ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Clear the vertex cache of this node if the insert order is
     * not IO_LAST, because the mapping of names to vertices may
     * have changed.
     */

    if (io != E4_IOLAST) {
	ClearVertexCache();
    }

    return TCL_OK;
}

/*
 * DeleteVertex --
 *
 *	Delete a given vertex from this node.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May delete a vertex from the persistent storage.
 */

int
T4Node::DeleteVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew, res;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node deletevertex vertexspec");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it.
     *
     * NOTE: If this vertex was not yet exported to Tcl, we cache it in
     * the vertex cache even though we are going to delete it right away.
     * This is so that the new hash entry can be deleted efficienctly by
     * the Delete() operation on the vertex.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }

    /*
     * Now call the Delete method of the obtained vertex.
     */

    res = vp->Delete(interp, objc-1, objv+1);

    /*
     * Clear the vertex cache of this node because things may change
     * incompatibly after the delete.
     */

    ClearVertexCache();

    /*
     * Finally return the result of deleting the vertex.
     */

    return res;
}

/*
 * VertexType --
 *
 *	Returns a string representing the type of the value stored in
 *	a specified vertex.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string representing the type of the value stored
 *	in the specified vertex.
 *
 * Side effects:
 *	None.
 */

int
T4Node::VertexType(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node vertextype vertexspec");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXUSED);

    /*
     * Now call the Type method of the obtained vertex.
     */

    return vp->Type(interp, objc-1, objv+1);
}

/*
 * VertexName --
 *
 *	Returns a string, the name of the vertex identified by the given
 *	rank in this node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the name of the chosen vertex.
 *
 * Side effects:
 *	None.
 */

int
T4Node::VertexName(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node vertexname rank");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXUSED);

    /*
     * Now call the Name method of the obtained vertex.
     */

    return vp->Name(interp, objc-1, objv+1);
}

/*
 * VertexRank --
 *
 *	Return the rank of a specified vertex in this node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains an integer representing the rank of the selected
 *	vertex in this node.
 *
 * Side effects:
 *	None.
 */

int
T4Node::VertexRank(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node vertexrank vertexspec");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXUSED);

    /*
     * Now call the Rank method of the obtained vertex. Note that
     * this is pretty inefficient...
     */

    return vp->Rank(interp, objc-1, objv+1);
}

/*
 * RenameVertex --
 *
 *	Rename a vertex in this node to a new name.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May rename a vertex in the persistent storage.
 */

int
T4Node::RenameVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *fn;
    char *vn;
    int i;
    T4VertexNameKinds vnk;

    /*
     * Expecting two arguments.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node renamevertex vertex newname");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Parse the name of the vertex to rename.
     */

    vn = Tcl_GetString(objv[0]);
    if (ParseVertexName(interp, vn, &fn, &i, &vnk) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * If the vertex is specified by index, convert it to a rank.
     */

    if (vnk == T4VNK_INDEX) {
	i = n.VertexRank(fn, i);
	if (i == E4_VERTEXNOTFOUND) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "can not rename vertex ",
			     Tcl_GetString(objv[0]),
			     " in node ", GetName(), NULL);
	    return TCL_ERROR;
	}
    }


    /*
     * No longer need "fn", so we can reset the interpreter result here.
     */

    Tcl_ResetResult(interp);

    /*
     * Attempt to rename the specified vertex.
     */

    if (!n.RenameVertex(i, Tcl_GetString(objv[1]))) {
	if (vnk == T4VNK_INDEX) {
	    Tcl_AppendResult(interp, "can not rename vertex ",
			     Tcl_GetString(objv[0]),
			     " in node ", GetName(), NULL);
	} else {
	    Tcl_AppendResult(interp, "can not rename vertex ranked ",
			     Tcl_GetString(objv[0]),
			     " in node ", GetName(), NULL);
	}
	return TCL_ERROR;
    }

    /*
     * Things may have changed incompatibly so we need to clear
     * the vertex cache.
     */

    ClearVertexCache();

    return TCL_OK;
}

/*
 * Exists --
 *
 *	Return a boolean value indicating whether the selected vertex
 *	exists. The vertex is specified either as a simple name or as
 *	an indexed name.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 */

int
T4Node::Exists(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node exists vertexspec");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertice we want is already cached. If not, obtain it,
     * if it exists.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {
	    Tcl_DeleteHashEntry(hPtr);
	    Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
	} else {
	    Tcl_SetHashValue(hPtr, vp);
	    Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
	    vp->SetFlag(T4_VERTEXUSED);
	}
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    }

    return TCL_OK;
}

/*
 * Parent --
 *
 *	Retrieve the node that is a specified (numbered) parent of this node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name by which the selected parent node is
 *	known in Tcl.
 *
 * Side effects:
 *	A node may be exported to Tcl.
 */

int
T4Node::Parent(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node p;
    e4_NodeUniqueID nuid;
    T4Node *np;
    int ii = 1;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node parent ?index?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If an index is specified, it denotes the nth parent. In that
     * case retrieve the nth parent, otherwise retrieve the first one.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &ii) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
    }

    /*
     * Attempt to obtain the selected parent.
     */

    if ((!n.GetParent(ii, p)) || (!p.IsValid())) {
	Tcl_AppendResult(interp, "can not retrieve selected parent node ",
			 " of node ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * See if we have exported this node already. If not, then
     * export it.
     */

    (void) p.GetUniqueID(nuid);
    np = s->GetNodeById(nuid);
    if (np == NULL) {
	np = new T4Node(p, NULL, s);
	np->SetName(MakeNodeCommand(interp, np, nuid));
	s->StoreNode(np, nuid);
    }
    Tcl_SetStringObj(Tcl_GetObjResult(interp), np->GetName(), -1);
    return TCL_OK;
}

/*
 * ParentCount --
 *
 *	Returns an integer representing the number of parents that
 *	this node has.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains an integer representing the number of parents that
 *	this node has.
 *
 * Side effects:
 *	None.
 */

int
T4Node::ParentCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *)  "$node parentcount");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to the number of parents of this node.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), n.ParentCount());
    return TCL_OK;
}

/*
 * OccurrenceCount --
 *
 *	Returns an integer representing either how many vertices in total
 *	have this node as their value, or how many vertices in a specified
 *	parent have this node as their value. In the second case, the
 *	specified parent is passed as an argument.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result contains
 *	an integer representing the requested occurrence count.
 *
 * Side effects:
 *	None.
 */

int
T4Node::OccurrenceCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *pn;
    T4Node *pp;
    e4_Node p;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node occurrencecount ?parent?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If the parent is specified, obtain it.
     */

    if (objc == 1) {
	pn = Tcl_GetString(objv[0]);
	pp = s->GetNodeByName(pn);
	if (pp == NULL) {
	    Tcl_AppendResult(interp, "invalid node ", pn, NULL);
	    return TCL_ERROR;
	}
	pp->ExternalizeNode(p);
	if (!p.IsValid()) {
	    Tcl_AppendResult(interp, "invalid node ", pn, NULL);
	    return TCL_ERROR;
	}

	/*
	 * Obtain the occurrence count of this node in the specified
	 * parent.
	 */

	Tcl_SetIntObj(Tcl_GetObjResult(interp), n.OccurrenceCount(p));
	return TCL_OK;
    }

    /*
     * Obtain the total occurrence count for this node.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), n.OccurrenceCount());
    return TCL_OK;
}

/*
 * ParentRank --
 *
 *	Given the name of a parent node, compute the rank of this parent
 *	in the parents of this node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result contains
 *	an integer representing the rank of the given parent node in the
 *	parents of this node.
 *
 * Side effects:
 *	None.
 */

int
T4Node::ParentRank(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *pn;
    T4Node *pp;
    e4_Node p;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node parentrank parent");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the parent node whose rank is to be computed.
     */

    pn = Tcl_GetString(objv[0]);
    pp = s->GetNodeByName(pn);
    if (pp == NULL) {
	Tcl_AppendResult(interp, "invalid node ", pn, NULL);
	return TCL_ERROR;
    }
    pp->ExternalizeNode(p);
    if (!p.IsValid()) {
	Tcl_AppendResult(interp, "invalid node ", pn, NULL);
	return TCL_ERROR;
    }

    /*
     * Retrieve the rank for this parent node.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), n.ParentRank(p));
    return TCL_OK;
}

/*
 * IsRoot --
 *
 *	Returns a boolean value indicating whether this node is
 *	the root of its marker.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a boolean value indicating whether this node is
 *	the root of its marker.
 *
 * Side effects:
 *	None.
 */

int
T4Node::IsRoot(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node isroot");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Check if this node is the root of its marker.
     */

    if (n.IsRoot()) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    }

    return TCL_OK;
}

/*
 * RankInParent --
 *
 *	Returns an integer representing the rank of the vertex containing
 *	this node in its parent.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains an integer, the rank of the vertex containing this
 *	node in its parent.
 *
 * Side effects:
 *	None.
 */

int
T4Node::RankInParent(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int ii = 1;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node rankinparent ?index?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If the index of the parent is supplied, extract it.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &ii) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
    }

    /*
     * Set the interpreter result to the rank of the vertex that contains
     * this node in its parent.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), n.GetRankInParent(ii));

    return TCL_OK;
}

/*
 * NameInParent --
 *
 *	Returns a string, the name of the vertex containing this node
 *	in its parent.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name of the vertex containing this node
 *	in its parent.
 *
 * Side effects:
 *	None.
 */

int
T4Node::NameInParent(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int ii = 1;

    /*
     * Expecting zero arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node nameinparent ?index?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If the index of the parent is specified, extract it.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &ii) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
    }

    /*
     * Set the interpreter result to a string, the name of the vertex
     * containing this node in its parent.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), 
		     (char *) n.GetNameInParent(ii), -1);

    return TCL_OK;
}

/*
 * AddMarker --
 *
 *	Add a named marker to this node. Create the marker.
 *
 * Results:
 *	A standard Tcl result. Upon success the interpreter result contains
 *	a string denoting the name by which the new marker is known in Tcl.
 *
 * Side effects:
 *	May create a new marker in the underlying persistent storage.
 *	May create a new Tcl command.
 */

int
T4Node::AddMarker(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Marker m;
    e4_MarkerUniqueID muid;
    T4Marker *mp;
    char *mn;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node addmarker marker");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the name for the new marker to be added.
     */

    mn = Tcl_GetString(objv[0]);

    /*
     * Attempt to add the new marker.
     */

    if (!n.AddMarker(mn, m)) {
	Tcl_AppendResult(interp, "can not add marker ", mn, " to node ",
			 GetName(), " marker may already exist", NULL);
	return TCL_ERROR;
    }

    /*
     * Export the new marker to Tcl.
     */

    (void) m.GetUniqueID(muid);
    mp = new T4Marker(m, NULL, s);
    mp->SetName(MakeMarkerCommand(interp, mp, muid));
    s->StoreMarker(mp, muid);

    /*
     * Return the name of the new marker command just created.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), mp->GetName(), -1);
    return TCL_OK;
}

/*
 * MarkWith --
 *
 *	Make the marked node of the supplied marker be this node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result is
 *	empty.
 *
 * Side effects:
 *	The supplied marker may be made to point to this node in the
 *	underlying persistent storage.
 */

int
T4Node::MarkWith(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    T4Marker *mp;
    e4_Marker m;
    char *mn;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node markwith marker");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the name of the marker whose marked node will become
     * this node.
     */

    mn = Tcl_GetString(objv[0]);

    /*
     * Attempt to retrieve the marker -- it must have already been
     * exported to Tcl.
     */

    mp = s->GetMarkerByName(mn);
    if (mp == NULL) {
	Tcl_AppendResult(interp, "invalid marker ", mn, NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to obtain the underlying marker.
     */

    mp->ExternalizeMarker(m);
    if (!m.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", mn, NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to mark this node with the obtained marker.
     */

    if (!n.MarkWith(m)) {
	Tcl_AppendResult(interp, "can not mark node ", GetName(),
			 " with marker ", mn, NULL);
	return TCL_ERROR;
    }

    Tcl_ResetResult(interp);
    return TCL_OK;
}

/*
 * Unmark --
 *
 *	Remove the specified marker from the set of markers marking this
 *	node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	is empty.
 *
 * Side effects:
 *	May remove a marker in the underlying persistent storage. May
 *	remove a Tcl command for the removed marker.
 */

int
T4Node::Unmark(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *mn;
    T4Marker *mp;
    e4_Marker m;
    e4_Node n1;

    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node unmark marker");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the name of the marker that is to be removed.
     */

    mn = Tcl_GetString(objv[0]);

    /*
     * Attempt to retrieve the marker -- it must have already been
     * exported to Tcl.
     */

    mp = s->GetMarkerByName(mn);
    if (mp == NULL) {
	Tcl_AppendResult(interp, "invalid marker ", mn, NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to obtain the underlying marker.
     */

    mp->ExternalizeMarker(m);
    if (!m.IsValid()) {
	Tcl_AppendResult(interp, "invalid marker ", mn, NULL);
	return TCL_ERROR;
    }

    /*
     * Check that this marker marks this node.
     */

    if (!m.GetMarkedNode(n1) || !n1.IsValid() || n1 != n) {
	Tcl_AppendResult(interp, "node ", GetName(), " is not marked by ",
			 "marker ", mn, NULL);
	return TCL_ERROR;
    }

    /*
     * Unmark the node. All needed cleanups are done in the auto-deletion
     * callbacks.
     */

    if (!n.Unmark(m)) {
	Tcl_AppendResult(interp, "can not unmark node ", GetName(),
			 " with marker ", mn, NULL);
	return TCL_ERROR;
    }

    /*
     * Do not leave any remnants in the interpreter result!
     */

    Tcl_ResetResult(interp);

    return TCL_OK;
}

/*
 * MarkerCount --
 *
 *	Returns an integer indicating how many markers are marking this
 *	node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result contains
 *	an integer indicating how many markers mark this node.
 *
 * Side effects:
 *	None.
 */

int
T4Node::MarkerCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node markercount");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the count of markers for this node.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), n.MarkerCount());
    return TCL_OK;
}

/*
 * IsMarked --
 *
 *	Returns a boolean indicating whether this node is marked.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result contains
 *	a boolean indicating whether this node is marked.
 *
 * Side effects:
 *	None.
 */

int
T4Node::IsMarked(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node ismarked");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * See if the node is marked.
     */

    Tcl_SetBooleanObj(Tcl_GetObjResult(interp),
		      (n.MarkerCount() == 0) ? 0 : 1);
    return TCL_OK;
}

/*
 * Marker --
 *
 *	Export the requested marker of this node to Tcl.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result contains
 *	the name by which the requested marker is known in Tcl.
 *
 * Side effects:
 *	May create a new Tcl command for the newly exported marker.
 */

int
T4Node::Marker(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int index = 1;
    T4Marker *mp;
    e4_Marker m, m1;
    e4_MarkerUniqueID muid;

    /*
     * Expecting one argument.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node marker ?index?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Obtain the index of the requested marker.
     */

    if ((objc == 1) && 
	(Tcl_GetIntFromObj(interp, objv[0], &index) == TCL_ERROR)) {
	return TCL_ERROR;
    }
    Tcl_ResetResult(interp);

    /*
     * Attempt to obtain the requested marker from this node.
     */

    if ((!n.GetMarker(index, m)) || (!m.IsValid())) {
	Tcl_AppendResult(interp, "can not obtain marker #",
			 Tcl_GetString(objv[0]),
			 " for node ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * See if this marker has already been exported to Tcl. If not,
     * export it and create the Tcl command to manipulate it.
     */

    (void) m.GetUniqueID(muid);
    mp = s->GetMarkerById(muid);
    if (mp == NULL) {
	mp = new T4Marker(m, NULL, s);
	mp->SetName(MakeMarkerCommand(interp, mp, muid));
	s->StoreMarker(mp, muid);
    }
    Tcl_SetStringObj(Tcl_GetObjResult(interp), mp->GetName(), -1);

    return TCL_OK;
}    

/*
 * Delete --
 *
 *	Delete this node and recursively all the vertices it contains
 *	and their values.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May delete a (sub)-graph in the persistent storage.
 */

int
T4Node::Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node delete");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }
    
    /*
     * Attempt to delete the node.
     */

    if (!n.Delete()) {
	Tcl_AppendResult(interp, "can not delete node ", GetName(),
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Do not leave any remnants in the interpreter result.
     */

    Tcl_ResetResult(interp);

    return TCL_OK;
}

/*
 * Dispose --
 *
 *	Disposes of the Tcl command for this node.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Disposes of the Tcl command for this node. After this call,
 *	the Tcl command for this node can no longer be used by Tcl
 *	programs. The node still exists in its permanent storage.
 */

int
T4Node::Dispose(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node dispose");
	return TCL_ERROR;
    }

    /*
     * Check that the node object is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "invalid node ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Delete the Tcl command that refers to this node and ask the storage
     * to forget this instance.
     */

    DeleteNodeCommand(interp, this);

    /*
     * And finally dispose of the tgraph instance.
     */

    delete this;

    /*
     * Do not leave any remnants in the interpreter result!
     */

    Tcl_ResetResult(interp);

    return TCL_OK;
}

/*
 * IsValid --
 *
 *	Returns a boolean indicating whether this node is valid.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 */

int
T4Node::IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node isvalid");
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to a boolean object indicating whether
     * this node is valid.
     */

    if (n.IsValid()) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    }

    return TCL_OK;
}

/*
 * Method --
 *
 *	Define a new method of this node (callable with Call).
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains an empty string.
 *
 * Side effects:
 *	Adds a new vertex named by the supplied method name as the
 *	last vertex of this node.
 */

int
T4Node::Method(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    Tcl_Obj **nobjv;
    int res;

    /*
     * Expecting three arguments.
     */

    if (objc != 3) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node method methodname args body");
	return TCL_ERROR;
    }

    /*
     * Compute the string value of the new vertex -- it is the list
     * composed of the args and body of the method.
     */

    nobjv = (Tcl_Obj **) Tcl_Alloc(2 * sizeof(Tcl_Obj *));

    nobjv[0] = objv[0];
    nobjv[1] = Tcl_NewListObj(objc-1, objv+1);

    res = Set(interp, 2, (Tcl_Obj *CONST *) nobjv);

    Tcl_Free((char *) nobjv);

    return TCL_OK;
}    

/*
 * Call --
 *
 *	Call a selected vertex as a Tcl procedure.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the result of calling the specified procedure.
 *
 * Side effects:
 *	Whatever the called procedure does.
 */

int
T4Node::Call(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int isnew;
    char *vn;
    T4Vertex *vp;
    Tcl_HashEntry *hPtr;

    /*
     * Expecting at least one argument.
     */

    if (objc < 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
			 (char *) "$node call vertex ?arg ...?");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Ensure that the vertices cache is initialized.
     */

    if (vertexcache == NULL) {
	vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
    }

    /*
     * See if the vertex we want is already cached. If not, obtain it.
     */

    vn = Tcl_GetString(objv[0]);
    hPtr = Tcl_CreateHashEntry(vertexcache, vn, &isnew);
    if (isnew) {
	vp = ObtainVertex(interp, vn);
	if (vp == NULL) {
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	Tcl_SetHashValue(hPtr, vp);
    } else {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
    }
    vp->SetFlag(T4_VERTEXUSED);

    /*
     * Now call the Call method on the vertex that we obtained.
     */

    return vp->Call(interp, objc, objv);
}

/*
 * RemoveVertexFromCache --
 *
 *	Removes one vertex from the cache.
 *
 * Results:
 *	None:
 *
 * Side effects:
 *	May remove one or more entries from the vertex cache.
 */

void
T4Node::RemoveVertexFromCache(T4Vertex *vp)
{
    Tcl_HashSearch search;
    Tcl_HashEntry *ePtr, *nPtr;
    T4Vertex *ep;

    /*
     * If the cache has not been initialized yet, we do not
     * have to clear it either.
     */

    if (vertexcache == NULL) {
	return;
    }

    /*
     * Iterate over the cache, look for entries whose value is the
     * vertex to remove, and remove them.
     */

    for (ePtr = Tcl_FirstHashEntry(vertexcache, &search);
	 ePtr != NULL;
	 ePtr = nPtr) {
	ep = (T4Vertex *) Tcl_GetHashValue(ePtr);
	nPtr = Tcl_NextHashEntry(&search);

	/*
	 * If the value is the vertex we want to remove, delete the
	 * hash entry.
	 */

	if (ep == vp) {
	    Tcl_DeleteHashEntry(ePtr);
	}
    }
}

/*
 * ClearVertexCache --
 *
 *	Discards the cache of vertices that are already known in Tcl.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	After this operation, the cache is empty, so access operations
 *	will take longer for the first access to each vertex.
 */

void
T4Node::ClearVertexCache()
{
    /*
     * If the cache has not been initialized yet, we do not
     * have to clear it either.
     */

    if (vertexcache == NULL) {
	return;
    }

    /*
     * Empty out the hash table efficiently.
     */

    Tcl_DeleteHashTable(vertexcache);

    /*
     * Deallocate the cache, allocate a new one and initialize it.
     */

    Tcl_Free((char *) vertexcache);
    vertexcache = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(vertexcache, TCL_STRING_KEYS);
}

/*
 * Foreach --
 *
 *	Visit selected components of this node, executing a Tcl command
 *	for each visited component.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the Tcl command does. May export new elements of this node
 *	to Tcl.
 */

static char *selectors[] = {
    (char *) "vertex",
    (char *) "marker",
    (char *) "parent",
    (char *) NULL
};
typedef enum SSelectors {
    SVertex = 0,
    SMarker,
    SParent
} SSelectors;

int
T4Node::Foreach(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    SSelectors index;

    /*
     * Expecting at least three arguments.
     */

    if (objc < 3) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node foreach sel var ?options? cmd");
	return TCL_ERROR;
    }

    /*
     * Figure out what selector was requested.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], selectors, (char *) "selector",
			    0, (int *) &index) != TCL_OK) {
	return TCL_ERROR;
    }

    objc--;
    objv++;

    switch (index) {
    case SVertex:
	return VisitVertices(interp, objc, objv);
    case SMarker:
	return VisitMarkers(interp, objc, objv);
    case SParent:
	return VisitParents(interp, objc, objv);
    }

    /*
     * Not reached, but some compilers insist on this..
     */

    return TCL_ERROR;
}

/*
 * VisitVertices --
 *
 *	Helper function to visit selected vertices in this node.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the Tcl command executed does. May export new vertices
 *	to Tcl, if -keep true is supplied in the arguments.
 */

static char *filters[] = {
    (char *) "-type",
    (char *) "-name",
    (char *) "-keep",
    (char *) NULL
};
typedef enum SFilters {
    SType = 0,
    SName,
    SKeep
} SFilters;

static char *typenames[] = {
    (char *) "node",
    (char *) "int",
    (char *) "double",
    (char *) "string",
    (char *) "binary",
    (char *) NULL
};

int
T4Node::VisitVertices(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Vertex v;
    e4_VertexUniqueID vuid;
    T4Vertex *vp;
    char *varname;
    char *namefilter = NULL;
    SFilters filter;
    e4_VertexType vt = E4_VTUNKNOWN;
    int ret = TCL_OK, retone;
    bool done = false;
    bool keep = false;
    bool isnew;

    /*
     * Expecting two, four, six or eight arguments, the name of a variable
     * to set to each vertex as it is visited, an optional type filter (two
     * args), an optional name filter (two args), an optional keep specifier
     * (two args), and the command to execute.
     */

    if ((objc != 2) && (objc != 4) && (objc != 6) && (objc != 8)) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *)
		"$node foreach vertex v ?-type t? ?-name n? ?-keep bool? cmd");
	return TCL_ERROR;
    }

    /*
     * Get the variable name.
     */

    varname = Tcl_GetString(objv[0]);

    /*
     * Parse the optional specifiers:
     */

    for (objc--, objv++; objc > 1; objc -= 2, objv += 2) {
	if (Tcl_GetIndexFromObj(interp, objv[0], filters, (char *) "filter",
				0, (int *) &filter) != TCL_OK) {
	    return TCL_ERROR;
	}

	switch (filter) {
	case SName:
	    namefilter = Tcl_GetString(objv[1]);
	    break;
	case SType:
	    if (Tcl_GetIndexFromObj(interp, objv[1], typenames, 
				    (char *) "typename", 0, (int *) &vt) 
		!= TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	case SKeep:
	    if (Tcl_GetBooleanFromObj(interp, objv[1], (int *) &keep) 
		!= TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	}
    }

    /*
     * Create the iterator:
     */

    e4_VertexVisitor vv(n, (const char *) namefilter, vt);

    /*
     * Iterate over the selected vertices.
     */

    while ((!done) && (vv.CurrentVertexAndAdvance(v))) {

	/*
	 * See if we already exported this vertex to Tcl.
	 */

	(void) v.GetUniqueID(vuid);
	vp = s->GetVertexById(vuid);

	/*
	 * If we did not export it, make a new T4Vertex and store it under
	 * the ID obtained from the e4Vertex object.
	 */

	if (vp == NULL) {
	    vp = new T4Vertex(v, NULL, s);
	    vp->SetName(MakeVertexCommand(interp, vp, vuid));
	    s->StoreVertex(vp, vuid);
	    isnew = true;
	} else {
	    isnew = false;
	}

	/*
	 * Now set the iteration variable to the name of this vertex.
	 */

	(void) Tcl_SetVar(interp, varname, vp->GetName(), 0);

	/*
	 * And finally execute the command.
	 */

	retone = Tcl_EvalObjEx(interp, objv[0], 0);
	switch (retone) {
	case TCL_OK:
	case TCL_CONTINUE:
	    if (isnew & !keep) {
		(void) vp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    break;
	case TCL_BREAK:
	    if (isnew & !keep) {
		(void) vp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    done = true;
	    break;
	case TCL_ERROR:
	default:
	    done = true;
	    ret = retone;
	    break;
	}
    }

    /*
     * If we are not keeping the values computed, then unset the iteration
     * variable also.
     */

    if (!keep) {
	(void) Tcl_UnsetVar(interp, varname, 0);
    }

    return ret;
}

/*
 * VisitMarkers --
 *
 *	Helper function to visit all markers marking this node.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the Tcl command executed does. May export new markers to
 *	Tcl.
 */

int
T4Node::VisitMarkers(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int i, l;
    bool done;
    bool wasExported;
    e4_Marker m;
    e4_MarkerUniqueID muid;
    T4Marker *mp;
    char *varname;
    int ret = TCL_OK, retone;

    /*
     * Expecting exacly two arguments, the name of a variable to set to
     * each marker visited, and a Tcl command to execute for each 
     * marker visited.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node foreach marker var cmd");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * The first argument is the name of the iteration variable.
     */

    varname = Tcl_GetString(objv[0]);

    /*
     * Visit each marker in turn, bind it to the iteration variable and
     * evaluate the cmd.
     */

    for (done = false, i = 1, l = n.MarkerCount(); (!done) && (i <= l); i++) {
	if ((!n.GetMarker(i, m)) || (!m.IsValid())) {
	    Tcl_AppendResult(interp,
			     "internal error: missing marker for node ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}

	/*
	 * See if this marker has already been exported to Tcl. If not,
	 * export it and create the Tcl command to manipulate it.
	 */

	(void) m.GetUniqueID(muid);
	mp = s->GetMarkerById(muid);
	if (mp == NULL) {
	    wasExported = false;
	    mp = new T4Marker(m, NULL, s);
	    mp->SetName(MakeMarkerCommand(interp, mp, muid));
	    s->StoreMarker(mp, muid);
	} else {
	    wasExported = true;
	}

	/*
	 * Now set the iteration variable to the name of this marker.
	 */

	(void) Tcl_SetVar(interp, varname, mp->GetName(), 0);

	/*
	 * And finally execute the command.
	 */

	retone = Tcl_EvalObjEx(interp, objv[1], 0);
	switch (retone) {
	case TCL_OK:
	case TCL_CONTINUE:
	    if (!wasExported) {
		(void) mp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    break;
	case TCL_BREAK:
	    if (!wasExported) {
		(void) mp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    done = true;
	    break;
	case TCL_ERROR:
	default:
	    done = true;
	    ret = retone;
	    break;
	}
    }

    /*
     * Unset the iteration variable.
     */

    (void) Tcl_UnsetVar(interp, varname, 0);

    return ret;
}

/*
 * VisitParents --
 *
 *	Helper function to visit all parent nodes of this node.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the Tcl command executed does. May export new nodes to
 *	Tcl.
 */

int
T4Node::VisitParents(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int i, l;
    bool done;
    bool wasExported;
    e4_Node p;
    e4_NodeUniqueID nuid;
    T4Node *np;
    char *varname;
    int ret = TCL_OK, retone;

    /*
     * Expecting exacly two arguments, the name of a variable to set to
     * each parent node visited, and a Tcl command to execute for
     * each node visited.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node foreach parent var cmd");
	return TCL_ERROR;
    }

    /*
     * Check that this node is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "node ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * The first argument is the name of the iteration variable.
     */

    varname = Tcl_GetString(objv[0]);

    /*
     * Visit each parent in turn, bind it to the iteration variable and
     * evaluate the cmd.
     */

    for (done = false, i = 1, l = n.ParentCount(); (!done) && (i <= l); i++) {
	if ((!n.GetParent(i, p)) || (!p.IsValid())) {
	    Tcl_AppendResult(interp,
			     "internal error: missing parent for node ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}

	/*
	 * See if we have exported this parent node already to Tcl. If not,
	 * then export it.
	 */

	(void) p.GetUniqueID(nuid);
	np = s->GetNodeById(nuid);
	if (np == NULL) {
	    wasExported = false;
	    np = new T4Node(p, NULL, s);
	    np->SetName(MakeNodeCommand(interp, np, nuid));
	    s->StoreNode(np, nuid);
	} else {
	    wasExported = true;
	}

	/*
	 * Now set the iteration variable to the name of this parent node.
	 */

	(void) Tcl_SetVar(interp, varname, np->GetName(), 0);

	/*
	 * And finally execute the command.
	 */

	/*
	 * And finally execute the command.
	 */

	retone = Tcl_EvalObjEx(interp, objv[1], 0);
	switch (retone) {
	case TCL_OK:
	case TCL_CONTINUE:
	    if (!wasExported) {
		(void) np->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    break;
	case TCL_BREAK:
	    if (!wasExported) {
		(void) np->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    done = true;
	    break;
	case TCL_ERROR:
	default:
	    done = true;
	    ret = retone;
	    break;
	}
    }

    /*
     * Unset the iteration variable.
     */

    (void) Tcl_UnsetVar(interp, varname, 0);

    return ret;
}

/*
 * Id --
 *
 *	Returns an identifier that is unique among all nodes in the
 *	storage containing this node.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result contains the integer ID.
 *
 * Side effects:
 *	None.
 */

int
T4Node::Id(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_NodeUniqueID nuid;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$node id");
	return TCL_ERROR;
    }

    /*
     * Check that the marker object is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "invalid node ", GetName(), NULL);
	return TCL_ERROR;
    }
    
    (void) n.GetUniqueID(nuid);
    Tcl_SetIntObj(Tcl_GetObjResult(interp), nuid.GetUniqueID());

    return TCL_OK;
}

/*
 * UserData --
 *
 *	Sets or returns the user data associated with this node.
 *
 * Results:
 *	A standard Tcl result. When retrieving the user data, upon
 *	success, the interpreter result contains an integer representing
 *	the stored user data.
 *
 * Side effects:
 *	May modify the user data associated with this node in its
 *	storage.
 */

int
T4Node::UserData(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int i;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node userdata ?newvalue?");
	return TCL_ERROR;
    }

    /*
     * Check that the node object is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "invalid node ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * If one argument is given, it must be an integer (the new value).
     * Attempt to extract the new value.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &i) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (!n.SetUserData(i)) {
	    Tcl_AppendResult(interp, "cannot set user data for node ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	return TCL_OK;
    }

    /*
     * No arguments given, so we want to retrieve the value.
     */

    if (!n.GetUserData(i)) {
	Tcl_AppendResult(interp, "cannot retrieve user data for node ",
			 GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to the retrieved value.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), i);

    return TCL_OK;
}

/*
 * DeleteFirstVertexWithNode --
 *
 *	Delete the first vertex in this node whose value is the given
 *	child node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result is
 *	empty.
 *
 * Side effects:
 *	May delete a vertex and its contents from this storage.
 */

int
T4Node::DeleteFirstVertexWithNode(Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[])
{
    T4Node *childp;
    e4_Node childn;

    /*
     * Expecting exactly one argument, the child node.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$node deletefirstvertexwithnode child");
	return TCL_ERROR;
    }

    /*
     * Check that the node object is valid.
     */

    if (!n.IsValid()) {
	Tcl_AppendResult(interp, "invalid node ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to retrieve the child node.
     */

    childp = s->GetNodeByName(Tcl_GetString(objv[0]));
    if (childp == NULL) {
	Tcl_AppendResult(interp, "could not retrieve node named ",
			 Tcl_GetString(objv[0]), NULL);
	return TCL_ERROR;
    }
    childp->ExternalizeNode(childn);
    if (!childn.IsValid()) {
	Tcl_AppendResult(interp, "node name ", Tcl_GetString(objv[0]),
			 " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to delete the first vertex containing the child node
     * in this node.
     */

    if (!n.DeleteFirstVertexWithNode(childn)) {
	Tcl_AppendResult(interp, "can not delete first vertex of ", GetName(),
			 " that contains the node ", Tcl_GetString(objv[0]),
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Success.
     */

    return TCL_OK;
}
