/*
 * t4open.cpp --
 *
 *	Implementation of the the T4Graph command to open a storage.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * OpenStorageProc --
 *
 *	This procedure is called when "tgraph::open" is invoked.
 *	Syntax:
 *		tgraph::open name	--OR--
 *		tgraph::open name driver
 *
 * Results:
 *	A standard Tcl result. The interpreter result will contain the
 *	name of the opened storage if successful.
 *
 * Side effects:
 *	May open a new storage.
 */

int
OpenStorageProc(ClientData cd, Tcl_Interp *interp, int objc,
		Tcl_Obj *CONST objv[])
{
    Tcl_HashTable *tblPtr;
    char *fnm = NULL;
    char *dnm = (char *) E4_METAKIT;
    T4Storage *sp;

    /*
     * Check arguments for conformance.
     */

    if ((objc != 2) && (objc != 3) && (objc != 4)) {
	Tcl_WrongNumArgs(interp, 1, objv, 
			 (char *) "filename ?driver? ?open-mode?");
	return TCL_ERROR;
    }

    /*
     * Get the storage filename from the first argument, and if present, the
     * driver name from the second argument.
     */

    fnm = Tcl_GetString(objv[1]);
    if (objc > 2) {
	dnm = Tcl_GetString(objv[2]);
    }

    /*
     * Check if the storage is already open in this interpreter. If it
     * is then return the same object.
     */

    tblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, T4_ASSOCKEY, NULL);
    if (tblPtr == NULL) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "internal error: storages table not found",
			       NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to find the requested storage among those registered.
     */

    sp = FindRegisteredStorage(tblPtr, fnm, dnm);
    if (sp != NULL) {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), sp->GetName(), -1);
	return TCL_OK;
    }

    /*
     * Attempt to open the requested storage.
     */

    e4_Storage s(fnm, dnm);

    /*
     * If the storage was not opened successfully, bail out
     * right away.
     */

    if (!s.IsValid()) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       fnm, ": cannot open with ", dnm,
			       NULL);
	return TCL_ERROR;
    }

    /*
     * Allocate a new storage record, initialize it,
     * and store it in the open storages table. Storing it
     * there allows tgraph::open to reuse it for subsequent
     * open requests on the same storage/driver arguments.
     */

    sp = new T4Storage(s, T4Storage::NextStorageName());

    /*
     * Associate the interpreter with this storage.
     */

    sp->SetAssociatedInterpreter(interp);

    /*
     * Return the Tcl name as the result and export it as a command.
     */

    if (MakeStorageCommand(interp, sp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    Tcl_ResetResult(interp);
    Tcl_SetStringObj(Tcl_GetObjResult(interp), sp->GetName(), -1);

    return TCL_OK;
}

/*
 * IsOpenStorageProc --
 *
 *	This procedure determines whether the given argument is the name
 *	of a physical storage open in this interpreter.
 *
 * Results:
 *	A standard Tcl result. The interpreter result will contain a boolean
 *	value denoting whether a physical storage by that name is open in this
 *	interpreter.
 *
 * Side effects:
 *	None.
 */

int
IsOpenStorageProc(ClientData cd, Tcl_Interp *interp, int objc,
		  Tcl_Obj *CONST objv[])
{
    Tcl_HashTable *tblPtr;
    Tcl_HashEntry *ePtr;
    char *snm;
    char *dnm = E4_METAKIT;
    char *key;

    /*
     * Expecting exactly one argument, the name of a physical storage.
     */

    if ((objc != 2) && (objc != 3)) {
	Tcl_WrongNumArgs(interp, 1, objv,
			 (char *) "storagename ?drivername?");
	return TCL_ERROR;
    }

    /*
     * Attempt to retrieve the storages hash table. If it's not found then
     * the storage is not open in this interpreter.
     */

    tblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp, T4_ASSOCKEY, NULL);
    if (tblPtr == NULL) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
	return TCL_OK;
    }

    /*
     * Extract the storage name, and if a driver name was supplied
     * use it in place of the default.
     */

    if (objc == 3) {
	dnm = Tcl_GetString(objv[2]);
    }
    snm = Tcl_GetString(objv[1]);

    /*
     * See if a storage is registered under the given name.
     */

    key = (char *) malloc(strlen(snm) + strlen(":") + strlen(dnm) + 1);
    sprintf(key, "%s:%s", snm, dnm);
    ePtr = Tcl_FindHashEntry(tblPtr, key);
    free(key);

    if (ePtr == NULL) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    }

    return TCL_OK;
}

/*
 * FindRegisteredStorage --
 *
 *	Helper procedure to retrieve a storage by name/driver pair.
 *
 * Results:
 *	A pointer to the T4Storage object if found, or NULL.
 *
 * Side effects:
 *	None.
 */

T4Storage *
FindRegisteredStorage(Tcl_HashTable *tblPtr, const char *fnm,
		      const char *dnm)
{
    Tcl_HashEntry *ePtr;
    char *buf;

    /*
     * Try to retrieve the hash key for the requested storage.
     */

    buf = (char *) malloc(strlen(fnm) + strlen(":") + strlen(dnm) + 1);
    sprintf(buf, "%s:%s", fnm, dnm);
    ePtr = Tcl_FindHashEntry(tblPtr, buf);
    free(buf);

    /*
     * If found, return the pointer to the T4Storage object.
     */

    if (ePtr != NULL) {
	return (T4Storage *) Tcl_GetHashValue(ePtr);
    }

    /*
     * Not found, return NULL.
     */

    return NULL;
}

/*
 * RegisterStorage --
 *
 *	Helper procedure to create hash entries for a storage in the
 *	table of known storages in the given interpreter.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	After this operation the storage is registered in the given
 *	interpreter.
 */

void
RegisterStorage(Tcl_HashTable *tblPtr, T4Storage *sp, const char *cmdname,
		const char *fnm, const char *dnm)
{
    Tcl_HashEntry *ePtr;
    int isnew;
    char *buf;

    /*
     * Try to create the hash key for the requested storage.
     */

    buf = (char *) malloc(strlen(fnm) + strlen(":") + strlen(dnm) + 1);
    sprintf(buf, "%s:%s", fnm, dnm);
    ePtr = Tcl_CreateHashEntry(tblPtr, buf, &isnew);
    free(buf);
    Tcl_SetHashValue(ePtr, sp);

    /*
     * Store the storage also under the name of its command:
     */

    ePtr = Tcl_CreateHashEntry(tblPtr, cmdname, &isnew);
    Tcl_SetHashValue(ePtr, sp);
}

/*
 * UnregisterStorage --
 *
 *	Helper procedure to delete the hash entries for a storage
 *	in the table of known storages for the given interpreter.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The entries for this storage in the table of known storages
 *	are deleted.
 */

void
UnregisterStorage(Tcl_HashTable *tblPtr, T4Storage *s, const char *cmdName,
		  const char *fnm, const char *dnm)
{
    Tcl_HashEntry *hPtr;
    char *key;

    /*
     * First delete the hash entry for the physical storage and driver name.
     */

    key = (char *) malloc(strlen(fnm) + strlen(":") + strlen(dnm) + 1);
    sprintf(key, "%s:%s", fnm, dnm);
    hPtr = Tcl_FindHashEntry(tblPtr, key);
    free(key);
    if (hPtr != NULL) {
	Tcl_DeleteHashEntry(hPtr);
    }

    /*
     * Also delete the entry for this storage under the Tcl command name.
     */

    hPtr = Tcl_FindHashEntry(tblPtr, cmdName);
    if (hPtr != NULL) {
	Tcl_DeleteHashEntry(hPtr);
    }
}
