/*
 * t4storagerep.cpp --
 *
 *	This file contains the implementation of the class T4Storage
 *	which is defined in .../include/t4graphrep.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * Private variable used to assign the next storage name:
 */

static int nextStorageID = 0;

/*
 * Element deletion callbacks.
 */

static void
MarkerDeletionAutoCallback(void *clientData, e4_Marker m)
{
    T4Storage *sp = (T4Storage *) clientData;

    sp->CleanupMarker(m);
}

static void
NodeDeletionAutoCallback(void *clientData, e4_Node n)
{
    T4Storage *sp = (T4Storage *) clientData;

    sp->CleanupNode(n);
}

static void
VertexDeletionAutoCallback(void *clientData, e4_Vertex v)
{
    T4Storage *sp = (T4Storage *) clientData;

    sp->CleanupVertex(v);
}

/*
 * These functions are declared for Tcl script callbacks.
 */

static void
MarkerAddCallbackFn(void *clientData, e4_Marker m)
{
    T4CallbackRecord *r = (T4CallbackRecord *) clientData;

    r->storage->AddMarkerCallback(r->interp, m);
}

static void
MarkerDelCallbackFn(void *clientData, e4_Marker m)
{
    T4CallbackRecord *r = (T4CallbackRecord *) clientData;

    r->storage->DelMarkerCallback(r->interp, m);
}

static void
NodeAddCallbackFn(void *clientData, e4_Node n)
{
    T4CallbackRecord *r = (T4CallbackRecord *) clientData;

    r->storage->AddNodeCallback(r->interp, n);
}

static void
NodeDelCallbackFn(void *clientData, e4_Node n)
{
    T4CallbackRecord *r = (T4CallbackRecord *) clientData;

    r->storage->DelNodeCallback(r->interp, n);
}

static void
VertexAddCallbackFn(void *clientData, e4_Vertex v)
{
    T4CallbackRecord *r = (T4CallbackRecord *) clientData;

    r->storage->AddVertexCallback(r->interp, v);
}

static void
VertexDelCallbackFn(void *clientData, e4_Vertex v)
{
    T4CallbackRecord *r = (T4CallbackRecord *) clientData;

    r->storage->DelVertexCallback(r->interp, v);
}

static void
VertexModCallbackFn(void *clientData, e4_Vertex v)
{
    T4CallbackRecord *r = (T4CallbackRecord *) clientData;

    r->storage->ModVertexCallback(r->interp, v);
}

/*
 * Constructor:
 */

T4Storage::T4Storage(e4_Storage ss, char *name)
{
    /*
     * Set the associated interpreter to NULL. We must do this manually
     * for the first time (instead of calling SetAssociatedInterpreter)
     * because that function uses the current value of the
     * associatedInterp field, which may be garbage at this stage.
     *
     * We set it to NULL because at this point we do not know which
     * interpreter this storage will be associated with.
     */

    associatedInterp = NULL;

    /*
     * Assign the given name and compute the string length.
     */

    SetName(name);

    /*
     * Copy the given e4_Storage object into our cache.
     */

    s = ss;

    /*
     * Initialize hash table storage for T4Graph objects that
     * were exported to Tcl:
     */

    exportedMarkers = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(exportedMarkers, TCL_ONE_WORD_KEYS);
    exportedNodes = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(exportedNodes, TCL_ONE_WORD_KEYS);
    exportedVertices = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(exportedVertices, TCL_ONE_WORD_KEYS);

    /*
     * Initialize the Tcl callback facility.
     */

    callbacks = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(callbacks, TCL_ONE_WORD_KEYS);

    cbaddmarker = 0;
    cbdelmarker = 0;
    cbaddnode = 0;
    cbdelnode = 0;
    cbaddvertex = 0;
    cbdelvertex = 0;
    cbmodvertex = 0;

    cbAddMarkerRecord = NULL;
    cbDelMarkerRecord = NULL;
    cbAddNodeRecord = NULL;
    cbDelNodeRecord = NULL;
    cbAddVertexRecord = NULL;
    cbDelVertexRecord = NULL;
    cbModVertexRecord = NULL;

    /*
     * Set up to be informed when an element is deleted from the underlying
     * storage.
     */

    s.DeclareMarkerDelCallback(MarkerDeletionAutoCallback, (void *) this);
    s.DeclareNodeDelCallback(NodeDeletionAutoCallback, (void *) this);
    s.DeclareVertexDelCallback(VertexDeletionAutoCallback, (void *) this);

    /*
     * No vertices are cached yet.
     */

    verticesCachedSinceLastSweep = 0;
}

/*
 * Destructor:
 */

T4Storage::~T4Storage()
{
    /*
     * Discard the hash tables for exported objects.
     */

    Tcl_DeleteHashTable(exportedMarkers);
    Tcl_Free((char *)exportedMarkers);
    Tcl_DeleteHashTable(exportedNodes);
    Tcl_Free((char *) exportedNodes);
    Tcl_DeleteHashTable(exportedVertices);
    Tcl_Free((char *) exportedVertices);

    /*
     * Clean up the callback facility.
     */

    RemoveAllCallbacks();
    Tcl_DeleteHashTable(callbacks);
    Tcl_Free((char *) callbacks);

    /*
     * Remove the auto callbacks for element deletion.
     */

    s.DeleteMarkerDelCallback(MarkerDeletionAutoCallback, (void *) this);
    s.DeleteNodeDelCallback(NodeDeletionAutoCallback, (void *) this);
    s.DeleteVertexDelCallback(VertexDeletionAutoCallback, (void *) this);

    /*
     * Free the allocated storage:
     */

    SetName(NULL);
    s = invalidStorage;
}

/*
 * RemoveAllCallbacks --
 *
 *	Clean up the callback facility.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	All callbacks registered on this storage are cancelled.
 */

void
T4Storage::RemoveAllCallbacks()
{
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;
    Tcl_Obj *o;

    /*
     * Clean up the callback facility.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {

        /*
	 * Discard the script.
	 */

	o = (Tcl_Obj *) Tcl_GetHashValue(ePtr);
	Tcl_DecrRefCount(o);

	/*
	 * Discard the callback record itself.
	 */

	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	switch (r->kind) {
	case E4_CBADDMARKER:
	    cbaddmarker--;
	    if (cbaddmarker <= 0) {
		s.DeleteMarkerAddCallback(MarkerAddCallbackFn,
					  (void *) cbAddMarkerRecord);
		delete cbAddMarkerRecord;
		cbAddMarkerRecord = NULL;
	    }
	    break;
	case E4_CBDELMARKER:
	    cbdelmarker--;
	    if (cbdelmarker <= 0) {
		s.DeleteMarkerDelCallback(MarkerDelCallbackFn,
					  (void *) cbDelMarkerRecord);
		delete cbDelMarkerRecord;
		cbDelMarkerRecord = NULL;
	    }
	    break;
	case E4_CBADDNODE:
	    cbaddnode--;
	    if (cbaddnode <= 0) {
		s.DeleteNodeAddCallback(NodeAddCallbackFn,
					(void *) cbAddNodeRecord);
		delete cbAddNodeRecord;
		cbAddNodeRecord = NULL;
	    }
	    break;
	case E4_CBDELNODE:
	    cbdelnode--;
	    if (cbdelnode <= 0) {
		s.DeleteNodeDelCallback(NodeDelCallbackFn,
					(void *) cbDelNodeRecord);
		delete cbDelNodeRecord;
		cbDelNodeRecord = NULL;
	    }
	    break;
	case E4_CBADDVERTEX:
	    cbaddvertex--;
	    if (cbaddvertex <= 0) {
		s.DeleteVertexAddCallback(VertexAddCallbackFn,
					  (void *) cbAddVertexRecord);
		delete cbAddVertexRecord;
		cbAddVertexRecord = NULL;
	    }
	    break;
	case E4_CBDELVERTEX:
	    cbdelvertex--;
	    if (cbdelvertex <= 0) {
		s.DeleteVertexDelCallback(VertexDelCallbackFn,
					  (void *) cbDelVertexRecord);
		delete cbDelVertexRecord;
		cbDelVertexRecord = NULL;
	    }
	    break;
	case E4_CBMODVERTEX:
	    cbmodvertex--;
	    if (cbmodvertex <= 0) {
		s.DeleteVertexModCallback(VertexModCallbackFn,
					  (void *) cbModVertexRecord);
		delete cbModVertexRecord;
		cbModVertexRecord = NULL;
	    }
	    break;
	}
	delete r;
    }
}

/*
 * RemoveAllExportedObjects --
 *
 *	Invalidate all T4Graph objects that have been exported to Tcl
 *	and that belong to this storage.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Invalidates all T4Graph objects that have been exported to Tcl
 *	and that belong to this storage.
 */

void
T4Storage::RemoveAllExportedObjects(Tcl_Interp *interp)
{
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    T4Marker *t;
    T4Node *n;
    T4Vertex *f;

    /*
     * Walk over the hash tables for each type of exported object and
     * delete all of them.
     */

    for (hPtr = Tcl_FirstHashEntry(exportedMarkers, &search);
	 hPtr != NULL;
	 hPtr = Tcl_FirstHashEntry(exportedMarkers, &search)) {
	t = (T4Marker *) Tcl_GetHashValue(hPtr);
	Tcl_DeleteHashEntry(hPtr);
	DeleteMarkerCommand(interp, t);
	delete t;
    }

    for (hPtr = Tcl_FirstHashEntry(exportedNodes, &search);
	 hPtr != NULL;
	 hPtr = Tcl_FirstHashEntry(exportedNodes, &search)) {
	n = (T4Node *) Tcl_GetHashValue(hPtr);
	Tcl_DeleteHashEntry(hPtr);
	DeleteNodeCommand(interp, n);
	delete n;
    }

    for (hPtr = Tcl_FirstHashEntry(exportedVertices, &search);
	 hPtr != NULL;
	 hPtr = Tcl_FirstHashEntry(exportedVertices, &search)) {
	f = (T4Vertex *) Tcl_GetHashValue(hPtr);
	Tcl_DeleteHashEntry(hPtr);
	DeleteVertexCommand(interp, f);
	delete f;
    }
}

/*
 * GetMarkerById --
 *
 *	Given a hash id, retrieve the associated T4Marker object.
 *
 * Results:
 *	The T4Marker object if found, NULL otherwise.
 *
 * Side effects:
 *	None.
 */

T4Marker *
T4Storage::GetMarkerById(e4_MarkerUniqueID muid)
{
   Tcl_HashEntry *hPtr;
   int id = muid.GetUniqueID();

   hPtr = Tcl_FindHashEntry(exportedMarkers, (char *) id);
   if (hPtr == NULL) {
	return NULL;
   }
   return (T4Marker *) Tcl_GetHashValue(hPtr);
}

/*
 * GetNodeById --
 *
 *	Given a hash id, retrieve the associated T4Node object.
 *
 * Results:
 *	The T4Node object if found, NULL otherwise.
 *
 * Side effects:
 *	None.
 */

T4Node *
T4Storage::GetNodeById(e4_NodeUniqueID nuid)
{
   Tcl_HashEntry *hPtr;
   int id = nuid.GetUniqueID();

   hPtr = Tcl_FindHashEntry(exportedNodes, (char *) id);
   if (hPtr == NULL) {
	return NULL;
   }
   return (T4Node *) Tcl_GetHashValue(hPtr);
}

/*
 * GetVertexById --
 *
 *	Given a hash id, retrieve the associated T4Vertex object.
 *
 * Results:
 *	The T4Vertex object if found, NULL otherwise.
 *
 * Side effects:
 *	None.
 */

T4Vertex *
T4Storage::GetVertexById(e4_VertexUniqueID vuid)
{
   Tcl_HashEntry *hPtr;
   int id = vuid.GetUniqueID();

   hPtr = Tcl_FindHashEntry(exportedVertices, (char *) id);
   if (hPtr == NULL) {
	return NULL;
   }
   return (T4Vertex *) Tcl_GetHashValue(hPtr);
}

/*
 * GetMarkerByName --
 *
 *	Given a marker name that has been previously exported to Tcl,
 *	retrieve its associated T4Marker object.
 *
 * Results:
 *	A T4Marker object or NULL if not found.
 *
 * Side effects:
 *	None.
 */

T4Marker *
T4Storage::GetMarkerByName(char *name)
{
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4Marker *m;

    for (ePtr = Tcl_FirstHashEntry(exportedMarkers, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	m = (T4Marker *) Tcl_GetHashValue(ePtr);
	if (strcmp(name, m->GetName()) == 0) {
	    return m;
	}
    }
    return NULL;
}

/*
 * GetNodeByName --
 *
 *	Given the name of a node previously exported to Tcl, retrieve
 *	its associated T4Node object.
 *
 * Results:
 *	A T4Node object or NULL if not found.
 *
 * Side effects:
 *	None.
 */

T4Node *
T4Storage::GetNodeByName(char *name)
{
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4Node *n;

    for (ePtr = Tcl_FirstHashEntry(exportedNodes, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	n = (T4Node *) Tcl_GetHashValue(ePtr);
	if (strcmp(name, n->GetName()) == 0) {
	    return n;
	}
    }
    return NULL;
}

/*
 * GetVertexByName --
 *
 *	Given the name of a vertex that has been previously exported
 *	to Tcl, retrieve its associated T4Vertex object.
 *
 * Results:
 *	A T4Vertex object or NULL if not found.
 *
 * Side effects:
 *	None.
 */

T4Vertex *
T4Storage::GetVertexByName(char *name)
{
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4Vertex *v;

    for (ePtr = Tcl_FirstHashEntry(exportedVertices, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	v = (T4Vertex *) Tcl_GetHashValue(ePtr);
	if (strcmp(name, v->GetName()) == 0) {
	    return v;
	}
    }
    return NULL;
}

/*
 * StoreMarker --
 *
 *	Make a new entry in the exported markers table for the given
 *	T4Marker object under the supplied id.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Subsequently the marker can be retrieved given the supplied id.
 */

void
T4Storage::StoreMarker(T4Marker *t, int id)
{
    Tcl_HashEntry *hPtr;
    int isnew;

    hPtr = Tcl_CreateHashEntry(exportedMarkers, (char *) id, &isnew);
    if (!isnew) {

	/*
	 * Should panic!
	 */

	fprintf(stderr, 
		"TGRAPH: duplicate marker storage for %d: 0x%x and 0x%x\n",
		id, t, (int) Tcl_GetHashValue(hPtr));
	return;
    }
    Tcl_SetHashValue(hPtr, t);
}

/*
 * StoreNode --
 *
 *	Make a new entry in the exported nodes table for the given
 *	T4Node object under the supplied id.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Subsequently the node can be retrieved given the supplied id.
 */

void
T4Storage::StoreNode(T4Node *n, int id)
{
    Tcl_HashEntry *hPtr;
    int isnew;

    hPtr = Tcl_CreateHashEntry(exportedNodes, (char *) id, &isnew);
    if (!isnew) {

	/*
	 * Should panic!
	 */

	fprintf(stderr,
		"TGRAPH: duplicate node storage for %d: 0x%x and 0x%x\n",
		id, n, (int) Tcl_GetHashValue(hPtr));
	return;
    }
    Tcl_SetHashValue(hPtr, n);
}

/*
 * StoreVertex --
 *
 *	Make a new entry in the exported vertices table for the given
 *	T4Vertex object under the supplied id.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Subsequently the vertex can be retrieved given the supplied id.
 */

void
T4Storage::StoreVertex(T4Vertex *f, int id)
{
    Tcl_HashEntry *hPtr;
    int isnew;

    hPtr = Tcl_CreateHashEntry(exportedVertices, (char *) id, &isnew);
    if (!isnew) {

	/*
	 * Should panic!
	 */

	fprintf(stderr, 
		"TGRAPH: duplicate vertex storage for %d: 0x%x and 0x%x\n",
		id, f, (int) Tcl_GetHashValue(hPtr));
	return;
    }
    Tcl_SetHashValue(hPtr, f);

    /*
     * Mark this vertex as used since last sweep, so that if we
     * do a sweep now we'll keep it.
     */

    f->SetFlag(T4_VERTEXUSED);

    /*
     * Increment the count of stored vertices. If the count exceeds a
     * threshhold then do a sweep through the vertex table, eliminating
     * all vertices that:
     *
     * - Have not been exported explicitly,
     * - and have not been used since the last sweep.
     *
     * For each such vertex we call its dispose method.
     */

    verticesCachedSinceLastSweep++;
    if (verticesCachedSinceLastSweep > T4_SWEEP_INTERVAL) {
	CleanupStoredVertices();
	verticesCachedSinceLastSweep = 0;
    }
}

/*
 * CleanupStoredVertices --
 *
 *	Sweep through the stored vertices and remove any that have not
 *	been since the last sweep and haven't been exported explicitly
 *	to Tcl (those are the programmer's responsibility to remove).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May cause some Tcl commands to become undefined, may free memory.
 */

void
T4Storage::CleanupStoredVertices()
{
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    T4Vertex *vp;

    for (hPtr = Tcl_FirstHashEntry(exportedVertices, &search);
	 hPtr != NULL;
	 ) {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
	hPtr = Tcl_NextHashEntry(&search);

	/*
	 * If this vertex was explicitly exported, skip over it.
	 */

	if (vp->IsFlagSet(T4_VERTEXEXPORTED)) {
	    continue;
	}

	/*
	 * If the vertex wasn't exported and wasn't used recently,
	 * remove it.
	 */

	if (!vp->IsFlagSet(T4_VERTEXUSED)) {
	    vp->Dispose(associatedInterp, 0, NULL);
	}
    }

    /*
     * Now sweep through once more to unset the used flags so that
     * subsequent sweeps can remove future unused vertices.
     */

    for (hPtr = Tcl_FirstHashEntry(exportedVertices, &search);
	 hPtr != NULL;
	 hPtr = Tcl_NextHashEntry(&search)) {
	vp = (T4Vertex *) Tcl_GetHashValue(hPtr);
	vp->UnsetFlag(T4_VERTEXUSED);
    }
}

/*
 * RemoveMarker --
 *
 *	Given an id, removes the entry for the associated marker
 *	from the hash table of exported markers in this storage.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Subsequently the marker can no longer be retrieved given the id.
 */

void
T4Storage::RemoveMarker(e4_MarkerUniqueID muid)
{
    Tcl_HashEntry *hPtr;
    int id = muid.GetUniqueID();

    hPtr = Tcl_FindHashEntry(exportedMarkers, (char *) id);
    if (hPtr == NULL) {
        return;
    }
    Tcl_DeleteHashEntry(hPtr);
}

/*
 * RemoveNode --
 *
 *	Given an id, removes the entry for the associated node
 *	from the hash table of exported nodes in this storage.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Subsequently the node can no longer be retrieved given the id.
 */

void
T4Storage::RemoveNode(e4_NodeUniqueID nuid)
{
    Tcl_HashEntry *hPtr;
    int id = nuid.GetUniqueID();

    hPtr = Tcl_FindHashEntry(exportedNodes, (char *) id);
    if (hPtr == NULL) {
        return;
    }
    Tcl_DeleteHashEntry(hPtr);
}

/*
 * RemoveVertex --
 *
 *	Given an id, removes the entry for the associated vertex
 *	from the hash table of exported vertices in this storage.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Subsequently the vertex can no longer be retrieved given the id.
 */

void
T4Storage::RemoveVertex(e4_VertexUniqueID vuid)
{
    Tcl_HashEntry *hPtr;
    int id = vuid.GetUniqueID();

    hPtr = Tcl_FindHashEntry(exportedVertices, (char *) id);
    if (hPtr == NULL) {
        return;
    }
    Tcl_DeleteHashEntry(hPtr);
}

/*
 * NextStorageName --
 *
 *	This procedure returns a string, the next name available for
 *	exporting a T4Graph storage object to Tcl.
 *
 * Results:
 *	The string name.
 *
 * Side effects:
 *	Allocates memory, increments a counter.
 */

char *
T4Storage::NextStorageName()
{
    char buf[32];
    char *name;
    int len = 8;	/* strlen("storage") + 1 */

    sprintf(buf, "%d", nextStorageID++);
    name = new char[len + strlen(buf)];
    sprintf(name, "storage%s", buf);

    return name;
}

/*
 * ExternalizeStorage --
 *
 *	This procedure gives the caller access to the underlying
 *	e4_Storage object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The caller can now manipulate the underlying e4_Storage
 *	object directly.
 */

void
T4Storage::ExternalizeStorage(e4_Storage &ss)
{
    ss = s;
}

/*
 ****************************************************************************
 *
 * The following methods implement Tcl sub-commands on a T4Storage
 * object.
 *
 ****************************************************************************
 */

/*
 * AutoCommit --
 *
 *	If called with one argument, sets auto-commit on or off. If called
 *	with no arguments, returns the current auto-commit setting.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May change the auto-commit behavior of the underlying storage.
 */

int
T4Storage::AutoCommit(Tcl_Interp *interp, int objc,
		      Tcl_Obj *CONST objv[])
{
    int bv;

    /*
     * Expecting either zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage auto-commit [onoff]");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If no arguments are given, return the current setting.
     */

    if (objc == 0) {
	if (s.AutoCommit()) {
	    Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
	} else {
	    Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
	}
	return TCL_OK;
    }
    
    /*
     * Extract the boolean value from the given argument.
     */

    if (Tcl_GetBooleanFromObj(interp, objv[0], &bv) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * Set the auto-commit value and also return the value to Tcl
     * to indicate the current setting.
     */

    s.AutoCommit(bv == 1 ? true : false);
    Tcl_SetObjResult(interp, objv[0]);

    return TCL_OK;
}

/*
 * Close --
 *
 *	Remove the T4Storage object from this interpreter without
 *	Deleteing it. If this is the last reference to the underlying
 *	e4Graph storage object then it will be closed.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The storage is no longer accessible in this interpreter, and
 *	all T4Graph objects that belong in this storage that have been
 *	exported to Tcl become invalid.
 */

int
T4Storage::Close(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage close");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    return InternalClose(interp);
}

/*
 * Helper function for closing the storage and cleaning up its state.
 */

int
T4Storage::InternalClose(Tcl_Interp *interp)
{
    Tcl_HashTable *storages;

    /*
     * Delete the Tcl command for this storage, and delete this T4Storage
     * object from the hash table for storages that are open in this interp.
     * Also remove all exported objects from this storage in this interp.
     */

    DeleteStorageCommand(interp, this);
    RemoveAllExportedObjects(interp);

    storages = (Tcl_HashTable *) Tcl_GetAssocData(interp, T4_ASSOCKEY, NULL);

    /*
     * The table of storages may be NULL because Tcl_GetAssocData does not
     * work during interpreter tear-down.
     */

    if (storages != NULL) {
	UnregisterStorage(storages, this, GetName(),
			  s.GetName(), s.GetDriver());
    }

    /*
     * No more references to this storage remain, so we delete its instance.
     * The destructor closes the storage by explicitly assigning 
     * invalidStorage to the instance variable.
     */

    delete this;

    return TCL_OK;
}

/*
 * Commit --
 *
 *	Commit any changes to this object to persistent storage.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	After this operation, the state of exported T4Graph objects
 *	exactly reflects the state in the persistent storage.
 */

int
T4Storage::Commit(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage commit");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to commit.
     */

    if (!s.Commit()) {
	Tcl_AppendResult(interp, "commit on storage ", GetName(), 
			 " failed", NULL);
	return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 * CopyTo --
 *
 *	Copies the contents of this e4Graph storage object to the given
 *	e4Graph storage object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	After this operation the other storage contains the same structure
 *	as this storage. All objects belonging to the other storage become
 *	invalid.
 */

int
T4Storage::CopyTo(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int forceCommit = 0;
    Tcl_HashTable *storages;
    Tcl_HashEntry *ePtr;
    T4Storage *otherStorage;
    e4_Storage os;

    /*
     * Expecting one or two arguments.
     */

    if ((objc != 1) && (objc != 2)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage copyto otherstorage ?commit?");
	return TCL_ERROR;
    }

    /*
     * If there are two arguments, attempt to obtain the boolean that
     * says whether we want to commit the other storage after the copy.
     */

    if (objc == 2) {
	if (Tcl_GetBooleanFromObj(interp, objv[1], &forceCommit) != TCL_OK) {
	    return TCL_ERROR;
	}
    }

    /*
     * Check that the underlying storage is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to get the storage from the given argument.
     */

    storages = (Tcl_HashTable *) Tcl_GetAssocData(interp, T4_ASSOCKEY, NULL);
    if (storages == NULL) {
	Tcl_AppendResult(interp, "internal error: invalid storage hash",
			 " table", NULL);
	return TCL_ERROR;
    }
    ePtr = Tcl_FindHashEntry(storages, Tcl_GetString(objv[0]));
    if (ePtr == NULL) {
	Tcl_AppendResult(interp, "unknown storage ", Tcl_GetString(objv[0]),
			 NULL);
	return TCL_ERROR;
    }
    otherStorage = (T4Storage *) Tcl_GetHashValue(ePtr);
    if (otherStorage == NULL) {
	Tcl_AppendResult(interp, "unknown storage ", Tcl_GetString(objv[0]),
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Disallow copying across interpreter boundaries, for security
     * reasons (the application may have "guessed" the name of another
     * storage and is trying to smash it). Behave as if the other storage
     * is unknown.
     */

    if (otherStorage->associatedInterp != associatedInterp) {
	Tcl_AppendResult(interp, "unknown storage ", Tcl_GetString(objv[0]),
			 NULL);
	return TCL_ERROR;
    }
    otherStorage->ExternalizeStorage(os);

    /*
     * Check the other storage is valid.
     */

    if (!os.IsValid()) {
	Tcl_AppendResult(interp, "invalid storage ", Tcl_GetString(objv[0]),
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to copy.
     */

    if (!s.CopyTo(os, (forceCommit == 0) ? false : true)) {
	Tcl_AppendResult(interp, "copy from ", GetName(), " to ",
			 Tcl_GetString(objv[0]), " failed", NULL);
	return TCL_ERROR;
    }

    /*
     * If we get here, success
     */

    Tcl_ResetResult(interp);
    return TCL_OK;
}

/*
 * Delete --
 *
 *	Deletes the underlying e4Graph storage object and the T4Graph
 *	storage object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The persistent storage for the T4Graph object may be Deleteed.
 *	After this operation, the T4Graph object can no longer be accessed
 *	in this interpreter and any T4Graph objects exported to Tcl that
 *	belong in this storage become invalid.
 */

int
T4Storage::Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Storage hold;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage Delete");
	return TCL_ERROR;
    }

    /*
     * Check that the underlying storage is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Hold onto the storage while we're closing this instance's reference.
     */

    hold = s;

    /*
     * Now clean up the in-core state.
     */

    if (InternalClose(interp) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * And finally delete the storage through the handle we held onto.
     */

    hold.Delete();

    return TCL_OK;
}

/*
 * MarkerExists --
 *
 *	Returns a boolean value indicating whether a marker by the given
 *	name exists in this storage.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result contains
 *	a boolean indicating whether the named marker exists.
 *
 * Side effects:
 *	None.
 */

int
T4Storage::MarkerExists(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting one argument, the name of the marker we are checking
     * for existence.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
			 (char *) "$storage markerexists markername");
	return TCL_ERROR;
    }

    /*
     * Check that the underlying storage is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Return whether the marker exists.
     */

    if (s.MarkerExists(Tcl_GetString(objv[0]))) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    }
    return TCL_OK;
}

/*
 * MarkerCount --
 *
 *	Returns the number of markers in this storage.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the number of markers in this storage.
 *
 * Side effects:
 *	None.
 */

int
T4Storage::MarkerCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage markercount");
	return TCL_ERROR;
    }

    /*
     * Check that the underlying storage is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Return the marker count.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), s.MarkerCount());
    return TCL_OK;
}

/*
 * Marker --
 *
 *	Exports the marker requested by name into Tcl.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the name of the exported marker.
 *
 * Side effects:
 *	May create a new T4Marker object.
 */

int
T4Storage::Marker(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Marker t;
    e4_MarkerUniqueID muid;
    T4Marker *tp;
    char *markername;

    /*
     * Expecting one argument, the name of the marker to be exported.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage marker name");
	return TCL_ERROR;
    }

    /*
     * Check that the underlying storage is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Get the requested marker.
     */

    markername = Tcl_GetString(objv[0]);
    if ((!s.GetMarker(markername, t)) || (!t.IsValid())) {
	Tcl_AppendResult(interp, "marker ", markername, " not found.", NULL);
	return TCL_ERROR;
    }

    /*
     * See if we already exported this marker to Tcl.
     */

    (void) t.GetUniqueID(muid);
    tp = GetMarkerById(muid);

    /*
     * If we did not export it, make a new T4Marker and store it under the
     * id obtained from the e4Marker object.
     */

    if (tp == NULL) {
	tp = new T4Marker(t, NULL, this);
	tp->SetName(MakeMarkerCommand(interp, tp, muid));
	StoreMarker(tp, muid);
    }

    Tcl_SetStringObj(Tcl_GetObjResult(interp), tp->GetName(), -1);
    return TCL_OK;
}

/*
 * MarkedNode --
 *
 *	Returns the node marked by a named marker. The marker is
 *	created if it does not exist yet.
 *
 * Results:
 *	Standard Tcl result. If successful, the interpreter result
 *	contains the name of the command for the node marked by the
 *	named marker.
 *
 * Side effects:
 *	May create a new T4Node object.
 */

int
T4Storage::MarkedNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node n;
    e4_NodeUniqueID nuid;
    T4Node *np;
    char *markername;

    /*
     * Expecting one argument, the name of the marker whose marked node
     * is to be exported.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage markednode name");
	return TCL_ERROR;
    }

    /*
     * Check that the underlying storage is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Get the requested marker.
     */

    markername = Tcl_GetString(objv[0]);
    if ((!s.GetMarkedNode(markername, n)) || (!n.IsValid())) {
	Tcl_AppendResult(interp, "marker ", markername, " not found.", NULL);
	return TCL_ERROR;
    }

    /*
     * See if we already exported this node to Tcl.
     */

    (void) n.GetUniqueID(nuid);
    np = GetNodeById(nuid);

    /*
     * If we did not export it, make a new T4Node and store it under the
     * id obtained from the e4_Node object.
     */

    if (np == NULL) {
	np = new T4Node(n, NULL, this);
	np->SetName(MakeNodeCommand(interp, np, nuid));
	StoreNode(np, nuid);
    }
    Tcl_SetStringObj(Tcl_GetObjResult(interp), np->GetName(), -1);
    return TCL_OK;
}

/*
 * IsEmpty --
 *
 *	Returns TRUE if the storage is empty, FALSE otherwise.
 *
 * Results:
 *	Standard Tcl result. If successful, the interpreter result
 *	contains a boolean value indicating whether the storage is
 *	empty.
 *
 * Side effects:
 *	None.
 */

int
T4Storage::IsEmpty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage isempty");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Return whether this storage is empty.
     */

    if (s.IsEmpty()) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    }
    return TCL_OK;
}

/*
 * MakeEmpty --
 *
 *	Makes this storage empty. As a side effect, all T4Graph objects
 *	exported to Tcl that belong to this storage become invalid.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Makes the persistent storage empty. As a side effect, all T4Graph
 *	objects exported to Tcl that belong to this storage become
 *	invalid.
 */

int
T4Storage::MakeEmpty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage makeempty");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Make the storage empty.
     */

    s.MakeEmpty();

    /*
     * Remove all T4Graph exported objects. This is not strictly needed
     * because we got callbacks when the exported elements were deleted,
     * in the MakeEmpty() call above.
     */

    RemoveAllExportedObjects(interp);

    return TCL_OK;
}

/*
 * IsValid --
 *
 *	Returns a boolean value indicating whether the T4Graph storage
 *	object is valid.
 *
 * Results:
 *	A standard Tcl result. Upon success the interpreter result
 *	contains a boolean value indicating whether the storage is
 *	valid.
 *
 * Side effects:
 *	None.
 */

int
T4Storage::IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage isvalid");
	return TCL_ERROR;
    }

    /*
     * Is this storage valid?
     */

    if (s.IsValid()) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    }
    return TCL_OK;
}

/*
 * IsDirty --
 *
 *	Returns a boolean value indicating whether the T4Graph storage
 *	object is dirty (it needs to be committed).
 *
 * Results:
 *	A standard Tcl result. Upon success the interpreter result
 *	contains a boolean value indicating whether the storage is
 *	dirty.
 *
 * Side effects:
 *	None.
 */

int
T4Storage::IsDirty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage isdirty");
	return TCL_ERROR;
    }

    /*
     * Is this storage valid?
     */

    if (s.IsDirty()) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    }
    return TCL_OK;
}

/*
 * MarkDirty --
 *
 *	Mark the storage as dirty (it needs to be committed).
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The storage is marked as dirty (it needs to be committed).
 */

int
T4Storage::MarkDirty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage markdirty");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Mark the storage as dirty.
     */

    s.MarkDirty();

    return TCL_OK;
}

/*
 * Name --
 *
 *	Sets the interpreter result to the name of this storage.
 *
 * Results:
 *	A standard Tcl result. Upon success the interpreter result
 *	contains a string, the name of this storage.
 *
 * Side effects:
 *	None.
 */

int
T4Storage::Name(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage name");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Return the string name of this storage. If the interpreter is
     * safe, do not tell him the real name of the storage. Instead tell
     * him only the name by which it is known in Tcl.
     */

    if (Tcl_IsSafe(interp)) {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), GetName(), -1);
    } else {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) s.GetName(), -1);
    }

    return TCL_OK;
}

/*
 * Driver --
 *
 *	Sets the interpreter result to the name of the driver used
 *	by this storage.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name of the driver used by this storage.
 *
 * Side effects:
 *	None.
 */

int
T4Storage::Driver(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage driver");
	return TCL_ERROR;
    }

    /*
     * Check that the storage object is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendResult(interp, "storage ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Return the name of the driver.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) s.GetDriver(), -1);

    return TCL_OK;
}

/*
 * Foreach --
 *
 *	Iterate over selected elements in this storage, executing an
 *	arbitrary piece of Tcl code for each selected element.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the evaluated Tcl code does. May export new elements of
 *	this storage to Tcl.
 */

static char *selectors[] = {
    (char *) "marker",
    (char *) "node",
    (char *) "vertex",
    (char *) NULL
};
typedef enum SSelectors {
    SMarker = 0,
    SNode,
    SVertex
} SSelectors;

int
T4Storage::Foreach(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    SSelectors index;

    /*
     * Expecting at least three arguments.
     */

    if (objc < 3) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
			 (char *) "$storage foreach sel var ?options? cmd");
	return TCL_ERROR;
    }

    /*
     * Figure out what selector was requested.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], selectors, (char *) "selector",
			    0, (int *) &index) != TCL_OK) {
	return TCL_ERROR;
    }

    objc--;
    objv++;

    switch (index) {
    case SMarker:
	return VisitMarkers(interp, objc, objv);
    case SNode:
	return VisitNodes(interp, objc, objv);
    case SVertex:
	return VisitVertices(interp, objc, objv);
    }

    /*
     * Not reached, but some compilers insist on this..
     */

    return TCL_ERROR;
}

/*
 * These filters help select markers, nodes and vertices to visit.
 */

static char *filters[] = {
    (char *) "-type",
    (char *) "-name",
    (char *) "-keep",
    (char *) NULL
};
typedef enum SFilters {
    SType = 0,
    SName,
    SKeep
} SFilters;

/*
 * VisitMarkers --
 *
 *	Helper function to iterate over all markers defined in this storage.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the evaluated Tcl code does. May export new elements of
 *	this storage to Tcl.
 */

int
T4Storage::VisitMarkers(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Marker m;
    e4_MarkerUniqueID muid;
    T4Marker *mp;
    char *varname;
    int ret = TCL_OK, retone;
    SFilters filter;
    bool done = false;
    bool keep = false;
    bool isnew;

    /*
     * Expecting two or four arguments, the name of a variable to set to
     * each marker as it is visited, an optional keep specifier (two args)
     * and the command to execute.
     */

    if ((objc != 2) && (objc != 4)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
	    (char *) "$storage foreach marker var ?-keep bool? cmd");
	return TCL_ERROR;
    }

    /*
     * Get the variable name.
     */

    varname = Tcl_GetString(objv[0]);

    /*
     * Parse the optional specifiers:
     */

    for (objc--, objv++; objc > 1; objc -= 2, objv += 2) {
	if (Tcl_GetIndexFromObj(interp, objv[0], filters, (char *) "filter",
				0, (int *) &filter) != TCL_OK) {
	    return TCL_ERROR;
	}

	switch (filter) {
	case SKeep:
	    if (Tcl_GetBooleanFromObj(interp, objv[1], (int *) &keep)
		!= TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	default:
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, 
			     "incorrect optional specifier \"", 
			     Tcl_GetString(objv[0]), "\"", NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Create the iterator.
     */

    e4_MarkerVisitor mv(s);

    /*
     * Iterate over every marker in the storage.
     */

    while ((!done) && (mv.CurrentMarkerAndAdvance(m))) {

        /*
	 * See if we already exported this marker to Tcl.
	 */

	(void) m.GetUniqueID(muid);
	mp = GetMarkerById(muid);

	/*
	 * If we did not export it, make a new T4Marker and store it under
	 * the ID obtained from the e4Marker object.
	 */

	if (mp == NULL) {
	    mp = new T4Marker(m, NULL, this);
	    mp->SetName(MakeMarkerCommand(interp, mp, muid));
	    StoreMarker(mp, muid);
	    isnew = true;
	} else {
	    isnew = false;
	}

	/*
	 * Now set the iteration variable to the name of this marker.
	 */

	(void) Tcl_SetVar(interp, varname, mp->GetName(), 0);

	/*
	 * And finally execute the command.
	 */

	retone = Tcl_EvalObjEx(interp, objv[0], 0);
	switch (retone) {
	case TCL_OK:
	case TCL_CONTINUE:
	    if (isnew & !keep) {
		(void) mp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    break;
	case TCL_BREAK:
	    if (isnew & !keep) {
		(void) mp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    done = true;
	    break;
	case TCL_ERROR:
	default:
	    done = true;
	    ret = retone;
	    break;
	}
    }

    /*
     * If we are not keeping newly exported elements, unset the iteration
     * variable at the end also.
     */

    if (!keep) {
	(void) Tcl_UnsetVar(interp, varname, 0);
    }

    return ret;
}

/*
 * VisitNodes --
 *
 *	Helper function to visit selected nodes in a storage.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the evaluated Tcl code does. May export new elements of
 *	this storage to Tcl.
 */

int
T4Storage::VisitNodes(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node n;
    e4_NodeUniqueID nuid;
    T4Node *np;
    char *varname;
    int ret = TCL_OK, retone;
    SFilters filter;
    bool done = false;
    bool keep = false;
    bool isnew;

    /*
     * Expecting two or four arguments, the name of a variable to set to
     * each node as it is visited, an optional keep specifier (two args),
     * and the command to execute.
     */

    if ((objc != 2) && (objc != 4)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
	    (char *) "$storage foreach node var ?-keep bool? cmd");
	return TCL_ERROR;
    }

    /*
     * Get the variable name.
     */

    varname = Tcl_GetString(objv[0]);

    /*
     * Parse the optional specifiers:
     */

    for (objc--, objv++; objc > 1; objc -= 2, objv += 2) {
	if (Tcl_GetIndexFromObj(interp, objv[0], filters, (char *) "filter",
				0, (int *) &filter) != TCL_OK) {
	    return TCL_ERROR;
	}

	switch (filter) {
	case SKeep:
	    if (Tcl_GetBooleanFromObj(interp, objv[1], (int *) &keep)
		!= TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	default:
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp,
			     "incorrect optional specifier \"", 
			     Tcl_GetString(objv[0]), "\"", NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Create the iterator.
     */

    e4_NodeVisitor nv(s);

    /*
     * Iterate over every node in the storage.
     */

    while ((!done) && (nv.CurrentNodeAndAdvance(n))) {

	/*
	 * See if we already exported this node to Tcl.
	 */

	(void) n.GetUniqueID(nuid);
	np = GetNodeById(nuid);

	/*
	 * If we did not export it, make a new T4Node and store it under
	 * the ID obtained from the e4Node object.
	 */

	if (np == NULL) {
	    np = new T4Node(n, NULL, this);
	    np->SetName(MakeNodeCommand(interp, np, nuid));
	    StoreNode(np, nuid);
	    isnew = true;
	} else {
	    isnew = false;
	}

	/*
	 * Now set the iteration variable to the name of this node.
	 */

	(void) Tcl_SetVar(interp, varname, np->GetName(), 0);

	/*
	 * And finally execute the command.
	 */

	retone = Tcl_EvalObjEx(interp, objv[0], 0);
	switch (retone) {
	case TCL_OK:
	case TCL_CONTINUE:
	    if (isnew & !keep) {
		(void) np->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    break;
	case TCL_BREAK:
	    if (isnew & !keep) {
		(void) np->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    done = true;
	    break;
	case TCL_ERROR:
	default:
	    done = true;
	    ret = retone;
	    break;
	}
    }

    /*
     * If we are not keeping newly exported elements, unset the iteration
     * variable at the end also.
     */

    if (!keep) {
	(void) Tcl_UnsetVar(interp, varname, 0);
    }

    return ret;
}

/*
 * VisitVertices --
 *
 *	Helper function to visit selected vertices in a storage.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result is left empty.
 *
 * Side effects:
 *	Whatever the evaluated Tcl command does. May export new elements
 *	of the storage to Tcl.
 */

static char *typenames[] = {
    (char *) "node",
    (char *) "int",
    (char *) "float",
    (char *) "string",
    (char *) "binary",
    (char *) NULL
};

int
T4Storage::VisitVertices(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Vertex v;
    e4_VertexUniqueID vuid;
    T4Vertex *vp;
    char *varname;
    char *namefilter = NULL;
    SFilters filter;
    e4_VertexType vt = E4_VTUNKNOWN;
    int ret = TCL_OK, retone;
    bool done = false;
    bool keep = false;
    bool isnew;

    /*
     * Expecting two, four, six or eight arguments, the name of a variable
     * to set to each vertex as it is visited, an optional type filter (two
     * args), an optional name filter (two args), an optional keep specifier
     * (two args), and the command to execute.
     */

    if ((objc != 2) && (objc != 4) && (objc != 6) && (objc != 8)) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *)
	    "$storage foreach vertex v ?-type t? ?-name n? ?-keep bool? cmd");
	return TCL_ERROR;
    }

    /*
     * Get the variable name.
     */

    varname = Tcl_GetString(objv[0]);

    /*
     * Parse the optional specifiers:
     */

    for (objc--, objv++; objc > 1; objc -= 2, objv += 2) {
	if (Tcl_GetIndexFromObj(interp, objv[0], filters, (char *) "filter",
				0, (int *) &filter) != TCL_OK) {
	    return TCL_ERROR;
	}

	switch (filter) {
	case SName:
	    namefilter = Tcl_GetString(objv[1]);
	    break;
	case SType:
	    if (Tcl_GetIndexFromObj(interp, objv[1], typenames, 
				    (char *) "typename", 0, (int *) &vt) 
		!= TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	case SKeep:
	    if (Tcl_GetBooleanFromObj(interp, objv[1], (int *) &keep)
		!= TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	}
    }

    /*
     * Create the iterator:
     */

    e4_VertexVisitor vv(s, (const char *) namefilter, vt);

    /*
     * Iterate over the selected vertices.
     */

    while ((!done) && (vv.CurrentVertexAndAdvance(v))) {

	/*
	 * See if we already exported this vertex to Tcl.
	 */

	(void) v.GetUniqueID(vuid);
	vp = GetVertexById(vuid);

	/*
	 * If we did not export it, make a new T4Vertex and store it under
	 * the ID obtained from the e4Vertex object.
	 */

	if (vp == NULL) {
	    vp = new T4Vertex(v, NULL, this);
	    vp->SetName(MakeVertexCommand(interp, vp, vuid));
	    StoreVertex(vp, vuid);
	    isnew = true;
	} else {
	    isnew = false;
	}

	/*
	 * Now set the iteration variable to the name of this vertex.
	 */

	(void) Tcl_SetVar(interp, varname, vp->GetName(), 0);

	/*
	 * And finally execute the command.
	 */

	retone = Tcl_EvalObjEx(interp, objv[0], 0);
	switch (retone) {
	case TCL_OK:
	case TCL_CONTINUE:
	    if (isnew & !keep) {
		(void) vp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    break;
	case TCL_BREAK:
	    if (isnew & !keep) {
		(void) vp->Dispose(interp, 0, NULL);
	    }
	    Tcl_ResetResult(interp);
	    done = true;
	    break;
	case TCL_ERROR:
	default:
	    done = true;
	    ret = retone;
	    break;
	}
    }

    /*
     * If we are not keeping the values computed, then unset the iteration
     * variable also.
     */

    if (!keep) {
	(void) Tcl_UnsetVar(interp, varname, 0);
    }

    return ret;
}

/*
 * Statistic --
 *
 *	Retrieve a statistic collected while running a TGraph application.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result contains
 *	the value of the statistic retrieved.
 *
 * Side effects:
 *	None.
 */

static char *statspaceselectors[] = {
    (char *) "marker",
    (char *) "node",
    (char *) "vertex",
    (char *) "name",
    (char *) "string",
    (char *) "int",
    (char *) "float",
    (char *) "binary",
    NULL
};
static char *statkindselectors[] = {
    (char *) "used",
    (char *) "available",
    (char *) "freed",
    (char *) "allocated",
    NULL
};

int
T4Storage::Statistic(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Space sp;
    e4_SpaceStat st;
    int v;

    /*
     * Expecting exactly two arguments.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage statistic space kind");
	return TCL_ERROR;
    }

    /*
     * Determine which space kind we want a statistic about.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], statspaceselectors,
			    (char *) "space", 0, (int *) &sp) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Determine which statistic we want about that space.
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], statkindselectors,
			    (char *) "kind", 0, (int *) &st) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Retrieve the actual statistic.
     */

    if (!s.GetStatistic(sp, st, v)) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "cannot retrieve statistic \"",
			       Tcl_GetString(objv[1]), 
			       "\" about space \"",
			       Tcl_GetString(objv[0]),
			       "\" in storage ",
			       GetName(), NULL);
	return TCL_ERROR;
    }

    Tcl_SetIntObj(Tcl_GetObjResult(interp), v);

    return TCL_OK;
}

/*
 * Get --
 *
 *	Get a TGraph entity identified by a given ID and kind string.
 *
 * Result:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains the name of the Tcl command to use to invoke
 *	operations on the TGraph entity.
 *
 * Side effects:
 *	May define a Tcl command for a newly exported TGraph entity.
 */

static char *objectkindselectors[] = {
    (char *) "marker",
    (char *) "node",
    (char *) "vertex",
    NULL
};

int
T4Storage::Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int id;
    T4ObjectKindSelector oksel;
    e4_MarkerUniqueID muid;
    e4_NodeUniqueID nuid;
    e4_VertexUniqueID vuid;
    e4_Marker m;
    e4_Node n;
    e4_Vertex v;
    T4Marker *tmp;
    T4Node *tnp;
    T4Vertex *tvp;

    /*
     * Expecting exactly two arguments, the kind of the object to retrieve
     * and its ID.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$storage get objkind id");
	return TCL_ERROR;
    }

    /*
     * Determine the kind of object to retrieve.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], objectkindselectors,
			    (char *) "objkind", 0, (int *) &oksel) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Attempt to get the integer ID of the object to retrieve:
     */

    if (Tcl_GetIntFromObj(interp, objv[1], &id) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Check that the storage is valid.
     */

    if (!s.IsValid()) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "invalid storage ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to retrieve the object requested:
     */
    switch (oksel) {
    case T4_OKMARKER:
        muid.SetUniqueID(id, s);

	/* 
	 * If the marker is already exported to Tcl, return its command name.
	 */

	tmp = GetMarkerById(muid);
	if (tmp != NULL) {
	    Tcl_SetStringObj(Tcl_GetObjResult(interp), tmp->GetName(), -1);
	    return TCL_OK;
	}

	/*
	 * Not exported yet. Retrieve the e4_Marker for this ID, then export
	 * the corresponding Tcl command.
	 */

	if ((!s.GetMarkerFromID(muid, m)) || (!m.IsValid())) {
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				   "invalid marker id ",
				   Tcl_GetString(objv[1]),
				   NULL);
	    return TCL_ERROR;
	}

	tmp = new T4Marker(m, NULL, this);
	tmp->SetName(MakeMarkerCommand(interp, tmp, muid));
	StoreMarker(tmp, muid);

	Tcl_SetStringObj(Tcl_GetObjResult(interp), tmp->GetName(), -1);

	return TCL_OK;

    case T4_OKNODE:
	nuid.SetUniqueID(id, s);

	/*
	 * If the node is already exported to Tcl, return its command name.
	 */

	tnp = GetNodeById(nuid);
	if (tnp != NULL) {
	    Tcl_SetStringObj(Tcl_GetObjResult(interp), tnp->GetName(), -1);
	    return TCL_OK;
	}

	/*
	 * Not exported yet. Retrieve the e4_Node for this ID, then export
	 * the corresponding Tcl command.
	 */

	if ((!s.GetNodeFromID(nuid, n)) || (!n.IsValid())) {
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				   "invalid node id ",
				   Tcl_GetString(objv[1]),
				   NULL);
	    return TCL_ERROR;
	}

	tnp = new T4Node(n, NULL, this);
	tnp->SetName(MakeNodeCommand(interp, tnp, nuid));
	StoreNode(tnp, nuid);

	Tcl_SetStringObj(Tcl_GetObjResult(interp), tnp->GetName(), -1);

	return TCL_OK;

    case T4_OKVERTEX:
	vuid.SetUniqueID(id, s);

	/*
	 * If the vertex is already exported to Tcl, return its command name.
	 */

	tvp = GetVertexById(vuid);
	if (tvp != NULL) {
	    Tcl_SetStringObj(Tcl_GetObjResult(interp), tvp->GetName(), -1);
	    return TCL_OK;
	}

	/*
	 * Not yet exported. Retrieve the e4_Vertex for this ID, then export
	 * its Tcl command.
	 */

	if ((!s.GetVertexFromID(vuid, v)) || (!v.IsValid())) {
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				   "invalid vertex id ",
				   Tcl_GetString(objv[1]),
				   NULL);
	    return TCL_ERROR;
	}

	tvp = new T4Vertex(v, NULL, this);
	tvp->SetName(MakeVertexCommand(interp, tvp, vuid));
	StoreVertex(tvp, vuid);

	Tcl_SetStringObj(Tcl_GetObjResult(interp), tvp->GetName(), -1);

	return TCL_OK;
    }

    /*
     * If we get here there's an internal error, because this code should
     * be unreachable.
     */

    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			   "internal error: \"$storage get\": unreachable",
			   " code", NULL);
    return TCL_ERROR;
}

/*
 * Callback --
 *
 *	Add or remove a callback for events of interest while running
 *	a TGraph application.
 *
 * Results:
 *	A standard Tcl result. Upon success, when a callback is added,
 *	the interpreter result contains the token to use when cancelling
 *	that callback later. Upon success, when a callback is removed,
 *	the interpreter result is empty.
 *
 * Side effects:
 *	Subsequently, when the specified event happens, a callback may
 *	occur or may stop to occur.
 */

static char *callbackeventselectors[] = {
    (char *) "add",
    (char *) "delete",
    (char *) "modify",
    NULL
};
static char *callbackactionselectors[] = {
    (char *) "add",
    (char *) "delete",
    (char *) "get",
    (char *) "set",
    (char *) "kind",
    (char *) "count",
    (char *) "exists",
    NULL
};

int
T4Storage::Callback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    T4CallbackActionSelector cbas;

    /*
     * Expecting at least two arguments.
     */

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage callback action arg ?arg ...?");
	return TCL_ERROR;
    }

    /*
     * Determine what kind of action (add or delete a callback) is requested.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], callbackactionselectors,
			    (char *) "action", 0, (int *) &cbas) != TCL_OK) {
	return TCL_ERROR;
    }

    switch (cbas) {
    case T4_CBASADD:
	return CBAddCallback(interp, objc-1, objv+1);
    case T4_CBASDEL:
	return CBDelCallback(interp, objc-1, objv+1);
    case T4_CBASGET:
	return CBGetCallback(interp, objc-1, objv+1);
    case T4_CBASSET:
	return CBSetCallback(interp, objc-1, objv+1);
    case T4_CBASKND:
	return CBKindCallback(interp, objc-1, objv+1);
    case T4_CBASCNT:
	return CBCountCallback(interp, objc-1, objv+1);
    case T4_CBASHAS:
	return CBHasCallback(interp, objc-1, objv+1);
    }

    /*
     * Some compilers (VC++) insist on this...
     */

    return TCL_ERROR;
}

/*
 * Helper functions to manage callbacks.
 */

int
T4Storage::CBAddCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    T4CallbackRecord *r, *savedr;
    Tcl_HashEntry *ePtr;
    int isnew;
    T4ObjectKindSelector cbos;
    T4CallbackEventSelector cbes;

    /*
     * Expecting three arguments.
     */

    if (objc != 3) {
	Tcl_WrongNumArgs(interp, 0, NULL,
		 (char *) "storage callback add objsel eventsel script");
	return TCL_ERROR;
    }

    /*
     * Attempt to parse the object selector.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], objectkindselectors,
			    (char *) "objsel", 0, (int *) &cbos) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Attempt to parse the event selector.
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], callbackeventselectors,
			    (char *) "eventsel", 0, (int *) &cbes) != TCL_OK){
	return TCL_ERROR;
    }

    r = new T4CallbackRecord;
    r->interp = interp;
    r->storage = this;

    switch (cbos) {
    case T4_OKMARKER:
	switch (cbes) {
	case T4_CBESMOD:
	    delete r;
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				   "modify callback not supported on markers",
				   NULL);
	    return TCL_ERROR;
	case T4_CBESADD:
	    if (cbaddmarker < 0) {
		cbaddmarker = 0;
	    }
	    if (cbaddmarker == 0) {
		savedr = new T4CallbackRecord;
		savedr->interp = interp;
		savedr->storage = this;
		savedr->kind = E4_CBADDMARKER;

		cbAddMarkerRecord = savedr;

		s.DeclareMarkerAddCallback(MarkerAddCallbackFn,
					   (void *) savedr);
	    }
	    cbaddmarker++;
	    r->kind = E4_CBADDMARKER;
	    break;
	case T4_CBESDEL:
	    if (cbdelmarker < 0) {
		cbdelmarker = 0;
	    }
	    if (cbdelmarker == 0) {
		savedr = new T4CallbackRecord;
		savedr->interp = interp;
		savedr->storage = this;
		savedr->kind = E4_CBDELMARKER;

		cbDelMarkerRecord = savedr;

		s.DeclareMarkerDelCallback(MarkerDelCallbackFn,
					   (void *) savedr);
	    }
	    cbdelmarker++;
	    r->kind = E4_CBDELMARKER;
	    break;
	}
	break;
    case T4_OKNODE:
	switch (cbes) {
	case T4_CBESMOD:
	    delete r;
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				   "modify callback not supported on nodes",
				   NULL);
	    return TCL_ERROR;
	case T4_CBESADD:
	    if (cbaddnode < 0) {
		cbaddnode = 0;
	    }
	    if (cbaddnode == 0) {
		savedr = new T4CallbackRecord;
		savedr->interp = interp;
		savedr->storage = this;
		savedr->kind = E4_CBADDNODE;
		
		cbAddNodeRecord = savedr;

		s.DeclareNodeAddCallback(NodeAddCallbackFn,
					 (void *) savedr);
	    }
	    cbaddnode++;
	    r->kind = E4_CBADDNODE;
	    break;
	case T4_CBESDEL:
	    if (cbdelnode < 0) {
		cbdelnode = 0;
	    }
	    if (cbdelnode == 0) {
		savedr = new T4CallbackRecord;
		savedr->interp = interp;
		savedr->storage = this;
		savedr->kind = E4_CBDELNODE;

		cbDelNodeRecord = savedr;

		s.DeclareNodeDelCallback(NodeDelCallbackFn,
					 (void *) savedr);
	    }
	    cbdelnode++;
	    r->kind = E4_CBDELNODE;
	    break;
	}
	break;
    case T4_OKVERTEX:
	switch (cbes) {
	case T4_CBESADD:
	    if (cbaddvertex < 0) {
		cbaddvertex = 0;
	    }
	    if (cbaddvertex == 0) {
		savedr = new T4CallbackRecord;
		savedr->interp = interp;
		savedr->storage = this;
		savedr->kind = E4_CBADDVERTEX;

		cbAddVertexRecord = savedr;

		s.DeclareVertexAddCallback(VertexAddCallbackFn,
					   (void *) savedr);
	    }
	    cbaddvertex++;
	    r->kind = E4_CBADDVERTEX;
	    break;
	case T4_CBESDEL:
	    if (cbdelvertex < 0) {
		cbdelvertex = 0;
	    }
	    if (cbdelvertex == 0) {
		savedr = new T4CallbackRecord;
		savedr->interp = interp;
		savedr->storage = this;
		savedr->kind = E4_CBDELVERTEX;

		cbDelVertexRecord = savedr;

		s.DeclareVertexDelCallback(VertexDelCallbackFn,
					   (void *) savedr);
	    }
	    cbdelvertex++;
	    r->kind = E4_CBDELVERTEX;
	    break;
	case T4_CBESMOD:
	    if (cbmodvertex < 0) {
		cbmodvertex = 0;
	    }
	    if (cbmodvertex == 0) {
		savedr = new T4CallbackRecord;
		savedr->interp = interp;
		savedr->storage = this;
		savedr->kind = E4_CBMODVERTEX;

		cbModVertexRecord = savedr;

		s.DeclareVertexModCallback(VertexModCallbackFn,
					   (void *) savedr);
	    }
	    cbmodvertex++;
	    r->kind = E4_CBMODVERTEX;
	    break;
	}
	break;
    }

    /*
     * Record the script.
     */

    ePtr = Tcl_CreateHashEntry(callbacks, (char *) r, &isnew);
    Tcl_SetHashValue(ePtr, objv[2]);
    Tcl_IncrRefCount(objv[2]);

    Tcl_SetIntObj(Tcl_GetObjResult(interp), (int) r);
    return TCL_OK;
}

int
T4Storage::CBDelCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    Tcl_HashEntry *ePtr;
    int i;
    T4CallbackRecord *r;
    Tcl_Obj *o;

    /*
     * Expecting one argument, the callback token.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage callback del callbacktoken");
	return TCL_ERROR;
    }

    /*
     * Extract the callback record.
     */

    if (Tcl_GetIntFromObj(interp, objv[0], &i) != TCL_OK) {
	return TCL_ERROR;
    }
    r = (T4CallbackRecord *) i;

    /*
     * Unrecord this script.
     */

    ePtr = E4_FINDHASHENTRY(callbacks, (char *) r);
    if (ePtr == NULL) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "invalid callback token ",
			       Tcl_GetString(objv[0]), NULL);
	return TCL_ERROR;
    }
    o = (Tcl_Obj *) Tcl_GetHashValue(ePtr);
    Tcl_DecrRefCount(o);

    /*
     * Delete the hash entry.
     */

    Tcl_DeleteHashEntry(ePtr);

    /*
     * Now figure out if this was the last recorded Tcl-level callback
     * of this kind. If so, then remove the e4Graph-level callback.
     */

    switch (r->kind) {
    case E4_CBADDMARKER:
	cbaddmarker--;
	if (cbaddmarker <= 0) {
	    s.DeleteMarkerAddCallback(MarkerAddCallbackFn,
				      (void *) cbAddMarkerRecord);
	    delete cbAddMarkerRecord;
	    cbAddMarkerRecord = NULL;
	}
	break;
    case E4_CBDELMARKER:
	cbdelmarker--;
	if (cbdelmarker <= 0) {
	    s.DeleteMarkerDelCallback(MarkerDelCallbackFn,
				      (void *) cbDelMarkerRecord);
	    delete cbDelMarkerRecord;
	    cbDelMarkerRecord = NULL;
	}
	break;
    case E4_CBADDNODE:
	cbaddnode--;
	if (cbaddnode <= 0) {
	    s.DeleteNodeAddCallback(NodeAddCallbackFn,
				    (void *) cbAddNodeRecord);
	    delete cbAddNodeRecord;
	    cbAddNodeRecord = NULL;
	}
	break;
    case E4_CBDELNODE:
	cbdelnode--;
	if (cbdelnode <= 0) {
	    s.DeleteNodeDelCallback(NodeDelCallbackFn,
				    (void *) cbDelNodeRecord);
	    delete cbDelNodeRecord;
	    cbDelNodeRecord = NULL;
	}
	break;
    case E4_CBADDVERTEX:
	cbaddvertex--;
	if (cbaddvertex <= 0) {
	    s.DeleteVertexAddCallback(VertexAddCallbackFn,
				      (void *) cbAddVertexRecord);
	    delete cbAddVertexRecord;
	    cbAddVertexRecord = NULL;
	}
	break;
    case E4_CBDELVERTEX:
	cbdelvertex--;
	if (cbdelvertex <= 0) {
	    s.DeleteVertexDelCallback(VertexDelCallbackFn,
				      (void *) cbDelVertexRecord);
	    delete cbDelVertexRecord;
	    cbDelVertexRecord = NULL;
	}
	break;
    case E4_CBMODVERTEX:
	cbmodvertex--;
	if (cbmodvertex <= 0) {
	    s.DeleteVertexModCallback(VertexModCallbackFn,
				      (void *) cbModVertexRecord);
	    delete cbModVertexRecord;
	    cbModVertexRecord = NULL;
	}
	break;
    }

    /*
     * Finally, delete the callback token.
     */

    delete r;

    return TCL_OK;
}

int
T4Storage::CBGetCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    Tcl_HashEntry *ePtr;
    int i;
    T4CallbackRecord *r;
    Tcl_Obj *o;

    /*
     * Expecting one argument, the callback token.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage callback get callbacktoken");
	return TCL_ERROR;
    }

    /*
     * Extract the callback record.
     */

    if (Tcl_GetIntFromObj(interp, objv[0], &i) != TCL_OK) {
	return TCL_ERROR;
    }
    r = (T4CallbackRecord *) i;

    /*
     * See if we have such a record.
     */

    ePtr = E4_FINDHASHENTRY(callbacks, (char *) r);
    if (ePtr == NULL) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "invalid callback token ",
			       Tcl_GetString(objv[0]), NULL);
	return TCL_ERROR;
    }
    o = (Tcl_Obj *) Tcl_GetHashValue(ePtr);

    /*
     * Now return the actual script.
     */

    Tcl_SetObjResult(interp, o);

    return TCL_OK;
}

int
T4Storage::CBSetCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    Tcl_HashEntry *ePtr;
    int i;
    T4CallbackRecord *r;
    Tcl_Obj *o;

    /*
     * Expecting two argument, the callback token and the new script.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage callback get callbacktoken");
	return TCL_ERROR;
    }

    /*
     * Extract the callback record.
     */

    if (Tcl_GetIntFromObj(interp, objv[0], &i) != TCL_OK) {
	return TCL_ERROR;
    }
    r = (T4CallbackRecord *) i;

    /*
     * See if we have such a record.
     */

    ePtr = E4_FINDHASHENTRY(callbacks, (char *) r);
    if (ePtr == NULL) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "invalid callback token ",
			       Tcl_GetString(objv[0]), NULL);
	return TCL_ERROR;
    }
    o = (Tcl_Obj *) Tcl_GetHashValue(ePtr);

    /*
     * Discard old script.
     */

    Tcl_DecrRefCount(o);

    /*
     * Install new script.
     */

    E4_SETHASHVALUE(ePtr, objv[1]);
    Tcl_IncrRefCount(objv[1]);
    
    /*
     * Now return the callback token, as we do when it was created.
     */

    Tcl_SetObjResult(interp, objv[0]);
    
    return TCL_OK;
}

int
T4Storage::CBKindCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    Tcl_HashEntry *ePtr;
    int i;
    T4CallbackRecord *r;

    /*
     * Expecting one argument, the callback token.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage callback kind callbacktoken");
	return TCL_ERROR;
    }

    /*
     * Extract the callback record.
     */

    if (Tcl_GetIntFromObj(interp, objv[0], &i) != TCL_OK) {
	return TCL_ERROR;
    }
    r = (T4CallbackRecord *) i;

    /*
     * See if we have such a record.
     */

    ePtr = E4_FINDHASHENTRY(callbacks, (char *) r);
    if (ePtr == NULL) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "invalid callback token ",
			       Tcl_GetString(objv[0]), NULL);
	return TCL_ERROR;
    }

    switch (r->kind) {
    case E4_CBADDMARKER:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "marker add", NULL);
	break;
    case E4_CBDELMARKER:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "marker delete", NULL);
	break;
    case E4_CBADDNODE:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "node add", NULL);
	break;
    case E4_CBDELNODE:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "node delete", NULL);
	break;
    case E4_CBADDVERTEX:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "vertex add", NULL);
	break;
    case E4_CBDELVERTEX:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "vertex delete", NULL);
	break;
    case E4_CBMODVERTEX:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "vertex modify", NULL);
	break;
    }

    return TCL_OK;
}

int
T4Storage::CBCountCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    T4ObjectKindSelector cbos;
    T4CallbackEventSelector cbes;

    /*
     * Expecting two arguments, an object selector and an event selector.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage callback count objsel eventsel");
	return TCL_ERROR;
    }

    /*
     * Attempt to parse the object selector.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], objectkindselectors,
			    (char *) "objsel", 0, (int *) &cbos) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Attempt to parse the event selector.
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], callbackeventselectors,
			    (char *) "eventsel", 0, (int *) &cbes) != TCL_OK){
	return TCL_ERROR;
    }

    switch (cbos) {
    case T4_OKMARKER:
	switch (cbes) {
	case T4_CBESMOD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), 0);
	    break;
	case T4_CBESADD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), cbaddmarker);
	    break;
	case T4_CBESDEL:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), cbdelmarker);
	    break;
	}
	break;
    case T4_OKNODE:
	switch (cbes) {
	case T4_CBESMOD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), 0);
	    break;
	case T4_CBESADD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), cbaddnode);
	    break;
	case T4_CBESDEL:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), cbdelnode);
	    break;
	}
	break;
    case T4_OKVERTEX:
	switch (cbes) {
	case T4_CBESMOD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), cbmodvertex);
	    break;
	case T4_CBESADD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), cbaddvertex);
	    break;
	case T4_CBESDEL:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), cbdelvertex);
	    break;
	}
	break;
    }

    return TCL_OK;
}

int
T4Storage::CBHasCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    T4ObjectKindSelector cbos;
    T4CallbackEventSelector cbes;

    /*
     * Expecting two arguments, an object selector and an event selector.
     */

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$storage callback exists objsel eventsel");
	return TCL_ERROR;
    }

    /*
     * Attempt to parse the object selector.
     */

    if (Tcl_GetIndexFromObj(interp, objv[0], objectkindselectors,
			    (char *) "objsel", 0, (int *) &cbos) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Attempt to parse the event selector.
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], callbackeventselectors,
			    (char *) "eventsel", 0, (int *) &cbes) != TCL_OK){
	return TCL_ERROR;
    }

    switch (cbos) {
    case T4_OKMARKER:
	switch (cbes) {
	case T4_CBESMOD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), 0);
	    break;
	case T4_CBESADD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), 
			  (cbaddmarker > 0) ? 1 : 0);
	    break;
	case T4_CBESDEL:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp),
			  (cbdelmarker > 0) ? 1 : 0);
	    break;
	}
	break;
    case T4_OKNODE:
	switch (cbes) {
	case T4_CBESMOD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), 0);
	    break;
	case T4_CBESADD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp),
			  (cbaddnode > 0) ? 1 : 0);
	    break;
	case T4_CBESDEL:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp),
			  (cbdelnode > 0) ? 1 : 0);
	    break;
	}
	break;
    case T4_OKVERTEX:
	switch (cbes) {
	case T4_CBESMOD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp),
			  (cbmodvertex > 0) ? 1 : 0);
	    break;
	case T4_CBESADD:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), 
			  (cbaddvertex > 0) ? 1 : 0);
	    break;
	case T4_CBESDEL:
	    Tcl_SetIntObj(Tcl_GetObjResult(interp),
			  (cbdelvertex > 0) ? 1 : 0);
	    break;
	}
	break;
    }

    return TCL_OK;
}

/*
 * These methods are called when a callback event occurs.
 */

void
T4Storage::AddMarkerCallback(Tcl_Interp *interp, e4_Marker m)
{
    e4_MarkerUniqueID muid;
    T4Marker *mp;
    char *objName;
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;
    bool wasCreated = false;

    if (!m.GetUniqueID(muid)) {
	return;
    }

    /*
     * Ensure the marker is exported to Tcl.
     */

    mp = GetMarkerById(muid);
    if (mp == NULL) {
	mp = new T4Marker(m, NULL, this);
	mp->SetName(MakeMarkerCommand(interp, mp, muid));
	StoreMarker(mp, muid);
	wasCreated = true;
    }

    /*
     * Get the name by which the marker is known in Tcl.
     */

    objName = mp->GetName();

    /*
     * Iterate over all callbacks, and invoke the ones for adding markers.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	if (r->kind != E4_CBADDMARKER) {
	    continue;
	}
	Tcl_ResetResult(interp);
	Tcl_VarEval(interp,
		    Tcl_GetString((Tcl_Obj *) Tcl_GetHashValue(ePtr)),
		    " ",
		    objName,
		    NULL);
    }

    if (wasCreated) {
	DeleteMarkerCommand(interp, mp);
	delete mp;
    }

    Tcl_ResetResult(interp);
}

void
T4Storage::DelMarkerCallback(Tcl_Interp *interp, e4_Marker m)
{
    e4_MarkerUniqueID muid;
    T4Marker *mp;
    char *objName;
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;
    bool wasCreated = false;

    if (!m.GetUniqueID(muid)) {
	return;
    }

    /*
     * Ensure the marker is exported to Tcl.
     */

    mp = GetMarkerById(muid);
    if (mp == NULL) {
	mp = new T4Marker(m, NULL, this);
	mp->SetName(MakeMarkerCommand(interp, mp, muid));
	StoreMarker(mp, muid);
	wasCreated = true;
    }

    /*
     * Get the name by which the marker is known in Tcl.
     */

    objName = mp->GetName();

    /*
     * Iterate over all callbacks and invoke the ones for deleting markers.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	if (r->kind != E4_CBDELMARKER) {
	    continue;
	}
	Tcl_ResetResult(interp);
	Tcl_VarEval(interp, 
		    Tcl_GetString((Tcl_Obj *) Tcl_GetHashValue(ePtr)),
		    " ",
		    objName,
		    NULL);
    }

    if (wasCreated) {
	DeleteMarkerCommand(interp, mp);
	RemoveMarker(muid);
	delete mp;
    }

    Tcl_ResetResult(interp);
}

void
T4Storage::AddNodeCallback(Tcl_Interp *interp, e4_Node n)
{
    e4_NodeUniqueID nuid;
    T4Node *np;
    char *objName;
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;

    if (!n.GetUniqueID(nuid)) {
	return;
    }

    /*
     * Ensure the node is exported to Tcl.
     */

    np = GetNodeById(nuid);
    if (np == NULL) {
	np = new T4Node(n, NULL, this);
	np->SetName(MakeNodeCommand(interp, np, nuid));
	StoreNode(np, nuid);
    }

    /*
     * Get the name by which the node is known in Tcl.
     */

    objName = np->GetName();

    /*
     * Iterate over all callbacks and invoke the ones for adding nodes.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	if (r->kind != E4_CBADDNODE) {
	    continue;
	}
	Tcl_ResetResult(interp);
	Tcl_VarEval(interp, 
		    Tcl_GetString((Tcl_Obj *) Tcl_GetHashValue(ePtr)),
		    " ",
		    objName,
		    NULL);
    }

    Tcl_ResetResult(interp);
}

void
T4Storage::DelNodeCallback(Tcl_Interp *interp, e4_Node n)
{
    e4_NodeUniqueID nuid;
    T4Node *np;
    char *objName;
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;
    bool wasCreated = false;

    if (!n.GetUniqueID(nuid)) {
	return;
    }

    /*
     * Ensure the node is exported to Tcl.
     */

    np = GetNodeById(nuid);
    if (np == NULL) {
	np = new T4Node(n, NULL, this);
	np->SetName(MakeNodeCommand(interp, np, nuid));
	StoreNode(np, nuid);
	wasCreated = true;
    }

    /*
     * Get the name by which the node is known in Tcl.
     */

    objName = np->GetName();

    /*
     * Iterate over all callbacks and invoke the ones for deleting nodes.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	if (r->kind != E4_CBDELNODE) {
	    continue;
	}
	Tcl_ResetResult(interp);
	Tcl_VarEval(interp, 
		    Tcl_GetString((Tcl_Obj *) Tcl_GetHashValue(ePtr)),
		    " ",
		    objName,
		    NULL);
    }

    if (wasCreated) {
	DeleteNodeCommand(interp, np);
	RemoveNode(nuid);
	delete np;
    }

    Tcl_ResetResult(interp);
}

void
T4Storage::AddVertexCallback(Tcl_Interp *interp, e4_Vertex v)
{
    e4_VertexUniqueID vuid;
    T4Vertex *vp;
    char *objName;
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;

    if (!v.GetUniqueID(vuid)) {
	return;
    }

    /*
     * Ensure the vertex is exported to Tcl.
     */

    vp = GetVertexById(vuid);
    if (vp == NULL) {
	vp = new T4Vertex(v, NULL, this);
	vp->SetName(MakeVertexCommand(interp, vp, vuid));
	StoreVertex(vp, vuid);
    }

    /*
     * Get the name by which this vertex is known in Tcl.
     */

    objName = vp->GetName();

    /*
     * Iterate over all callbacks and invoke the ones for adding vertices.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	if (r->kind != E4_CBADDVERTEX) {
	    continue;
	}
	Tcl_ResetResult(interp);
	Tcl_VarEval(interp,
		    Tcl_GetString((Tcl_Obj *) Tcl_GetHashValue(ePtr)),
		    " ",
		    objName,
		    NULL);
    }

    Tcl_ResetResult(interp);
}

void
T4Storage::DelVertexCallback(Tcl_Interp *interp, e4_Vertex v)
{
    e4_VertexUniqueID vuid;
    T4Vertex *vp;
    char *objName;
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;
    bool wasCreated = false;
    int res;

    if (!v.GetUniqueID(vuid)) {
	return;
    }

    /*
     * Ensure the vertex is exported to Tcl.
     */

    vp = GetVertexById(vuid);
    if (vp == NULL) {
	vp = new T4Vertex(v, NULL, this);
	vp->SetName(MakeVertexCommand(interp, vp, vuid));
	StoreVertex(vp, vuid);
	wasCreated = true;
    }
    
    /*
     * Get the name by which this vertex is known in Tcl.
     */

    objName = vp->GetName();

    /*
     * Iterate over all callbacks and invoke the ones for deleting vertices.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	if (r->kind != E4_CBDELVERTEX) {
	    continue;
	}
	Tcl_ResetResult(interp);
	res = Tcl_VarEval(interp, 
		    Tcl_GetString((Tcl_Obj *) Tcl_GetHashValue(ePtr)),
		    " ",
		    objName,
		    NULL);
    }

    if (wasCreated) {
	DeleteVertexCommand(interp, vp);
	RemoveVertex(vuid);
	delete vp;
    }

    Tcl_ResetResult(interp);
}

void
T4Storage::ModVertexCallback(Tcl_Interp *interp, e4_Vertex v)
{
    e4_VertexUniqueID vuid;
    T4Vertex *vp;
    char *objName;
    Tcl_HashEntry *ePtr;
    Tcl_HashSearch search;
    T4CallbackRecord *r;
    bool wasCreated = false;

    if (!v.GetUniqueID(vuid)) {
	return;
    }

    /*
     * Ensure the vertex is exported to Tcl.
     */

    vp = GetVertexById(vuid);
    if (vp == NULL) {
	vp = new T4Vertex(v, NULL, this);
	vp->SetName(MakeVertexCommand(interp, vp, vuid));
	StoreVertex(vp, vuid);
	wasCreated = true;
    }
    
    /*
     * Get the name by which this vertex is known in Tcl.
     */

    objName = vp->GetName();

    /*
     * Iterate over all callbacks and invoke the ones for deleting vertices.
     */

    for (ePtr = Tcl_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = Tcl_NextHashEntry(&search)) {
	r = (T4CallbackRecord *) Tcl_GetHashKey(callbacks, ePtr);
	if (r->kind != E4_CBMODVERTEX) {
	    continue;
	}
	Tcl_ResetResult(interp);
	Tcl_VarEval(interp, 
		    Tcl_GetString((Tcl_Obj *) Tcl_GetHashValue(ePtr)),
		    " ",
		    objName,
		    NULL);
    }

    if (wasCreated) {
	DeleteVertexCommand(interp, vp);
	RemoveVertex(vuid);
	delete vp;
    }

    Tcl_ResetResult(interp);
}

/*
 * Record the interpreter in which this storage is accessible from Tcl.
 */

void
T4Storage::SetAssociatedInterpreter(Tcl_Interp *interp)
{
    Tcl_HashTable *tp;

    if (associatedInterp != NULL) {
	tp = (Tcl_HashTable *) Tcl_GetAssocData(associatedInterp,
						T4_ASSOCKEY, NULL);
	if (tp == NULL) {
	    return;
	}
	UnregisterStorage(tp, this, GetName(), s.GetName(), s.GetDriver());
    }
    associatedInterp = interp;
    if ((interp != NULL) && (s.IsValid())) {
	tp = (Tcl_HashTable *) Tcl_GetAssocData(associatedInterp,
						T4_ASSOCKEY, NULL);

	/*
	 * If this interpreter does not have a registry of open storages,
	 * create it on the fly.
	 */

	if (tp == NULL) {
	    tp = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
	    Tcl_InitHashTable(tp, TCL_STRING_KEYS);
	    Tcl_SetAssocData(associatedInterp, T4_ASSOCKEY,
			     OpenStorageInterpDelProc, tp);
	}

	/*
	 * Now register the storage in the new interpreter.
	 */

	RegisterStorage(tp, this, GetName(), s.GetName(), s.GetDriver());
    }
}

/*
 * Retrieve the interpreter in which this storage is accessible from Tcl.
 */

Tcl_Interp *
T4Storage::GetAssociatedInterpreter()
{
    return associatedInterp;
}

/*
 * These methods are auto-called from a callback for element deletion.
 */

void
T4Storage::CleanupMarker(e4_Marker m)
{
    e4_MarkerUniqueID muid;
    T4Marker *mp;

    /*
     * Get the unique ID from the marker.
     */

    if (!m.GetUniqueID(muid)) {
	return;
    }

    /*
     * Check if this marker is currently exported to Tcl.
     */

    mp = GetMarkerById(muid);
    if (mp == NULL) {
	return;
    }

    /*
     * It is, so remove the command.
     */

    DeleteMarkerCommand(associatedInterp, mp);

    /*
     * And finally free up the storage for the marker object.
     */

    delete mp;
}

void
T4Storage::CleanupNode(e4_Node n)
{
    e4_NodeUniqueID nuid;
    T4Node *np;

    /*
     * Get the unique ID from the node.
     */

    if (!n.GetUniqueID(nuid)) {
	return;
    }

    /*
     * Check if this node is currently exported to Tcl.
     */

    np = GetNodeById(nuid);
    if (np == NULL) {
	return;
    }

    /*
     * It is, so remove the command for this node.
     */

    DeleteNodeCommand(associatedInterp, np);

    /*
     * And finally free up the storage for the node object.
     */

    delete np;
}

void
T4Storage::CleanupVertex(e4_Vertex v)
{
    e4_VertexUniqueID vuid;
    T4Vertex *vp;

    /*
     * Get the unique ID from the vertex.
     */

    if (!v.GetUniqueID(vuid)) {
	return;
    }

    /*
     * Check if this vertex is currently exported to Tcl.
     */

    vp = GetVertexById(vuid);
    if (vp == NULL) {
	return;
    }

    /*
     * It is, so remove the vertex from the cache of the containing node and
     * remove the command for this vertex.
     */

    vp->NodeRemoveVertexFromCache();
    DeleteVertexCommand(associatedInterp, vp);

    /*
     * And finally free up the storage for the vertex object.
     */

    delete vp;
}
