/*
 * t4stored.cpp --
 *
 *	This file implements procedures that manipulate "stored procedures",
 *	Tcl procedures stored as strings values of e4Graph vertices.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * DefineStoredProcedure --
 *
 *	Define a new Tcl command named "cmdname" that passes "nodename"
 *	as the value of the parameter "this" and has the body given in
 *	"storedproc".
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result is
 *	set to "cmdname".
 *
 * Side effects:
 *	May define a new Tcl procedure.
 */

int
DefineStoredProcedure(Tcl_Interp *interp, char *cmdname, char *nodename,
		      const char *storedproc)
{
    Tcl_Obj *obj, *args;
    int objc, res;
    Tcl_Obj **objv;
    char *buf, *argstr, *bodystr;

    /*
     * Parse the stored procedure into args and body:
     */

    obj = Tcl_NewStringObj(storedproc, strlen(storedproc));
    if (Tcl_ListObjGetElements(interp, obj, &objc, &objv) == TCL_ERROR) {
	return TCL_ERROR;
    }
    if (objc != 2) {
	Tcl_AppendResult(interp, cmdname, ": malformed stored procedure",
			 NULL);
	return TCL_ERROR;
    }

    args = objv[0];
    bodystr = Tcl_GetString(objv[1]);

    /*
     * Modify the args to pass "this" with value the name of the
     * containing node.
     */

    buf = Tcl_Alloc(10 + strlen(nodename));
    sprintf(buf, " {this %s}", nodename);
    Tcl_AppendStringsToObj(args, buf, NULL);
    Tcl_Free(buf);

    /*
     * Now prepare a new command to create a proc. Then invoke it with
     * Tcl_Eval.
     */

    argstr = Tcl_GetString(args);

    buf = Tcl_Alloc(4 + 1 + strlen(cmdname) + 1 + 2 + strlen(argstr) + 1 + 2 +
		    strlen(bodystr) + 1);
    sprintf(buf, "proc %s {%s} {%s}", cmdname, argstr, bodystr);
    res = Tcl_Eval(interp, buf);
    Tcl_Free(buf);
    Tcl_DecrRefCount(obj);
    return res;
}

/*
 * InvokeStoredProcedure --
 *
 *	Invokes a stored procedured defined by the passed-in cmdInfo.
 *
 * Results:
 *	A standard Tcl result. The interpreter result contains the value
 *	computed by the stored procedure itself.
 *
 * Side effects:
 *	Whatever the stored procedure does.
 */

int
InvokeStoredProcedure(Tcl_Interp *interp, char *cmdname,
		      int objc, Tcl_Obj *CONST objv[])
{
    Tcl_Obj *stackobjv[32];
    Tcl_Obj **nobjv = stackobjv;
    int i, res;

    /*
     * Expecting any number of arguments.
     */

    /*
     * Create a new objv with the zeroth slot the command name and the
     * rest of the slots the original arguments, then invoke Tcl_EvalObjv.
     */

    if (objc > 31) {
	nobjv = (Tcl_Obj **) Tcl_Alloc((objc + 1) * sizeof(Tcl_Obj *));
    }
    nobjv[0] = Tcl_NewStringObj(cmdname, strlen(cmdname));
    for (i = 0; i < objc; i++) {
	nobjv[i+1] = objv[i];
    }
    res = Tcl_EvalObjv(interp, objc + 1, nobjv, TCL_EVAL_GLOBAL);
    Tcl_DecrRefCount(nobjv[0]);
    if (nobjv != stackobjv) {
	Tcl_Free((char *) nobjv);
    }
    return res;
}
