/*
 * t4vertex.cpp --
 *
 *	This file implements the command procedure invoked when the
 *	Tcl procedure "t4graph::vertex" is called.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * The various sub-commands for "t4graph::vertex":
 */

static char *subCommands[] = {
    (char *) "kind",
    (char *) "get",
    (char *) "set",
    (char *) "setnode",
    (char *) "rank",
    (char *) "delete",
    (char *) "dispose",
    (char *) "isvalid",
    (char *) "type",
    (char *) "name",
    (char *) "rename",
    (char *) "node",
    (char *) "storage",
    (char *) "move",
    (char *) "next",
    (char *) "prev",
    (char *) "call",
    (char *) "id",
    (char *) "userdata",
    (char *) NULL
};
typedef enum FSubCommands {
    FKind = 0, 
    FGet,
    FSet,
    FSetNode,
    FRank,
    FDelete,
    FDispose,
    FIsValid,
    FType,
    FName,
    FRename,
    FNode,
    FStorage,
    FMove,
    FNext,
    FPrev,
    FCall,
    FId,
    FUserData
} FSubCommands;

/*
 * VertexCmdProc --
 *
 *	This procedure is invoked when the "t4graph::vertex" command is
 *	called.
 *
 * Results:
 *	A standard Tcl result. The interpreter result may contain various
 *	results depending on the sub-command selected.
 *
 * Side effects:
 *	Whatever the selected sub-command does.
 */

int
VertexCmdProc(ClientData cd, Tcl_Interp *interp, int objc,
	      Tcl_Obj *CONST objv[])
{
    T4Vertex *f;
    FSubCommands index;

    /*
     * There must be at least two arguments.
     */

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 0, NULL,
		 (char *) "$vertex cmd ?arg ...?");
	return TCL_ERROR;
    }

    /*
     * The client data is the T4Vertex for this instance.
     */

    f = (T4Vertex *) cd;

    /*
     * Figure out what operation was requested. There is no default --
     * the caller must specify an operation.
     */

    if (Tcl_GetIndexFromObj(interp, objv[1], subCommands, (char *) "cmd",
			    0, (int *) &index) != TCL_OK) {
	return TCL_ERROR;
    }

    objc -= 2;
    objv += 2;
    
    /*
     * Invoke the requested operation:
     */

    switch (index) {
    case FKind:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "vertex", -1);
	return TCL_OK;
    case FGet:
	return f->Get(interp, objc, objv);
    case FSet:
	return f->Set(interp, objc, objv);
    case FSetNode:
	return f->SetNode(interp, objc, objv);
    case FRank:
	return f->Rank(interp, objc, objv);
    case FDelete:
	return f->Delete(interp, objc, objv);
    case FDispose:
	return f->Dispose(interp, objc, objv);
    case FIsValid:
	return f->IsValid(interp, objc, objv);
    case FType:
	return f->Type(interp, objc, objv);
    case FName:
	return f->Name(interp, objc, objv);
    case FRename:
	return f->Rename(interp, objc, objv);
    case FNode:
	return f->Node(interp, objc, objv);
    case FStorage:
	Tcl_SetStringObj(Tcl_GetObjResult(interp),
			 f->GetStorage()->GetName(), -1);
	return TCL_OK;
    case FMove:
	return f->Move(interp, objc, objv);
    case FNext:
	return f->Next(interp, objc, objv);
    case FPrev:
	return f->Prev(interp, objc, objv);
    case FCall:
	return f->Call(interp, objc+1, objv-1);
    case FId:
	return f->Id(interp, objc, objv);
    case FUserData:
	return f->UserData(interp, objc, objv);
    default:
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			       "VertexProc: unreachable code!",
			       (char *) NULL);
	return TCL_ERROR;
    }
}
