/*
 * t4vertexrep.cpp --
 *
 *	This file contains the implementation of the class T4Vertex
 *	which is defined in .../include/t4graphrep.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "t4graph.h"

/*
 * The various (optional) type selector values for the T4Vertex::Set
 * operation.
 */

static char *typeSelectors[] = {
    (char *) "-int",
    (char *) "-double",
    (char *) "-string",
    (char *) "-node",
    (char *) "-binary",
    NULL
};
typedef enum TSelector {
    TSInteger = 0,
    TSDouble,
    TSString,
    TSNode,
    TSBinary
} TSelector;    

/*
 * Constructor:
 */

T4Vertex::T4Vertex(e4_Vertex ff, char *name, T4Storage *ss)
{
    SetName(name);
    flags = 0;
    f = ff;
    s = ss;
    storedcmdinfo = NULL;
    storedobj = NULL;
}

/*
 * Destructor:
 */

T4Vertex::~T4Vertex()
{
    f = invalidVertex;
    s = NULL;
    SetName(NULL);
    if (storedcmdinfo != NULL) {
	Tcl_Free((char *) storedcmdinfo);
    }
    if (storedobj != NULL) {
	Tcl_DecrRefCount(storedobj);
    }
}

/*
 * ExternalizeVertex --
 *
 *	This procedure gives the caller access to the underlying
 *	e4_Vertex object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The caller can now manipulate the underlying e4_Vertex
 *	object directly.
 */

void
T4Vertex::ExternalizeVertex(e4_Vertex &ff)
{
    ff = f;
}

/*
 * NodeRemoveVertexFromCache --
 *
 *	Remove this vertex from the cache in the node that contains
 *	the vertex.
 */

void
T4Vertex::NodeRemoveVertexFromCache()
{
    T4Node *np;
    e4_Node n;
    e4_NodeUniqueID nuid;

    /*
     * If the node is invalid, something is wrong, so we just bail out.
     */

    if ((!f.GetNode(n)) || (!n.IsValid())) {
	return;
    }

    /*
     * If the node has not been exported then this vertex is not cached,
     * so no need to do anything.
     */

    (void) n.GetUniqueID(nuid);
    np = s->GetNodeById(nuid);
    if (np == NULL) {
	return;
    }

    /*
     * Finally tell the node to remove this vertex from its cache.
     */

    np->RemoveVertexFromCache(this);
}

/*
 * NodeClearVertexCache --
 *
 *	Clears the vertex cache of the node containing this vertex.
 */

void
T4Vertex::NodeClearVertexCache()
{
    T4Node *np;
    e4_Node n;
    e4_NodeUniqueID nuid;

    if ((!f.GetNode(n)) || (!n.IsValid())) {
	return;
    }

    /*
     * See if we have exported this node to Tcl before. If not, then
     * the vertex is not cached, and we need not do anything.
     */

    (void) n.GetUniqueID(nuid);
    np = s->GetNodeById(nuid);
    if (np == NULL) {
	return;
    }

    /*
     * Finally tell the node to clear its cache.
     */

    np->ClearVertexCache();
}

/*
 ****************************************************************************
 *
 * The following methods implement Tcl sub-commands on a T4Vertex
 * object.
 *
 ****************************************************************************
 */

/*
 * Get --
 *
 *	Return the value of this vertex as a Tcl object.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains (an object representing) the value of this vertex.
 *
 * Side effects:
 *	May export more T4Graph objects (a T4Node) to Tcl.
 */

int
T4Vertex::Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    T4Node *np;
    int ii;
    double ff;
    const void *bytes;
    int nbytes;
    const char *ss;
    e4_Node nn;
    e4_NodeUniqueID nuid;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex get");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If there is a cached value, use that as the result.
     */

    if (storedobj != NULL) {
	Tcl_SetObjResult(interp, storedobj);
	return TCL_OK;
    }

    /*
     * Depending on the vertex type, get the value using different
     * access methods.
     */

    switch (f.Type()) {
    case E4_VTBINARY:
	if (!f.Get(bytes, nbytes)) {
	    Tcl_AppendResult(interp, "cannot retrieve value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) bytes, nbytes);
	return TCL_OK;
    case E4_VTINT:
	if (!f.Get(ii)) {
	    Tcl_AppendResult(interp, "cannot retrieve value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	Tcl_SetIntObj(Tcl_GetObjResult(interp), ii);
	break;
    case E4_VTDOUBLE:
	if (!f.Get(ff)) {
	    Tcl_AppendResult(interp, "cannot retrieve value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	Tcl_SetDoubleObj(Tcl_GetObjResult(interp), (double) ff);
	break;
    case E4_VTSTRING:
	if (!f.Get(ss)) {
	    Tcl_AppendResult(interp, "cannot retrieve value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) ss, -1);
	break;
    case E4_VTNODE:
	if (!f.Get(nn)) {
	    Tcl_AppendResult(interp, "cannot retrieve value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	(void) nn.GetUniqueID(nuid);
	np = s->GetNodeById(nuid);
	if (np == NULL) {
	    np = new T4Node(nn, NULL, s);
	    np->SetName(MakeNodeCommand(interp, np, nuid));
	    s->StoreNode(np, nuid);
	}
	Tcl_SetStringObj(Tcl_GetObjResult(interp), np->GetName(), -1);
	break;
    default:
	Tcl_AppendResult(interp, "t4graph::vertex storage get -- ",
			 " unreachable code!", NULL);
	return TCL_ERROR;
    }

    /*
     * Cache the Tcl object representing this value.
     */

    storedobj = Tcl_GetObjResult(interp);
    Tcl_IncrRefCount(storedobj);

    return TCL_OK;
}

/*
 * SetAs --
 *
 *	Sets the value of this vertex to the supplied value converted
 *	to the requested type.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May set the value of this vertex. T4Graph objects that were
 *	previously accessible through this vertex may become invalid.
 */

int
T4Vertex::SetAs(Tcl_Interp *interp, Tcl_Obj *val, int rt)
{
    TSelector reqtype = (TSelector) rt;
    int i;
    double d;
    char *sv;
    int len;
    T4Node *np;
    e4_Node nn;

    switch (reqtype) {
    case TSInteger:
	if (Tcl_GetIntFromObj(interp, val, &i) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (!f.Set(i)) {
	    Tcl_AppendResult(interp, "could not set value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	break;
    case TSDouble:
	if (Tcl_GetDoubleFromObj(interp, val, &d) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (!f.Set(d)) {
	    Tcl_AppendResult(interp, "could not set value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	break;
    case TSString:
	sv = Tcl_GetString(val);
	if (!f.Set(sv)) {
	    Tcl_AppendResult(interp, "could not set value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	break;
    case TSNode:
	sv = Tcl_GetString(val);
	np = s->GetNodeByName(sv);
	if (np == NULL) {
	    Tcl_AppendResult(interp, "could not retrieve node named ", sv,
			     NULL);
	    return TCL_ERROR;
	}
	np->ExternalizeNode(nn);
	if (!nn.IsValid()) {
	    Tcl_AppendResult(interp, "node name ", sv, " is invalid", NULL);
	    return TCL_ERROR;
	}
	if (!f.Set(nn)) {
	    Tcl_AppendResult(interp, "could not set value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	break;
    case TSBinary:
	sv = Tcl_GetStringFromObj(val, &len);
	if (!f.Set((void *) sv, len)) {
	    Tcl_AppendResult(interp, "could not set value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	break;
    default:
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "invalid requested type for assignment to ",
			 "vertex ", GetName(), NULL);
	return TCL_ERROR;
    }

    Tcl_ResetResult(interp);

    if (storedcmdinfo != NULL) {
	Tcl_Free((char *) storedcmdinfo);
	storedcmdinfo = NULL;
    }
    if (storedobj != NULL) {
	Tcl_DecrRefCount(storedobj);
    }
    storedobj = val;
    Tcl_IncrRefCount(storedobj);

    return TCL_OK;
}

/*
 * Set --
 *
 *	Sets the value of this vertex to the value supplied as argument.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May set the value of this vertex. T4Graph objects that were
 *	previously accessible through this vertex may become invalid.
 */

int
T4Vertex::Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *newval;
    int i;
    double d;
    T4Node *np;
    e4_Node nn;
    TSelector index;

    /*
     * Expecting one argument.
     */

    if ((objc != 1) && (objc != 2)) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
			 (char *) "$vertex set newval ?typesel?");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If a specific value type is requested, handle it specially by
     * attempting to convert the supplied value to the requested type
     * before assigning it to the vertex.
     */

    if (objc == 2) {
	if (Tcl_GetIndexFromObj(interp, objv[1], typeSelectors, 
				(char *) "typeselector", 0, (int *) &index)
	    != TCL_OK) {
	    return TCL_ERROR;
	}
	return SetAs(interp, objv[0], (int) index);
    }

    /*
     * Extract the value as a string.
     */

    newval = Tcl_GetString(objv[0]);

    /*
     * First of all see if the new value contains a '.'; if it does then
     * attempt to convert it to a double.
     */

    if (Tcl_StringMatch(newval, "?*.*")) {
	if (Tcl_GetDoubleFromObj(interp, objv[0], &d) == TCL_OK) {
	    if (!f.Set(d)) {
		Tcl_AppendResult(interp, "could not set value of vertex ",
				 GetName(), NULL);
		return TCL_ERROR;
	    }
	    goto end;
	}
    }

    /*
     * Attempt to get an integer value out of the argument. If that
     * works, set the vertex value to that integer.
     */

    if (Tcl_GetIntFromObj(interp, objv[0], &i) == TCL_OK) {
	if (!f.Set(i)) {
	    Tcl_AppendResult(interp, "could not set value of vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	goto end;
    }
    Tcl_ResetResult(interp);

    /*
     * Attempt to convert the string into the name of a node.
     */

    np = s->GetNodeByName(newval);
    if (np != NULL) {
	np->ExternalizeNode(nn);
	if (nn.IsValid()) {
	    if (!f.Set(nn)) {
		Tcl_AppendResult(interp, "could not set value of vertex ",
				 GetName(), NULL);
		return TCL_ERROR;
	    }
	    goto end;
	}
    }

    /*
     * Last resort - set the value as a string.
     */

    if (!f.Set(newval)) {
	Tcl_AppendResult(interp, "could not set value of vertex ",
			 GetName(), NULL);
	return TCL_ERROR;
    }
end:
    if (storedcmdinfo != NULL) {
	Tcl_Free((char *) storedcmdinfo);
	storedcmdinfo = NULL;
    }
    if (storedobj != NULL) {
	Tcl_DecrRefCount(storedobj);
    }
    storedobj = objv[0];
    Tcl_IncrRefCount(storedobj);

    return TCL_OK;
}

/*
 * SetNode --
 *
 *	Set the value of this vertex to a new node. Also export the
 *	new node to Tcl.
 *
 * Results:
 *	A standard Tcl result. If successful, the interpreter result
 *	contains the name by which the new node is known in Tcl.
 *
 * Side effects:
 *	May create a new node. May export a new T4Node object to Tcl.
 *	T4Graph objects that were previously accessible through this
 *	vertex may become invalid.
 */

int
T4Vertex::SetNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node n;
    e4_NodeUniqueID nuid;
    T4Node *np;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex setnode");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to set the vertex to a new node.
     */

    if (!f.SetNode(n)) {
	Tcl_AppendResult(interp, "could not set vertex ", GetName(),
			 " to a new node", NULL);
	return TCL_ERROR;
    }

    /*
     * Create the new T4Node object and export it to Tcl.
     */

    (void) n.GetUniqueID(nuid);
    np = new T4Node(n, NULL, s);
    np->SetName(MakeNodeCommand(interp, np, nuid));
    s->StoreNode(np, nuid);
    Tcl_SetStringObj(Tcl_GetObjResult(interp), np->GetName(), -1);

    if (storedcmdinfo != NULL) {
	Tcl_Free((char *) storedcmdinfo);
	storedcmdinfo = NULL;
    }
    if (storedobj != NULL) {
	Tcl_DecrRefCount(storedobj);
    }
    storedobj = Tcl_GetObjResult(interp);
    Tcl_IncrRefCount(storedobj);

    return TCL_OK;
}

/*
 * Rank --
 *
 *	Return the rank of this vertex in its containing node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter
 *	result contains an integer, the rank of this vertex in
 *	its containing node.
 *
 * Side effects:
 *	None.
 */

int
T4Vertex::Rank(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex rank");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to contain the rank.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), f.Rank());

    return TCL_OK;
}

/*
 * Delete --
 *
 *	Delete this vertex from its containing node.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May delete this vertex from the persistent storage. T4Graph objects
 *	that were previously accessible through the value of this vertex
 *	may become invalid.
 */

int
T4Vertex::Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex delete");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Clear the vertex cache of the node containing this vertex, because
     * the mapping of names to vertices may have changed.
     */

    NodeClearVertexCache();

    /*
     * Attempt to delete this vertex.
     */

    if (!f.Delete()) {
	Tcl_AppendResult(interp, "could not delete vertex ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Do not leave remnant results in the interpreter.
     */

    Tcl_ResetResult(interp);

    return TCL_OK;
}

/*
 * Dispose --
 *
 *	Disposes of the Tcl command for this vertex.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Disposes of the Tcl command for this vertex. After this call,
 *	the Tcl command for this vertex can no longer be used by Tcl
 *	programs. The vertex still exists in its persistent storage.
 *
 *	This call may (re)create the tgraph instance and Tcl command for
 *	the node containing this vertex, if that instance and command
 *	have not yet been exported to Tcl.
 */

int
T4Vertex::Dispose(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex dispose");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex object is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "invalid vertex ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Delete the Tcl command that refers to this vertex and ask the storage
     * to forget this instance.
     */

    DeleteVertexCommand(interp, this);

    /*
     * Remove this vertex from the cache in its node because it is no
     * longer accessible from Tcl.
     */

    NodeRemoveVertexFromCache();

    /*
     * Dispose of the tgraph vertex instance.
     */

    delete this;

    /*
     * Returns success.
     */

    return TCL_OK;
}

/*
 * IsValid --
 *
 *	Return a boolean indicating whether this vertex is valid.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a boolean value indicating whether this vertex is valid.
 *
 * Side effects:
 *	None.
 */

int
T4Vertex::IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex isvalid");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
    } else {
	Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
    }

    return TCL_OK;
}

/*
 * Type --
 *
 *	Returns a string representing the type of the value stored
 *	in this vertex.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string representing the type of the value stored
 *	in this vertex.
 *
 * Side effects:
 *	None.
 */

int
T4Vertex::Type(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int userData = 0;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex type");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    switch (f.Type()) {
    case E4_VTUNKNOWN:
	Tcl_AppendResult(interp, "could not retrieve type of vertex ",
			 GetName(), NULL);
	return TCL_ERROR;
    case E4_VTNODE:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "node", -1);
	return TCL_OK;
    case E4_VTINT:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "int", -1);
	return TCL_OK;
    case E4_VTDOUBLE:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "double", -1);
	return TCL_OK;
    case E4_VTSTRING:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "string", -1);
	return TCL_OK;
    case E4_VTBINARY:
	Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) "binary", -1);
	return TCL_OK;
    default:
	Tcl_AppendResult(interp, "$vertex type -- unreachable code!", NULL);
	return TCL_ERROR;
    }
}

/*
 * Name --
 *
 *	Returns the name of this vertex in its containing node.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name of this vertex in its containing
 *	node.
 *
 * Side effects:
 *	None.
 */

int
T4Vertex::Name(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex name");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to the name of this vertex.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), (char *) f.Name(), -1);

    return TCL_OK;
}

/*
 * Rename --
 *
 *	Rename this vertex within its containing node to the value of
 *	the argument.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The vertex is renamed in the persistent storage.
 */

int
T4Vertex::Rename(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /*
     * Expecting one argument.
     */

    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex rename newname");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to rename the vertex.
     */

    if (!f.Rename(Tcl_GetString(objv[0]))) {
	Tcl_AppendResult(interp, "could not rename vertex ", GetName(),
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Remove this vertex from the cache in the node which contains this
     * vertex, because the vertex is no longer accessible from Tcl.
     */

    NodeRemoveVertexFromCache();

    return TCL_OK;
}

/*
 * Node --
 *
 *	Set the interpreter result to the name assigned to the node
 *	containing this vertex.
 *
 * Results:
 *	A standard Tcl result. Upon success, the interpreter result
 *	contains a string, the name assigned to the node containing
 *	this vertex.
 *
 * Side effects:
 *	May export a T4Node object to Tcl.
 */

int
T4Vertex::Node(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Node n;
    e4_NodeUniqueID nuid;
    T4Node *np;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex node");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Attempt to retrieve the node containing this vertex.
     */

    if ((!f.GetNode(n)) || (!n.IsValid())) {
	Tcl_AppendResult(interp, "could not get containing node for vertex ",
			 GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * See if we have exported this node to Tcl before. If not, create
     * a new T4Node and make it ready to be exported.
     */

    (void) n.GetUniqueID(nuid);
    np = s->GetNodeById(nuid);
    if (np == NULL) {
	np = new T4Node(n, NULL, s);
	np->SetName(MakeNodeCommand(interp, np, nuid));
	s->StoreNode(np, nuid);
    }

    /*
     * Set the interpreter result to the name by which this node is known
     * in Tcl.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), np->GetName(), -1);

    return TCL_OK;
}

/*
 * Move --
 *
 *	Move a given vertex into the node containing this vertex at the
 *	position determined by the given insertOrder and offset. If
 *	insertOrder is "before" or "at", offset is subtracted from the
 *	rank of this vertex plus one to get the rank at which to insert
 *	the given vertex. If insertOrder is "after", offset is added to
 *	the rank of this vertex to get the rank at which to insert the
 *	vertex. If insertOrder is "first" or "last", the vertex is inserted
 *	as the first or last vertex in the node containing this vertex; in
 *	these cases offset is ignored.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	The vertex is moved in the persisent storage.
 */

int
T4Vertex::Move(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   e4_Vertex of;
   T4Vertex *ofp;
   int offset;
   e4_InsertOrder io;

   /*
    * Expecting three arguments.
    */

    if (objc != 3) {
	Tcl_WrongNumArgs(interp, 0, NULL, 
			 (char *) "$vertex move vertex insertorder offset");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Try to extract the offset:
     */

    if (Tcl_GetIntFromObj(interp, objv[2], &offset) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * Try to extract the insertOrder:
     */

    if (ParseInsertOrder(interp, objv[1], &io) == TCL_ERROR) {
	return TCL_ERROR;
    }
    
    /*
     * Try to locate the other vertex.
     */

    ofp = s->GetVertexByName(Tcl_GetString(objv[0]));
    if (ofp == NULL) {

	/*
	 * The other vertex is not known in Tcl. We cannot move it.
	 */

	Tcl_AppendResult(interp, "unknown vertex ",
			 Tcl_GetString(objv[0]), NULL);
	return TCL_ERROR;
    }

    ofp->ExternalizeVertex(of);

    /*
     * Clear the vertex cache of the node containing the vertex being
     * moved, because the mapping of names to vertices may have changed.
     * We must do this before the move, because the vertex may be moving
     * to another node.
     */

    ofp->NodeClearVertexCache();

    /*
     * Attempt the move.
     */

    if (!f.MoveVertex(of, io, offset)) {
	Tcl_AppendResult(interp, "could not move vertex ",
			 Tcl_GetString(objv[0]),
			 " relative to vertex ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Clear the vertex cache of the node containing this vertex, because
     * the mapping of names to vertices may have changed.
     */

    NodeClearVertexCache();

    return TCL_OK;
}

/*
 * Next --
 *
 *	Return the next vertex after this one in the node containing
 *	this vertex.
 *
 * Results:
 *	A standard Tcl result. Upon success the interpreter result
 *	contains the name by which the next vertex is known in Tcl or
 *	the empty string if this vertex is not followed by any vertices.
 *
 * Side effects:
 *	May export a vertex to Tcl.
 */

int
T4Vertex::Next(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Vertex vv;
    e4_VertexUniqueID vuid;
    T4Vertex *nv;
    int num = 1;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex next ?num?");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If there is one argument, it is the number of vertices to go
     * forward.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &num) == TCL_ERROR) {
	    return TCL_ERROR;
	}
    }

    /*
     * Attempt to retrieve the next vertex.
     */

    if ((!f.Next(num, vv)) || (!vv.IsValid())) {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), "", -1);
	return TCL_OK;
    }

    /*
     * See if we have exported this vertex to Tcl before. If not, create
     * a new T4Vertex and make it ready to be exported.
     */

    (void) vv.GetUniqueID(vuid);
    nv = s->GetVertexById(vuid);
    if (nv == NULL) {
	nv = new T4Vertex(vv, NULL, s);
	nv->SetName(MakeVertexCommand(interp, nv, vuid));
	s->StoreVertex(nv, vuid);
    }

    /*
     * Set the interpreter result to the name by which this vertex is
     * known in Tcl.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), nv->GetName(), -1);

    return TCL_OK;
}

/*
 * Prev --
 *
 *	Return the previous vertex after this one in the node containing
 *	this vertex.
 *
 * Results:
 *	A standard Tcl result. Upon success the interpreter result
 *	contains the name by which the next vertex is known in Tcl or
 *	the empty string if this vertex is not preceded by any vertices.
 *
 * Side effects:
 *	May export a vertex to Tcl.
 */

int
T4Vertex::Prev(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_Vertex vv;
    e4_VertexUniqueID vuid;
    T4Vertex *nv;
    int num = 1;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex prev ?num?");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * If there is one argument, it is the number of vertices to go
     * backwards.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &num) == TCL_ERROR) {
	    return TCL_ERROR;
	}
    }

    /*
     * Attempt to retrieve the next vertex.
     */

    if ((!f.Prev(num, vv)) || (!vv.IsValid())) {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), "", -1);
	return TCL_OK;
    }

    /*
     * See if we have exported this vertex to Tcl before. If not, create
     * a new T4Vertex and make it ready to be exported.
     */

    (void) vv.GetUniqueID(vuid);
    nv = s->GetVertexById(vuid);
    if (nv == NULL) {
	nv = new T4Vertex(vv, NULL, s);
	nv->SetName(MakeVertexCommand(interp, nv, vuid));
	s->StoreVertex(nv, vuid);
    }

    /*
     * Set the interpreter result to the name by which this vertex is
     * known in Tcl.
     */

    Tcl_SetStringObj(Tcl_GetObjResult(interp), nv->GetName(), -1);

    return TCL_OK;
}

/*
 * Call --
 *
 *	Call the value of this vertex (which should be a string) as
 *	a Tcl command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Whatever the command does.
 */

int
T4Vertex::Call(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int id, len;
    T4Node *np;
    e4_Node n;
    e4_NodeUniqueID nuid;
    e4_VertexUniqueID vuid;
    const char *storedproc;
    char *storedprocname;

    /*
     * There are no limits on what arguments may be given.
     */

    /*
     * Check that the vertex is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " is invalid", NULL);
	return TCL_ERROR;
    }

    /*
     * Make sure that the value of this vertex is a string. If not, we can
     * not use it as the body of a stored procedure.
     */

    if (f.Type() != E4_VTSTRING) {
	Tcl_AppendResult(interp, "vertex ", GetName(), " cannot be used as ",
			 "a stored procedure", NULL);
	return TCL_ERROR;
    }

    /*
     * See if the stored command procedure is already exported to Tcl. If
     * not, define the stored command procedure.
     */

    if (storedcmdinfo == NULL) {

        /*
	 * Check that the vertex can be used as a stored procedure.
	 * We need the T4Node representing the node containing this
	 * vertex so that we can pass its name to the stored procedure
	 * as the value of its "this" argument.
	 */

        if ((!f.GetNode(n)) || (!n.IsValid())) {
	    Tcl_AppendResult(interp, "vertex ", f.Name(), " cannot be used ",
			     "as a stored procedure", NULL);
	    return TCL_ERROR;
	}

	/*
	 * Check that the node was already exported to Tcl, and if not,
	 * export it now.
	 */

	(void) n.GetUniqueID(nuid);
	np = s->GetNodeById(nuid);
	if (np == NULL) {
	    np = new T4Node(n, NULL, s);
	    np->SetName(MakeNodeCommand(interp, np, nuid));
	    s->StoreNode(np, nuid);
	}

	/*
	 * Compute the name to be assigned to the stored procedure.
	 */

	(void) f.GetUniqueID(vuid);
	id = vuid.GetUniqueID();
	len = 64 + s->GetNameLen();
	storedprocname = Tcl_Alloc(len);
	sprintf(storedprocname, "::tgraph::%s::storedproc%d", 
		s->GetName(), id);

	/*
	 * Define the stored procedure command.
	 */

	if ((!f.Get(storedproc)) || (storedproc == NULL) ||
	    (DefineStoredProcedure(interp, storedprocname, np->GetName(), 
				   (char *) storedproc) != TCL_OK)) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "(1) cannot invoke stored procedure ",
			     f.Name(), " in node ", np->GetName(),
			     NULL);
	    Tcl_Free(storedprocname);
	    return TCL_ERROR;
	}

	/*
	 * Cache the command info for the stored procedure.
	 */

	storedcmdinfo = (Tcl_CmdInfo *) Tcl_Alloc(sizeof(Tcl_CmdInfo));
	if (Tcl_GetCommandInfo(interp, storedprocname, storedcmdinfo) == 0) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "(2) cannot invoke stored procedure ",
			     f.Name(), " in node ", np->GetName(), NULL);
	    Tcl_Free(storedprocname);
	    Tcl_Free((char *) storedcmdinfo);
	    storedcmdinfo = NULL;
	    return TCL_ERROR;
	}
    }

    /*
     * Invoke the stored command procedure and return whatever that
     * returns. Assume that it is always an object command.
     */

    return (storedcmdinfo->objProc)(storedcmdinfo->objClientData, interp,
				    objc, objv);
}

/*
 * Id --
 *
 *	Returns an identifier that is unique among all vertices in
 *	the storage containing this vertex.
 *
 * Results:
 *	A standard Tcl result. Upon successful completion, the interpreter
 *	result contains the integer ID.
 *
 * Side effects:
 *	None.
 */

int
T4Vertex::Id(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    e4_VertexUniqueID vuid;

    /*
     * Expecting zero arguments.
     */

    if (objc != 0) {
	Tcl_WrongNumArgs(interp, 0, NULL, (char *) "$vertex id");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex object is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "invalid vertex ", GetName(), NULL);
	return TCL_ERROR;
    }
    
    (void) f.GetUniqueID(vuid);
    Tcl_SetIntObj(Tcl_GetObjResult(interp), vuid.GetUniqueID());

    return TCL_OK;
}

/*
 * UserData --
 *
 *	Sets or returns the user data associated with this vertex.
 *
 * Results:
 *	A standard Tcl result. When retrieving the user data, upon
 *	success, the interpreter result contains an integer representing
 *	the stored user data.
 *
 * Side effects:
 *	May modify the user data associated with this vertex in its
 *	storage.
 */

int
T4Vertex::UserData(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    int i;

    /*
     * Expecting zero or one arguments.
     */

    if ((objc != 0) && (objc != 1)) {
	Tcl_WrongNumArgs(interp, 0, NULL,
			 (char *) "$vertex userdata ?newvalue?");
	return TCL_ERROR;
    }

    /*
     * Check that the vertex object is valid.
     */

    if (!f.IsValid()) {
	Tcl_AppendResult(interp, "invalid vertex ", GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * If one argument is given, it must be an integer (the new value).
     * Attempt to extract the new value.
     */

    if (objc == 1) {
	if (Tcl_GetIntFromObj(interp, objv[0], &i) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (!f.SetUserData(i)) {
	    Tcl_AppendResult(interp, "cannot set user data for vertex ",
			     GetName(), NULL);
	    return TCL_ERROR;
	}
	Tcl_ResetResult(interp);
	return TCL_OK;
    }

    /*
     * No arguments given, so we want to retrieve the value.
     */

    if (!f.GetUserData(i)) {
	Tcl_AppendResult(interp, "cannot retrieve user data for vertex ",
			 GetName(), NULL);
	return TCL_ERROR;
    }

    /*
     * Set the interpreter result to the retrieved value.
     */

    Tcl_SetIntObj(Tcl_GetObjResult(interp), i);

    return TCL_OK;
}

