# testnode.tcl --
#
# Test the TGraph node object's commands:
#
# Copyright (c) 2000-2001, JYL Software Inc.
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
# JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.

# Require the tcltest package:

package require tcltest

# Load some utilities:

source [file join [file dirname [info script]] util.tcl]

# Ensure the TGraph package is loaded:

if {[catch {loadTGraph} err]} {
    puts "error loading library TGraph: $err"
    return -code error ""
}
    
# The tests:

# Test coverage panel:
#
# $n kind
# $n vertexcount
# $n set
# $n add
# $n get
# $n setnode
# $n addnode
# $n getvertex
# $n movevertex
# $n deletevertex
# $n renamevertex
# $n vertextype
# $n vertexrank
# $n vertexname
# $n exists
# $n parent
# $n parentcount
# $n occurrencecount
# $n parentrank
# $n isroot
# $n addmarker
# $n markwith
# $n unmark
# $n markercount
# $n ismarked
# $n marker
# $n delete
# 
# Not tested:
#
# $n rankinparent
# $n nameinparent
# $n storage
# $n dispose
# $n isvalid
# $n method
# $n call
# $n foreach
# $n id

# Test that [$n marker] returns the same marker as was used to
# create the node.

tcltest::test node-1.0.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set m [$x marker m]
    set n [$m node]

    if {![string equal [$n marker] $m]} {
	deleteAndError $x "Marker $m and \[$n marker\] are different
    }
    $x delete
    set empty
} ""

# Test that enclosed nodes are removed when a storage is closed.

tcltest::test node-1.0.2 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set m [$x marker m]
    set n [$m node]
    $x close
    if {![string equal [info commands $m] ""] ||
	![string equal [info commands $n] ""]} {
	error "Entities $n and/or $m were not deleted when $x was closed"
    }
    set empty
} ""

tcltest::test node-1.0.3 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set m [$x marker m]
    set n [$m node]
    $x delete
    if {![string equal [info commands $m] ""] ||
	![string equal [info commands $n] ""]} {
	error "Entities $n and/or $m were not deleted when $x was deleted"
    }
    set empty
} ""

# Test the ismarked operation:

tcltest::test node-1.0.4 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    set n1 [$n addnode frob last]
    if {![$n ismarked] || [$n1 ismarked]} {
	deleteAndError $x \
		"Entity $n is not marked or $n1 is marked"
    }
    $x delete
    set empty
} ""

# Test the addmarker operation:

tcltest::test node-1.0.5 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    set n1 [$n addnode frob last]
    set m1 [$n addmarker m1]
    set m2 [$n1 addmarker m2]
    if {([$n markercount] != 2) || ([$n1 markercount] != 1)} {
	deleteAndError $x \
		"Incorrect marker count for $n and/or $n1"
    }
    $x delete
    set empty
} ""

tcltest::test node-1.0.6 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	set m [$n addmarker m$i]
	if {(![string equal m$i [$m name]]) || \
		(![string equal [$m node] $n]) || \
		([$n markercount] != ($i + 2))} {
	    set j [$n markercount]
	    deleteAndError $x \
		    "Marker count incorrect, expecting [expr $i + 2], got $j"
	}
    }
    if {([$x markercount] != 11) || ([$n markercount] != 11)} {
	set mc1 [$x markercount]
	set mc2 [$n markercount]
	deleteAndError $x \
		"Marker count incorrect, expecting 11, got $mc1, $mc2"
    }
    $x delete
    set empty
} ""

# Test the unmark operation:

tcltest::test node-1.0.7 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	set markers($i) [$n addmarker m$i]
	if {(![string equal [$markers($i) name] m$i]) || \
		(![string equal [$markers($i) node] $n]) || \
		([$n markercount] != ($i + 2))} {
	    set j [$n markercount]
	    deleteAndError $x \
		    "Marker count incorrect, expecting [expr $i + 2], got $j"
	}
    }
    if {([$x markercount] != 11) || ([$n markercount] != 11)} {
	set mc1 [$x markercount]
	set mc2 [$n markercount]
	deleteAndError $x \
		"Marker count incorrect, expecting 11, got $mc1, $mc2"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n unmark $markers($i)
	set c [expr 10 - $i]
	if {[$n markercount] != $c} {
	    set j [$n markercount]
	    deleteAndError $x \
		    "Marker count incorrect, expecting $c, got $j"
	}
    }
    if {([$n markercount] != 1) || ([$x markercount] != 1)} {
	set j [$n markercount]
	set k [$x markercount]
	deleteAndError $x "Marker count incorrect, expecting 1, got $j, $k"
    }
    $x delete
    set empty
} ""

# Test the markwith operation:

tcltest::test node-1.0.8 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	set markers($i) [$n addmarker m$i]
	if {(![string equal [$markers($i) name] m$i]) || \
		(![string equal [$markers($i) node] $n]) || \
		([$n markercount] != ($i + 2))} {
	    set j [$n markercount]
	    deleteAndError $x \
		    "Marker count incorrect, expecting [expr $i + 2], got $j"
	}
    }
    if {([$x markercount] != 11) || ([$n markercount] != 11)} {
	set mc1 [$x markercount]
	set mc2 [$n markercount]
	deleteAndError $x \
		"Marker count incorrect, expecting 11, got $mc1, $mc2"
    }
    set n1 [$n addnode frob last]
    if {[$n1 markercount] != 0} {
	set j [$n1 markercount]
	deleteAndError $x \
		"Marker count for $n1 incorrect, expecting 0, got $j"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n1 markwith $markers($i)
	set np [$markers($i) node]
	if {[string equal $np $n] || ![string equal $np $n1]} {
	    set m $markers($i)
	    deleteAndError $x \
		    "Marker confusion, $m should mark $n1, not $n, [$m node]"
	}
    }
    if {([$n markercount] != 1) || ([$n1 markercount] != 10)} {
	set i [$n markercount]
	set j [$n1 markercount]
	deleteAndError $x \
		"Marker count incorrect, expecting 1 and 10, got $i and $j"
    }
    $x delete
    set empty
} ""

# Test the isroot operation:

tcltest::test node-1.0.9 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    if {![$n isroot]} {
	deleteAndError $x \
		"Isroot confusion, should be 1, got [$n isroot]"
    }
    set n1 [$n addnode frob last]
    if {[$n1 isroot]} {
	deleteAndError $x \
		"Isroot confusion, should be 0, got [$n1 isroot]"
    }
    $x delete
    set empty

} ""

tcltest::test node-1.0.10 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    if {![$n isroot]} {
	deleteAndError $x \
		"Isroot confusion, should be 1, got [$n isroot]"
    }
    set n1 [$n addnode frob last]
    if {[$n1 isroot]} {
	deleteAndError $x \
		"Isroot confusion, should be 0, got [$n1 isroot]"
    }

    # Make $n1 a parent of $n, see if [$n isroot] is still true (shouldn't be).

    $n1 add frob last $n -node
    if {[$n isroot]} {
	deleteAndError $x \
		"Isroot confusion, should be 0, got [$n isroot]"
    }

    # But $n is still marked, and $n1 is not.

    if {[$n1 ismarked] || [$n1 isroot] || ![$n ismarked]} {
	deleteAndError $x \
		"Isroot and ismarked confusion"
    }

    $x delete
    set empty

} ""

# New nodes should have zero vertices initially.

tcltest::test node-1.1.0 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    if {[$n vertexcount] != 0} {
	deleteAndError $x \
		"initial vertex count for node $n is [$n vertexcount], not 0"
    }
    $x delete
    set empty
} ""

# New top-level nodes should have only one marker initially.

tcltest::test node-1.1.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    if {[$n markercount] != 1} {
	deleteAndError $x \
		"initial markercount for node $n is [$n markercount], not 1"
    }
    $x delete
    set empty
} ""

# Nodes marked by different markers should be different.

tcltest::test node-1.1.2 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n1 [$x markednode m1]
    set n2 [$x markednode m2]
    if {[string equal $n1 $n2]} {
	deleteAndError $x \
		"nodes marked by m1 and m2 {$n1 and $n2} are equal"
    }
    if {[string equal [$n1 marker] [$n2 marker]]} {
	deleteAndError $x \
		"markers of $n1 and $n2 are equal"
    }
    $x delete
    set empty
} ""

# The "kind" attribute of a node is "node"

tcltest::test node-1.2.0 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    if {![string equal node [$n kind]]} {
	deleteAndError $x \
		"node kind of $n is [$n kind], not \"node\""
    }
    $x delete
    set empty
} ""

# Add an integer vertex to a node, and see if it is there.

tcltest::test node-1.3.0 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    $n add frob last 12
    if {[$n vertexcount] != 1} {
	deleteAndError $x \
	   "after adding one vertex the vertexcount is [$n vertexcount], not 1"
    }
    set v [$n get frob]
    if {$v != 12} {
	deleteAndError $x \
	   "set vertex frob to 12, and got back $v"
    }
    $x delete
    set empty
} ""

# Add ten integer vertices to a node, and see if they're all there.

tcltest::test node-1.3.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    for {set i 0} {$i < 10} {incr i} {
	set v [$n get frob$i]
	if {$v != $i} {
	    deleteAndError $x \
		    "set vertex frob$i to $i and got back $v"
	}
    }
    $x delete
    set empty
} ""

# Create ten integer vertices, then set all ten of them to a string

tcltest::test node-1.3.2 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    for {set i 0} {$i < 10} {incr i} {
	$n set frob$i abcde$i
    }
    for {set i 0} {$i < 10} {incr i} {
	set v [$n get frob$i]
	if {![string equal $v abcde$i]} {
	    deleteAndError $x \
		    "set vertex frob$i to abcde$i and got back $v"
	}
    }
    $x delete
    set empty
} ""

# Create string vertices and see if they are all there.

tcltest::test node-1.3.3 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last abcde$i
    }
    for {set i 0} {$i < 10} {incr i} {
	set v [$n get frob$i]
	if {![string equal $v abcde$i]} {
	    deleteAndError $x \
		    "set vertex frob$i to abcde$i and got back $v"
	}
    }
    $x delete
    set empty
} ""

# Test vertex types:

tcltest::test node-1.3.4 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	if {$i < 3} {
	    $n add frob$i last $i
	} elseif {$i < 6} {
	    $n add frob$i last abcde$i
	} else {
	    $n add frob$i last $i.0
	}
    }
    for {set i 0} {$i < 10} {incr i} {
	set type [$n vertextype frob$i]
	if {$i < 3} {
	    if {![string equal $type int]} {
		deleteAndError $x \
			"set frob$i to integer value but type is $type"
	    }
	} elseif {$i < 6} {
	    if {![string equal $type string]} {
		deleteAndError $x \
			"set frob$i to string value but type is $type"
	    }
	} else {
	    if {![string equal $type double]} {
		deleteAndError $x \
			"set frob$i to double value but type is $type"
	    }
	}
    }
    $x delete
    set empty
} ""

# Test setting vertices:

tcltest::test node-1.3.5 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	if {$i < 3} {
	    $n add frob$i last $i
	} elseif {$i < 6} {
	    $n add frob$i last abcde$i
	} else {
	    $n add frob$i last $i.0
	}
    }
    for {set i 0} {$i < 10} {incr i} {
	set type [$n vertextype frob$i]
	if {$i < 3} {
	    if {![string equal $type int]} {
		deleteAndError $x \
			"set frob$i to integer value but type is $type"
	    }
	} elseif {$i < 6} {
	    if {![string equal $type string]} {
		deleteAndError $x \
			"set frob$i to string value but type is $type"
	    }
	} else {
	    if {![string equal $type double]} {
		deleteAndError $x \
			"set frob$i to double value but type is $type"
	    }
	}
    }
    for {set i 0} {$i < 10} {incr i} {
	$n set frob$i 23 -double
	set type [$n vertextype frob$i]
	if {![string equal $type double]} {
	    deleteAndError $x \
		    "set frob$i 23 double but type is $type"
	}
    }
    $x delete
    set empty
} ""

# Test the addnode operation.

tcltest::test node-1.4.0 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n addnode frob$i last
    }
    for {set i 0} {$i < 10} {incr i} {
	set n1 [$n get frob$i]
	set type [$n vertextype frob$i]
	set p [$n1 parent]
	if {![string equal $p $n] || ![string equal $type node]} {
	    deleteAndError $x \
		    "set frob$i to node and got $type, $n1"
	}
    }
    $x delete
    set empty
} ""

tcltest::test node-1.4.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    for {set i 0} {$i < 10} {incr i} {
	$n addnode frob$i first
    }
    if {[$n vertexcount] != 20} {
	deleteAndError $x \
		"node $n should have 20 vertices, but has [$n vertexcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	set n1 [$n get frob$i]
	set type [$n vertextype frob$i]
	set p [$n1 parent]
	if {![string equal $p $n] || ![string equal $type node]} {
	    deleteAndError $x \
		    "set frob$i to node and got $type, $n1"
	}
    }
    $x delete
    set empty
} ""

# Test the setnode operation:

tcltest::test node-1.4.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    for {set i 0} {$i < 10} {incr i} {
	if {![string equal [$n vertextype frob$i] int]} {
	    deleteAndError $x \
		    "set frob$i to integer and got type [$n vertextype frob$i]"
	}
    }
    for {set i 0} {$i < 10} {incr i} {
	$n setnode frob$i
    }
    for {set i 0} {$i < 10} {incr i} {
	set n1 [$n get frob$i]
	set type [$n vertextype frob$i]
	set p [$n1 parent]
	if {![string equal $p $n] || ![string equal $type node]} {
	    deleteAndError $x \
		    "set frob$i to node and got type $type, $n1"
	}
    }
    $x delete
    set empty
} ""

# Test vertex deletion for scalar vertices:

tcltest::test node-1.5.0 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Added ten vertices but vertex count is [$n vertexcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n deletevertex frob$i
    }
    if {[$n vertexcount] != 0} {
	deleteAndError $x \
		"Deleted ten vertices but vertex count is [$n vertexcount]"
    }
    $x delete
    set empty
} ""

tcltest::test node-1.5.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last abcde$i
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Added ten vertices but vertex count is [$n vertexcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n deletevertex frob$i
    }
    if {[$n vertexcount] != 0} {
	deleteAndError $x \
		"Deleted ten vertices but vertex count is [$n vertexcount]"
    }
    $x delete
    set empty
} ""

tcltest::test node-1.5.2 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i.0
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Added ten vertices but vertex count is [$n vertexcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n deletevertex frob$i
    }
    if {[$n vertexcount] != 0} {
	deleteAndError $x \
		"Deleted ten vertices but vertex count is [$n vertexcount]"
    }
    $x delete
    set empty
} ""

# Test vertex deletion with node values:

tcltest::test node-1.5.3 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n addnode frob$i last
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Added ten vertices but vertex count is [$n vertexcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n deletevertex frob$i
    }
    if {[$n vertexcount] != 0} {
	deleteAndError $x \
		"Deleted ten vertices but vertex count is [$n vertexcount]"
    }
    $x delete
    set empty
} ""

# Test vertex deletion with non-empty node values:

tcltest::test node-1.5.4 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	set n1 [$n addnode frob$i last]
	for {set j 0} {$j < 10} {incr j} {
	    $n1 add jay$i last $i
	}
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Added ten vertices but vertex count is [$n vertexcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n deletevertex frob$i
    }
    if {[$n vertexcount] != 0} {
	deleteAndError $x \
		"Deleted ten vertices but vertex count is [$n vertexcount]"
    }
    $x delete
    set empty
} ""

# Test vertex deletion but with non-empty node-valued vertices.

tcltest::test node-1.5.5 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	set n1 [$n addnode frob$i last]
	for {set j 0} {$j < 10} {incr j} {
	    $n1 addnode jay$i last
	}
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Added ten vertices but vertex count is [$n vertexcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n deletevertex frob$i
    }
    if {[$n vertexcount] != 0} {
	deleteAndError $x \
		"Deleted ten vertices but vertex count is [$n vertexcount]"
    }
    if {![string equal [info commands $n1] ""]} {
	deleteAndError $x \
		"Deleted all vertices but recursive node $n1 not deleted"
    }
    $x delete
    set empty
} ""

# Test the movevertex operation, within a node:

tcltest::test node-1.6.0 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Expecting vertex count of 10, got [$n vertexcount]"
    }
    if {[$n vertexrank frob3] != 4} {
	deleteAndError $x \
		"Vertex frob3 should be rank 4, is [$n vertexrank frob3]"
    }
    set v [$n getvertex frob3]
    for {set i 0} {$i < 9} {incr i} {
	$n movevertex $v at [expr $i + 1]
	if {[$n vertexrank frob3] != ($i + 1)} {
	    set j [expr $i + 1]
	    deleteAndError $x \
		    "Vertex frob3 should be rank $j, is [$n vertexrank frob3]"
	}
    }
    for {set i 10} {$i > 0} {incr i -1} {
	$n movevertex $v at $i
	if {[$n vertexrank frob3] != $i} {
	    deleteAndError $x \
		    "Vertex frob3 should be rank $i, is [$n vertexrank frob3]"
	}
    }
    $x delete
    set empty
} ""

# Test the movevertex operation between two nodes:

tcltest::test node-1.6.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Expecting vertex count of 10, got [$n vertexcount]"
    }
    if {[$n vertexrank frob3] != 4} {
	deleteAndError $x \
		"Vertex frob3 should be rank 4, is [$n vertexrank frob3]"
    }
    set v [$n getvertex frob3]
    set n1 [$x markednode m1]
    for {set i 0} {$i < 10} {incr i} {
	$n1 add foo$i last $i
    }
    if {[$n1 vertexcount] != 10} {
	deleteAndError $x \
		"Expecting vertex count of 10, got [$n1 vertexcount]"
    }
    if {![string equal [$v node] $n]} {
	deleteAndError $x \
		"Vertex frob3 should be in $n, actually is in [$v node]"
    }
    for {set i 0} {$i < 11} {incr i} {
	$n1 movevertex $v at [expr $i + 1]
	if {[$n1 vertexrank frob3] != ($i + 1)} {
	    set c1 [expr $i + 1]
	    set c2 [$n1 vertexrank frob3]
	    deleteAndError $x \
		    "Vertex frob3 expected at rank $c1, got $c2"
	}
    }
    if {([$n vertexcount] != 9) || ([$n1 vertexcount] != 11)} {
	set c1 [$n vertexcount]
	set c2 [$n1 vertexcount]
	deleteAndError $x \
		"Vertex count confusion, expected 9 and 11, got $c1 and $c2"
    }
    $x delete
    set empty
} ""

# Test the movevertex operation for moving a node around:

tcltest::test node-1.6.2 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Expecting vertex count of 10, got [$n vertexcount]"
    }
    if {[$n vertexrank frob3] != 4} {
	deleteAndError $x \
		"Vertex frob3 should be rank 4, is [$n vertexrank frob3]"
    }
    set v [$n getvertex frob3]
    set n1 [$n setnode frob3]
    for {set i 0} {$i < 10} {incr i} {
	$n1 add foo$i last $i.0
    }
    if {[$n1 vertexcount] != 10} {
	deleteAndError $x \
		"Expecting vertex count of 10, got [$n1 vertexcount]"
    }
    if {![string equal [$v node] $n]} {
	deleteAndError $x \
		"Vertex frob3 should be in $n, actually is in [$v node]"
    }
    for {set i 0} {$i < 10} {incr i} {
	$n movevertex $v at [expr $i + 1]
	if {[$n vertexrank frob3] != ($i + 1)} {
	    set c1 [expr $i + 1]
	    set c2 [$n vertexrank frob3]
	    deleteAndError $x \
		    "Vertex frob3 expected at rank $c1, got $c2"
	}
	if {![string equal $n [$n1 parent]]} {
	    deleteAndError $x \
		    "Node value of frob3 lost parent, $n vs. [$n1 parent]"
	}
	if {[$n1 parentcount] != 1} {
	    deleteAndError $x \
		    "Parent count confusion, expected 1, got [$n1 parentcount]"
	}
	if {[$n1 vertexcount] != 10} {
	    set c [$n1 vertexcount]
	    deleteAndError $x \
		    "Vertex count confusion, expected 10, got $c"
	}
    }
    $x delete
    set empty
} ""

# Test the add operation with a node value; see that the occurrence count
# increases. Also tests the parentcount and occurrencecount operations.

tcltest::test node-1.6.2 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    if {[$n vertexcount] != 10} {
	deleteAndError $x \
		"Expecting vertex count of 10, got [$n vertexcount]"
    }
    if {[$n vertexrank frob3] != 4} {
	deleteAndError $x \
		"Vertex frob3 should be rank 4, is [$n vertexrank frob3]"
    }
    set v [$n getvertex frob3]
    set n1 [$n setnode frob3]
    if {[$n1 parentcount] != 1} {
	deleteAndError $x \
		"Parent count confusion, expected 1, got [$n1 parentcount]"
    }
    set n2 [$x markednode m2]
    for {set i 0} {$i < 10} {incr i} {
	$n2 add foo$i last $n1 -node
	if {[$n1 occurrencecount] != ($i + 2)} {
	    set c1 [expr $i + 2]
	    set c2 [$n1 occurrencecount]
	    deleteAndError $x \
		    "Occurrence count confusion, expected $c1, got $c2"
	}
    }
    if {[$n1 occurrencecount $n2] != 10} {
	set c1 [$n1 occurrencecount $n2]
	deleteAndError $x \
		"Occurrence count confusion, expected 10, got $c1"
    }
    if {[$n1 occurrencecount $n] != 1} {
	set c1 [$n1 occurrencecount $n]
	deleteAndError $x \
		"Occurrence count confusion, expected 1, got $c1"
    }
    if {[$n1 parentcount] != 2} {
	deleteAndError $x \
		"Parent count confusion, expected 2, got [$n1 parentcount]"
    }
    $x delete
    set empty
} ""

# Test parent counts and the delete operation:

tcltest::test node-1.6.3 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    set n1 [$x markednode m1]
    set n2 [$n addnode foo last]
    $n1 add foo last $n2 -node
    if {[$n2 parentcount] != 2} {
	deleteAndError $x \
		"Parent count confusion, expected 2, got [$n1 parentcount]"
    }

    # Now delete $n, see [$n2 parentcount] go to 1

    $n delete
    if {[$n2 parentcount] != 1} {
	deleteAndError $x \
		"Parent count confusion, expected 1, got [$n1 parentcount]"
    }

    # Now recreate $n, make ten node valued vertices in it and make each
    # of these nodes a parent of $n2.

    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	set n3 [$n addnode frob$i last]
	$n3 add foo first $n2 -node
	if {[$n2 parentcount] != ($i + 2)} {
	    set c1 [expr $i + 2]
	    set c2 [$n2 parentcount]
	    deleteAndError $x \
		    "Parent count confusion, expected $c1, got $c2"
	}
    }

    # Now delete $n, watch [$n2 parentcount] go back to 1.

    $n delete
    if {[$n2 parentcount] != 1} {
	deleteAndError $x \
		"Parent count confusion, expected 1, got [$n2 parentcount]"
    }

    $x delete
    set empty
} ""

# Test the parentrank operation:

tcltest::test node-1.6.4 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    for {set i 0} {$i < 10} {incr i} {
	set nodes($i) [$x markednode m$i]
    }
    set c [$nodes(0) addnode frob last]
    for {set i 1} {$i < 10} {incr i} {
	$nodes($i) add frob last $c -node
    }
    if {[$c parentcount] != 10} {
	deleteAndError $x \
		"Parent count confusion, expected 10, got [$c parentcount]"
    }
    for {set i 0} {$i < 10} {incr i} {
	set r [$c parentrank $nodes($i)]
	if {$r != ($i + 1)} {
	    deleteAndError $x \
		    "Parent rank confusion, expected [expr $i + 1], got $r"
	}
    }
    for {set i 0} {$i < 10} {incr i} {
	$nodes($i) delete
	if {$i < 9} {
	    set remaining [expr 9 - $i]
	    if {[$c parentcount] != $remaining} {
		deleteAndError $x \
			"Parent count confusion, expected $remaining"
	    }
	}
    }
    if {![$x isempty]} {
	deleteAndError $x "Storage should be empty but isn't"
    }
    $x delete
    set empty
} ""
    
# Test the rankinparent and nameinparent operations.

tcltest::test node-1.6.5 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	set nodes($i) [$x markednode m$i]
	for {set j 0} {$j < 10} {incr j} {
	    if {$i == $j} {
		$nodes($i) add frob$i last $n -node
	    } else {
		$nodes($i) add frob$i last $i
	    }
	}
    }

    # TODO: test the rankinparent and nameinparent operations:

    $x delete
    set empty
} ""
    
# Test exists operation:

tcltest::test node-1.7.0 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    if {[$n exists foo]} {
	deleteAndError $x \
		"Vertex foo exists when it shouldn't"
    }
    if {![$n exists frob3]} {
	deleteAndError $x \
		"Vertex frob3 doesn't exist when it should"
    }
    $n deletevertex frob3
    if {[$n exists frob3]} {
	deleteAndError $x \
		"Vertex frob3 exists when it shouldn't"
    }
    $x delete
    set empty
} ""

# Test the vertextype, vertexrank and vertexname operations.

tcltest::test node-1.7.1 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    for {set i 1} {$i < 11} {incr i} {
	set nm [$n vertexname $i]
	if {![string equal $nm frob[expr $i - 1]]} {
	    set en frob[expr $i - 1]
	    deleteAndError $x \
		    "Vertex name confusion, expected $en, got $nm"
	}
	if {![string equal int [$n vertextype $nm]]} {
	    set tp [$n vertextype $nm]
	    deleteAndError $x \
		    "Vertex type confusion, expected int, got $tp"
	}
	if {![string equal int [$n vertextype $i]]} {
	    set tp [$n vertextype $i]
	    deleteAndError $x \
		    "Vertex type confusion, expected int, got $tp"
	}
    }
    for {set i 0} {$i < 10} {incr i} {
	$n set frob$i abcde$i
	set nm [$n vertexname [expr $i + 1]]
	if {![string equal $nm frob$i]} {
	    deleteAndError $x \
		    "Vertex name confusion, expected frob$i, got $nm"
	}
	set tp [$n vertextype frob$i]
	if {![string equal $tp string]} {
	    deleteAndError $x \
		    "Vertex type confusion, expected string, got $tp"
	}
	if {![string equal [$n vertextype [expr $i + 1]] string]} {
	    set tp [$n vertextype [expr $i + 1]]
	    deleteAndError $x \
		    "Vertex type confusion, expected string, got $tp"
	}
    }
    for {set i 0} {$i < 10} {incr i} {
	set r [$n vertexrank frob$i]
	if {$r != ($i + 1)} {
	    set c [expr $i + 1]
	    deleteAndError $x \
		    "Vertex rank confusion, expected $c, got, $r"
	}
    }
    $x delete
    set empty
} ""

# Test the renamevertex operation

tcltest::test node-1.7.2 {node tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set n [$x markednode m]
    for {set i 0} {$i < 10} {incr i} {
	$n add frob$i last $i
    }
    for {set i 0} {$i < 10} {incr i} {
	$n renamevertex frob$i foo
    }
    for {set i 1} {$i < 11} {incr i} {
	$n renamevertex foo(1) hello$i
    }
    for {set i 1} {$i < 11} {incr i} {
	if {![$n exists hello$i]} {
	    deleteAndError $x \
		    "Vertex name confusion, expected hello$i to exist"
	}
    }
    for {set i 1} {$i < 11} {incr i} {
	$n renamevertex $i hello
    }
    for {set i 1} {$i < 11} {incr i} {
	if {![$n exists hello($i)]} {
	    deleteAndError $x \
		    "Vertex name confusion, expected hello($i) to exist"
	}
    }
    $x delete
    set empty
} ""
