# teststorage.tcl --
#
# Test the TGraph storage object's commands:
#
# Copyright (c) 2000-2001, JYL Software Inc.
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
# JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.

# Require the tcltest package:

package require tcltest

# Load some utilities:

source [file join [file dirname [info script]] util.tcl]

# Ensure the TGraph package is loaded:

if {[catch {loadTGraph} err]} {
    puts "error loading library TGraph: $err"
    return -code error ""
}

# The tests:

# Storages should be created empty and should not be dirty initially:

tcltest::test storage-1.0.0 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x is not empty"
    }
    if {[$x isdirty]} {
	deleteAndError $x "Storage $x is dirty but unchanged"
    }
    $x delete
    set empty
} ""

# Storages should have auto-commit on initially:

tcltest::test storage-1.0.1 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {[$x autocommit] == 0} {
	deleteAndError $x "Storage $x should have autocommit 1"
    }
    $x autocommit 0
    if {[$x autocommit] == 1} {
	deleteAndError $x "Storage $x should have autocommit 0"
    }
    $x delete
    set empty
} ""

# Storages should be empty after deletion:

tcltest::test storage-1.0.2 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x should be empty initially"
    }
    set m [$x marker newmarker]
    $x commit
    $x delete
    set x [tgraph::open foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x should be empty after delete"
    }
    $x delete
    set empty
} ""

# Storages should be non-empty after a commit:

tcltest::test storage-1.0.3 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x should be empty initially"
    }
    set m [$x marker newmarker]
    $x commit
    $x close
    set x [tgraph::open foo.db]
    if {[$x isempty]} {
	deleteAndError $x "Storage $x should be non-empty after commit"
    }
    $x delete
    set empty
} ""

# Storages should auto-commit on close if auto-commit is on:

tcltest::test storage-1.0.4 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x should be empty initially"
    }
    set m [$x marker newmarker]
    $x close
    set x [tgraph::open foo.db]
    if {[$x isempty]} {
	deleteAndError $x "Storage $x should be non-empty after auto-commit"
    }
    $x delete
    set empty
} ""

# Storages should have the same contents before and after commit-close-open

tcltest::test storage-1.0.5 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x should be empty initially"
    }
    set m [$x marker newmarker]
    $x close
    set x [tgraph::open foo.db]
    if {[$x markercount] != 1} {
	deleteAndError $x "Storage $x should have one marker after auto-commit"
    }
    set m [$x marker newmarker]
    if {[$x markercount] != 1} {
	deleteAndError $x "Storage $x should have the same content"
    }
    $x delete
    set empty
} ""

# Test $storage name, $storage driver and $storage mode:

tcltest::test storage-1.0.6 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x should be empty initially"
    }
    if {![string equal [$x name] foo.db]} {
	deleteAndError $x "Storage $x should have the name \"foo.db\""
    }
    if {![string equal [$x mode] w]} {
	deleteAndError $x "Storage $x should have mode \"w\""
    }
    if {![string equal [$x driver] "Metakit 2.0"]} {
	deleteAndError $x "Storage $x should have driver \"Metakit 2.0\""
    }
    $x delete
    set empty
} ""

# Mark a storage dirty and see if it is marked dirty.

tcltest::test storage-1.0.7 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {![$x isempty]} {
	deleteAndError $x "Storage $x should be empty initially"
    }
    if {[$x isdirty]} {
	deleteAndError $x "Storage $x should not be marked dirty initially"
    }
    $x markdirty
    if {![$x isdirty]} {
	deleteAndError $x "Storage $x should be marked dirty"
    }
    $x commit
    if {[$x isdirty]} {
	deleteAndError $x "Storage $x should not be marked dirty after commit"
    }
    $x delete
    set empty
} ""

# Test that $x markednode returns the same node as $m node.

tcltest::test storage-1.0.8 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    set m [$x marker newmarker]
    set n1 [$m node]
    set n2 [$x markednode newmarker]
    if {![string equal $n1 $n2]} {
	deleteAndError $x "\"$storage markednode\" returns different node"
    }
    if {[$x markercount] != 1} {
	deleteAndError $x "Storage $x contains [$x markercount] markers"
    }
    $x delete
    set empty
} ""

# Test $storage get operation

tcltest::test storage-1.0.9 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    catch {unset markers}
    catch {unset nodes}
    catch {unset vertices}

    # Collect some markers and nodes:

    for {set i 0} {$i < 10} {incr i} {
	set markers($i) [$x marker m$i]
	set m $markers($i)

	set markers($i,id) [$m id]

	set nodes($i) [$m node]
	set n $nodes($i)
	set nodes($i,id) [$n id]

	$n add first first first
	set v [$n getvertex first]

	set vertices($i) $v
	set vertices($i,id) [$v id]
    }

    # Now see if the $storage get marker operation with an $id returns the
    # Same marker as stored in the slot for $id

    for {set i 0} {$i < 10} {incr i} {
	set m [$x get marker $markers($i,id)]
	if {![string equal $m $markers($i)]} {
	    deleteAndError $x "\"storage get marker $i\" didn't work"
	}
	set n [$x get node $nodes($i,id)]
	set n1 [$m node]
	if {![string equal $n $nodes($i)] || ![string equal $n1 $nodes($i)]} {
	    deleteAndError $x "\"storage get node $i\" didn't work"
	}
	set v [$x get vertex $vertices($i,id)]
	set v1 [$n getvertex first]
	if {![string equal $v $vertices($i)] || \
	    ![string equal $v1 $vertices($i)] || \
	    ![string equal first [$v get]] || \
	    ![string equal first [$v1 get]]} {
	    deleteAndError $x "\"storage get vertex $i\" didn't work"
	}
    }
    catch {unset markers}
    catch {unset nodes}
    catch {unset vertices}
    $x delete
    set empty
} ""

# Test $storage statistic operation on an empty storage.

tcltest::test storage-1.0.10 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    if {[$x statistic marker used] != 0} {
	deleteAndError $x "Storage $x statistic marker used != 0"
    }
    if {[$x statistic marker freed] != 0} {
	deleteAndError $x "Storage $x statistic marker freed != 0"
    }
    if {[$x statistic marker available] != 0} {
	deleteAndError $x "Storage $x statistic marker available != 0"
    }
    if {[$x statistic marker allocated] != 0} {
	deleteAndError $x "Storage $x statistic marker allocated != 0"
    }
    if {[$x statistic node used] != 0} {
	deleteAndError $x "Storage $x statistic node used != 0"
    }
    if {[$x statistic node freed] != 0} {
	deleteAndError $x "Storage $x statistic node freed != 0"
    }
    if {[$x statistic node available] != 0} {
	deleteAndError $x "Storage $x statistic node available != 0"
    }
    if {[$x statistic node allocated] != 0} {
	deleteAndError $x "Storage $x statistic node allocated != 0"
    }
    if {[$x statistic vertex used] != 0} {
	deleteAndError $x "Storage $x statistic vertex used != 0"
    }
    if {[$x statistic vertex freed] != 0} {
	deleteAndError $x "Storage $x statistic vertex freed != 0"
    }
    if {[$x statistic vertex available] != 0} {
	deleteAndError $x "Storage $x statistic vertex available != 0"
    }
    if {[$x statistic vertex allocated] != 0} {
	deleteAndError $x "Storage $x statistic vertex allocated != 0"
    }
    $x delete
    set empty
} ""

# Test $storage statistic after some allocations.

tcltest::test storage-1.0.11 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]
    for {set i 0} {$i < 10} {incr i} {
	$x marker m$i
    }
    
    # There should be ten markers and nodes allocated, and space for 128.
    # No frees have happened, so the statistics for those should be 0.

    if {[$x statistic marker used] != 10} {
	deleteAndError $x "Storage $x statistic marker used != 10"
    }
    if {[$x statistic marker freed] != 0} {
	deleteAndError $x "Storage $x statistic marker freed != 0"
    }
    if {[$x statistic marker available] != 128} {
	deleteAndError $x "Storage $x statistic marker available != 128"
    }
    if {[$x statistic marker allocated] != 10} {
	deleteAndError $x "Storage $x statistic marker allocated != 10"
    }
    if {[$x statistic node used] != 10} {
	deleteAndError $x "Storage $x statistic node used != 10"
    }
    if {[$x statistic node freed] != 0} {
	deleteAndError $x "Storage $x statistic node freed != 0"
    }
    if {[$x statistic node available] != 128} {
	deleteAndError $x "Storage $x statistic node available != 128"
    }
    if {[$x statistic node allocated] != 10} {
	deleteAndError $x "Storage $x statistic node allocated != 10"
    }
    if {[$x statistic vertex used] != 0} {
	deleteAndError $x "Storage $x statistic vertex used != 0"
    }
    if {[$x statistic vertex freed] != 0} {
	deleteAndError $x "Storage $x statistic vertex freed != 0"
    }
    if {[$x statistic vertex available] != 0} {
	deleteAndError $x "Storage $x statistic vertex available != 0"
    }
    if {[$x statistic vertex allocated] != 0} {
	deleteAndError $x "Storage $x statistic vertex allocated != 0"
    }

    # Now delete all the markers and see if the freed statistic is
    # correctly updated.

    for {set i 0} {$i < 10} {incr i} {
	set m [$x get marker $i]
	$m delete
    }
    
    if {[$x statistic marker used] != 0} {
	deleteAndError $x "Storage $x statistic marker used != 0"
    }
    if {[$x statistic marker freed] != 10} {
	deleteAndError $x "Storage $x statistic marker freed != 10"
    }
    if {[$x statistic marker available] != 128} {
	deleteAndError $x "Storage $x statistic marker available != 128"
    }
    if {[$x statistic marker allocated] != 10} {
	deleteAndError $x "Storage $x statistic marker allocated != 10"
    }
    if {[$x statistic node used] != 0} {
	deleteAndError $x "Storage $x statistic node used != 0"
    }
    if {[$x statistic node freed] != 10} {
	deleteAndError $x "Storage $x statistic node freed != 10"
    }
    if {[$x statistic node available] != 128} {
	deleteAndError $x "Storage $x statistic node available != 128"
    }
    if {[$x statistic node allocated] != 10} {
	deleteAndError $x "Storage $x statistic node allocated != 10"
    }
    if {[$x statistic vertex used] != 0} {
	deleteAndError $x "Storage $x statistic vertex used != 0"
    }
    if {[$x statistic vertex freed] != 0} {
	deleteAndError $x "Storage $x statistic vertex freed != 0"
    }
    if {[$x statistic vertex available] != 0} {
	deleteAndError $x "Storage $x statistic vertex available != 0"
    }
    if {[$x statistic vertex allocated] != 0} {
	deleteAndError $x "Storage $x statistic vertex allocated != 0"
    }
    $x delete
    set empty
} ""

# Test storage foreach operation

tcltest::test storage-1.0.12 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]

    # Allocate some entities into the storage.

    for {set i 0} {$i < 100} {incr i} {
	set m [$x marker m$i]
	set n [$m node]
	for {set j 0} {$j < 10} {incr j} {
	    $n add v$j last $j
	}
    }

    # We now have 100 markers, 100 nodes and 1000 vertices.

    set mcount 0
    set ncount 0
    set vcount 0

    if {[$x markercount] != 100} {
	deleteAndError $x "Storage $x should contain 100 markers"
    }

    $x foreach marker m {incr mcount}
    $x foreach node n {incr ncount}
    $x foreach vertex v {incr vcount}

    if {$mcount != 100} {
	deleteAndError $x "\"storage foreach marker\" didn't visit all markers"
    }
    if {$ncount != 100} {
	deleteAndError $x "\"storage foreach node\" didn't visit all nodes"
    }
    if {$vcount != 1000} {
	deleteAndError $x \
		"\"storage foreach vertex\" didn't visit all vertices"
    }

    # One more check: sum the value of all vertices -- we should get
    # (0+1+2+3+4+5+6+7+8+9) * 100 which is 4500.

    set sum 0
    $x foreach vertex v {incr sum [$v get]}
    if {$sum != 4500} {
	deleteAndError $x \
		"\"storage foreach vertex\" didn't visit all vertices once"
    }
    $x delete
    set empty
} ""

# Test storage makeempty operation

tcltest::test storage-1.0.13 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]

    # Allocate some entities in the storage.

    for {set i 0} {$i < 100} {incr i} {
	set m [$x marker m$i]
	set n [$m node]
	for {set j 0} {$j < 10} {incr j} {
	    $n add v$j last $j
	}
    }

    # We now have 100 markers, 100 nodes and 1000 vertices.

    set mcount 0
    set ncount 0
    set vcount 0

    if {[$x markercount] != 100} {
	deleteAndError $x "Storage $x should contain 100 markers"
    }

    $x foreach marker m {incr mcount}
    $x foreach node n {incr ncount}
    $x foreach vertex v {incr vcount}

    if {$mcount != 100} {
	deleteAndError $x "\"storage foreach marker\" didn't visit all markers"
    }
    if {$ncount != 100} {
	deleteAndError $x "\"storage foreach node\" didn't visit all nodes"
    }
    if {$vcount != 1000} {
	deleteAndError $x \
		"\"storage foreach vertex\" didn't visit all vertices"
    }

    # One more check: sum the value of all vertices -- we should get
    # (0+1+2+3+4+5+6+7+8+9) * 100 which is 4500.

    set sum 0
    $x foreach vertex v {incr sum [$v get]}
    if {$sum != 4500} {
	deleteAndError $x \
		"\"storage foreach vertex\" didn't visit all vertices once"
    }

    # Now make the storage empty:

    $x makeempty

    # Now we have no markers, nodes or vertices:

    set mcount 0
    set ncount 0
    set vcount 0

    $x foreach marker m {incr mcount}
    $x foreach node n {incr ncount}
    $x foreach vertex v {incr vcount}

    if {($mcount != 0) || ($ncount != 0) || ($vcount != 0)} {
	deleteAndError "\"storage makeempty\" didn't work"
    }
    $x delete
    set empty
} ""

# Test storage callback operation various operations

proc cb {v en} {
    upvar #0 $v counter

    incr counter
}

tcltest::test storage-1.0.14 {storage tests} {
    showActive
    set empty ""
    set x [ensureFreshStorage foo.db]

    # Install some callbacks:

    catch {unset callbacks}

    set callbacks($x,marker,add) [$x callback add marker add "cb maddcb"]
    set callbacks($x,marker,del) [$x callback add marker del "cb mdelcb"]
    set callbacks($x,node,add) [$x callback add node add "cb naddcb"]
    set callbacks($x,node,del) [$x callback add node del "cb ndelcb"]
    set callbacks($x,vertex,add) [$x callback add vertex add "cb vaddcb"]
    set callbacks($x,vertex,del) [$x callback add vertex del "cb vdelcb"]
    set callbacks($x,vertex,mod) [$x callback add vertex mod "cb vmodcb"]

    # Initialize all counters to zero:

    set maddcb 0
    set mdelcb 0
    set naddcb 0
    set ndelcb 0
    set vaddcb 0
    set vdelcb 0
    set vmodcb 0

    # Allocate some entities in the storage.

    for {set i 0} {$i < 100} {incr i} {
	set m [$x marker m$i]
	set n [$m node]
	for {set j 0} {$j < 10} {incr j} {
	    $n add v$j last $j
	}
    }

    # We now have 100 markers, 100 nodes and 1000 vertices.

    set mcount 0
    set ncount 0
    set vcount 0

    if {[$x markercount] != 100} {
	deleteAndError $x "Storage $x should contain 100 markers"
    }

    $x foreach marker m {incr mcount}
    $x foreach node n {incr ncount}
    $x foreach vertex v {incr vcount}

    if {$mcount != 100} {
	deleteAndError $x "\"storage foreach marker\" didn't visit all markers"
    }
    if {$ncount != 100} {
	deleteAndError $x "\"storage foreach node\" didn't visit all nodes"
    }
    if {$vcount != 1000} {
	deleteAndError $x \
		"\"storage foreach vertex\" didn't visit all vertices"
    }

    # One more check: sum the value of all vertices -- we should get
    # (0+1+2+3+4+5+6+7+8+9) * 100 which is 4500.

    set sum 0
    $x foreach vertex v {incr sum [$v get]}
    if {$sum != 4500} {
	deleteAndError $x \
		"\"storage foreach vertex\" didn't visit all vertices once"
    }

    # Now check we had the correct number of callbacks:

    if {($maddcb != 100) || ($naddcb != 100) || ($vaddcb != 1000) || \
	($mdelcb != 0) || ($ndelcb != 0) || ($vdelcb != 0)} {
	deleteAndError $x "\"storage callback\" incorrect number of callbacks"
    }

    # Check vertex modification callback works correctly (you no longer
    # get a mod callback for a vertex addition).

    if {$vmodcb != 0} {
	deleteAndError $x "0 vs $vmodcb: incorrect mod callbacks"
    }

    # Now modify all vertices:

    $x foreach vertex v {$v set 100}

    if {$vmodcb != 1000} {
	deleteAndError $x "1000 vs $vmodcb: incorrect mod callbacks here"
    }

    # Now make the storage empty.

    $x makeempty

    # And check the number of callbacks again:

    if {($maddcb != 100) || ($naddcb != 100) || ($vaddcb != 1000) || \
	($mdelcb != 100) || ($ndelcb != 100) || ($vdelcb != 1000)} {
	deleteAndError $x "\"storage callback\" incorrect number of callbacks"
    }

    # There should not have been any additional mod callbacks:

    if {$vmodcb != 1000} {
	deleteAndError $x "1000 vs $vmodcb: incorrect mod callbacks"
    }

    $x delete
    set empty
} ""
