/*
 * testcallback.cpp --
 *
 * Tests of the callback facility.
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#ifndef	_WIN32
#include <unistd.h>
#endif
#include "e4graph.h"
#include "test.h"

/*
 * These counters are incremented by the callbacks.
 */

static int cbaddmarker;
static int cbdelmarker;
static int cbaddnode;
static int cbdelnode;
static int cbaddvertex;
static int cbdelvertex;
static int cbmodvertex;

/*
 * This function resets the callback counters.
 */

static void
ResetCallbackCounters()
{
    cbaddmarker = 0;
    cbdelmarker = 0;
    cbaddnode = 0;
    cbdelnode = 0;
    cbaddvertex = 0;
    cbdelvertex = 0;
    cbmodvertex = 0;
}

/*
 * Test how many callbacks actually occurred, against a set of expected
 * values.
 */

static bool
TestCallbackCounters(int eam, int edm, int ean, int edn, int eav, int edv)
{
    if ((eam != cbaddmarker) ||
	(edm != cbdelmarker) ||
	(ean != cbaddnode) ||
	(edn != cbdelnode) ||
	(eav != cbaddvertex) ||
	(edv != cbdelvertex)) {
	return false;
    }
    return true;
}

/*
 * These functions are the actual callbacks:
 */

static void MarkerAddCallback(void *cd, e4_Marker m)
{
    cbaddmarker++;
}
static void MarkerDelCallback(void *cd, e4_Marker m)
{
    cbdelmarker++;
}
static void NodeAddCallback(void *cd, e4_Node n)
{
    cbaddnode++;
}
static void NodeDelCallback(void *cd, e4_Node n)
{
    cbdelnode++;
}
static void VertexAddCallback(void *cd, e4_Vertex v)
{
    cbaddvertex++;
}
static void VertexDelCallback(void *cd, e4_Vertex v)
{
    cbdelvertex++;
}

/*
 * These are special callbacks that test that the storage is locked during
 * the callback.
 */

static void MarkerCLAddCallback(void *cd, e4_Marker m)
{
    e4_Storage s;

    if (!m.GetStorage(s) || !s.IsValid()) {
	return;
    }
    if (!s.IsLocked()) {
	return;
    }
    cbaddmarker++;
}

static void MarkerCLDelCallback(void *cd, e4_Marker m)
{
    e4_Storage s;

    if (!m.GetStorage(s) || !s.IsValid()) {
	return;
    }
    if (!s.IsLocked()) {
	return;
    }
    cbdelmarker++;
}

static void NodeCLAddCallback(void *cd, e4_Node n)
{
    e4_Storage s;

    if (!n.GetStorage(s) || !s.IsValid()) {
	return;
    }
    if (!s.IsLocked()) {
	return;
    }
    cbaddnode++;
}

static void NodeCLDelCallback(void *cd, e4_Node n)
{
    e4_Storage s;

    if (!n.GetStorage(s) || !s.IsValid()) {
	return;
    }
    if (!s.IsLocked()) {
	return;
    }
    cbdelnode++;
}

static void VertexCLAddCallback(void *cd, e4_Vertex v)
{
    e4_Storage s;

    if (!v.GetStorage(s) || !s.IsValid()) {
	return;
    }
    if (!s.IsLocked()) {
	return;
    }
    cbaddvertex++;
}

static void VertexCLDelCallback(void *cd, e4_Vertex v)
{
    e4_Storage s;

    if (!v.GetStorage(s) || !s.IsValid()) {
	return;
    }
    if (!s.IsLocked()) {
	return;
    }
    cbdelvertex++;
}

static void VertexCLModCallback(void *cd, e4_Vertex v)
{
    e4_Storage s;

    if (!v.GetStorage(s) || !s.IsValid()) {
	return;
    }
    if (!s.IsLocked()) {
	return;
    }
    cbmodvertex++;
}

/*
 * These are the tests:
 */

static int
test_callback1()
{
    ResetCallbackCounters();

    {
	e4_Storage s("foo.db", E4_METAKIT);

	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback1 failed (step 1)\n");
	    return 1;
	}

	s.Delete();
    }

    if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	fprintf(stderr, "test callback1 failed (step 2)\n");
	return 1;
    }

    return 0;
}

static int
test_callback2()
{
    e4_Marker m;

    ResetCallbackCounters();

    /*
     * Test basic callback functionality.
     */

    {
	e4_Storage s("foo.db", E4_METAKIT);

	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback2 failed (step 1)\n");
	    return 2;
	}
	if (!s.GetMarker("mymarker", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback2 failed (step 2)\n");
	    return 2;
	}
	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback2 failed (step 3)\n");
	    return 2;
	}
	s.DeclareMarkerAddCallback(MarkerAddCallback, (void *) 0);
	if (!s.GetMarker("mymarker", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback2 failed (step 4)\n");
	    return 2;
	}
	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback2 failed (step 5)\n");
	    return 2;
	}
	if (!s.GetMarker("mynewmarker", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback2 failed (step 6)\n");
	    return 2;
	}
	if (!TestCallbackCounters(1, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback2 failed (step 7)\n");
	    return 2;
	}

	s.Delete();
    }

    if (!TestCallbackCounters(1, 0, 0, 0, 0, 0)) {
	fprintf(stderr, "test callback2 failed (step 8)\n");
	return 2;
    }

    return 0;
}

static int
test_callback3()
{
    e4_Marker m;

    ResetCallbackCounters();

    /*
     * Test basic callback functionality.
     */

    {
	e4_Storage s("foo.db", E4_METAKIT);

	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback3 failed (step 1)\n");
	    return 3;
	}
	if (!s.GetMarker("mymarker", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback3 failed (step 2)\n");
	    return 3;
	}
	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback3 failed (step 3)\n");
	    return 3;
	}
	s.DeclareMarkerAddCallback(MarkerAddCallback, (void *) 0);
	if (!s.GetMarker("m1", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback3 failed (step 4)\n");
	    return 3;
	}
	if (!TestCallbackCounters(1, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback3 failed (step 5)\n");
	    return 3;
	}
	if (!s.GetMarker("m2", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback3 failed (step 6)\n");
	    return 3;
	}
	if (!TestCallbackCounters(2, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback3 failed (step 7)\n");
	    return 3;
	}
	s.DeclareMarkerAddCallback(MarkerAddCallback, (void *) 1);
	if (!s.GetMarker("m3", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback3 failed (step 8)\n");
	    return 3;
	}
	if (!TestCallbackCounters(4, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback3 failed (step 9)\n");
	    return 3;
	}
	s.DeleteMarkerAddCallback(MarkerAddCallback, (void *) 1);
	if (!s.GetMarker("m4", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback3 failed (step 8)\n");
	    return 3;
	}
	if (!TestCallbackCounters(5, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback3 failed (step 9)\n");
	    return 3;
	}
	s.DeleteMarkerAddCallback(MarkerAddCallback, (void *) 1);
	if (!s.GetMarker("m5", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback3 failed (step 10)\n");
	    return 3;
	}
	if (!TestCallbackCounters(6, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback3 failed (step 11)\n");
	    return 3;
	}

	s.Delete();
    }

    if (!TestCallbackCounters(6, 0, 0, 0, 0, 0)) {
	fprintf(stderr, "test callback3 failed (step 12)\n");
	return 3;
    }

    return 0;
}

static int
test_callback4()
{
    e4_Marker m, m2;
    e4_Node n;

    ResetCallbackCounters();

    /*
     * Test various conditions under which marker add and delete callbacks
     * are called.
     */

    {
	e4_Storage s("foo.db", E4_METAKIT);

	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 1)\n");
	    return 4;
	}
	if (!s.GetMarker("m1", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 2)\n");
	    return 4;
	}
	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 3)\n");
	    return 4;
	}
	s.DeclareMarkerAddCallback(MarkerAddCallback, (void *) 0);
	if (!s.GetMarker("m2", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 4)\n");
	    return 4;
	}
	if (!TestCallbackCounters(1, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 5)\n");
	    return 4;
	}
	s.DeclareMarkerDelCallback(MarkerDelCallback, (void *) 0);
	if (!m.GetMarkedNode(n) || (!n.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 6)\n");
	    return 4;
	}
	if (!n.Unmark(m) || (m.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 7)\n");
	    return 4;
	}
	if (!TestCallbackCounters(1, 1, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 8), %d\n",
		    cbdelmarker);
	    return 4;
	}
	if (!s.GetMarker("m2", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 9)\n");
	    return 4;
	}
	if (!TestCallbackCounters(2, 1, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 10)\n");
	    return 4;
	}
	if (!m.GetMarkedNode(n) || (!n.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 11)\n");
	    return 4;
	}
	if (!n.Unmark(m) || (m.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 12)\n");
	    return 4;
	}
	if (!TestCallbackCounters(2, 2, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 13)\n");
	    return 4;
	}
	if (!s.GetMarker("m2", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 14)\n");
	    return 4;
	}
	if (!TestCallbackCounters(3, 2, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 15)\n");
	    return 4;
	}
	if (!m.Delete() || m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 16)\n");
	    return 4;
	}
	if (!TestCallbackCounters(3, 3, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 17)\n");
	    return 4;
	}
	if (!s.GetMarker("m2", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 18)\n");
	    return 4;
	}
	if (!TestCallbackCounters(4, 3, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 19)\n");
	    return 4;
	}
	if (!m.GetMarkedNode(n) || (!n.IsValid())) {
	    fprintf(stderr, "test callback4 failed (step 20)\n");
	    return 4;
	}
	if (!n.Delete() || n.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 21)\n");
	    return 4;
	}

	/*
	 * Delete callback should also happen when a node with a marker
	 * is deleted.
	 */

	if (!TestCallbackCounters(4, 4, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 22)\n");
	    return 4;
	}

	/*
	 * When a marker is added to a node, it causes an add marker
	 * callback.
	 */

	if (!s.GetMarker("m1", m) || !m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 23)\n");
	    return 4;
	}
	if (!m.GetMarkedNode(n) || !n.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 24)\n");
	    return 4;
	}
	if (!n.AddMarker("m2", m2) || !m2.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 25)\n");
	    return 4;
	}
	if (!TestCallbackCounters(5, 4, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 26)\n");
	    return 4;
	}

	/*
	 * Declare another add marker and delete marker callback, and see
	 * if there are now two callbacks per add marker and delete marker
	 * events.
	 */

	s.DeclareMarkerAddCallback(MarkerAddCallback, (void *) 1);
	s.DeclareMarkerDelCallback(MarkerDelCallback, (void *) 1);

	if (!s.GetMarker("m3", m) || !m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 27)\n");
	    return 4;
	}
	if (!m.Delete() || m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 28)\n");
	    return 4;
	}
	if (!TestCallbackCounters(7, 6, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 29)\n");
	    return 4;
	}

	/*
	 * Delete the add marker and del marker callbacks just added, and
	 * ensure the number of callbacks happening for each event goes
	 * back to one.
	 */

	s.DeleteMarkerAddCallback(MarkerAddCallback, (void *) 1);
	s.DeleteMarkerDelCallback(MarkerDelCallback, (void *) 1);

	if (!s.GetMarker("m3", m) || !m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 30)\n");
	    return 4;
	}
	if (!m.Delete() || m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 31)\n");
	    return 4;
	}
	if (!TestCallbackCounters(8, 7, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 32)\n");
	    return 4;
	}
	
	/*
	 * Attempt to delete non-existent add marker and del marker
	 * callbacks, and ensure this has no impact on existing callbacks.
	 */

	s.DeleteMarkerAddCallback(MarkerAddCallback, (void *) 3);
	s.DeleteMarkerDelCallback(MarkerDelCallback, (void *) 3);

	if (!s.GetMarker("m3", m) || !m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 33)\n");
	    return 4;
	}
	if (!m.Delete() || m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 34)\n");
	    return 4;
	}
	if (!TestCallbackCounters(9, 8, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 35)\n");
	    return 4;
	}

	/*
	 * Add a callback combination that already exists. This should not
	 * affect the number of callbacks that happen per add marker or
	 * del marker event.
	 */

	s.DeclareMarkerAddCallback(MarkerAddCallback, (void *) 0);
	s.DeclareMarkerDelCallback(MarkerDelCallback, (void *) 0);

	if (!s.GetMarker("m3", m) || !m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 36)\n");
	    return 4;
	}
	if (!m.Delete() || m.IsValid()) {
	    fprintf(stderr, "test callback4 failed (step 37)\n");
	    return 4;
	}
	if (!TestCallbackCounters(10, 9, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback4 failed (step 38)\n");
	    return 4;
	}

	s.Delete();
    }

    if (!TestCallbackCounters(10, 9, 0, 0, 0, 0)) {
	fprintf(stderr, "test callback4 failed (step 39)\n");
	return 4;
    }

    return 0;
}

static int
test_callback5()
{
    e4_Marker m;
    e4_Node n1, n2;
    int rank;

    /*
     * Test various conditions under which add and delete node callbacks
     * should occur.
     */

    ResetCallbackCounters();

    {
	e4_Storage s("foo.db", E4_METAKIT);

	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 1)\n");
	    return 5;
	}
	if (!s.GetMarker("m1", m) || (!m.IsValid())) {
	    fprintf(stderr, "test callback5 failed (step 2)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 3)\n");
	    return 5;
	}
	s.DeclareNodeAddCallback(NodeAddCallback, (void *) 0);
	s.DeclareNodeDelCallback(NodeDelCallback, (void *) 0);

	/*
	 * If a marker is deleted and that causes a node to be deleted,
	 * the delete node callback should happen.
	 */

	if (!m.Delete() || m.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 4)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 0, 1, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 5)\n");
	    return 5;
	}

	/*
	 * Adding a marker causes an add node callback to happen.
	 */

	if (!s.GetMarker("m1", m) || !m.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 6)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 1, 1, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 7)\n");
	    return 5;
	}

	if (!m.GetMarkedNode(n1) || !n1.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 8)\n");
	    return 5;
	}

	/*
	 * Deleting a node explicitly causes a delete node callback.
	 */

	if (!n1.Delete() || n1.IsValid() || m.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 9)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 1, 2, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 10)\n");
	    return 5;
	}

	/*
	 * Preliminaries to get a node.
	 */

	if (!s.GetMarker("m1", m) || !m.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 11)\n");
	    return 5;
	}
	if (!m.GetMarkedNode(n1) || !n1.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 12)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 2, 2, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 13)\n");
	    return 5;
	}

	/*
	 * Adding a node field should cause an add node callback.
	 */

	if (!n1.AddNode("foo", E4_IOFIRST, rank, n2) || !n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 14)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 3, 2, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 15)\n");
	    return 5;
	}

	/*
	 * Deleting a vertex whose value is a node should not cause
	 * a delete node callback.
	 */

	if (!n1.DeleteVertex("foo") || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 16)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 3, 3, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 17)\n");
	    return 5;
	}

	/*
	 * Setting an existing vertex to a new node should cause an add node
	 * callback to happen.
	 */

	if (!n1.AddVertex("foo", E4_IOFIRST, rank, 32)) {
	    fprintf(stderr, "test callback5 failed (step 18)\n");
	    return 5;
	}
	if (!n1.SetNode("foo", n2) || !n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 19)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 4, 3, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 20)\n");
	    return 5;
	}
	if (!n2.Delete() || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 21)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 4, 4, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 22)\n");
	    return 5;
	}
	if (n1.GetVertex("foo", n2) || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 23)\n");
	    return 5;
	}

	/*
	 * Same exercise except we're setting the vertex by rank.
	 */

	if (!n1.AddVertex("foo", E4_IOFIRST, rank, 32)) {
	    fprintf(stderr, "test callback5 failed (step 24)\n");
	    return 5;
	}
	if (!n1.SetNodeByRank(rank, n2) || !n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 25)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 5, 4, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 26)\n");
	    return 5;
	}
	if (!n2.Delete() || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 27)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 5, 5, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 28)\n");
	    return 5;
	}
	if (n1.GetVertex("foo", n2) || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 29)\n");
	    return 5;
	}

	/*
	 * Adding another add node callback causes two callbacks to happen,
	 * one for each declared callback, when a new node is created.
	 */

	s.DeclareNodeAddCallback(NodeAddCallback, (void *) 1);

	if (!n1.AddVertex("foo", E4_IOFIRST, rank, 34)) {
	    fprintf(stderr, "test callback5 failed (step 30)\n");
	    return 5;
	}
	if (!n1.SetNodeByRank(rank, n2) || !n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 31)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 7, 5, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 32)\n");
	    return 5;
	}
	if (!n2.Delete() || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 33)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 7, 6, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 34)\n");
	    return 5;
	}
	if (n1.GetVertex("foo", n2) || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 35)\n");
	    return 5;
	}

	/*
	 * Removing one of the callbacks means there is only one
	 * callback when a node is created.
	 */

	s.DeleteNodeAddCallback(NodeAddCallback, (void *) 1);

	if (!n1.AddVertex("foo", E4_IOFIRST, rank, 34)) {
	    fprintf(stderr, "test callback5 failed (step 36)\n");
	    return 5;
	}
	if (!n1.SetNodeByRank(rank, n2) || !n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 37)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 8, 6, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 38)\n");
	    return 5;
	}
	if (!n2.Delete() || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 39)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 8, 7, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 40)\n");
	    return 5;
	}
	if (n1.GetVertex("foo", n2) || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 41)\n");
	    return 5;
	}

	/*
	 * Try deleting a non-existent add node callback and ensure that
	 * this does not impact existing callbacks.
	 */

	s.DeleteNodeAddCallback(NodeAddCallback, (void *) 3);

	if (!n1.AddVertex("foo", E4_IOFIRST, rank, 34)) {
	    fprintf(stderr, "test callback5 failed (step 42)\n");
	    return 5;
	}
	if (!n1.SetNodeByRank(rank, n2) || !n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 43)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 9, 7, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 44)\n");
	    return 5;
	}
	if (!n2.Delete() || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 45)\n");
	    return 5;
	}
	if (!TestCallbackCounters(0, 0, 9, 8, 0, 0)) {
	    fprintf(stderr, "test callback5 failed (step 46)\n");
	    return 5;
	}
	if (n1.GetVertex("foo", n2) || n2.IsValid()) {
	    fprintf(stderr, "test callback5 failed (step 47)\n");
	    return 5;
	}

	s.Delete();
    }

    if (!TestCallbackCounters(0, 0, 9, 8, 0, 0)) {
	fprintf(stderr, "test callback5 failed (step 48)\n");
	return 5;
    }

    return 0;
}

static int
test_callback6()
{
    e4_Node n1, n2;
    e4_Vertex v1, v2;
    int rank, i;

    /*
     * Test various conditions under which add and delete node callbacks
     * should occur.
     */

    ResetCallbackCounters();

    {
	e4_Storage s("foo.db", E4_METAKIT);

	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback6 failed (step 1)\n");
	    return 6;
	}
	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback6 failed (step 3)\n");
	    return 6;
	}
	s.DeclareVertexAddCallback(VertexAddCallback, (void *) 0);
	s.DeclareVertexDelCallback(VertexDelCallback, (void *) 0);

	if (!s.GetMarkedNode("m1", n1) || !n1.IsValid()) {
	    fprintf(stderr, "test callback6 failed (step 4)\n");
	    return 6;
	}

	/*
	 * Add a bunch of vertices and see if the callbacks happened.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback6 failed (step 5.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, i+1, 0)) {
		fprintf(stderr, "test callback6 failed (step 5.2.%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Delete a bunch of vertices and see if the callbacks happen.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.DeleteVertex("foo")) {
		fprintf(stderr, "test callback6 failed (step 6.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 32, i+1)) {
		fprintf(stderr, "test callback6 failed (step 6.2,%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Add another add vertex and delete vertex callback, and see if the
	 * number of callbacks happening is 2 when a vertex is added and
	 * deleted.
	 */

	s.DeclareVertexAddCallback(VertexAddCallback, (void *) 1);
	s.DeclareVertexDelCallback(VertexDelCallback, (void *) 1);

	/*
	 * Add a bunch of vertices and see if the callbacks happened.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback6 failed (step 7.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 32+((i+1)*2), 32)) {
		fprintf(stderr, "test callback6 failed (step 7.2.%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Delete a bunch of vertices and see if the callbacks happen.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.DeleteVertex("foo")) {
		fprintf(stderr, "test callback6 failed (step 8.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 96, 32+((i+1)*2))) {
		fprintf(stderr, "test callback6 failed (step 8.2,%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Delete the add vertex and delete vertex callbacks just added,
	 * and check that the number of callbacks per add vertex and delete
	 * vertex event goes back to 1.
	 */

	s.DeleteVertexAddCallback(VertexAddCallback, (void *) 1);
	s.DeleteVertexDelCallback(VertexDelCallback, (void *) 1);

	/*
	 * Add a bunch of vertices and see if the callbacks happened.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback6 failed (step 9.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 96+(i+1), 96)) {
		fprintf(stderr, "test callback6 failed (step 9.2.%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Delete a bunch of vertices and see if the callbacks happen.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.DeleteVertex("foo")) {
		fprintf(stderr, "test callback6 failed (step 10.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 128, 96+(i+1))) {
		fprintf(stderr, "test callback6 failed (step 10.2,%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Same exercise except use n.DeleteVertexByRank to delete the
	 * vertices, and see if the delete vertex callback happens.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback6 failed (step 11.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 128+(i+1), 128)) {
		fprintf(stderr, "test callback6 failed (step 11.2.%d)\n", i);
		return 6;
	    }
	}

	for (i = 0; i < 32; i++) {
	    if (!n1.DeleteVertexByRank(1)) {
		fprintf(stderr, "test callback6 failed (step 12.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 160, 128+(i+1))) {
		fprintf(stderr, "test callback6 failed (step 12.2,%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Try removing non-existent callbacks, ensure that that has no
	 * impact on the installed callbacks.
	 */

	s.DeleteVertexAddCallback(VertexAddCallback, (void *) 3);
	s.DeleteVertexDelCallback(VertexDelCallback, (void *) 3);

	/*
	 * Add a bunch of vertices and see if the callbacks happened.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback6 failed (step 13.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 160+(i+1), 160)) {
		fprintf(stderr, "test callback6 failed (step 13.2.%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Delete a bunch of vertices and see if the callbacks happen.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.DeleteVertex("foo")) {
		fprintf(stderr, "test callback6 failed (step 14.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 192, 160+(i+1))) {
		fprintf(stderr, "test callback6 failed (step 14.2,%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Add a bunch of vertices and see if the callbacks happened.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback6 failed (step 15.1.%d)\n", i);
		return 6;
	    }
	    if (!TestCallbackCounters(0, 0, 0, 0, 192+(i+1), 192)) {
		fprintf(stderr, "test callback6 failed (step 15.2.%d)\n", i);
		return 6;
	    }
	}

	/*
	 * Delete the node and see if the vertex deletion callbacks happen.
	 */

	n1.Delete();

	if (!TestCallbackCounters(0, 0, 0, 0, 224, 224)) {
	    fprintf(stderr, "test callback6 failed (step 16), %d %d\n",
		    cbaddvertex, cbdelvertex);
	    return 6;
	}

	s.Delete();
    }

    if (!TestCallbackCounters(0, 0, 0, 0, 224, 224)) {
	fprintf(stderr, "test callback6 failed (step 17)\n");
	return 6;
    }

    return 0;
}

static int
test_callback7()
{
    e4_Node n1;
    e4_Vertex v1, v2;
    int rank, i;

    /*
     * Test that the storage is set to locked during callbacks.
     */

    ResetCallbackCounters();
    {
	e4_Storage s("foo.db", E4_METAKIT);

	if (!TestCallbackCounters(0, 0, 0, 0, 0, 0)) {
	    fprintf(stderr, "test callback7 failed (step 1)\n");
	    return 7;
	}
	s.DeclareMarkerAddCallback(MarkerCLAddCallback, (void *) 0);
	s.DeclareMarkerDelCallback(MarkerCLDelCallback, (void *) 0);
	s.DeclareNodeAddCallback(NodeCLAddCallback, (void *) 0);
	s.DeclareNodeDelCallback(NodeCLDelCallback, (void *) 0);
	s.DeclareVertexAddCallback(VertexCLAddCallback, (void *) 0);
	s.DeclareVertexDelCallback(VertexCLDelCallback, (void *) 0);
	s.DeclareVertexModCallback(VertexCLModCallback, (void *) 0);

	if (!s.GetMarkedNode("m1", n1) || !n1.IsValid()) {
	    fprintf(stderr, "test callback7 failed (step 2)\n");
	    return 7;
	}
	if (!TestCallbackCounters(1, 0, 1, 0, 0, 0)) {
	    fprintf(stderr, "test callback7 failed (step 3)\n");
	    return 7;
	}
	if (!n1.Delete() || n1.IsValid()) {
	    fprintf(stderr, "test callback7 failed (step 4)\n");
	    return 7;
	}
	if (!TestCallbackCounters(1, 1, 1, 1, 0, 0)) {
	    fprintf(stderr, "test callback7 failed (step 5)\n");
	    return 7;
	}
	if (!s.GetMarkedNode("m1", n1) || !n1.IsValid()) {
	    fprintf(stderr, "test callback7 failed (step 6)\n");
	    return 7;
	}
	if (!TestCallbackCounters(2, 1, 2, 1, 0, 0)) {
	    fprintf(stderr, "test callback7 failed (step 7)\n");
	    return 7;
	}

	/*
	 * Add a bunch of vertices and see if the callbacks happened.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback7 failed (step 8.1.%d)\n", i);
		return 7;
	    }
	    if (!TestCallbackCounters(2, 1, 2, 1, i+1, 0)) {
		fprintf(stderr, "test callback7 failed (step 8.2.%d)\n", i);
		return 7;
	    }
	}

	/*
	 * Delete the node, see if all the callbacks happen.
	 */

	if (!n1.Delete() || n1.IsValid()) {
	    fprintf(stderr, "test callback7 failed (step 9)\n");
	    return 7;
	}
	if (!TestCallbackCounters(2, 2, 2, 2, 32, 32)) {
	    fprintf(stderr, "test callback7 failed (step 10)\n");
	    return 7;
	}

	/*
	 * Recreate the node and see if the callbacks happen.
	 */

	if (!s.GetMarkedNode("m1", n1) || !n1.IsValid()) {
	    fprintf(stderr, "test callback7 failed (step 11)\n");
	    return 7;
	}
	if (!TestCallbackCounters(3, 2, 3, 2, 32, 32)) {
	    fprintf(stderr, "test callback7 failed (step 12)\n");
	    return 7;
	}

	/*
	 * Add a bunch of vertices and see if the callbacks happened.
	 */

	for (i = 0; i < 32; i++) {
	    if (!n1.AddVertex("foo", E4_IOLAST, rank, 34)) {
		fprintf(stderr, "test callback7 failed (step 13.1.%d)\n", i);
		return 7;
	    }
	    if (!TestCallbackCounters(3, 2, 3, 2, 33+i, 32)) {
		fprintf(stderr, "test callback7 failed (step 13.2.%d)\n", i);
		return 7;
	    }
	}

	/*
	 * Grab a vertex and modify it, see if the callback happens.
	 */

	if (!n1.GetVertexRefByRank(11, v1) || !v1.IsValid()) {
	    fprintf(stderr, "test callback7 failed (step 14)\n");
	    return 7;
	}
	if (!v1.Set((float) 34.45)) {
	    fprintf(stderr, "test callback7 failed (step 15)\n");
	    return 7;
	}
	if (cbmodvertex != 1) {
	    fprintf(stderr, "test callback7 failed (step 16)\n");
	    return 7;
	}

	/*
	 * Delete the node, see if all the callbacks happen.
	 */

	if (!n1.Delete() || n1.IsValid()) {
	    fprintf(stderr, "test callback7 failed (step 17)\n");
	    return 7;
	}
	if (!TestCallbackCounters(3, 3, 3, 3, 64, 64)) {
	    fprintf(stderr, "test callback7 failed (step 18)\n");
	    return 7;
	}

	s.Delete();
    }

    return 0;
}

int
test_callback()
{
    int result;

    fprintf(stderr, "Running callback tests: ");

    result = test_callback1();
    if (result != 0) {
	return result;
    }
    fprintf(stderr, ".");
    result = test_callback2();
    if (result != 0) {
        return result;
    }
    fprintf(stderr, ".");
    result = test_callback3();
    if (result != 0) {
        return result;
    }
    fprintf(stderr, ".");
    result = test_callback4();
    if (result != 0) {
        return result;
    }
    fprintf(stderr, ".");
    result = test_callback5();
    if (result != 0) {
        return result;
    }
    fprintf(stderr, ".");
    result = test_callback6();
    if (result != 0) {
        return result;
    }
    fprintf(stderr, ".");
    result = test_callback7();
    if (result != 0) {
	return result;
    }
    fprintf(stderr, ".\n");
    
    return 0;
}
