/*
 * e4xmlgenerator.cpp --
 *
 *	This file contains the implementation of the e4_XMLGenerator class
 *	defined in e4xml.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 *	Copyright: JYL Software, Inc., (c) 2000-2001.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4xml.h"

/*
 * This constructs an empty XML generator.
 */

e4_XMLGenerator::e4_XMLGenerator()
    : sn(invalidNode),
      se(NULL),
      nodesSeen(NULL),
      nodeCounter(0),
      ready(false),
      base64str(NULL)
{
    dsptr = new e4_DString();
    Reset();
}

/*
 * This constructor produces a fully initialized (ready to produce)
 * XML Generator.
 */

e4_XMLGenerator::e4_XMLGenerator(e4_Node nn, char *sen)
    : sn(invalidNode),
      se(NULL),
      nodeCounter(0),
      nodesSeen(NULL),
      ready(false),
      base64str(NULL)
{
    dsptr = new e4_DString();
    SetElementNameAndNode(sen, nn);
}

/*
 * Destructor:
 */

e4_XMLGenerator::~e4_XMLGenerator()
{
    delete dsptr;
    if (se != NULL) {
	free(se);
    }
    if (base64str != NULL) {
	free(base64str);
    }
    e4_DeleteHashTable(nodesSeen);
}

/*
 * Set the starting node.
 */

void
e4_XMLGenerator::SetNode(e4_Node nn)
{
    sn = nn;
    Reset();
}

/*
 * Set the element name.
 */

void
e4_XMLGenerator::SetElementName(char *sen)
{
    if (se != NULL) {
	free(se);
    }
    if (sen == NULL) {
	se = NULL;
    } else {
	se = (char *) malloc(strlen(sen) + 1);
	strcpy(se, sen);
    }
    Reset();
}

/*
 * Set both the element name and the node.
 */

void
e4_XMLGenerator::SetElementNameAndNode(char *sen, e4_Node nn)
{
    if (se != NULL) {
	free(se);
    }
    if (sen == NULL) {
	se = NULL;
    } else {
	se = (char *) malloc(strlen(sen) + 1);
	strcpy(se, sen);
    }
    sn = nn;
    Reset();
}

/*
 * Retrieve the associated node.
 */

void
e4_XMLGenerator::GetNode(e4_Node &nn) const
{
    nn = sn;
}

/*
 * Retrieve the wrapping XML tag name.
 */

char *
e4_XMLGenerator::GetElementName() const
{
    if (!ready) {
	return NULL;
    }
    return se;
}

/*
 * Reset the generator.
 */

void
e4_XMLGenerator::Reset()
{
    ready = ((sn == invalidNode) || (se == NULL)) ? false : true;
    dsptr->SetLength(0);
    if (nodesSeen != NULL) {
	e4_DeleteHashTable(nodesSeen);
    }
    nodesSeen = e4_NewHashTable(E4_ONE_WORD_KEYS);
}

/*
 * Generate XML.
 */

char *
e4_XMLGenerator::Generate()
{
    /*
     * If we're not ready, return NULL right away.
     */

    if (!ready) {
	return NULL;
    }

    /*
     * Reset the generator to an empty state.
     */

    Reset();

    /*
     * Generate from the starting element and node.
     */

    GenerateNode(se, sn);

    /*
     * Finally return the accumulated XML output.
     */

    return Get();
}

/*
 * Return the accumulated XML output.
 */

char *
e4_XMLGenerator::Get()
{
    if (!ready) {
	return NULL;
    }
    return dsptr->Get();
}

/*
 * Generate XML output from a node and a wrapping element.
 */

void
e4_XMLGenerator::GenerateNode(char *elname, e4_Node nn)
{
    e4_HashEntry *ePtr;
    e4_NodeUniqueID nuid;
    int id, isnew, vcount, i, lim;
    char intbuf[32];
    bool attributesSeen = false;
    e4_Node attributes;
    e4_Vertex v, vv;
    e4_Marker mm;

    /*
     * First of all determine if this is a node we've already produced
     * XML output for. If so, generate a back reference.
     */

    if (!nn.GetUniqueID(nuid)) {
	return;
    }
    id = nuid.GetUniqueID();
    ePtr = E4_CREATEHASHENTRY(nodesSeen, (char *) id, &isnew);

    /*
     * If it's not new, we've generated this node already. In that
     * case generate a back pointer to the entity containing this
     * node. That entity will be marked with a __multiparent__
     * attribute with the token that's the hash value of this entry.
     */

    if (!isnew) {
	dsptr->Append("<__nodebackref__ __nodeid__=\"", -1);
	sprintf(intbuf, "%d", (int) E4_GETHASHVALUE(ePtr));
	dsptr->Append(intbuf, -1);
	dsptr->Append("\" __name__=\"", -1);
	dsptr->Append(elname, -1);
	dsptr->Append("\"/>\n", -1);

	return;
    }

    dsptr->Append("<", -1);
    dsptr->Append(elname, -1);

    /*
     * If this node has any parents, it's possible that we have to
     * generate XML for a cyclical graph. Tag this node with a 
     * __nodeid__ attribute so that when it gets deserialized, this
     * node can be associated with subsequent __nodebackref__ elements.
     */

    if (nn.ParentCount() > 0) {
	E4_SETHASHVALUE(ePtr, nodeCounter);
	nodeCounter++;

	dsptr->Append(" __nodeid__=\"", -1);
	sprintf(intbuf, "%d", (int) E4_GETHASHVALUE(ePtr));
	dsptr->Append(intbuf, -1);
	dsptr->Append("\"", -1);
    } else {
	e4_DeleteHashEntry(ePtr);
    }

    /*
     * If the node has any markers attached, generate __marker__ attributes
     * for all of them.
     */

    for (i = 0, lim = nn.MarkerCount(); i < lim; i++) {
	dsptr->Append(" __marker__=\"", -1);
	nn.GetMarker(i+1, mm);
	dsptr->Append(mm.GetName(), -1);
	dsptr->Append("\"", -1);
    }

    /*
     * Figure out how many vertices this node contains. If there are exactly
     * zero, produce an empty XML element with the given name.
     */

    vcount = nn.VertexCount();
    if (vcount == 0) {
	dsptr->Append("/>\n", -1);
	return;
    }

    /*
     * If the first vertex is named "__attributes__" and has a node value,
     * that sub-node contains attributes of the node we're producing XML
     * output for. If so, produce the attributes.
     */

    if ((strcmp(nn.VertexName(1), "__attributes__") == 0) &&
	(nn.VertexTypeByRank(1) == E4_VTNODE)) {
	(void) nn.GetVertexByRank(1, attributes);

	if (GenerateAttributes(attributes)) {
	    vcount--;
	    attributesSeen = true;
	}

	if (vcount <= 0) {
	    dsptr->Append("/>\n", -1);
	    return;
	}
    }

    /*
     * Finish off the open-element construct.
     */

    dsptr->Append(">\n", -1);
    
    /*
     * Now iterate over the vertices in turn.
     */

    (void) nn.GetVertexRefByRank((attributesSeen ? 2 : 1), v);
    for (; vcount > 0; vcount--) {
	GenerateVertex(v);
	(void) v.Next(1, vv);
	v = vv;
    }

    dsptr->Append("</", -1);
    dsptr->Append(elname, -1);
    dsptr->Append(">\n", -1);
}

/*
 * Generate XML output for a given vertex.
 */

void
e4_XMLGenerator::GenerateVertex(e4_Vertex v)
{
    e4_Node nn;
    e4_Vertex vv1, vv2, vv3;
    const void *bytes;
    int nbytes;
    const char *s;
    int vi;
    double vf;
    char *base64str;
    char buf[32];

    /*
     * Special case for vertex with name "__data__" and type "string". This
     * generates character data in the XML output.
     */

    if ((v.Type() == E4_VTSTRING) &&
	(strcmp(v.Name(), "__data__") == 0)) {
	(void) v.Get(s);
	dsptr->Append(s, -1);
	dsptr->Append("\n", -1);
	return;
    }

    /*
     * Special case for vertex with name "__comment__" and type "string". This
     * generates an XML comment in the output.
     */

    if ((v.Type() == E4_VTSTRING) && (strcmp(v.Name(), "__comment__") == 0)) {
	(void) v.Get(s);
	dsptr->Append("<-- ", -1);
	dsptr->Append(s, -1);
	dsptr->Append(" -->", -1);
	return;
    }

    /*
     * Special case for vertex with name "__cdata__" and type "node". This
     * generates an XML CDATA section in the output, if that node contains
     * a vertex named "__data__" with a string value.
     */

    if ((v.Type() == E4_VTNODE) &&
	(strcmp(v.Name(), "__cdata__") == 0)) {
	(void) v.Get(nn);
	if ((nn.GetVertexRef("__data__", vv1)) &&
	    (vv1.Type() == E4_VTSTRING)) {
	    (void) vv1.Get(s);
	    dsptr->Append("<![CDATA[", -1);
	    dsptr->Append(s, -1);
	    dsptr->Append("]]>", -1);
	    return;
	}
    }

    /*
     * Special case for vertex with name "__processinginstruction__" and
     * type "node". This generates an XML processing instruction element
     * in the output, if that node has two vertices with names "__target__"
     * and "__data__" with string values.
     */

    if ((v.Type() == E4_VTNODE) &&
	(strcmp(v.Name(), "__processinginstruction__") == 0)) {
	(void) v.Get(nn);
	if ((nn.GetVertexRef("__target__", vv1)) &&
	    (vv1.Type() == E4_VTSTRING) &&
	    (nn.GetVertexRef("__data__", vv2)) &&
	    (vv2.Type() == E4_VTSTRING)) {
	    dsptr->Append("<?", -1);
	    (void) vv1.Get(s);
	    dsptr->Append(s, -1);
	    dsptr->Append(" ", -1);
	    (void) vv2.Get(s);
	    dsptr->Append(s, -1);
	    dsptr->Append(" ?>", -1);
	    return;
	}
    }

    /*
     * Special case for a vertex named "__xml__" with a node value. This
     * generates an XML declaration element in the output, if that node has
     * three vertices with names "__version__", "__encoding__" and
     * "__standalone__" with string, string and int values.
     */

    if ((v.Type() == E4_VTNODE) && (strcmp(v.Name(), "__xml__") == 0)) {
	(void) v.Get(nn);
	dsptr->Append("<?xml ", -1);
	if ((nn.GetVertexRef("__version__", vv1)) &&
	    (vv1.Type() == E4_VTSTRING)) {
	    dsptr->Append(" version=\"", -1);
	    (void) vv1.Get(s);
	    dsptr->Append(s, -1);
	    dsptr->Append("\"", -1);
	}
	if ((nn.GetVertexRef("__encoding__", vv2)) &&
	    (vv2.Type() == E4_VTSTRING)) {
	    dsptr->Append(" encoding=\"", -1);
	    (void) vv2.Get(s);
	    dsptr->Append(s, -1);
	    dsptr->Append("\"", -1);
	}
	if ((nn.GetVertexRef("__standalone__", vv3)) &&
	    (vv3.Type() == E4_VTINT)) {
	    (void) vv3.Get(vi);
	    switch (vi) {
	    case -1:
		break;
	    case 0:
		dsptr->Append(" standalone=no", -1);
		break;
	    case 1:
		dsptr->Append(" standalone=yes", -1);
		break;
	    }
	}
    }

    /*
     * Special case for a vertex with a node value and named
     * "__doctypedecl__". This node produces a DOCTYPE section in
     * the XML output, with special handling for vertices named
     * "__doctypename__", "__sysid__", "__pubid__" and
     * "__hasinternalsubset__".
     */

    if ((v.Type() == E4_VTNODE) && 
	(strcmp(v.Name(), "__doctypedecl__") == 0)) {
	(void) v.Get(nn);
	dsptr->Append("<!DOCTYPE ", -1);
	if (nn.GetVertex("__doctypename__", s)) {
	    dsptr->Append(s, -1);
	    if (nn.GetVertex("__sysid__", s)) {
		dsptr->Append(" ", -1);
		dsptr->Append(s, -1);
	    }
	    if (nn.GetVertex("__pubid__", s)) {
		dsptr->Append(" \"", -1);
		dsptr->Append(s, -1);
		dsptr->Append("\"", -1);
	    }

	    /*
	     * NOT CLEAR what to do with the __hasinternalsubset__
	     * vertex. Ignore it for now.
	     */

	    dsptr->Append(" [\n", -1);

	    /*
	     * Produce output for element declarations.
	     */

	    GenerateElementDecls(nn);

	    /*
	     * All done, close the DOCTYPE declaration.
	     */

	    dsptr->Append("\n]>", -1);
	    return;
	}
    }
	    
    /*
     * Special case for a vertex with a node value and none of the
     * preceding special case names.
     */

    if (v.Type() == E4_VTNODE) {
	(void) v.Get(nn);
	GenerateNode((char *) v.Name(), nn);
	return;
    }

    /*
     * Generate a regular vertex.
     */

    dsptr->Append("<__vertex__", -1);

    /*
     * Produce the name attribute.
     */

    dsptr->Append(" name=\"", -1);
    dsptr->Append(v.Name(), -1);
    dsptr->Append("\"", -1);

    /*
     * Produce the type and value attributes.
     */

    switch (v.Type()) {
    case E4_VTNODE:
	/*
	 * Should not occur!
	 */

	break;

    case E4_VTINT:
	dsptr->Append(" type=\"int\" value=\"", -1);
	(void) v.Get(vi);
	sprintf(buf, "%d", vi);
	dsptr->Append(buf, -1);
	dsptr->Append("\"", -1);
	break;
    case E4_VTDOUBLE:
	dsptr->Append(" type=\"double\" value=\"", -1);
	(void) v.Get(vf);
	sprintf(buf, "%e", vf);
	dsptr->Append(buf, -1);
	dsptr->Append("\"", -1);
	break;
    case E4_VTSTRING:
	dsptr->Append(" type=\"string\" value=\"", -1);
	(void) v.Get(s);
	dsptr->Append(s, -1);
	dsptr->Append("\"", -1);
	break;
    case E4_VTBINARY:
	dsptr->Append(" type=\"binary\" value=\"", -1);
	(void) v.Get(bytes, nbytes);
	base64str = base64_encode((byte *) bytes, nbytes);
	dsptr->Append(base64str, -1);
	dsptr->Append("\"", -1);

	/*
	 * Free the buffer returned by base64_encode.
	 */

	free(base64str);
	break;
    default:
	/*
	 * Should not occur!
	 */

	break;
    }

    dsptr->Append("/>\n", -1);
}

/*
 * Generate XML output for a list of attributes.
 */

bool
e4_XMLGenerator::GenerateAttributes(e4_Node attrs) const
{
    e4_Vertex v, vv;
    int len;
    const char *s;
    bool first;

    /*
     * First of all a sanity check: the node must contain only
     * vertices of type E4_VTSTRING. If not, we generate a regular
     * node for it.
     */

    for (len = attrs.VertexCount(), first = true, 
	   (void) attrs.GetVertexRefByRank(1, v);
	 len > 0;
	 len--, v.Next(1, vv), first = false) {
	if (first == false) {
	    v = vv;
	}
	if (v.Type() != E4_VTSTRING) {
	    return false;
	}
    }

    /*
     * Now generate an attribute for each vertex.
     */

    for (len = attrs.VertexCount(), first = true,
	   (void) attrs.GetVertexRefByRank(1, v);
	 len > 0;
	 len--, v.Next(1, vv), first = false) {
	if (first == false) {
	    v = vv;
	}
	(void) v.Get(s);
	dsptr->Append(" ", -1);
	dsptr->Append(v.Name(), -1);
	dsptr->Append("=\"", -1);
	dsptr->Append(s, -1);
	dsptr->Append("\"", -1);
    }

    return true;
}

/*
 * This operation generates output for a series of DOCTYPE element
 * declarations.
 *
 * NOTE: For now this does nothing because I don't understand what
 * it all means.. :(
 */

bool
e4_XMLGenerator::GenerateElementDecls(e4_Node doctype) const
{
   return true;
}

/*
 * This operation encodes a binary value of a given length into a BASE64
 * encoded string.
 */

char *
e4_XMLGenerator::Base64_Encode(byte *bytes, int len)
{
    if (base64str != NULL) {
	free(base64str);
    }
    base64str = base64_encode(bytes, len);
    return base64str;
}
