/*
 * e4xmlparser.cpp --
 *
 *	This file contains the implementation of the e4_XMLParser class
 *	defined in e4xml.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 *	Copyright: JYL Software, Inc., (c) 2000-2001.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4xml.h"

/*
 * Static functions defined in this file:
 */

static bool		e4Graph_IsBlank(const char *, int);
static void		e4Graph_StartElement(void *, const char *,
					     const char **);
static void		e4Graph_EndElement(void *, const char *);
static void		e4Graph_DataHandler(void *, const char *, int);
static void		e4Graph_CommentHandler(void *, const char *);
static void		e4Graph_StartCData(void *);
static void		e4Graph_EndCData(void *);
static void		e4Graph_ProcessingInstructionHandler(void *,
							     const char *,
							     const char *);
static void		e4Graph_XMLDeclHandler(void *, const char *,
					       const char *, int);
static void		e4Graph_StartDocTypeHandler(void *,
						    const char *,
						    const char *,
						    const char *,
						    int);
static void		e4Graph_EndDocTypeHandler(void *);

/*
 * This routine determines whether the data is all whitespace.
 */

static bool
e4Graph_IsBlank(const char *data, int len)
{
    int i;

    for (i = 0; i < len; i++) {
	if ((data[i] != '\n') && (data[i] != ' ') && (data[i] != '\t')) {
	    return false;
	}
    }
    return true;
}

/*
 * This procedure is called for the start of each XML element. It creates
 * a new node with the given name in the current node and populates the
 * new node with the supplied attributes. Finally it sets the user data of
 * the parser to the new node so that nested elements will be inserted into
 * the new node.
 */

static void
e4Graph_StartElement(void *userData, const char *name, const char **attributes)
{
    int id = -1, startattr = 0, rank;
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n;
    e4_Node nn;
    e4_Node nna;
    e4_Marker mm;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if (p == NULL) {
	return;
    }
    if (p->HasError() == true) {
	return;
    }
    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * Finish any open character data first of all.
     */

    p->FinishCharData();

    /*
     * If this element represents a vertex, handle it specially.
     */

    if ((strcmp(name, "__vertex__") == 0) && 
	(attributes != NULL) &&
	(attributes[0] != NULL) && (strcmp(attributes[0], "name") == 0) &&
	(attributes[1] != NULL) &&
	(attributes[2] != NULL) && (strcmp(attributes[2], "type") == 0) &&
	(attributes[3] != NULL) &&
	(attributes[4] != NULL) && (strcmp(attributes[4], "value") == 0) &&
	(attributes[5] != NULL) &&
	(attributes[6] == NULL)) {
	p->AddVertex(attributes[1], attributes[3], attributes[5]);
	return;
    }

    /*
     * If this element represents a backwards reference to a node
     * that was already parsed, handle it specially.
     */

    if ((strcmp(name, "__nodebackref__") == 0) &&
	(attributes != NULL) &&
	(attributes[0] != NULL) && 
	(strcmp(attributes[0], "__nodeid__") == 0) &&
	(attributes[1] != NULL) &&
	(attributes[2] != NULL) && 
	(strcmp(attributes[2], "__name__") == 0) &&
	(attributes[3] != NULL) &&
	(attributes[4] == NULL)) {
	p->AddNodeBackRef(attributes[3], attributes[1]);
	return;
    }

    /*
     * Otherwise the new element will be represented as a new node. Increment
     * the depth counter.
     */

    if (!n.AddNode(name, E4_IOLAST, rank, nn)) {
	p->FlagError("Can't add node");
	return;
    }

    /*
     * Parse an optional __nodeid__ attribute.
     */

    if ((attributes != NULL) && (attributes[startattr] != NULL)) {
	if (strcmp(attributes[startattr], "__nodeid__") == 0) {
	    id = atoi(attributes[startattr + 1]);
	    p->HashNode(nn, id);
	    startattr += 2;
	}
    }

    /*
     * Parse any __marker__ attributes. These get turned into markers
     * attached to this node.
     */

    if ((attributes != NULL) && (attributes[startattr] != NULL)) {
	for (; 
	     (attributes[startattr] != NULL) &&
	       (strcmp(attributes[startattr], "__marker__") == 0);
	     startattr += 2) {
	    if (!nn.AddMarker(attributes[startattr + 1], mm) ||
		!mm.IsValid()) {
		p->FlagError("Can't add markers");
		return;
	    }
	}
    }

    if ((attributes != NULL) && (attributes[startattr] != NULL)) {
	if (!nn.AddNode("__attributes__", E4_IOLAST, rank, nna)) {
	    p->FlagError("Can't add attributes");
	    return;
	}
	for (; attributes[startattr] != NULL; startattr += 2) {
	    if (!nna.AddVertex(attributes[startattr], E4_IOLAST, rank,
			       attributes[startattr + 1])) {
		p->FlagError("Can't add attribute");
		return;
	    }
	}
    }

    /*
     * Record the current depth of recursion.
     */

    p->IncrDepth();

    /*
     * Set the current node (where nested elements are added) to 'nn'.
     */

    p->SetNode(nn);
}

/*
 * This procedure is called for the end of each XML element. It sets the user
 * data of the parser to the parent of the current node so that subsequent
 * elements will be inserted into the parent.
 */

static void
e4Graph_EndElement(void *userData, const char *name)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node nn;
    e4_Node n;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if (p == NULL) {
	return;
    }
    if (p->HasError() == true) {
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * If the current element being closed was a vertex, handle
     * specially.
     */

    if ((strcmp(name, "__vertex__") == 0) && (p->InVertex())) {
	p->ExitVertex();
	return;
    }

    /*
     * If the current element being closed is a back reference,
     * don't pop the node stack.
     */

    if (strcmp(name, "__nodebackref__") == 0) {
	return;
    }

    /*
     * It's a regular node. Parse it and all sub-elements.
     */

    if (!p->GetNode(nn)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!nn.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }
    if (!nn.GetParent(n)) {
	p->FlagError("Can't get parent of current node");
	return;
    }

    /*
     * We're done with the child node, reset the current node to its parent.
     * Also decrement the nesting level.
     */

    p->DecrDepth();
    p->SetNode(n);
}

/*
 * This procedure is called to parse an XML comment. It creates a vertex
 * with name "__comment__" and string value the contents of the comment.
 */

static void
e4Graph_CommentHandler(void *userData, const char *comment)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n;
    int rank;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * If we are in a vertex-add, do not allow adding a comment!
     */

    if (p->InVertex()) {
	p->FlagError("In vertex-add, cannot add comment");
	return;
    }

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * Add a "__comment__" vertex with the comment string as value.
     */

    if (!n.AddVertex("__comment__", E4_IOLAST, rank, comment)) {
	p->FlagError("Could not add comment");
    }
}

/*
 * This procedure handles the start of a CDATA section in the XML input.
 * It creates a node named "__cdata__" in the current node and sets that
 * node as the current node.
 */

static void
e4Graph_StartCData(void *userData)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n, nn;
    int rank;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * If we are in a vertex-add, do not allow adding a comment!
     */

    if (p->InVertex()) {
	p->FlagError("In vertex-add, cannot add CDATA section");
	return;
    }

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * Add a vertex named "__cdata__" with a new node.
     */

    if ((!n.AddNode("__cdata__", E4_IOLAST, rank, nn)) ||
	(!nn.IsValid())) {
	p->FlagError("Could not add CDATA section");
	return;
    }

    /*
     * Record the current depth of recursion.
     */

    p->IncrDepth();

    /*
     * Set the current node (where nested elements are added) to 'nn'.
     */

    p->SetNode(nn);
}

/*
 * This procedure handles the end of a CDATA section. It simply pops
 * the current node off the stack and sets the current node to its
 * parent.
 */

static void
e4Graph_EndCData(void *userData)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n, nn;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * Get the parent of this node.
     */

    if ((!n.GetParent(nn)) || (!nn.IsValid())) {
	p->FlagError("Could not close CDATA section");
	return;
    }

    /*
     * We're done with the child node, reset the current node to its parent.
     * Also decrement the nesting level.
     */

    p->DecrDepth();
    p->SetNode(nn);
}

/*
 * This procedure handles a processing instruction XML input element.
 * It stores the data in a node called "__processinginstruction__" with
 * two vertices named "__target__" and "__data__".
 */

static void
e4Graph_ProcessingInstructionHandler(void *userData,
				     const char *target,
				     const char *data)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n, nn;
    int rank;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * If we are in a vertex-add, do not allow adding a comment!
     */

    if (p->InVertex()) {
	p->FlagError(
		"In vertex-add, cannot add PROCESSINGINSTRUCTION section");
	return;
    }

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * Add a vertex named "__processinginstruction__" with a new node.
     */

    if ((!n.AddNode("__processinginstruction__", E4_IOLAST, rank, nn)) ||
	(!nn.IsValid())) {
	p->FlagError("Could not add PROCESSINGINSTRUCTION section");
	return;
    }

    /*
     * Add two vertices, named "__target__" and "__data__".
     */

    if ((!nn.AddVertex("__target__", E4_IOLAST, rank, target)) ||
	(!nn.AddVertex("__data__", E4_IOLAST, rank, data))) {
	p->FlagError("Could not add PROCESSINGINSTRUCTION section");
	return;
    }
}

/*
 * This procedure is called to handle an XML declaration. It creates a
 * node named "__xml__" with three vertices named "__version__",
 * "__encoding__" and "__standalone__" with string, string and int values.
 */

static void
e4Graph_XMLDeclHandler(void *userData, const char *version,
		       const char *encoding, int standalone)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n, nn;
    int rank;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * If we are in a vertex-add, do not allow adding data!
     */

    if (p->InVertex()) {
	p->FlagError("In vertex-add, cannot add data");
	return;
    }

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * Create a node named "__xml__".
     */

    if ((!n.AddNode("__xml__", E4_IOLAST, rank, nn)) || (!nn.IsValid())) {
	p->FlagError("Could not add XML declaration");
	return;
    }

    /*
     * Potentially add three vertices named "__version__", "__encoding__" and
     * "__standalone__".
     */

    if (version != NULL) {
	if (!nn.AddVertex("__version__", E4_IOLAST, rank, version)) {
	    p->FlagError("Could not add XML declaration");
	    return;
	}
    }
    if (encoding != NULL) {
	if (!nn.AddVertex("__encoding__", E4_IOLAST, rank, encoding)) {
	    p->FlagError("Could not add XML declaration");
	    return;
	}
    }
    if (!nn.AddVertex("__standalone__", E4_IOLAST, rank, standalone)) {
	p->FlagError("Could not add XML declaration");
    }
}

/*
 * This procedure is called to handle the start of a document decl XML
 * input element. This creates a node named "__doctypedecl__" with
 * several vertices named "__doctypename__", "__sysid__", "__pubid__"
 * and "__hasinternalsubset__" with string, string, string and int
 * types.
 */

static void
e4Graph_StartDocTypeHandler(void *userData, const char *doctypeName,
			    const char *sysid, const char *pubid,
			    int hasinternalsubset)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n, nn;
    int rank;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * If we are in a vertex-add, do not allow adding data!
     */

    if (p->InVertex()) {
	p->FlagError("In vertex-add, cannot add DOCTYPE declaration");
	return;
    }

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * Create a node named "__doctypedecl__".
     */

    if ((!n.AddNode("__doctypedecl__", E4_IOLAST, rank, nn)) ||
	(!nn.IsValid())) {
	p->FlagError("Could not add DOCTYPE declaration");
	return;
    }

    /*
     * Add the specified vertices.
     */

    if (doctypeName != NULL) {
	if (!nn.AddVertex("__doctypename__", E4_IOLAST, rank, doctypeName)) {
	    p->FlagError("Could not add DOCTYPE declaration");
	    return;
	}
    }
    if (sysid != NULL) {
	if (!nn.AddVertex("__sysid__", E4_IOLAST, rank, sysid)) {
	    p->FlagError("Could not add DOCTYPE declaration");
	    return;
	}
    }
    if (pubid != NULL) {
	if (!nn.AddVertex("__pubid__", E4_IOLAST, rank, pubid)) {
	    p->FlagError("Could not add DOCTYPE declaration");
	    return;
	}
    }
    if (!nn.AddVertex("__hasinternalsubset__", E4_IOLAST, rank, 
		      hasinternalsubset)) {
	p->FlagError("Could not add DOCTYPE declaration");
	return;
    }
    
    /*
     * Record the current depth of recursion.
     */

    p->IncrDepth();

    /*
     * Set the current node (where nested elements are added) to 'nn'.
     */

    p->SetNode(nn);
}

/*
 * This procedure is called to handle the end of a doctype declaration.
 * It closes the current node and installs its parent as the current
 * node.
 */

static void
e4Graph_EndDocTypeHandler(void *userData)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n, nn;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * If we are in a vertex-add, do not allow adding data!
     */

    if (p->InVertex()) {
	p->FlagError("In vertex-add, cannot add DOCTYPE declaration");
	return;
    }

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * First of all ensure any currently open character data is saved.
     */

    p->FinishCharData();

    /*
     * Get the parent of the current node.
     */

    if ((!n.GetParent(nn)) || (!nn.IsValid())) {
	p->FlagError("Could not close DOCTYPE declaration section");
	return;
    }

    /*
     * We're done with the child node, reset the current node to its parent.
     * Also decrement the nesting level.
     */

    p->DecrDepth();
    p->SetNode(nn);
}

/*
 * This procedure is called to handle character data contained within
 * XML elements. It stores the data into the current node as the value
 * of a field named "__data__".
 */

static void
e4Graph_DataHandler(void *userData, const char *data, int len)
{
    e4_XMLParser *p = (e4_XMLParser *) userData;
    e4_Node n;
    char *buf;

    /*
     * If there was an error before this call, do nothing more so that
     * the state will be at least partially correct!
     */

    if ((p == NULL) || (p->HasError() == true)) {
	return;
    }

    /*
     * If we are in a vertex-add, do not allow adding data!
     */

    if (p->InVertex()) {
	p->FlagError("In vertex-add, cannot add data");
	return;
    }

    /*
     * Check the current node is valid.
     */

    if (!p->GetNode(n)) {
	p->FlagError("Can't get current node from parser");
	return;
    }
    if (!n.IsValid()) {
	p->FlagError("Invalid node");
	return;
    }

    /*
     * If the data is "artificial", ignore it.
     */

    if (e4Graph_IsBlank(data, len)) {
	return;
    }

    /*
     * The data is not NULL terminated, so we must copy it and null-terminate
     * the copy. Alternatively we could insert the data as a binary value, but
     * that would lose the type information.
     */

    buf = new char[len + 1];
    strncpy(buf, data, len);
    buf[len] = '\0';

    /*
     * Now store it.
     */

    p->StoreCharData(buf, len);
    delete[] buf;
}

/*
 * This constructor creates an empty XML parser.
 */

e4_XMLParser::e4_XMLParser()
    : ready(false),
      n(invalidNode),
      s(invalidStorage),
      inVertex(false),
      parser(NULL),
      depth(0),
      started(false),
      error(false),
      errorString(NULL),
      base64bytes(NULL),
      dsp(NULL)
{
    nodesSeen = e4_NewHashTable(E4_ONE_WORD_KEYS);
}

/*
 * This constructor creates an XML parser from a given node and
 * stores all XML input in an e4Graph under the given node.
 */

e4_XMLParser::e4_XMLParser(e4_Node nn)
    : ready(true),
      n(nn),
      inVertex(false),
      parser(NULL),
      depth(0),
      started(false),
      error(false),
      errorString(NULL),
      base64bytes(NULL),
      dsp(NULL)
{
    ConstructParser();
    (void) n.GetStorage(s);
    nodesSeen = e4_NewHashTable(E4_ONE_WORD_KEYS);
}

/*
 * The destructor deletes the instance parser if one was created, and
 * sets the instance node to invalidNode to ensure that the refcount is
 * decremented properly.
 */

e4_XMLParser::~e4_XMLParser()
{
    if (parser != NULL) {
	XML_ParserFree(parser);
    }
    n = invalidNode;
    s = invalidStorage;
    e4_DeleteHashTable(nodesSeen);
    if (base64bytes != NULL) {
	free(base64bytes);
    }
    if (dsp != NULL) {
	delete dsp;
    }
}

/*
 * Is this parse finished?
 */

bool
e4_XMLParser::Finished()
{
    if ((started == true) && (depth <= 0)) {
	return true;
    }
    return false;
}

/*
 * Increment the depth of nesting of XML elements entered.
 */

void
e4_XMLParser::IncrDepth()
{
    depth++;

    /*
     * Also record the parse has started.
     */

    started = true;
}

/*
 * Decrement the depth of nesting of XML elements entered.
 */

void
e4_XMLParser::DecrDepth()
{
    depth--;
}

/*
 * This operation stores a unique ID for the new node in the nodesSeen
 * hash table so that subsequent back references to it can be made.
 */

void
e4_XMLParser::HashNode(e4_Node nn, int id) const
{
   e4_NodeUniqueID nuid;
   e4_HashEntry *ePtr;
   int isnew, i;

   ePtr = E4_CREATEHASHENTRY(nodesSeen, (char *) id, &isnew);

   (void) nn.GetUniqueID(nuid);
   i = nuid.GetUniqueID();
   E4_SETHASHVALUE(ePtr, i);
}

/*
 * This operation assigns a node to the parser. This allows the parser
 * to switch where it is going to insert new elements.
 *
 * NOTE: Care must be taken to only switch nodes between parse operations
 * and not during a parse operation, or unpredictable behavior may result.
 */

void
e4_XMLParser::SetNode(e4_Node nn)
{
    if (inVertex == false) {
	n = nn;
	n.GetStorage(s);
    } else {
	FlagError("Can't set node while inside vertex!");
    }
}

/*
 * This operation returns the current node in nn.
 */

bool
e4_XMLParser::GetNode(e4_Node &nn)
{
    nn = n;
    return true;
}

/*
 * This operation adds a new vertex to the current node.
 */

bool
e4_XMLParser::AddVertex(const char *vname, const char *vtype, const char *vval)
{
    int rank;
    int iv;
    double fv;
    byte *bytes;
    int len;

    if (inVertex == true) {
	FlagError("Already inside a vertex!");
	return false;
    }

    if (strcmp(vtype, "int") == 0) {
	iv = atoi(vval);
	if (!n.AddVertex(vname, E4_IOLAST, rank, iv)) {
	    FlagError("Can't add int vertex");
	    return false;
	}
	inVertex = true;
	return true;
    }
    if (strcmp(vtype, "double") == 0) {
	fv = (double) atof(vval);
	if (!n.AddVertex(vname, E4_IOLAST, rank, fv)) {
	    FlagError("Can't add double vertex");
	    return false;
	}
	inVertex = true;
	return true;
    }
    if (strcmp(vtype, "string") == 0) {
	if (!n.AddVertex(vname, E4_IOLAST, rank, vval)) {
	    FlagError("Can't add string vertex");
	    return false;
	}
	inVertex = true;
	return true;
    }
    if (strcmp(vtype, "binary") == 0) {
	bytes = base64_decode(vval, &len);
	if (bytes == NULL) {
	    FlagError("Can't parse binary value");
	    return false;
	}
	if (!n.AddVertex(vname, E4_IOLAST, rank, bytes, len)) {
	    FlagError("Can't add binary vertex");
	    return false;
	}
	inVertex = true;
	return true;
    }

    FlagError("Can't add vertex -- illegal type");
    return false;
}

/*
 * This operation checks whether we're currently inside a vertex-add.
 */

bool
e4_XMLParser::InVertex() const
{
    return inVertex;
}

/*
 * This operation finishes a vertex-add.
 */

void
e4_XMLParser::ExitVertex()
{
    if (inVertex == false) {
	FlagError("Not in vertex-add!");
    }
    inVertex = false;
}

/*
 * This operation adds a back reference to a node that was already
 * parsed.
 */

bool
e4_XMLParser::AddNodeBackRef(const char *nm, const char *nid)
{
   int index, rank, id;
   e4_NodeUniqueID nuid;
   e4_HashEntry *ePtr;
   e4_Node bn;

   index = atoi(nid);
   ePtr = E4_FINDHASHENTRY(nodesSeen, (char *) index);
   if (ePtr == NULL) {
	FlagError("Invalid node back reference!");
	return false;
   }
   id = (int) E4_GETHASHVALUE(ePtr);
   nuid.SetUniqueID(id, s);
   if ((!s.GetNodeFromID(nuid, bn)) || (!bn.IsValid())) {
	FlagError("Invalid node back reference!");
	return false;
   }
   if (!n.AddVertex(nm, E4_IOLAST, rank, bn)) {
	FlagError("Invalid node back reference!");
	return false;
   }
   return true;
}

/*
 * This operation constructs a new parser from the file pointer and
 * node stored already in the instance. It returns true if the construction
 * succeeded.
 */

bool
e4_XMLParser::ConstructParser()
{
    error = false;
    errorString = NULL;
    if (!n.IsValid()) {
	ready = false;
    } else {
        if (parser != NULL) {
	    XML_ParserFree(parser);
	}
	parser = XML_ParserCreate(NULL);
	if (parser == NULL) {
	    ready = false;
	} else {
	    ready = true;
	    XML_SetUserData(parser, this);
	    XML_SetElementHandler(parser, 
				  e4Graph_StartElement,
				  e4Graph_EndElement);
	    XML_SetCommentHandler(parser, e4Graph_CommentHandler);
	    XML_SetCdataSectionHandler(parser,
				       e4Graph_StartCData,
				       e4Graph_EndCData);
	    XML_SetProcessingInstructionHandler(
				parser,
				e4Graph_ProcessingInstructionHandler);
	    XML_SetXmlDeclHandler(parser, e4Graph_XMLDeclHandler);
	    XML_SetCharacterDataHandler(parser, e4Graph_DataHandler);
	    XML_SetDoctypeDeclHandler(parser,
				      e4Graph_StartDocTypeHandler,
				      e4Graph_EndDocTypeHandler);
	}
    }
    return ready;
}

/*
 * This operation parses a buffer of data.
 */

bool
e4_XMLParser::Parse(char *buf, size_t len)
{
    if (!ready) {
	return false;
    }
    error = false;
    if (!XML_Parse(parser, buf, len, false)) {
	error = true;
	if (errorString == NULL) {
	    FlagError("Input following XML expression");
	}
	return false;
    }
    return true;
}
    
/*
 * This operation flags an error and stores an error string into the
 * parser.
 */

void
e4_XMLParser::FlagError(const char *msg)
{
    char *format = "Error: %s (line %d, column %d, byte %d)";

    error = true;
    errorString = new char[50 /* strlen(format) */ + strlen(msg) + 64];
    sprintf(errorString, format, msg,
	    XML_GetCurrentLineNumber(parser),
	    XML_GetCurrentColumnNumber(parser),
	    XML_GetCurrentByteIndex(parser));
}

/*
 * This operation returns true if there was an error in a previous
 * operation that has not yet been cleared.
 */

bool
e4_XMLParser::HasError()
{
    return error;
}

/*
 * This operation retrieves the error string if there was an error.
 */

const char *
e4_XMLParser::ErrorString()
{
    if (errorString == NULL) {
	return "";
    }
    return (const char *) errorString;
}

/*
 * This operation clears an error state from the parser.
 */

void
e4_XMLParser::ClearError()
{
    error = false;
    errorString = NULL;
}

/*
 * This operation decodes a BASE64 encoded string into a binary value.
 */

byte *
e4_XMLParser::Base64_Decode(const char *base64str, int *len)
{
    if (base64bytes != NULL) {
	free(base64bytes);
    }
    base64bytes = base64_decode(base64str, len);
    return base64bytes;
}

/*
 * Store a character data string.
 */

void
e4_XMLParser::StoreCharData(char *data, int len)
{
    /*
     * Auto-initialize.
     */

    if (dsp == NULL) {
	dsp = new e4_DString();
    }

    /*
     * Add the current string to the character data being collected.
     */

    dsp->Append(data, len);
}

/*
 * Finish collecting a sequence of character data and save it in the
 * persistent storage, inside the current node as the last vertex.
 * The vertex name will be "__data__".
 */

void
e4_XMLParser::FinishCharData()
{
    int rank;

    /*
     * If we're not collecting character data, return immediately.
     */

    if (dsp == NULL) {
	return;
    }

    /*
     * Assume the current node is valid.
     */

    if (!n.AddVertex("__data__", E4_IOLAST, rank, dsp->Get())) {
	FlagError("Can't add data to current node");
    }

    /*
     * Free the dynamic string and ensure that the parser state reflects
     * that this collection of character data sequence is done, by setting
     * dsp to NULL.
     */

    delete dsp;
    dsp = NULL;
}
